Package: dataRetrieval
Type: Package
Title: Retrieval Functions for USGS and EPA Hydrology and Water Quality
        Data
Version: 2.7.25
Authors@R: c(
    person("Laura", "DeCicco", role = c("aut","cre"),
            email = "ldecicco@usgs.gov",
            comment=c(ORCID="0000-0002-3915-9487")),
    person("Robert", "Hirsch", role = c("aut"),
            comment=c(ORCID = "0000-0002-4534-075X")),
    person("David","Lorenz", role=c("aut")),
    person("Jordan", "Read", role = c("ctb")),
    person("Jordan", "Walker", role = c("ctb")),
    person("Lindsay","Platt", role=c("ctb")),
    person("David","Watkins", role=c("aut"),
            email = "wwatkins@usgs.gov",
            comment=c(ORCID = "0000-0002-7544-0700")),
    person("David", "Blodgett", role="aut",
            comment=c(ORCID = "0000-0001-9489-1710"),
            email = "dblodgett@usgs.gov"),
    person("Mike", "Johnson", role=c("aut"),
            email = "mikecp11@gmail.com",
            comment=c(ORCID = "0000-0002-5288-8350")),
    person("Aliesha", "Krall", role="ctb",
            email = "akrall@usgs.gov",
            comment=c(ORCID = "0000-0003-2521-5043")),
    person("Lee", "Stanish", role="ctb",
            comment=c(ORCID = "0000-0002-9775-6861")),
    person("Joseph", "Zemmels", role="aut",
            email = "jzemmels@usgs.gov",
            comment=c(ORCID = "0009-0008-1463-6313")),
    person("Elise", "Hinman", role="aut",
            email = "ehinman@usgs.gov",
            comment=c(ORCID = "0000-0001-5396-1583")),  
    person("Michael", "Mahoney", role="ctb",
            email = "mjmahoney@usgs.gov",
            comment=c(ORCID = "0000-0003-2402-304X"))  
            )
Description: Collection of functions to help retrieve U.S. Geological Survey
    and U.S. Environmental Protection Agency water quality and
    hydrology data from web services. 
License: CC0
Copyright: This software is in the public domain because it contains
        materials that originally came from the United States
        Geological Survey, an agency of the United States Department of
        Interior.
Depends: R (>= 4.1.0)
Imports: curl (>= 7.0.0), lubridate (>= 1.5.0), stats, utils, xml2,
        readr (>= 1.4.0), jsonlite, httr2, whisker, sf, data.table,
        rlang
Suggests: covr, dplyr, knitr, rmarkdown, testthat
Encoding: UTF-8
BuildVignettes: true
VignetteBuilder: knitr
BugReports: https://github.com/DOI-USGS/dataRetrieval/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-27 18:01:15 UTC; ldecicco
Author: Laura DeCicco [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3915-9487>),
  Robert Hirsch [aut] (ORCID: <https://orcid.org/0000-0002-4534-075X>),
  David Lorenz [aut],
  Jordan Read [ctb],
  Jordan Walker [ctb],
  Lindsay Platt [ctb],
  David Watkins [aut] (ORCID: <https://orcid.org/0000-0002-7544-0700>),
  David Blodgett [aut] (ORCID: <https://orcid.org/0000-0001-9489-1710>),
  Mike Johnson [aut] (ORCID: <https://orcid.org/0000-0002-5288-8350>),
  Aliesha Krall [ctb] (ORCID: <https://orcid.org/0000-0003-2521-5043>),
  Lee Stanish [ctb] (ORCID: <https://orcid.org/0000-0002-9775-6861>),
  Joseph Zemmels [aut] (ORCID: <https://orcid.org/0009-0008-1463-6313>),
  Elise Hinman [aut] (ORCID: <https://orcid.org/0000-0001-5396-1583>),
  Michael Mahoney [ctb] (ORCID: <https://orcid.org/0000-0003-2402-304X>)
Maintainer: Laura DeCicco <ldecicco@usgs.gov>
Repository: CRAN
Date/Publication: 2026-05-28 05:11:18 UTC
Built: R 4.5.2; ; 2026-05-28 05:48:29 UTC; unix

# Generated by roxygen2: do not edit by hand

export(addWaterYear)
export(calcWaterYear)
export(checkWQPdates)
export(check_OGC_requests)
export(check_waterdata_sample_params)
export(constructNWISURL)
export(constructWQPURL)
export(construct_api_requests)
export(construct_waterdata_sample_request)
export(countyCd)
export(countyCdLookup)
export(create_NWIS_bib)
export(create_WQP_bib)
export(findNLDI)
export(getWebServiceData)
export(get_nldi_sources)
export(get_ogc_params)
export(importNGWMN)
export(importRDB1)
export(importWQP)
export(importWaterML1)
export(importWaterML2)
export(is_dataRetrieval_user)
export(pCodeToName)
export(parameterCdFile)
export(parse_WQP)
export(pcode_to_name)
export(readNGWMNdata)
export(readNGWMNlevels)
export(readNGWMNsites)
export(readNWISdata)
export(readNWISdv)
export(readNWISpCode)
export(readNWISpeak)
export(readNWISrating)
export(readNWISsite)
export(readNWISstat)
export(readNWISuse)
export(readNWISuv)
export(readWQPdata)
export(readWQPqw)
export(readWQPsummary)
export(read_USGS_samples)
export(read_waterdata)
export(read_waterdata_channel)
export(read_waterdata_combined_meta)
export(read_waterdata_continuous)
export(read_waterdata_daily)
export(read_waterdata_field_measurements)
export(read_waterdata_field_meta)
export(read_waterdata_latest_continuous)
export(read_waterdata_latest_daily)
export(read_waterdata_metadata)
export(read_waterdata_monitoring_location)
export(read_waterdata_parameter_codes)
export(read_waterdata_samples)
export(read_waterdata_stats_daterange)
export(read_waterdata_stats_por)
export(read_waterdata_ts_meta)
export(renameNWISColumns)
export(setAccess)
export(stateCd)
export(stateCdLookup)
export(summarize_USGS_samples)
export(summarize_waterdata_samples)
export(whatNWISdata)
export(whatNWISsites)
export(whatWQPdata)
export(whatWQPmetrics)
export(whatWQPsamples)
export(whatWQPsites)
export(wqp_check_status)
export(zeroPad)
importFrom(data.table,":=")
importFrom(data.table,.BY)
importFrom(data.table,.EACHI)
importFrom(data.table,.GRP)
importFrom(data.table,.I)
importFrom(data.table,.N)
importFrom(data.table,.NGRP)
importFrom(data.table,.SD)
importFrom(data.table,data.table)

# dataretrieval: Download hydrologic data

:warning: USGS data availability and format are changing on Water Quality Portal (WQP). Beginning in February 2024 data obtained from WQP legacy profiles will not include new USGS data or recent updates to existing data. 
To view the status of changes in data availability and code functionality, visit: https://doi-usgs.github.io/dataRetrieval/articles/Status.html

## What is dataretrieval?
`dataretrieval` was created to simplify the process of loading hydrologic data into the Python environment.
Like the original R version [`dataRetrieval`](https://github.com/DOI-USGS/dataRetrieval),
it is designed to retrieve the major data types of U.S. Geological Survey (USGS) hydrology
data that are available on the Web, as well as data from the Water
Quality Portal (WQP), which currently houses water quality data from the
Environmental Protection Agency (EPA), U.S. Department of Agriculture
(USDA), and USGS. Direct USGS data is obtained from a service called the
National Water Information System (NWIS).

Note that the python version is not a direct port of the original: it attempts to reproduce the functionality of the R package,
though its organization and interface often differ.

If there's a hydrologic or environmental data portal that you'd like dataretrieval to 
work with, raise it as an [issue](https://github.com/USGS-python/dataretrieval/issues).

Here's an example using `dataretrieval` to retrieve data from the National Water Information System (NWIS).

```python
# first import the functions for downloading data from NWIS
import dataretrieval.nwis as nwis

# specify the USGS site code for which we want data.
site = '03339000'


# get instantaneous values (iv)
df = nwis.get_record(sites=site, service='iv', start='2017-12-31', end='2018-01-01')

# get water quality samples (qwdata)
df2 = nwis.get_record(sites=site, service='qwdata', start='2017-12-31', end='2018-01-01')

# get basic info about the site
df3 = nwis.get_record(sites=site, service='site')
```
Services available from NWIS include:
- instantaneous values (iv)
- daily values (dv)
- statistics (stat)
- site info (site)
- discharge peaks (peaks)
- discharge measurements (measurements)
* water quality samples (qwdata)

To access the full functionality available from NWIS web services, nwis.get record appends any additional kwargs into the REST request. For example
```python
nwis.get_record(sites='03339000', service='dv', start='2017-12-31', parameterCd='00060')
```
will download daily data with the parameter code 00060 (discharge).

## Accessing the "Internal" NWIS
If you're connected to the USGS network, dataretrieval call pull from the internal (non-public) NWIS interface.
Most dataretrieval functions pass kwargs directly to NWIS's REST API, which provides simple access to internal data; simply specify "access='3'".
For example
```python
nwis.get_record(sites='05404147',service='iv', start='2021-01-01', end='2021-3-01', access='3')
```

More services and documentation to come!

## Quick start

dataretrieval can be installed using pip:
	
    $ python3 -m pip install -U dataretrieval

or conda:

    $ conda install -c conda-forge dataretrieval

More examples of use are include in [`demos`](https://github.com/USGS-python/dataretrieval/tree/master/demos).

## Issue tracker

Please report any bugs and enhancement ideas using the dataretrieval issue
tracker:

  https://github.com/USGS-python/dataretrieval/issues

Feel free to also ask questions on the tracker.


## Contributing

Any help in testing, development, documentation and other tasks is welcome.
For more details, see the file [CONTRIBUTING.md](CONTRIBUTING.md).


[![Coverage Status](https://coveralls.io/repos/github/thodson-usgs/data_retrieval/badge.svg?branch=master)](https://coveralls.io/github/thodson-usgs/data_retrieval?branch=master)

## Package Support
The Water Mission Area of the USGS supports the development and maintenance of `dataretrieval`
and most likely further into the future.
Resources are available primarily for maintenance and responding to user questions.
Priorities on the development of new features are determined by the `dataretrieval` development team.


## Acknowledgments
This material is partially based upon work supported by the National Science Foundation (NSF) under award 1931297.
Any opinions, findings, conclusions, or recommendations expressed in this material are those of the authors and do not necessarily reflect the views of the NSF.

## Disclaimer

This software is preliminary or provisional and is subject to revision. 
It is being provided to meet the need for timely best science.
The software has not received final approval by the U.S. Geological Survey (USGS).
No warranty, expressed or implied, is made by the USGS or the U.S. Government as to the functionality of the software and related material nor shall the fact of release constitute any such warranty. 
The software is provided on the condition that neither the USGS nor the U.S. Government shall be held liable for any damages resulting from the authorized or unauthorized use of the software.

## Citation

Hodson, T.O., Hariharan, J.A., Black, S., and Horsburgh, J.S., 2023, dataretrieval (Python): a Python package for discovering
and retrieving water data available from U.S. federal hydrologic web services:
U.S. Geological Survey software release,
https://doi.org/10.5066/P94I5TX3.

# Retrieve data from the National Water Quality Network (NWQN)

> This usage example is for demonstration and not for research or
> operational use.

This example uses Lithops to retrieve data from every NWQN
monitoring site, then writes the results to Parquet files on S3. Each
retrieval also searches the NLDI for neighboring sites with NWQN data and
merges those data. In the streamflow retrieval, the neighborhood search
progressively fill in gaps in the record by taking data from the
nearest streamgage and rescaling it by the drainage area ratio.

1. Set up a Python environment
```bash
conda create --name dataretrieval-lithops -y python=3.11
conda activate dataretrieval-lithops
pip install -r requirements.txt
```

2. Configure compute and storage backends for [lithops](https://lithops-cloud.github.io/docs/source/configuration.html).
The configuration in `lithops.yaml` uses AWS Lambda for [compute](https://lithops-cloud.github.io/docs/source/compute_config/aws_lambda.html) and AWS S3 for [storage](https://lithops-cloud.github.io/docs/source/storage_config/aws_s3.html).
To use those backends, simply edit `lithops.yaml` with your `bucket` and `execution_role`.

3. Build a runtime image for Cubed
```bash
export LITHOPS_CONFIG_FILE=$(pwd)/lithops.yaml
lithops runtime build -b aws_lambda -f Dockerfile_dataretrieval dataretrieval-runtime
```

4. Download the site list from ScienceBase using `wget` or navigate to the URL and copy the CVS into `nwqn_data_pull/`.
```bash
wget https://www.sciencebase.gov/catalog/file/get/655d2063d34ee4b6e05cc9e6?f=__disk__b3%2F3e%2F5b%2Fb33e5b0038f004c2a48818d0fcc88a0921f3f689 -O NWQN_sites.csv
```

5. Create a s3 bucket for the output, then set it as an environmental variable
```bash
export DESTINATION_BUCKET=<path/to/bucket>
```

6. Run the scripts
```bash
python retrieve_nwqn_samples.py

python retrieve_nwqn_streamflow.py
```

## Cleaning up
To rebuild the Lithops image, delete the existing one by running
```bash
lithops runtime delete -b aws_lambda -d dataretrieval-runtime
```

## ----setup, include=FALSE, message=FALSE------------------
library(knitr)
library(dataRetrieval)

options(continue = " ")
options(width = 60)
knitr::opts_chunk$set(
  echo = TRUE,
  warning = FALSE,
  message = FALSE,
  fig.height = 7,
  fig.width = 7
)

## ----workflow, echo=TRUE,eval=FALSE-----------------------
# library(dataRetrieval)
# # Choptank River near Greensboro, MD
# siteNumber <- "USGS-01491000"
# ChoptankInfo <- read_waterdata_monitoring_location(siteNumber)
# parameterCd <- "00060"
# 
# # Raw daily data:
# rawDailyData <- read_waterdata_daily(monitoring_location_id = siteNumber,
#                                      parameter_code = parameterCd,
#                                      time = c("1980-01-01", "2010-01-01"))
# 
# 
# pCode <- read_waterdata_parameter_codes(parameter_code = parameterCd)

## ----echo=FALSE-------------------------------------------
Functions <- c(
  "read_waterdata",
  "read_waterdata_daily",
  "read_waterdata_continuous",
  "read_waterdata_rating",
  "read_waterdata_field_measurements",
  "read_waterdata_peaks",
  "read_waterdata_field_measurements",
  "readNWISuse",
  "read_waterdata_stats_por, read_waterdata_stats_daterange",
  "read_waterdata_parameter_codes",
  "read_waterdata_monitoring_location",
  "read_waterdata_samples",
  "summarize_waterdata_samples",
  "read_waterdata_combined_meta",
  "read_waterdata_ts_meta",
  "readWQPdata",
  "readWQPqw",
  "whatWQPsites",
  "whatWQPdata",
  "readWQPsummary",
  "whatWQPmetrics",
  "whatWQPsamples"
)

Description <- c(
  "Time series data using user-specified queries", # read_waterdata
  "Daily values", # read_waterdata_daily
  "Instantaneous values", # read_waterdata_continuous
  "Rating table for active streamgage", # readNWISrating
  "Surface-water measurements", # read_waterdata_field_measurements
  "Peak flow", # readNWISpeak
  "Groundwater levels", # read_waterdata_field_measurements
  "Water use", # readNWISuse
  "Statistical service", # readNWISstat
  "Parameter code information", # read_waterdata_parameter_codes
  "Site information", # read_waterdata_monitoring_location
  "Discrete UGSS water quality data", # read_waterdata_samples
  "Discrete USGS water quality summary",
  "Site search using user-specified queries",
  "Data availability",
  "User-specified queries",
  "Water quality data",
  "Site search",
  "Data availability",
  "Summary data",
  "Metric availability",
  "Sample availability"
)
Source <- c("USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "NWIS",
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Samples Data",
            "USGS Samples Data",
            "USGS Water Data API",
            "USGS Water Data API",
            rep("WQP", 7))


data.df <- data.frame(
  Name = Functions,
  `Data Returned` = Description,
  Source, stringsAsFactors = FALSE
)

kable(data.df, col.names = c("Name", "Data Returned", "Source"),
  caption = "Table 1: dataRetrieval functions"
)

## ----tableParameterCodes, echo=FALSE----------------------


pCode <- c("00060", "00065", "00010", "00045", "00400")
shortName <- c(
  "Discharge [ft<sup>3</sup>/s]",
  "Gage height [ft]",
  "Temperature [C]",
  "Precipitation [in]",
  "pH"
)

data.df <- data.frame(pCode, shortName, stringsAsFactors = FALSE)

kable(data.df,
  caption = "Table 2: Common USGS Parameter Codes"
)

## ----tableStatCodes, echo=FALSE---------------------------
StatCode <- c("00001", "00002", "00003", "00008")
shortName <- c("Maximum", "Minimum", "Mean", "Median")

data.df <- data.frame(StatCode, shortName, stringsAsFactors = FALSE)

kable(data.df,
  caption = "Table 3: Commonly used USGS Stat Codes"
)

## ----getSite, echo=TRUE, eval=FALSE-----------------------
# siteNumbers <- c("USGS-01491000", "USGS-01645000")
# siteINFO <- read_waterdata_monitoring_location(siteNumbers)

## ----getSiteExtended, echo=TRUE, eval=FALSE---------------
# # Continuing from the previous example:
# # This pulls out just the daily, mean data:
# 
# dailyDataAvailable <- read_waterdata_ts_meta(
#   monitoring_location_id = siteNumbers,
#   computation_period_identifier = "Daily",
#   statistic_id = "00003"
# )

## ----echo=FALSE, eval=FALSE-------------------------------
# 
# tableData <- dailyDataAvailable[c("monitoring_location_id",
#                          "parameter_description",
#                          "unit_of_measure",
#                          "begin", "end")]
# 
# tableData$begin <- as.Date(tableData$begin)
# tableData$end <- as.Date(tableData$end)
# tableData <- sf::st_drop_geometry(tableData)
# 
# 
# knitr::kable(tableData,
#              caption = "Table 4: Reformatted version of output from the read_waterdata_ts_meta function for the Choptank River near Greensboro, MD, and from Seneca Creek at Dawsonville, MD from the daily values service [Some columns deleted for space considerations]")
# 
# 
# # nolint end

## ----label=getPCodeInfo, echo=TRUE, eval=FALSE------------
# # Using defaults:
# parameterCd <- "00618"
# parameterINFO <- read_waterdata_parameter_codes(parameter_code = parameterCd)

## ----label=getNWISDaily, echo=TRUE, eval=FALSE------------
# 
# # Choptank River near Greensboro, MD:
# siteNumber <- "USSG-01491000"
# parameterCd <- "00060" # Discharge
# startDate <- "2009-10-01"
# endDate <- "2012-09-30"
# 
# discharge <- read_waterdata_daily(monitoring_location_id = siteNumber,
#                                   parameter_code = parameterCd,
#                                   time = c(startDate, endDate))

## ----label=getNWIStemperature, echo=TRUE, eval=FALSE------
# siteNumber <- "USGS-01491000"
# parameterCd <- c("00010", "00060") # Temperature and discharge
# statCd <- c("00001", "00003") # Mean and maximum
# startDate <- "2012-01-01"
# endDate <- "2012-05-01"
# 
# temperatureAndFlow <- read_waterdata_daily(monitoring_location_id = siteNumber,
#                                   parameter_code = parameterCd,
#                                   statistic_id = statCd,
#                                   time = c(startDate, endDate))

## ----label=getNWIStemperature2, echo=FALSE, eval=TRUE-----
filePath <- system.file("extdata", package = "dataRetrieval")
fileName <- "temperatureAndFlow.RData"
fullPath <- file.path(filePath, fileName)
load(fullPath)

## ---------------------------------------------------------

temperature <- temperatureAndFlow[temperatureAndFlow$parameter_code == "00010",]
temperature <- temperature[temperature$statistic_id == "00001",]

flow <- temperatureAndFlow[temperatureAndFlow$parameter_code == "00060",]

par(mar = c(5, 5, 5, 5)) # sets the size of the plot window

plot(temperature$time, temperature$value,
  ylab = "Maximum Temperture [C]",
  xlab = ""
)
par(new = TRUE)
plot(flow$time,
  flow$value,
  col = "red", type = "l",
  xaxt = "n", yaxt = "n",
  xlab = "", ylab = "",
  axes = FALSE
)
axis(4, col = "red", col.axis = "red")
mtext("Discharge [ft3/s]", side = 4, line = 3, col = "red")
title("CHOPTANK RIVER NEAR GREENSBORO, MD")
legend("topleft", unique(temperatureAndFlow$unit_of_measure),
  col = c("black", "red"), lty = c(NA, 1),
  pch = c(1, NA)
)

## ----label=readNWISuv, eval=FALSE-------------------------
# 
# parameterCd <- "00060" # Discharge
# startDate <- "2012-05-12"
# endDate <- "2012-05-13"
# dischargeUnit <- read_waterdata_continuous(monitoring_location_id = siteNumber,
#                                            parameter_code = parameterCd,
#                                            time = c(startDate, endDate))
# 

## ----gwlexample, echo=TRUE, eval=FALSE--------------------
# siteNumber <- "USGS-434400121275801"
# groundWater <- read_waterdata_field_measurements(monitoring_location_id = siteNumber)

## ----peakexample, echo=TRUE, eval=FALSE-------------------
# siteNumber <- "USGS-01594440"
# peakData <- read_waterdata_peaks(monitoring_location_id = siteNumber)

## ----ratingexample, echo=TRUE, eval=FALSE-----------------
# ratingData <- read_waterdata_ratings(monitoring_location_id = siteNumber,
#                                     file_type = "base")
# comment(ratingData[[1]])

## ----surfexample, echo=TRUE, eval=FALSE-------------------
# surfaceData <- read_waterdata_field_measurements(monitoring_location_id = "USGS-01594440")

## ----eval=FALSE-------------------------------------------
# discharge_stats_por <- read_waterdata_stats_por(
#   monitoring_location_id = "USGS-05428500",
#   parameter_code = "00060"
# )
# 
# discharge_stats_daterange <- read_waterdata_stats_daterange(
#   monitoring_location_id = "USGS-05428500",
#   parameter_code = "00060"
# )

## ----label=getQWData, echo=TRUE, eval=FALSE---------------
# specificCond <- readWQPqw(
#   "WIDNR_WQX-10032762",
#   "Specific conductance",
#   "2011-05-01", "2011-09-30"
# )

## ----NJChloride, eval=FALSE-------------------------------
# 
# sitesNJ <- whatWQPsites(
#   statecode = "US:34",
#   characteristicName = "Chloride"
# )

## ----phData, eval=FALSE-----------------------------------
# dataPH <- readWQPdata(
#   statecode = "US:55",
#   characteristicName = "pH"
# )

## ----eval=FALSE-------------------------------------------
# type <- "Stream"
# sites <- whatWQPdata(countycode = "US:55:025", siteType = type)

## ----eval=FALSE-------------------------------------------
# site <- whatWQPsamples(siteid = "USGS-01594440")

## ----eval=FALSE-------------------------------------------
# type <- "Stream"
# sites <- whatWQPmetrics(countycode = "US:55:025", siteType = type)

## ----seeVignette,eval = FALSE-----------------------------
# # to see all available vignettes:
# vignette(package="dataRetrieval")
# 
# #to open a specific vignette:
# vignette(topic = "qwdata_changes", package = "dataRetrieval")

## ----cite, eval=TRUE--------------------------------------
citation(package = "dataRetrieval")

## ----nwisCite, eval=FALSE---------------------------------
# 
# dv <- read_waterdata_daily(monitoring_location_id = "USGS-04085427",
#                            parameter_code = "00060",
#                            time = c("2012-01-01", "2012-06-30"))
# 
# NWIScitation <- create_NWIS_bib(dv)
# NWIScitation

## ----show1, eval=FALSE------------------------------------
# print(NWIScitation, style = "Bibtex")

## ----show2, eval=FALSE------------------------------------
# print(NWIScitation, style = "citation")

## ----WQPcite, eval=FALSE----------------------------------
# SC <- readWQPqw(siteNumbers = "USGS-05288705",
#                 parameterCd = "00300")
# 
# WQPcitation <- create_WQP_bib(SC)
# WQPcitation

## ----show3, eval=FALSE------------------------------------
# print(WQPcitation, style = "Bibtex")

## ----show4, eval=FALSE------------------------------------
# print(WQPcitation, style = "citation")

## ----setup, include=FALSE, message=FALSE------------------
library(knitr)
library(dataRetrieval)
library(dplyr)

options(continue = " ")

knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  fig.height = 7,
  fig.width = 7
)

## ----eval=FALSE-------------------------------------------
# wqpData <- readWQPqw(paste0("USGS-", site_ids), parameterCd)

## ----echo=FALSE-------------------------------------------
nwisData <- readRDS("nwisData.rds")
wqpData <- readRDS("wqpData.rds")

## ---------------------------------------------------------
nrow(nwisData)
nrow(wqpData)

## ---------------------------------------------------------
ncol(nwisData)
ncol(wqpData)

## ---------------------------------------------------------
names(attributes(nwisData))
names(attributes(wqpData))

## ---------------------------------------------------------
site_NWIS <- attr(nwisData, "siteInfo")
site_WQP <- attr(wqpData, "siteInfo")

## ---------------------------------------------------------
library(dplyr)

nwisData_relevant <- nwisData |> 
  select(
    site_no, startDateTime, parm_cd,
    remark_cd, result_va
  ) |> 
  arrange(startDateTime, parm_cd)

knitr::kable(head(nwisData_relevant))

## ---------------------------------------------------------
wqpData_relevant <- wqpData |> 
  select(
    site_no = Location_Identifier,
    startDateTime = Activity_StartDateTime,
    parm_cd = USGSpcode,
    remark_cd = Result_ResultDetectionCondition,
    result_va = Result_Measure,
    detection_level = DetectionLimit_MeasureA
  ) |> 
  arrange(startDateTime, parm_cd)
knitr::kable(head(wqpData_relevant))

## ---------------------------------------------------------

censored_text <- c(
  "Not Detected",
  "Non-Detect",
  "Non Detect",
  "Detected Not Quantified",
  "Below Quantification Limit"
)

wqpData_relevant <- wqpData |> 
  mutate(left_censored = grepl(paste(censored_text, collapse = "|"),
    Result_ResultDetectionCondition,
    ignore.case = TRUE
  )) |> 
  select(
    site_no = Location_Identifier,
    startDateTime = Activity_StartDateTime,
    parm_cd = USGSpcode,
    left_censored,
    result_va = Result_Measure,
    detection_level = DetectionLimit_MeasureA,
    dl_units = DetectionLimit_MeasureUnitA
  ) |> 
  arrange(startDateTime, parm_cd)

knitr::kable(head(wqpData_relevant))

## ----whatdata, eval=FALSE---------------------------------
# whatNWIS <- whatNWISdata(
#   siteNumber = site_ids,
#   service = "qw"
# )

## ----whatdatanew, eval=FALSE------------------------------
# whatWQP <- whatWQPdata(siteNumber = paste0("USGS-", site_ids))

## ----echo=TRUE, eval=TRUE---------------------------------
schema <- readr::read_csv("https://www.epa.gov/system/files/other-files/2024-07/schema_outbound_wqx3.0.csv")

## ----echo=TRUE, eval=TRUE---------------------------------
sub_schema <- schema |> 
  select(WQX3 = FieldName3.0,
         WQX2 = FieldName2.0.Narrow) |> 
  filter(!is.na(WQX2))

knitr::kable(sub_schema)

#' Coordinate the request and retrieval of OGC calls
#'
#' @param args arguments from individual functions
#' @param output_id Name of id column to return
#' @param service Endpoint name.
#'
#' @noRd
#' @return data.frame with attributes
get_ogc_data <- function(args, output_id, service) {
  chunk_size <- args[["chunk_size"]]
  args[["..."]] <- NULL

  if (
    !is.na(chunk_size) & length(args[["monitoring_location_id"]]) > chunk_size
  ) {
    ml_splits <- split(
      args[["monitoring_location_id"]],
      ceiling(seq_along(args[["monitoring_location_id"]]) / chunk_size)
    )

    rl <- lapply(ml_splits, function(x) {
      args[["monitoring_location_id"]] <- x
      get_ogc_data(args = args, output_id = output_id, service = service)
    })

    rl_filtered <- rl[
      vapply(rl, FUN = function(x) dim(x)[1], FUN.VALUE = NA_integer_) > 0
    ]

    return_list <- data.frame(data.table::rbindlist(
      rl_filtered,
      use.names = TRUE,
      ignore.attr = TRUE
    ))
  } else {
    args[["output_id"]] <- output_id
    args[["service"]] <- service

    req <- do.call(construct_api_requests, args)

    no_paging <- grepl("f=csv", req$url)

    message("Requesting:\n", req$url)

    if (no_paging) {
      return_list <- get_csv(req, limit = args[["limit"]])
    } else {
      return_list <- walk_pages(req)
    }

    return_list <- deal_with_empty(
      return_list,
      args[["properties"]],
      service,
      isTRUE(args[["skipGeometry"]]),
      args[["convertType"]],
      no_paging
    )

    return_list <- rejigger_cols(return_list, args[["properties"]], output_id)

    if (args[["convertType"]]) {
      return_list <- cleanup_cols(return_list, service)
      return_list <- order_results(return_list)

      # Mostly drop the id column except ts-meta, monitoring location:
      if (
        !service %in%
          c(
            "monitoring-locations",
            "time-series-metadata",
            "field-measurements-metadata",
            "combined-metadata",
            "parameter-codes"
          )
      ) {
        return_list <- return_list[, names(return_list)[
          names(return_list) != output_id
        ]]
      }
      # Move other id columns:
      return_list <- move_id_col(return_list, output_id)
    }

    if (args[["attach_request"]]) {
      attr(return_list, "request") <- req
    }
  }

  if (
    !isTRUE(args[["skipGeometry"]]) &
      "geometry" %in% names(return_list)
  ) {
    if (
      all(sf::st_is_empty(return_list[["geometry"]])) &
        !"geometry" %in% args[["properties"]]
    ) {
      return_list <- sf::st_drop_geometry(return_list)
    } else {
      return_list <- sf::st_as_sf(return_list)
    }
  }

  attr(return_list, "queryTime") <- Sys.time()
  return(return_list)
}

order_results <- function(df) {
  if (all(c("time", "monitoring_location_id") %in% names(df))) {
    df <- df[order(df$time, df$monitoring_location_id), ]
  } else if ("time" %in% names(df)) {
    df <- df[order(df$time), ]
  }

  return(df)
}

move_id_col <- function(df, output_id) {
  if ("time_series_id" %in% names(df)) {
    df <- df[, c(names(df)[names(df) != "time_series_id"], "time_series_id")]
  }

  if ("field_series_id" %in% names(df)) {
    df <- df[, c(names(df)[names(df) != "field_series_id"], "field_series_id")]
  }

  return(df)
}

#' Switch properties id
#'
#' If a user asks for either "id" or "output_id", it is only included in the
#' properties if that's the only column requested. "id" will always come back,
#' so it is not needed in the properties call.
#'
#' @noRd
#' @return list
#' @examples
#'
#' properties <- c("id", "state_name", "country_name")
#' dataRetrieval:::switch_properties_id(properties,
#'                               id = "monitoring_location_id")
#'
#' properties2 <- c("monitoring_location_id", "state_name", "country_name")
#' dataRetrieval:::switch_properties_id(properties2,
#'                               id = "monitoring_location_id")
#'
#' properties3 <- c("monitoring_locations_id", "state_name", "country_name")
#' dataRetrieval:::switch_properties_id(properties3,
#'                               id = "monitoring_location_id")
#'
#' properties4 <- c("monitoring_location_id")
#' dataRetrieval:::switch_properties_id(properties4,
#'                               id = "monitoring_location_id")
#'
#' properties5 <- c("monitoring_location_id", "geometry")
#' dataRetrieval:::switch_properties_id(properties5,
#'                               id = "monitoring_location_id")
#'
switch_properties_id <- function(properties, id) {
  orig_properties <- properties
  if (!all(is.na(properties))) {
    if ("id" %in% properties) {
      properties <- properties[properties != "id"]
    } else if (id %in% properties) {
      properties <- properties[properties != id]
    }

    if ("geometry" %in% properties) {
      properties <- properties[properties != "geometry"]
    }

    if (length(properties) == 0) {
      # If a user requested only id and/or geometry, properties would now be empty
      # geometry is taken care of with skipGeometry
      properties <- "id"
    }
  }

  return(properties)
}

#' Check non-API arguments
#'
#' Function to check types and create parameter descriptions.
#'
#' @param convertType logical, defaults to `r getOption("dataRetrieval.convertType")`.
#' If `TRUE`, the function will convert the data to dates, any qualifiers to string
#' vector and reorder the returned data frame.
#' @param no_paging logical, defaults to `r getOption("dataRetrieval.no_paging")`.
#' If `TRUE`, the data will
#' be requested from a native csv format. This can be dangerous because the
#' data will cut off at 50,000 rows without indication that more data
#' is available. Use `TRUE` with caution.
#' @param limit numeric, The optional limit parameter is used to control the subset of the
#' selected features that should be returned in each page. The maximum allowable
#' limit is 50,000. It may be beneficial to set this number lower if your internet
#' connection is spotty. The default (`NA`) will set the limit to the maximum
#' allowable limit for the service.
#' @param attach_request logical, defaults to `r getOption("dataRetrieval.attach_request")`.
#' If set to `TRUE`, the full request sent to the Water Data API is attached
#' as an attribute to the data set.
#' @param chunk_size Number of monitoring_location_ids to chunk requests into.
#' The default for functions that don't generally return long-term data records
#' is `r getOption("dataRetrieval.site_chunk_size_meta")`, while
#' the default for time series functions is
#' `r getOption("dataRetrieval.site_chunk_size_data")`.
#' Setting to `NA` will eliminate site chunking, giving users full control.
#' @param \dots Not used. Included to help differentiate official Water Data API arguments
#' from more seldom used, optional dataRetrieval-specific arguments.
#' @keywords internal
check_arguments_non_api <- function(
  convertType,
  no_paging,
  limit,
  attach_request,
  chunk_size,
  ...
) {
  if (!is.null(convertType)) {
    if (!is.na(convertType) & !is.logical(convertType)) {
      stop("convertType should be a logical TRUE/FALSE")
    }
  }

  if (!is.null(no_paging)) {
    if (!is.na(no_paging) & !is.logical(no_paging)) {
      stop("no_paging should be a logical TRUE/FALSE")
    }
  }

  if (!is.null(attach_request)) {
    if (!is.na(attach_request) & !is.logical(attach_request)) {
      stop("attach_request should be a logical TRUE/FALSE")
    }
  }

  if (!is.null(limit)) {
    if (!is.na(limit) & !is.numeric(limit)) {
      stop("limit should be an integer")
    }
  }

  if (!is.null(chunk_size)) {
    if (!is.na(chunk_size) & !is.numeric(chunk_size)) {
      stop("chunk_size should be an integer")
    }
  }
}

#' Check other arguments
#'
#' Additional functions to check types and create parameter descriptions.
#'
#' @param bbox Only features that have a geometry that intersects the bounding
#' box are selected.The bounding box is provided as four or six numbers, depending
#' on whether the coordinate reference system includes a vertical axis (height or
#' depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
#' vector structured: c(xmin,ymin,xmax,ymax).
#' Another way to think of it is c(Western-most longitude,
#' Southern-most latitude, Eastern-most longitude, Northern-most longitude).
#' @param skipGeometry This parameter can be used to skip response geometries for
#' each feature. The returning object will be a data frame with no spatial
#' information. The default `NA` will not specify the argument in the request.
#'
#' @keywords internal
check_arguments_api <- function(bbox, skipGeometry) {
  if (!is.null(skipGeometry)) {
    if (!is.na(skipGeometry) & !is.logical(skipGeometry)) {
      stop("skipGeometry should be a logical TRUE/FALSE")
    }
  }

  if (!is.null(bbox)) {
    if (!all(is.na(bbox))) {
      if (!length(bbox) %in% c(1, 4)) {
        stop("bbox is not set up correctly")
      }
    }
  }
}

#' Get USGS daily data statistics
#'
#' @description
#'
#' This service provides endpoints for access to computations on the historical
#' record regarding water conditions. For more information regarding the
#' calculation of statistics and other details, please visit the Statistics
#' documentation page.
#'
#' Note: The /statistics API is under active beta development and subject to change.
#' Improved handling of significant figures will be addressed in a future
#' release.
#'
#' \code{read_waterdata_stats_por} Returns day-of-year and month-of-year
#' statistics matching your query.
#'
#' \code{read_waterdata_stats_daterange} Returns monthly and annual statistics matching
#' your query.
#'
#' @export
#'
#' @param approval_status Whether to include approved and/or provisional
#'   observations. At this time, only approved observations are returned.
#' @param computation_type Desired statistical computation method. Available
#'   values: "arithmetic_mean", "maximum", "median", "minimum", "percentile".
#' @param country_code Country Query Parameter. Accepts multiple values (see
#'   examples). If one of country, county, or state code is supplied then the
#'   other two arguments do not need to be specified.
#' @param state_code State Query Parameter. Accepts multiple values in a
#'   character vector.
#' @param county_code County Query Parameter. Accepts multiple values in a
#'   character vector.
#' @param start_date Start Date Query Parameter. The logic is inclusive i.e., it
#'   will also return records that match the date. If an end date is supplied,
#'   but no start date is supplied, then statistics will be supplied for the
#'   entire period of record ending with the end date. If an end date is not
#'   supplied, but a start date is supplied then statistics will be supplied for
#'   the period of record following the start date. If no start or end date are
#'   supplied then statistics will be supplied for the entire period of record.
#' @param end_date End Date Query Parameter. The logic is inclusive i.e., it will
#'   also return records that match the date.
#' @param normal_type Normal Type Query Parameter. If unspecified, all matching data
#'   will be returned. Otherwise, it will filter the results to one of the following
#'   normals: day-of-year, month-of-year. Available values: "DOY", "MOY".
#' @param interval_type Interval Type Query Parameter. If unspecified, all matching
#'   data will be returned. Otherwise, it will filter the results to one or more of
#'   the following intervals: month, calendar year, water year.
#'   Available values: "M", "CY", "WY".
#' @param monitoring_location_id Each monitoring location has been assigned a
#'   unique station number that places them in downstream order. Accepts
#'   multiple values in a character vector.
#' @param parent_time_series_id The parent_time_series_id returns statistics
#'   tied to a particular database entry. Accepts multiple values in a character
#'   vector. If no parent time series identifier is supplied, then all records
#'   matching the rest of the provided criteria will be returned.
#' @param site_type_code Site Type Code Query Parameter. Accepts multiple values
#'   in a character vector. If no Site Type code is specified, statistics of all
#'   site types with the matching Monitoring Location Identifier will be
#'   returned.
#' @param site_type_name Site Type Name Query Parameter. If no Site Type name is
#'   specified, statistics of all site types with the matching Monitoring
#'   Location Identifier will be returned.
#' @param parameter_code USGS Parameter Code Query Parameter. Accepts multiple
#'   values in a character vector. If no USGS parameter code is specified, but a
#'   Monitoring Location Identifier is supplied, then all statistics and their
#'   parameter codes with a matching monitoring location identifier will be
#'   returned. All statistics within the period of record will be returned if no
#'   parameter code or monitoring location identifier are specified.
#' @param page_size Return a defined number of results (default: 1000).
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' # All day-of-year and month-of-year statistics for two sites
#' x1 <- read_waterdata_stats_por(
#'   monitoring_location_id = c("USGS-02319394", "USGS-02171500")
#' )
#'
#' # Request only month-of-year statistics using normal_type arg
#' x1 <- read_waterdata_stats_por(
#'   monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
#'   normal_type = "MOY"
#' )
#'
#' # Request temperature percentiles for specific month-day range
#' # Returns:
#' # - Day-of-year temperature percentiles for each day between June 1 through June 15.
#' # - Month-of-year percentiles for June, computed using
#' #    all June data (not just June 1 through June 15).
#' # Note: the month-of-year percentiles are returned only when the month-day range includes
#' #  the beginning of the month (e.g., "06-01")
#' x2 <- read_waterdata_stats_por(
#'   monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
#'   parameter_code = "00010",
#'   start_date = "06-01", end_date = "06-15",
#'   computation_type = "percentile"
#' )
#'
#' # All calendar month, calendar year, and water year statistics for two sites
#' x3 <- read_waterdata_stats_daterange(
#'   monitoring_location_id = c("USGS-02319394", "USGS-02171500")
#' )
#'
#' # Request only calendar year statistics
#' x3 <- read_waterdata_stats_daterange(
#'   monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
#'   interval_type = "CY"
#' )
#'
#' # Request specific gage height and discharge summaries for a limited date range
#' # Returns:
#' # - calendar month summaries for each month between January, 2010 through December, 2011
#' # - calendar year summaries for 2010 and 2011
#' # - water year summaries for WY2010, WY2011, and WY2012
#' x4 <- read_waterdata_stats_daterange(
#'   monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
#'   parameter_code = c("00065", "00060"),
#'   start_date = "2010-01-01", end_date = "2011-12-31",
#'   computation_type = c("minimum", "median", "maximum")
#' )
#'
#' }
#'
#' @rdname read_waterdata_stats
#' @seealso \url{https://api.waterdata.usgs.gov/statistics/v0/docs}
read_waterdata_stats_por <- function(
  approval_status = NA,
  computation_type = NA_character_,
  country_code = NA_character_,
  state_code = NA_character_,
  county_code = NA_character_,
  start_date = NA_character_,
  end_date = NA_character_,
  normal_type = NA_character_,
  monitoring_location_id = NA_character_,
  parent_time_series_id = NA_character_,
  site_type_code = NA_character_,
  site_type_name = NA_character_,
  parameter_code = NA_character_,
  page_size = NA
) {
  args <- mget(names(formals()))

  get_statistics_data(args = args, service = "Normals")
}

#' @export
#' @rdname read_waterdata_stats
read_waterdata_stats_daterange <- function(
  approval_status = NA,
  computation_type = NA_character_,
  country_code = NA_character_,
  state_code = NA_character_,
  county_code = NA_character_,
  start_date = NA_character_,
  end_date = NA_character_,
  interval_type = NA_character_,
  monitoring_location_id = NA_character_,
  parent_time_series_id = NA_character_,
  site_type_code = NA_character_,
  site_type_name = NA_character_,
  parameter_code = NA_character_,
  page_size = NA
) {
  args <- mget(names(formals()))

  get_statistics_data(args = args, service = "Intervals")
}

#' Create a request object for the /statistics service
#'
#' @param service chr; "Normals" or "Intervals"
#'
#' @return a httr2 request object
#'
#' @noRd
construct_statistics_request <- function(service = "Normals") {
  base_request <- httr2::request(
    "https://api.waterdata.usgs.gov/statistics/"
  ) |>
    httr2::req_user_agent(default_ua()) |>
    httr2::req_headers(`Accept-Encoding` = "gzip") |>
    httr2::req_timeout(seconds = 180) |>
    httr2::req_url_path_append(getOption("dataRetrieval.api_version_stat")) |>
    httr2::req_url_path_append(paste0("observation", service))

  base_request <- add_api_token(base_request)

  return(base_request)
}

#' Retrieve data from /statistics API
#'
#' @param args arguments from individual functions.
#' @param service Endpoint name.
#'
#' @noRd
#' @return data.frame with attributes
get_statistics_data <- function(args, service) {
  rid <- data <- ts_id <- values <- NULL

  base_request <- construct_statistics_request(service = service)

  full_request <- explode_query(base_request, POST = FALSE, x = args)

  message("Requesting:\n", full_request$url)

  return_list <- data.table::as.data.table(walk_pages(full_request))

  if (nrow(return_list) == 0) {
    return(deal_with_empty_stats(return_list))
  }

  return_list[, rid := .I]

  parsed_data <- lapply(return_list$data, jsonlite::parse_json)
  return_list[, data := NULL]

  combined_list <- list()
  for (i in seq_along(parsed_data)) {
    # 1. One row per time series
    ts_meta <- data.table::rbindlist(parsed_data[[i]], fill = TRUE)
    ts_meta[, ts_id := .I] # local key

    # 2. One row per observation, keyed to ts_id
    obs <- data.table::rbindlist(ts_meta$values, fill = TRUE, idcol = "ts_id")
    # Some stats responses contain 0 values, which need to be treated differently
    if (nrow(obs) > 0) {
      obs <- cleanup_cols_stats(obs)
    } else {
      obs <- data.table::data.table(ts_id = unique(ts_meta$ts_id))
    }

    # 3. Drop nested list column
    ts_meta[, values := NULL]

    # 4. Join
    out <- ts_meta[obs, on = "ts_id", allow.cartesian = TRUE]

    out[, `:=`(
      ts_id = NULL,
      rid = i
    )]

    combined_list[[i]] <- out
  }

  combined <- data.table::rbindlist(
    combined_list,
    use.names = TRUE,
    fill = TRUE # needed if any response has 0 values
  )
  combined <- combined[return_list, on = "rid"]
  combined[, rid := NULL]

  combined <- sf::st_as_sf(as.data.frame(combined))

  attr(combined, "request") <- full_request
  attr(combined, "queryTime") <- Sys.time()

  return(combined)
}

#' Clean up "value", "values", and "percentiles" columns in a data.table
#'
#' @description If the input data.table has "values" and "percentiles" columns,
#'   then it unnests both, making the data.table longer (one row per value)
#'
#'   If the input data.table *also* has a "value" column, then it consolidates
#'   with "values", yielding a single "value" column
#'
#'   The "percentiles" column is replaced by "percentile" (matching
#'   singularization of "value"). The function also checks whether the
#'   "computation" column contains "minimum", "median", or "maximum" and sets
#'   the corresponding "percentile" to 0, 50, or 100, respectively. Note that
#'   the percentile column might only contain NAs if the input data.table only
#'   includes arithmetic_mean values.
#'
#'   Lastly, the value and percentile columns are converted from character to
#'   numeric
#'
#' @note This function is intended to evaluate as a data.table j expression,
#'   which is why it doesn't accept any arguments. j expressions are typically
#'   written inline with the data.table definition (e.g., \code{DT[, { do
#'   something }]}). However, this expression would have been excessively long.
#'   Instead, a substitute() is used so this could be pulled out into a separate
#'   function for readability, but we still gain the computational benefits of
#'   using an expression.
#'
#' @noRd
#' @return data.table object with "value" and "percentile" columns
#'
#' @seealso
#' \url{https://stat.ethz.ch/CRAN/web/packages/data.table/vignettes/datatable-programming.html}
cleanup_cols_stats <- function(df) {
  percentiles <- percentile <- values <- computation <- value <- percent <- NULL
  ## ---- detect column presence ----
  has_value <- "value" %in% names(df)
  has_values <- "values" %in% names(df)
  has_percentiles <- "percentiles" %in% names(df)

  if (has_value) {
    df[, value := as.numeric(value)]
  } else {
    df[, value := NA_real_]
  }

  if (has_percentiles) {
    df[, percent := NA_integer_]
    df[is.na(value), percent := as.integer(percentiles)][, percentiles := NULL]
    data.table::setnames(df, "percent", "percentile")
  } else {
    df[, percentile := NA_integer_]
  }

  # value has already been created, so this will only move values if needed
  if (has_values) {
    df[is.na(value), value := as.numeric(values)][, values := NULL]
  }

  df$value[is.nan(df$value)] <- NA_real_

  ## ---- percentiles ----
  if (
    any(data.table::`%chin%`(df$computation, c("minimum", "median", "maximum")))
  ) {
    df[,
      percentile := data.table::fcase(
        computation == "minimum" ,   0L ,
        computation == "median"  ,  50L ,
        computation == "maximum" , 100L ,
        default = percentile
      )
    ]
  }

  return(df)
}

#' Handle empty responses from the /statistics service
#'
#' @param return_list data.frame returned from walk_pages
#' @param properties character vector of requested columns (or NA)
#' @param convertType logical, whether to convert numeric columns
#'
#' @return data.frame or sf object with expected columns
#' @noRd
deal_with_empty_stats <- function(
  return_list,
  properties = NA,
  convertType = TRUE
) {
  # Define default columns for stats service
  default_columns <- c(
    "monitoring_location_id",
    "monitoring_location_name",
    "site_type",
    "site_type_code",
    "country_code",
    "state_code",
    "county_code",
    "parameter_code",
    "unit_of_measure",
    "parent_time_series_id",
    "parent_statistics_id",
    "parent_statistics_name",
    "value",
    "percentile",
    "start_date",
    "end_date",
    "sample_count",
    "approval_status",
    "computation_id",
    "computation"
  )

  if (all(is.na(properties))) {
    properties <- default_columns
  }

  # create empty data.frame
  return_list <- as.data.frame(matrix(nrow = 0, ncol = length(properties)))
  names(return_list) <- properties
  return_list <- lapply(return_list, as.character)
  return_list <- as.data.frame(return_list)

  # convert numeric columns if requested
  if (convertType) {
    numeric_cols <- c("value", "percentile", "sample_count")
    for (col in numeric_cols) {
      if (col %in% names(return_list)) {
        return_list[[col]] <- as.numeric(return_list[[col]])
      }
    }
  }

  # ensure geometry column exists if not skipped
  return_list <- sf::st_as_sf(
    return_list,
    geometry = sf::st_sfc(),
    crs = sf::st_crs("EPSG:4326")
  )

  return(return_list)
}

# ------------------------------------------------------------------------------
# cleanup_cols_stats()
# ------------------------------------------------------------------------------

test_that("values column is unnested and converted to numeric", {
  DT <- data.table(
    values = c("1.2", "3.4"),
    percentiles = c("10", "90"),
    computation = "percentile",
    start_date = "2000-01-01",
    end_date = "2000-12-31",
    interval_type = "month",
    ts_id = "1",
    sample_count = 10,
    approval_status = "approved",
    computation_id = "abc"
  )

  out <- cleanup_cols_stats(DT)

  expect_equal(out$value, c(1.2, 3.4))
  expect_equal(out$percentile, c(10, 90))
})

test_that("value column is used when values is absent", {
  DT <- data.table(
    value = "2.5",
    computation = "arithmetic_mean",
    start_date = "2000-01-01",
    end_date = "2000-12-31",
    interval_type = "year",
    sample_count = 1,
    ts_id = "1",
    approval_status = "approved",
    computation_id = "xyz"
  )

  out <- cleanup_cols_stats(DT)

  expect_equal(out$value, 2.5)
  expect_true(is.na(out$percentile))
})

test_that("'nan' values are converted to NA_real_", {
  DT <- data.table(
    values = "nan",
    computation = "arithmetic_mean",
    start_date = "2000-01-01",
    end_date = "2000-12-31",
    interval_type = "month",
    ts_id = "1",
    sample_count = 1,
    approval_status = "approved",
    computation_id = "nan-test"
  )

  out <- cleanup_cols_stats(DT)

  expect_true(is.na(out$value))
})

test_that("percentile is inferred from computation type", {
  DT <- data.table(
    value = "5",
    computation = c("minimum", "median", "maximum"),
    start_date = "2000-01-01",
    end_date = "2000-12-31",
    interval_type = "year",
    ts_id = "1",
    sample_count = 1,
    approval_status = "approved",
    computation_id = 1:3
  )

  out <- cleanup_cols_stats(DT)

  expect_equal(
    out[order(computation)]$percentile,
    c(100, 50, 0) # maximum, median, minimum
  )
})

test_that("scalar percentile is recycled to match values length", {
  DT <- data.table(
    values = c("1", "2", "3"),
    percentiles = "50",
    computation = "percentile",
    start_date = "2000-01-01",
    end_date = "2000-12-31",
    interval_type = "month",
    ts_id = "1",
    sample_count = 3,
    approval_status = "approved",
    computation_id = "recycle"
  )

  out <- cleanup_cols_stats(DT)

  expect_equal(out$percentile, c(50, 50, 50))
})

test_that("missing value and values yields NA value", {
  DT <- data.table(
    computation = "arithmetic_mean",
    start_date = "2000-01-01",
    end_date = "2000-12-31",
    interval_type = "year",
    ts_id = "1",
    sample_count = 1,
    approval_status = "approved",
    computation_id = "missing"
  )

  out <- cleanup_cols_stats(DT)

  expect_true(is.na(out$value))
})

# ------------------------------------------------------------------------------
# construct_statistics_request()
# ------------------------------------------------------------------------------

test_that("construct_statistics_request builds correct URL path", {
  req <- construct_statistics_request(service = "Normals")

  expect_true(grepl("statistics/v0/observationNormals", req$url))
})

test_that("Intervals service path is correct", {
  req <- construct_statistics_request(service = "Intervals")

  expect_true(grepl("statistics/v0/observationIntervals", req$url))
})

# ------------------------------------------------------------------------------
# explode_query integration (no HTTP)
# ------------------------------------------------------------------------------

test_that("arguments are passed as query parameters", {
  args <- list(
    state_code = "WI",
    computation_type = "median",
    page_size = 10
  )

  req <- construct_statistics_request("Normals")
  full <- explode_query(req, POST = FALSE, x = args)

  qs <- httr2::req_get_url(full)

  expect_match(qs, "state_code=WI")
  expect_match(qs, "computation_type=median")
  expect_match(qs, "page_size=10")
})

# ------------------------------------------------------------------------------
# get_statistics_data() (mocked)
# ------------------------------------------------------------------------------

test_that("get_statistics_data returns sf object with attributes", {
  fake_page <-
    sf::st_sf(
      monitoring_location_id = c("USGS-01432160", "USGS-01432110"),
      monitoring_location_name = c(
        "DELAWARE RIVER AT BARRYVILLE NY",
        "Lackawaxen River at Rowland, PA"
      ),
      site_type = c("Stream", "Stream"),
      site_type_code = c("ST", "ST"),
      country_code = c("US", "US"),
      state_code = c("42", "42"),
      county_code = c("103", "103"),
      geometry = sf::st_sfc(
        sf::st_point(c(0, 0)),
        sf::st_point(c(1, 1))
      ),
      data = c(
        # First site
        jsonlite::toJSON(
          list(
            list(
              parameter_code = "00060",
              unit_of_measure = "ft^3/s",
              parent_time_series_id = "1692e9a29c8c4276add4497c5da872fa",
              values = list(
                list(
                  start_date = "2018-10-01",
                  end_date = "2018-10-31",
                  interval_type = "month",
                  value = "8032.903",
                  sample_count = 31,
                  approval_status = "approved",
                  computation_id = "d98ebe80-d476-4144-b7ce-c6ca32f015de",
                  computation = "arithmetic_mean"
                ),
                list(
                  start_date = "2018-11-01",
                  end_date = "2018-11-30",
                  interval_type = "month",
                  value = "12521.667",
                  sample_count = 30,
                  approval_status = "approved",
                  computation_id = "93f45097-93ad-4b48-83d4-b27b5fc137dc",
                  computation = "arithmetic_mean"
                )
              )
            )
          ),
          auto_unbox = TRUE
        ),
        # Second site
        jsonlite::toJSON(
          list(
            list(
              parameter_code = "00060",
              unit_of_measure = "ft^3/s",
              parent_time_series_id = "7dda388ea270420dbe7324e56b6f907f",
              values = list(
                list(
                  start_date = "2007-07-27",
                  end_date = "2007-07-31",
                  interval_type = "month",
                  values = c("1", "2"),
                  percentiles = c("50", "100"),
                  sample_count = 5,
                  approval_status = "approved",
                  computation_id = "4579653d-f26f-4e9f-99ea-c3fa3fc68686",
                  computation = "percentile"
                )
              )
            )
          ),
          auto_unbox = TRUE
        )
      ),
      stringsAsFactors = FALSE
    )

  with_mocked_bindings(
    walk_pages = function(...) list(fake_page),
    {
      out <- get_statistics_data(list(), "Normals")

      expect_s3_class(out, "sf")
      expect_true("request" %in% names(attributes(out)))
      expect_true("queryTime" %in% names(attributes(out)))

      expect_equal(nrow(out), 4)
      expect_equal(out$value, c(8032.903, 12521.667, 1, 2))
      expect_equal(out$percentile, c(NA, NA, 50, 100))
      expect_equal(out$sample_count, c(31, 30, 5, 5))
    }
  )
})

test_that("get_statistics_data handles empty response", {
  with_mocked_bindings(
    walk_pages = function(...) list(),
    {
      out <- get_statistics_data(list(), "Normals")

      expect_s3_class(out, "sf")
      expect_equal(nrow(out), 0)
    }
  )
})

# ------------------------------------------------------------------------------
# High-level API tests (skipped on CRAN)
# ------------------------------------------------------------------------------

test_that("read_waterdata_stats_por returns data", {
  skip_on_cran()
  skip_if_offline()

  out <- read_waterdata_stats_por(
    monitoring_location_id = "USGS-01646500",
    parameter_code = "00060",
    computation_type = "median",
    page_size = 5
  )

  expect_s3_class(out, "sf")
  expect_true(nrow(out) > 0)
})

test_that("normal_type arg works in read_waterdata_stats_por", {
  skip_on_cran()
  skip_if_offline()

  out <- read_waterdata_stats_por(
    monitoring_location_id = "USGS-01646500",
    parameter_code = "00060",
    computation_type = "median",
    page_size = 5,
    normal_type = "MOY"
  )

  expect_s3_class(out, "sf")
  expect_true(nrow(out) > 0)
  # time_of_year should have 12 months of data
  expect_true(length(out$time_of_year) == 12)
})

test_that("read_waterdata_stats_daterange returns data", {
  skip_on_cran()
  skip_if_offline()

  out <- read_waterdata_stats_daterange(
    monitoring_location_id = "USGS-01646500",
    parameter_code = "00060",
    computation_type = "maximum",
    page_size = 5
  )

  expect_s3_class(out, "sf")
  expect_true(nrow(out) > 0)

  # setting interval_type arg returns fewer rows than unset
  out1 <- read_waterdata_stats_daterange(
    monitoring_location_id = "USGS-01646500",
    parameter_code = "00060",
    computation_type = "maximum",
    interval_type = "CY",
    page_size = 5
  )

  expect_true(nrow(out1) < nrow(out))
})

#' @title Trim and Cull NULLs
#' @description Remove NULL arguments from a named list
#' @param x a list
#' @keywords nldi internal
#' @return a list
#' @noRd

tc <- function(x) {
  Filter(Negate(is.null), x)
}

#' @title Find Good Names
#' @description minimize names of returned features
#' @param input data returned from NLDI
#' @param type type of return
#' @keywords nldi internal
#' @return a list
#' @noRd

find_good_names <- function(input, type) {
  # The features names are different across features, navigation, and basin returns
  # This sets the environment for what to expect
  if (type == "nav") {
    grep("comid", names(input), value = TRUE)
  } else if (type == "feature") {
    c(
      "sourceName",
      "identifier",
      "comid",
      names(input)[
        names(input) %in% c("name", "reachcode", "measure", "mainstem")
      ]
    )
  } else {
    NULL
  }
}

#' @title Get current NLDI offerings
#' @description Used to query the current resources available through the NLDI
#' @param url URL for NLDI sources. Default is supplied by package environment.
#' @return data.frame
#' @export
#' @keywords nldi
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' get_nldi_sources()
#' }
get_nldi_sources <- function(url = getOption("dataRetrieval.nldi_base")) {
  res <- httr2::request(url)
  res <- httr2::req_user_agent(res, default_ua())
  res <- httr2::req_error(res, is_error = \(x) FALSE)
  res <- httr2::req_perform(res)

  if (res$status_code == 200) {
    jsonlite::fromJSON(httr2::resp_body_string(res), simplifyDataFrame = TRUE)
  } else {
    message("Error in: ", url)
  }
}

#' @title Query NLDI
#' @description Queries the NLDI for a given URL. If local sf install is
#' available the function returns a sf data.frame.
#' If the object requested is a POINT object, the XY coordinates are added
#' as columns. Otherwise the columns returned are "sourceName" and
#' "identifier" for features, and "nhdplus_comid".
#' @param url the URL to retrieve
#' @param type the type of data being returned (nav or feature)
#' @param use_sf should a local sf install be usedto parse data
#' @keywords nldi internal
#' @noRd
#' @return a data.frame
get_nldi <- function(url, type = "", use_sf = FALSE, warn = TRUE) {
  # Query
  res <- httr2::request(url)
  res <- httr2::req_user_agent(res, default_ua())
  res <- httr2::req_error(res, is_error = \(x) FALSE)
  res <- httr2::req_perform(res)

  if (!is.null(res$headers$`X-Ratelimit-Remaining`)) {
    if (
      (as.numeric(res$headers$`X-Ratelimit-Limit`) -
        as.numeric(res$headers$`X-Ratelimit-Remaining`)) /
        as.numeric(res$headers$`X-Ratelimit-Limit`) >
        0.9
    ) {
      message(
        "Approaching NLDI rate limit. ",
        res$headers$`X-Ratelimit-Remaining`,
        " of ",
        res$headers$`X-Ratelimit-Limit`,
        " remaining"
      )
      if (as.numeric(res$headers$`X-Ratelimit-Remaining`) < 20) {
        message("Sleeping to try to avoid going over rate limit.")
        Sys.sleep(3600 / ((as.numeric(res$headers$`X-Ratelimit-Limit`) / 3)))
      }
    }
  }

  # If successful ...
  if (res$status_code == 200) {
    # Interpret as text
    if (length(res$body) > 0) {
      d <- httr2::resp_body_string(res)
    } else {
      if (warn) {
        warning("No data returned for: ", url, call. = FALSE)
      }

      return(NULL)
    }

    if (use_sf) {
      # Parse with sf

      tmp <- tryCatch(
        {
          sf::read_sf(d)
        },
        error = function(e) {
          message("No data found for: ", basename(url))
          return(NULL)
        },
        warning = function(w) {
          message("No data found for: ", basename(url))
          return(NULL)
        }
      )

      if (nrow(tmp) == 0) {
        if (warn) {
          warning("No data returned for: ", url, call. = FALSE)
        }
        return(NULL)
      }

      good_name <- find_good_names(tmp, type)

      if (!is.null(tmp)) {
        # if of type POINT at the X,Y coordinates as columns
        if (sf::st_geometry_type(tmp)[1] == "POINT") {
          tmp$X <- sf::st_coordinates(tmp)[, 1]
          tmp$Y <- sf::st_coordinates(tmp)[, 2]

          tmp <- tmp[, c(good_name, "X", "Y")]
        } else {
          # If line/polygon then keep geometry but don't expand
          tmp <- tmp[, c(good_name, attr(tmp, "sf_column"))]
        }
      }

      # Returning as data.frame drops the geometry column ...
      return(tmp)
    } else {
      # Parse as simplified JSON
      d <- jsonlite::fromJSON(d, simplifyDataFrame = TRUE)

      input <- d$features$properties

      good_name <- find_good_names(input, type)

      if (is.null(input) && type != "basin") {
        if (warn) {
          warning("No data returned for: ", url, call. = FALSE)
        }
        tmp <- NULL
      } else {
        # if of type POINT at the X,Y coordinates as columns
        if (d$features$geometry$type[1] == "Point") {
          geom <- d$features$geometry$coordinates

          tmp <- cbind(input[, good_name], do.call(rbind, geom))

          names(tmp) <- c(good_name, "X", "Y")

          return(tmp)
        } else {
          # If line/polygon then keep geometry but don't expand
          return(input[, good_name])
        }
      }
    }
  } else {
    message("Error in: ", url)
  }
}

#' @title NLDI Validity Check
#' @description tests if NLDI feature is available. Is vectorized and works with partial string matching.
#' @param all a data.frame of available features (see get_nldi_sources)
#' @param type type(s) to check (character)
#' @return a list with good and bad entries
#' @keywords nldi internal
#' @noRd
valid_ask <- function(all, type) {
  # those where the requested pattern is included in a nldi_source ...
  # means we will catch nwis - not just nwissite ...
  # means we will catch both wqp and WQP ...

  ### WOW! This is hacky and will hopefully be unneeded latter on....
  type <- ifelse(type == "nwis", "nwissite", type)
  all <- rbind(
    all,
    data.frame(
      source = "flowlines",
      sourceName = "NHDPlus comid",
      features = NA
    )
  )

  good <- grepl(
    paste0(tolower(type), collapse = "|"),
    tolower(c(all$source))
  )

  bad <- grepl(
    paste0(tolower(all$source), collapse = "|"),
    tolower(c(all$source))
  )

  list(good = all[good, ], bad = type[!bad])
}

#' @title R Client for the Network Linked Data Index
#' @description Provides a formal client to the USGS
#' Network Linked Data Index.
#' @details The function is useful for topology and location based
#' feature discovery. A user must specify an origin feature, optional navigation
#' direction(s) along the network, as well as features to identify along the
#' navigated paths. Valid starting options can be given by one of the following
#' arguments: comid, nwis, huc12, wqp, location, and start.
#' @param comid numeric or character. An NHDPlusV2 COMID
#' @param nwis  numeric or character. A USGS NWIS surface water siteID
#' @param wqp   numeric or character. A water quality point ID
#' @param huc12 numeric or character. A WBD HUC12 unit ID
#' @param location numeric vector. Coordinate pair in WGS84
#' SRS ordered lng/lat (X,Y)
#' @param origin named list. Specifying a feature type and ID
#' (e.g. list("comid" = 101))
#' @param nav character vector. where to navigate from the starting point.
#' Options include along the upper mainsteam (UM), upstream tributary (UT),
#' downstream mainstem (DM) and downstream divergences (DD). You may select
#' one or more of the abbreviations ("UM", "UT", DM", "DD").
#' @param find character vector. Define what resources to find along the
#' navigation path(s) (see get_nldi_sources()$source). Can also include 'basin'
#' or 'flowline', which will return the upstream basin of the starting feature
#' or flowlines along the navigation respectively. The default is "flowlines".
#' If you provide any other resource, AND want flowlines, then flowlines must
#' be explicitly requested.
#' @param distance_km numeric. Define how far to look along the navigation path in
#' kilometers (default = 100)
#' @param no_sf if available, should `sf` be used for parsing,
#' defaults to `TRUE` if `sf` is locally installed
#' @param warn (default TRUE) should warnings be printed
#' @return a list of data.frames if sf is not installed, a list of sf objects if it is
#' @export
#' @keywords nldi
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' # Find Features / Define origin features
#'
#' ## Find feature by COMID
#' findNLDI(comid = 101)
#'
#' ## Find feature by NWIS ID
#' findNLDI(nwis = "11120000")
#'
#'
#' ## Find feature by LOCATION
#' findNLDI(location = c(-115, 40))
#'
#' ## GENERAL ORIGIN: COMID
#' findNLDI(origin = list("comid" = 101))
#'
#' # Navigation (flowlines will be returned if find is unspecified)
#' # UPPER MAINSTEM of USGS-11120000
#' findNLDI(nwis = "11120000", nav = "UM")
#'
#' # MULTI-REQUEST
#' # UPPER MAINSTEM and TRIBUTARY of USGS-11120000
#' findNLDI(nwis = "11120000", nav = c("UT", "UM"))
#'
#' # Discover Features(flowlines will not be returned unless included in find)
#'
#' ## Find feature(s) on the upper tributary of USGS-11120000
#' findNLDI(nwis = "11120000", nav = "UT", find = c("nwis"))
#'
#' ## Find upstream basin boundary and  of USGS-11120000
#' findNLDI(nwis = "11120000", find = "basin")
#'
#' # Control Distance
#' ## Limit search to 50 km
#' findNLDI(comid = 101, nav = "DM", find = c("nwis", "flowlines"), distance_km = 50)
#' }

findNLDI <- function(
  comid = NULL,
  nwis = NULL,
  wqp = NULL,
  huc12 = NULL,
  location = NULL,
  origin = NULL,
  nav = NULL,
  find = c("flowlines"),
  distance_km = 100,
  no_sf = FALSE,
  warn = TRUE
) {
  # Should sf be used? Both no_sf and pkg.env must agree
  use_sf <- all(pkg.env$local_sf, !no_sf)

  # Should the basin be identified?
  getBasin <- ("basin" %in% find)

  # From the collection of possible origins, pick 1 and remove NULLS
  starter <- tc(c(
    list(
      comid = comid,
      nwis = nwis,
      wqp = wqp,
      huc12 = huc12,
      location = location
    ),
    origin
  ))

  # a single starting location must be given ...
  if (is.null(starter) || length(starter) > 1) {
    stop(
      "Define a single starting point. Use `find` to identify other resources."
    )
  }

  # Ensure nav types are valid
  bad_nav <- !nav %in% c("UM", "UT", "DD", "DM")

  if (any(bad_nav)) {
    stop(nav[bad_nav], " not a valid navigation. Chose from: UM, UT, DD, DM")
  }

  # name of starter
  start_type <- names(starter)

  # If location, ensure lng is first argument (hack for USA features)
  if (start_type == "location") {
    if (any(grepl("sfc$|sf$", class(location))) && use_sf) {
      if (sf::st_geometry_type(location) != "POINT") {
        stop("Only POINT objects can be passed to location")
      }

      location <- sf::st_coordinates(location)
    } else {
      if (location[1] > 0) {
        stop("Provide location in the form c(lng,lat)")
      }
    }

    # Must convert location to COMID for tracing and discovery ...
    tmp_url <- paste0(
      getOption("dataRetrieval.nldi_base"),
      "comid/position?f=json&coords=POINT%28",
      location[1],
      "%20",
      location[2],
      "%29"
    )

    tmp_return <- get_nldi(tmp_url, type = "feature", use_sf = use_sf)

    # Override starter with location based COMID
    starter <- list("comid" = tmp_return$identifier)
  }

  # Reset (if needed)
  start_type <- names(starter)

  if (
    is.null(pkg.env$current_nldi) ||
      !identical(
        pkg.env$current_nldi_base,
        getOption("dataRetrieval.nldi_base")
      )
  ) {
    pkg.env$current_nldi <- get_nldi_sources()
    pkg.env$current_nldi_base <- getOption("dataRetrieval.nldi_base")
  }

  # Defining the origin URL.
  #  Align request with formal name from sources
  #  If NWIS, add "USGS-" prefix
  start_url <- paste0(
    valid_ask(all = pkg.env$current_nldi, type = start_type)$good$features,
    "/",
    utils::URLencode(as.character(ifelse(
      start_type == "nwis" & !grepl("-", starter),
      paste0("USGS-", starter),
      starter
    ))),
    "/"
  )

  # Makes sure that all requested features to `find` are valid and name-aligned
  if (!is.null(find)) {
    find <- valid_ask(pkg.env$current_nldi, type = find)$good$source
  }

  # Set empty lists to store origin, navigation, features, and basin requests ...
  navigate <- features <- list()

  # Build navigation URLs
  for (i in seq_along(nav)) {
    navigate[[nav[i]]] <- paste0(start_url, "navigation/", nav[i])
  }

  # Build find URLs
  if (length(find) > 0) {
    features <- lapply(
      navigate,
      paste0,
      paste0("/", find),
      paste0("?f=json&distance=", distance_km)
    )
  }

  names <- unlist(lapply(nav, paste0, paste0("_", find)))

  search <- data.frame(
    url = unlist(c(
      start_url,
      if (getBasin) {
        paste0(start_url, "basin?f=json")
      },
      features
    )),
    type = c(
      "feature",
      if (getBasin) {
        "basin"
      },
      ifelse(
        rep(find, length(nav)) == "flowlines",
        "nav",
        "feature"
      )
    ),
    name = c(
      "origin",
      if (getBasin) {
        "basin"
      },
      names[!names %in% c("UM_", "UT_", "DD_", "DM_")]
    )
  )

  # Send NLDI queries ...
  shp <- lapply(seq_len(nrow(search)), function(x) {
    get_nldi(
      url = search$url[x],
      type = search$type[x],
      use_sf = use_sf,
      warn = warn
    )
  })

  # Set the names of the parsed URL list
  names(shp) <- search$name

  return(shp)
}

context("NLDI...")

test_that("NLDI messageing NULL", {
  skip_on_cran()
  skip_on_ci()
  xx <- findNLDI(
    wqp = "TCEQMAIN-10016",
    nav = "UM",
    find = "nwissite",
    distance_km = 2,
    warn = FALSE
  )

  expect_warning(findNLDI(
    wqp = "TCEQMAIN-10016",
    nav = "UM",
    find = "nwissite",
    distance_km = 2,
    warn = TRUE
  ))

  expect_true(class(xx) == "list")
  expect_true(nrow(xx[[1]]) == 1)
})


test_that("NLDI offerings...", {
  skip_on_cran()
  skip_on_ci()
  expect_true(nrow(get_nldi_sources()) > 1)
})


test_that("NLDI starting sources...", {
  skip_on_cran()
  skip_on_ci()

  # LINESTRING GEOMETERY
  expect_equal(
    sum(
      names(findNLDI(comid = 101, warn = FALSE)$origin) %in%
        c("sourceName", "identifier", "comid", "geometry")
    ),
    4
  )
  # POINT GEOMETERY
  expect_true(all(
    names(findNLDI(nwis = "11120000", warn = FALSE)$origin) %in%
      c(
        "sourceName",
        "identifier",
        "comid",
        "name",
        "mainstem",
        "reachcode",
        "measure",
        "X",
        "Y",
        "geometry"
      )
  ))
  # COMID
  expect_equal(
    findNLDI(comid = 101, warn = FALSE)$origin$sourceName,
    "NHDPlus comid"
  )
  # NWIS
  expect_equal(
    findNLDI(nwis = "11120000", warn = FALSE)$origin$sourceName,
    "NWIS Surface Water Sites"
  )
  # WQP
  expect_equal(
    findNLDI(wqp = "USGS-04024315", warn = FALSE)$origin$sourceName,
    "Water Quality Portal"
  )
  expect_equal(
    findNLDI(wqp = "CCU_EQL-2 H9", warn = FALSE)$origin$sourceName,
    "Water Quality Portal"
  )
  # LOCATION
  expect_equal(
    findNLDI(location = c(-115, 40), warn = FALSE)$origin$sourceName,
    "NHDPlus comid"
  )
  # ERROR: LOCATION COORDINATES FLIPPED
  expect_error(findNLDI(location = c(40, -115), warn = FALSE))
  # GENERAL START: STABLE
  expect_equal(
    findNLDI(origin = list("comid" = 101), warn = FALSE)$origin$sourceName,
    "NHDPlus comid"
  )
  # GENERAL START: NON-STABLE
  expect_equal(
    findNLDI(
      origin = list("nwissite" = "USGS-05427850"),
      warn = FALSE
    )$origin$sourceName,
    "NWIS Surface Water Sites"
  )
  # ERROR: TWO STARTS
  expect_error(findNLDI(nwis = 1000, comid = 101, warn = FALSE))
  # NON EXISTING SITE
  expect_message(findNLDI(comid = 1, warn = FALSE))
})

test_that("NLDI navigation sources...", {
  skip_on_cran()
  skip_on_ci()

  # UPPER TRIBUTARY
  expect_equal(
    length(findNLDI(nwis = "11120000", nav = "UT", warn = FALSE)$UT_flowlines),
    2
  )
  # UPPER MAIN
  expect_equal(
    length(findNLDI(nwis = "11120000", nav = "UM", warn = FALSE)$UM_flowlines),
    2
  )
  # DOWNSTREAM MAIN
  expect_equal(
    length(findNLDI(nwis = "11120000", nav = "DM", warn = FALSE)$DM_flowlines),
    2
  )
  # MULTI-REQUEST
  expect_equal(
    length(findNLDI(nwis = "11120000", nav = c("UT", "UM"), warn = FALSE)),
    3
  )
  # ERRORS: Bad NAV REQUEST
  expect_error(findNLDI(nwis = "11120000", nav = c("DT"), warn = FALSE))
  expect_error(findNLDI(nwis = "11120000", nav = c("DT", "UM"), warn = FALSE))
  # WARNING: Data not found
  # expect_warning(findNLDI(comid = 101, nav = "UM", find = "nwis", warn = TRUE))
})

test_that("NLDI find sources...", {
  skip_on_cran()
  skip_on_ci()

  expect_equal(
    length(findNLDI(
      nwis = "11120000",
      nav = "UT",
      distance_km = 999,
      find = "nwis",
      warn = FALSE
    )),
    2
  )
  # expect_equal(length(findNLDI(nwis = "11120000", nav = c("UT", "UM"), find = c("nwis", "wade", "flowlines"), warn = FALSE)), 7)
})

test_that("sf not installed...", {
  skip_on_cran()
  skip_on_ci()

  expect_true(
    !"geometry" %in%
      findNLDI(nwis = "11120000", no_sf = TRUE, warn = FALSE)[[1]]
  )
  expect_equal(
    class(findNLDI(
      nwis = "11120000",
      nav = "UT",
      find = c("nwis"),
      no_sf = TRUE,
      warn = FALSE
    )[[2]]),
    "data.frame"
  )
  expect_true(
    c("X") %in%
      names(findNLDI(
        nwis = "11120000",
        nav = "UT",
        find = c("nwis"),
        no_sf = TRUE,
        warn = FALSE
      )[[2]])
  )
})


test_that("Distance...", {
  skip_on_cran()
  skip_on_ci()

  full <- findNLDI(
    comid = 17595429,
    nav = "UT",
    find = "nwis",
    distance_km = 999,
    warn = FALSE
  )
  part <- findNLDI(
    comid = 17595429,
    nav = "UT",
    find = "nwis",
    distance_km = 2,
    warn = FALSE
  )
  expect_true(nrow(full$UT_nwissite) > nrow(part$UT_nwissite))
})

test_that("basin", {
  skip_on_cran()
  skip_on_ci()

  xx <- findNLDI(comid = 101, nav = "UT", find = "basin", warn = FALSE)
  xx2 <- findNLDI(
    comid = 101,
    nav = "UT",
    find = "basin",
    no_sf = TRUE,
    warn = FALSE
  )
  expect_true(sf::st_geometry_type(xx$basin) == "POLYGON")
  expect_equal(ncol(xx2$basin), 0)
})


test_that("ignore flowlines", {
  skip_on_cran()
  skip_on_ci()

  xx <- findNLDI(
    comid = 101,
    nav = "DM",
    find = c("nwis", "flowlines"),
    warn = FALSE
  )
  xx2 <- findNLDI(comid = 101, nav = "DM", find = "nwis", warn = FALSE)
  expect_gt(length(xx), length(xx2))
  expect_true("DM_flowlines" %in% names(xx))
  expect_true(!"DM_flowlines" %in% names(xx2))
})


test_that("sf points", {
  skip_on_cran()
  skip_on_ci()
  library(sf)
  p2 <- st_sfc(st_point(c(-119.8458, 34.4146)), crs = 4326)
  expect_equal(
    findNLDI(location = p2, warn = FALSE),
    findNLDI(location = sf::st_as_sf(p2))
  )
  expect_error(findNLDI(location = st_buffer(p2, .01), warn = FALSE))
})


test_that("warn_flag", {
  skip_on_ci()
  skip_on_cran()

  expect_warning(
    findNLDI(
      wqp = "TCEQMAIN-10016",
      nav = "UM",
      find = "nwissite",
      distance_km = 2,
      no_sf = TRUE,
      warn = FALSE
    ),
    regexp = NA
  )

  expect_warning(
    findNLDI(
      wqp = "TCEQMAIN-10016",
      nav = "UM",
      find = "nwissite",
      distance_km = 2,
      no_sf = TRUE,
      warn = TRUE
    )
  )
})

#' Return a data frame if there's an empty response
#'
#' @param return_list data frame returned from walk_pages
#' @param properties A vector of requested columns
#' @param service character, can be any existing collection such
#' as "daily", "monitoring-locations", "time-series-metadata"
#' @param skipGeometry A logical for whether to return geometry
#' @param convertType A logical for whether to convert value to numeric
#'
#' @return data.frame
#' @noRd
#' @examples
#'
#' df <- dataRetrieval:::deal_with_empty(data.frame(NULL),
#'                                       properties = c("time", "value"),
#'                                       service = "daily")
#'
#' df2 <- dataRetrieval:::deal_with_empty(data.frame(NULL),
#'                                       properties = NA,
#'                                       service = "daily")
#'
deal_with_empty <- function(
  return_list,
  properties,
  service,
  skipGeometry,
  convertType,
  no_paging = FALSE
) {
  if (nrow(return_list) == 0) {
    if (all(is.na(properties))) {
      schema <- check_OGC_requests(endpoint = service, type = "schema")
      properties <- names(schema$properties)
    }
    return_list <- data.frame(matrix(nrow = 0, ncol = length(properties)))
    return_list <- lapply(return_list, as.character)
    names(return_list) <- properties

    single_params <- c("datetime", "last_modified", "begin", "end", "time")

    for (i in single_params) {
      if (i %in% names(return_list)) {
        return_list[[i]] <- as.POSIXct(as.character(), origin = "1970-01-01")
      }
    }

    if (convertType && service == "daily") {
      return_list$time <- as.Date(as.character())
    }

    if (convertType && "value" %in% names(return_list)) {
      return_list$value <- as.numeric()
    }

    if (convertType && "contributing_drainage_area" %in% names(return_list)) {
      return_list$contributing_drainage_area <- as.numeric()
    }

    return_list <- data.frame(return_list)
    return_list$geometry <- NULL

    if (!skipGeometry) {
      if (!no_paging) {
        return_list <- sf::st_as_sf(return_list, geometry = sf::st_sfc())
      } else {
        return_list$x <- numeric()
        return_list$y <- numeric()
      }
    }
  }

  return(return_list)
}

#' Get USGS Monitoring Location Metadata
#'
#' @description `r get_description("combined-metadata")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("combined-metadata")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param agency_code `r get_ogc_params("combined-metadata")$agency_code`
#' @param agency_name `r get_ogc_params("combined-metadata")$agency_name`
#' @param monitoring_location_number `r get_ogc_params("combined-metadata")$monitoring_location_number`
#' @param monitoring_location_name `r get_ogc_params("combined-metadata")$monitoring_location_name`
#' @param district_code `r get_ogc_params("combined-metadata")$district_code`
#' @param state_name `r get_ogc_params("combined-metadata")$state_name`
#' @param county_code `r get_ogc_params("combined-metadata")$county_code`
#' @param county_name `r get_ogc_params("combined-metadata")$county_name`
#' @param country_code `r get_ogc_params("combined-metadata")$country_code`
#' @param country_name `r get_ogc_params("combined-metadata")$country_name`
#' @param state_code `r get_ogc_params("combined-metadata")$state_code`
#' @param minor_civil_division_code `r get_ogc_params("combined-metadata")$minor_civil_division_code`
#' @param site_type_code `r get_ogc_params("combined-metadata")$site_type_code`
#' @param site_type `r get_ogc_params("combined-metadata")$site_type`
#' @param hydrologic_unit_code `r get_ogc_params("combined-metadata")$hydrologic_unit_code`
#' @param basin_code `r get_ogc_params("combined-metadata")$basin_code`
#' @param altitude `r get_ogc_params("combined-metadata")$altitude`
#' @param altitude_accuracy `r get_ogc_params("combined-metadata")$altitude_accuracy`
#' @param altitude_method_code `r get_ogc_params("combined-metadata")$altitude_method_code`
#' @param altitude_method_name `r get_ogc_params("combined-metadata")$altitude_method_name`
#' @param vertical_datum `r get_ogc_params("combined-metadata")$vertical_datum`
#' @param vertical_datum_name `r get_ogc_params("combined-metadata")$vertical_datum_name`
#' @param horizontal_positional_accuracy_code `r get_ogc_params("combined-metadata")$horizontal_positional_accuracy_code`
#' @param horizontal_positional_accuracy `r get_ogc_params("combined-metadata")$horizontal_positional_accuracy`
#' @param horizontal_position_method_code `r get_ogc_params("combined-metadata")$horizontal_position_method_code`
#' @param horizontal_position_method_name `r get_ogc_params("combined-metadata")$horizontal_position_method_name`
#' @param original_horizontal_datum `r get_ogc_params("combined-metadata")$original_horizontal_datum`
#' @param original_horizontal_datum_name `r get_ogc_params("combined-metadata")$original_horizontal_datum_name`
#' @param drainage_area `r get_ogc_params("combined-metadata")$drainage_area`
#' @param contributing_drainage_area `r get_ogc_params("combined-metadata")$contributing_drainage_area`
#' @param time_zone_abbreviation `r get_ogc_params("combined-metadata")$time_zone_abbreviation`
#' @param uses_daylight_savings `r get_ogc_params("combined-metadata")$uses_daylight_savings`
#' @param construction_date `r get_ogc_params("combined-metadata")$construction_date`
#' @param aquifer_code `r get_ogc_params("combined-metadata")$aquifer_code`
#' @param national_aquifer_code `r get_ogc_params("combined-metadata")$national_aquifer_code`
#' @param aquifer_type_code `r get_ogc_params("combined-metadata")$aquifer_type_code`
#' @param well_constructed_depth `r get_ogc_params("combined-metadata")$well_constructed_depth`
#' @param hole_constructed_depth `r get_ogc_params("combined-metadata")$hole_constructed_depth`
#' @param depth_source_code `r get_ogc_params("combined-metadata")$depth_source_code`
#' @param thresholds `r get_ogc_params("combined-metadata")$thresholds`
#' @param parameter_code `r get_ogc_params("combined-metadata")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param parameter_name `r get_ogc_params("combined-metadata")$parameter_name`
#' Multiple parameter_names can be requested as a character vector.
#' @param unit_of_measure `r get_ogc_params("combined-metadata")$unit_of_measure`
#' @param parameter_description `r get_ogc_params("combined-metadata")$parameter_description`
#' @param data_type `r get_ogc_params("combined-metadata")$data_type`
#' @param primary
#' A flag identifying if the time series is a "primary" time series. "Primary" time
#' series (which have this flag) are standard observations which undergo Bureau
#' review and approval processes. Non-primary time series, which will have missing
#' values for "primary", are provisional datasets made available to meet the need
#' for timely best science and to assist with daily operations which need
#' real-time information. Non-primary time series data are only retained by
#' this system for 120 days.
#' @param web_description `r get_ogc_params("combined-metadata")$web_description`
#' @param parent_time_series_id `r get_ogc_params("combined-metadata")$parent_time_series_id`
#' @param statistic_id `r get_ogc_params("combined-metadata")$statistic_id`
#' Multiple statistic_ids can be requested as a character vector.
#' @param computation_identifier `r get_ogc_params("combined-metadata")$computation_identifier`
#' Multiple computation_identifiers can be requested as a character vector.
#' @param computation_period_identifier `r get_ogc_params("combined-metadata")$computation_period_identifier`
#' Multiple computation_period_identifiers can be requested as a character vector.
#' @param sublocation_identifier `r get_ogc_params("combined-metadata")$sublocation_identifier`
#' @param last_modified `r get_ogc_params("combined-metadata")$last_modified`
#'
#' See also Details below for more information.
#' @param begin `r get_ogc_params("combined-metadata")$begin`
#'
#' See also Details below for more information.
#' @param end `r get_ogc_params("combined-metadata")$end`
#'
#' See also Details below for more information.
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("combined-metadata", "field_measurement_id")`.
#' The default (`NA`) will return all columns of the data.
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#' @inherit read_waterdata_continuous details
#'
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-05407000"
#' available_data_sf <- read_waterdata_combined_meta(monitoring_location_id = site)
#'
#' groundwater <- read_waterdata_combined_meta(monitoring_location_id = "USGS-375907091432201")
#'
#' date_wi_data <- read_waterdata_combined_meta(state_name = "Wisconsin",
#'                                         county_name = "Dane County")
#'
#'
#' multi_site <- read_waterdata_combined_meta(
#'                               monitoring_location_id =  c("USGS-451605097071701",
#'                                                           "USGS-263819081585801"),
#'                               parameter_code = c("62611", "72019"))
#'
#'
#' surface_water <- read_waterdata_combined_meta(
#'                          monitoring_location_id = c("USGS-07069000",
#'                                                     "USGS-07064000",
#'                                                     "USGS-07068000"),
#'                          end = "P1M",
#'                          parameter_code = "00060")
#'
#' hucs <- read_waterdata_combined_meta(
#'                          hydrologic_unit_code = c("11010008", "11010009"),
#'                          site_type = c("Stream", "Spring")
#' )
#'
#' site_list <- read_waterdata_combined_meta(
#'                          monitoring_location_id = hucs$monitoring_location_id
#' )
#'
#' # Query for instantaneous gage height data for a site in Iowa
#' sites_inst <- read_waterdata_combined_meta(
#'   monitoring_location_id = "USGS-05418400",
#'   parameter_code = "00065"
#' )
#'
#' # parse individual thresholds lists:
#' threshold_1 <- jsonlite::fromJSON(sites_inst$thresholds[3])
#'
#' }
read_waterdata_combined_meta <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  parameter_name = NA_character_,
  unit_of_measure = NA_character_,
  statistic_id = NA_character_,
  parameter_description = NA_character_,
  data_type = NA_character_,
  computation_identifier = NA_character_,
  computation_period_identifier = NA_character_,
  thresholds = NA_character_,
  sublocation_identifier = NA_character_,
  primary = NA_character_,
  web_description = NA_character_,
  parent_time_series_id = NA_character_,
  begin = NA_character_,
  end = NA_character_,
  last_modified = NA_character_,
  agency_code = NA_character_,
  agency_name = NA_character_,
  monitoring_location_number = NA_character_,
  monitoring_location_name = NA_character_,
  district_code = NA_character_,
  country_code = NA_character_,
  country_name = NA_character_,
  state_code = NA_character_,
  state_name = NA_character_,
  county_code = NA_character_,
  county_name = NA_character_,
  minor_civil_division_code = NA_character_,
  site_type_code = NA_character_,
  site_type = NA_character_,
  hydrologic_unit_code = NA_character_,
  basin_code = NA_character_,
  altitude = NA_character_,
  altitude_accuracy = NA_character_,
  altitude_method_code = NA_character_,
  altitude_method_name = NA_character_,
  vertical_datum = NA_character_,
  vertical_datum_name = NA_character_,
  horizontal_positional_accuracy_code = NA_character_,
  horizontal_positional_accuracy = NA_character_,
  horizontal_position_method_code = NA_character_,
  horizontal_position_method_name = NA_character_,
  original_horizontal_datum = NA_character_,
  original_horizontal_datum_name = NA_character_,
  drainage_area = NA_character_,
  contributing_drainage_area = NA_character_,
  time_zone_abbreviation = NA_character_,
  uses_daylight_savings = NA_character_,
  construction_date = NA_character_,
  aquifer_code = NA_character_,
  national_aquifer_code = NA_character_,
  aquifer_type_code = NA_character_,
  well_constructed_depth = NA_character_,
  hole_constructed_depth = NA_character_,
  depth_source_code = NA_character_,
  properties = NA_character_,
  skipGeometry = NA,
  bbox = NA,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  limit = getOption("dataRetrieval.limit"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "combined-metadata"
  output_id <- "combined_meta_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

#' Date Check for Water Quality Portal
#'
#' Checks date format for inputs to the Water Quality Portal. Used in `readWQPqw`
#' and `readWQPdata`.
#'
#' @param values named list with arguments to send to the Water Quality Portal
#' @return values named list with corrected arguments to send to the Water Quality Portal
#' @export
#' @examples
#' values <- list(
#'   startDateLo = "01-01-2002",
#'   characteristicName = "Phosphorous",
#'   endDate = as.Date("2014-01-01")
#' )
#' values <- checkWQPdates(values)
checkWQPdates <- function(values) {
  dateNames <- c(
    "startDateLo",
    "startDateHi",
    "startDate",
    "endDate",
    "activityStartDateLower",
    "activityStartDateUpper"
  )

  if (any(names(values) %in% dateNames)) {
    index <- which(names(values) %in% dateNames)

    if ("" %in% values[index]) {
      values <- values[-index[values[index] == ""]]
      index <- which(names(values) %in% dateNames)
    }

    if (length(index) > 0) {
      # If a valid R date was put in, the format needs to be changed to mm-dd-yyyy for the WQP:
      for (i in index) {
        dateInput <- as.character(values[[i]])
        splitDates <- unlist(strsplit(dateInput, "-"))
        if (length(splitDates) == 3) {
          if (nchar(splitDates[1]) == 4) {
            # R object
            dates <- as.Date(lubridate::parse_date_time(dateInput, "%Y-%m-%d"))
            dates <- format(dates, format = "%m-%d-%Y")
            values[i] <- dates
          } else if (nchar(splitDates[3]) != 4) {
            # The way WQP wants it == 4, so this is probably a 2 digit year or something
            warning(
              "Please check the date format for the arguments: ",
              paste(names(values)[i], values[i], collapse = ", ")
            )
          }
        } else {
          # Probably something wrong
          warning(
            "Please check the date format for the arguments: ",
            paste(names(values)[i], values[i], collapse = ", ")
          )
        }
      }

      names(values)[names(values) == "startDate"] <- "startDateLo"
      names(values)[names(values) == "endDate"] <- "startDateHi"
    }
  }

  return(values)
}

#' Create NWIS data citation
#'
#' Uses attributes from the NWIS functions to create data citations.
#'
#' See `?bibentry` for more information.
#'
#' @param x Any data returned from an NWIS function, must
#' include "queryTime" and "url" attributes, which should
#' come with the data by default.
#' @return bibentry object to use for citing the data.
#'
#' @export
#'
#' @examples
#' \donttest{
#' nwisData <- read_waterdata_daily(monitoring_location_id = "USGS-04085427",
#'                                  parameter_code = "00060",
#'                                  time = c("2012-01-01", "2012-06-30"))
#' nwis_citation <- create_NWIS_bib(nwisData)
#' nwis_citation
#'
#' print(nwis_citation, style = "Bibtex")
#' print(nwis_citation, style = "citation")
#' }
create_NWIS_bib <- function(x) {
  textVersion <- paste0(
    "U.S. Geological Survey, ",
    format(attr(x, "queryTime"), "%Y"),
    ", USGS Water Data for the Nation: U.S. Geological Survey National Water Information System database, accessed ",
    format(attr(x, "queryTime"), "%b %d, %Y"),
    ", at https://dx.doi.org/10.5066/F7P55KJN"
  )

  ref <- utils::bibentry(
    bibtype = "Manual",
    textVersion = textVersion,
    title = "USGS Water Data for the Nation: U.S. Geological Survey National Water Information System database",
    author = utils::person("U.S. Geological Survey"),
    doi = "10.5066/F7P55KJN",
    note = paste("Accessed", format(attr(x, "queryTime"), "%b %d, %Y")),
    year = format(attr(x, "queryTime"), "%Y"),
    url = attr(x, "url")
  )

  return(ref)
}


#' Create WQP data citation
#'
#' Uses attributes from the WQP functions to create data citations.
#'
#' See `?bibentry` for more information.
#'
#' @param x Any data returned from an NWIS function, must
#' include "queryTime" and "url" attributes, which should
#' come with the data by default.
#' @return bibentry object to use for citing the data.
#'
#' @export
#'
#' @examples
#' \donttest{
#' WQPData <- readWQPqw("USGS-05288705",
#'                      parameterCd = "00300")
#' wqp_citation <- create_WQP_bib(WQPData)
#' wqp_citation
#'
#' print(wqp_citation, style = "Bibtex")
#' print(wqp_citation, style = "citation")
#' }
create_WQP_bib <- function(x) {
  textVersion <- paste0(
    "National Water Quality Monitoring Council, ",
    format(attr(x, "queryTime"), "%Y"),
    ", Water Quality Portal, accessed ",
    format(attr(x, "queryTime"), "%m, %d, %Y"),
    ", ",
    attr(x, "url"),
    ", https://doi.org/10.5066/P9QRKUVJ."
  )

  ref <- utils::bibentry(
    bibtype = "Manual",
    textVersion = textVersion,
    title = " Water Quality Portal",
    author = utils::person("National Water Quality Monitoring Council"),
    doi = "10.5066/P9QRKUVJ",
    note = paste("Accessed", format(attr(x, "queryTime"), "%b %d, %Y")),
    year = format(attr(x, "queryTime"), "%Y"),
    url = attr(x, "url")
  )

  return(ref)
}

#' Create API url
#'
#' Main documentation: <https://api.waterdata.usgs.gov/ogcapi/v0/>,
#' Swagger docs: <https://api.waterdata.usgs.gov/ogcapi/v0/openapi?f=html>.
#'
#' @export
#' @param service Which service available on <https://api.waterdata.usgs.gov/ogcapi/v0/>.
#' @param output_id Name of id column to return
#' @param ... Extra parameters from the specific services.
#' @param bbox Only features that have a geometry that intersects the bounding
#' box are selected.The bounding box is provided as four or six numbers, depending
#' on whether the coordinate reference system includes a vertical axis (height or
#' depth).
#' @keywords internal
#'
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @examples
#' site <- "USGS-02238500"
#' pcode <- "00060"
#' req_dv <- construct_api_requests("daily",
#'                                  output_id = "daily_id",
#'                                  monitoring_location_id = site,
#'                                  parameter_code = "00060")
#'
#' req_dv <- construct_api_requests("daily",
#'                                  output_id = "daily_id",
#'                                  monitoring_location_id = site,
#'                                  parameter_code = c("00060", "00065"))
#'
#' sites <- c("USGS-01491000", "USGS-01645000")
#' start_date <- "2018-01-01"
#' end_date <- "2022-01-01"
#' req_dv <- construct_api_requests("daily",
#'                                  output_id = "daily_id",
#'                                  monitoring_location_id = sites,
#'                                  parameter_code = c("00060", "00065"),
#'                                  datetime = c(start_date, end_date))
#'
construct_api_requests <- function(
  service,
  output_id,
  ...,
  bbox = NA,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_data"),
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  POST <- FALSE

  full_list <- list(...)
  full_list[["limit"]] <- limit

  check_arguments_non_api(
    convertType = convertType,
    no_paging = no_paging,
    limit = full_list[["limit"]],
    attach_request = attach_request,
    chunk_size = chunk_size
  )

  check_arguments_api(
    bbox = full_list[["bbox"]],
    skipGeometry = full_list[["skipGeometry"]]
  )

  full_list <- switch_arg_id(
    full_list,
    id_name = output_id,
    service = service
  )

  # Clean out non-API arguments:
  properties <- switch_properties_id(
    properties = full_list[["properties"]],
    id = output_id
  )

  full_list[["properties"]] <- NULL

  if (any(time_periods %in% names(full_list))) {
    for (i in time_periods[time_periods %in% names(full_list)]) {
      dates <- FALSE
      if (all(service == "daily" & i != "last_modified")) {
        dates <- TRUE
      }
      full_list[[i]] <- format_api_dates(full_list[[i]], date = dates)
    }
  }

  single_params <- c(
    "datetime",
    "last_modified",
    "begin",
    "end",
    "time",
    "begin_utc",
    "end_utc",
    "limit",
    "skipGeometry"
  )

  comma_params <- c(
    "monitoring_location_id",
    "parameter_code",
    "statistic_id",
    "time_series_id",
    "computation_period_identifier",
    "computation_identifier",
    "data_type"
  )

  if (
    service %in%
      c(
        "monitoring-locations",
        "parameter-codes",
        "time-series-metadata",
        "field-measurements-metadata",
        "combined-metadata"
      )
  ) {
    comma_params <- c(comma_params, "id")
  }

  if (all(is.na(full_list)) & all(is.na(bbox))) {
    warning("No filtering arguments specified.")
  }
  # Figure out if the GET request will be > 2048 characters
  # and remove NA's from the comma parameters
  comma_params_filtered <- Filter(
    Negate(anyNA),
    lapply(full_list[comma_params], function(x) x[!is.na(x)])
  )

  comma_params_filtered <- comma_params_filtered[
    !sapply(comma_params_filtered, is.null)
  ]

  single_params_filtered <- Filter(Negate(anyNA), full_list[single_params])
  single_params_filtered <- single_params_filtered[
    !sapply(single_params_filtered, is.null)
  ]

  force_post <- nchar(paste0(unlist(comma_params_filtered), collapse = ",")) >
    2048

  if (force_post) {
    get_list <- single_params_filtered
  } else {
    # GET list refers to arguments that will go in the URL no matter what (not POST)
    get_list <- c(single_params_filtered, comma_params_filtered)
  }

  get_list <- get_list[!is.na(get_list)]

  format_type <- ifelse(isTRUE(no_paging), "csv", "json")

  baseURL <- setup_api(service, format = format_type)
  baseURL <- explode_query(baseURL, POST = FALSE, get_list, multi = "comma")

  if (all(!is.na(bbox))) {
    baseURL <- httr2::req_url_query(
      baseURL,
      bbox = as.numeric(bbox),
      .multi = "comma"
    )
  }

  if (!all(is.na(properties))) {
    available_properties <- property_list[[service]]

    if (!all(properties %in% available_properties)) {
      # Check again:
      schema <- check_OGC_requests(endpoint = service, type = "schema")
      properties_fresh <- names(schema$properties)
      if (!all(properties %in% properties_fresh)) {
        stop(
          "Invalid properties: ",
          paste0(properties[!properties %in% properties_fresh], collapse = ", ")
        )
      }
    }

    baseURL <- httr2::req_url_query(
      baseURL,
      properties = properties,
      .multi = "comma"
    )
  }

  #POST list are the arguments that need to be in the POST body
  post_list <- full_list[!names(full_list) %in% names(get_list)]

  post_params <- explode_post(post_list)

  # Should we do a POST?
  POST = length(post_params) > 0

  if (POST) {
    baseURL <- baseURL |>
      httr2::req_headers(`Content-Type` = "application/query-cql-json")

    if (length(post_params) > 1) {
      post_params <- list(
        "params" = unname(post_params)
      )

      template_path_post <- system.file(
        "templates/post.CQL2",
        package = "dataRetrieval"
      )
      template_post <- readChar(
        template_path_post,
        file.info(template_path_post)$size
      )

      x <- whisker::whisker.render(template_post, post_params)
    } else {
      x <- post_params[[1]]
    }

    baseURL <- httr2::req_body_raw(baseURL, x)
  } else {
    baseURL <- explode_query(baseURL, POST = FALSE, full_list, multi = "comma")
  }

  return(baseURL)
}


#' Setup the request for the OGC API requests
#'
#' @noRd
#' @return httr2 request
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' request <- dataRetrieval:::base_url()
#' request
#' }
base_url <- function() {
  httr2::request("https://api.waterdata.usgs.gov/ogcapi/") |>
    httr2::req_url_path_append(getOption("dataRetrieval.api_version"))
}

#' Setup the request for a particular endpoint collection
#'
#' @noRd
#' @return httr2 request
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' request <- dataRetrieval:::setup_api("daily")
#' request
#' }
setup_api <- function(service, format = "json") {
  baseURL <- base_url() |>
    httr2::req_url_path_append("collections") |>
    httr2::req_url_path_append(service, "items") |>
    basic_request(format = format)
}

#' Switch endpoint id arg
#'
#' @noRd
#' @return list
#' @examples
#'
#' l1 <- list("id" = "1234")
#' dataRetrieval:::switch_arg_id(l1,
#'                               id_name = "monitoring_location_id",
#'                               service = "monitoring-locations")
#'
#' l2 <- list("monitoring_location_id" = "1234")
#' dataRetrieval:::switch_arg_id(l2,
#'                               id_name = "monitoring_location_id",
#'                               service = "monitoring-locations")
#'
#' l3 <- list("monitoring_locations_id" = "1234")
#' dataRetrieval:::switch_arg_id(l3,
#'                               id_name = "monitoring_location_id",
#'                               service = "monitoring-locations")
#'
switch_arg_id <- function(ls, id_name, service) {
  service_id <- paste0(gsub("-", "_", service), "_id")
  if (!"id" %in% names(ls)) {
    if (service_id %in% names(ls)) {
      ls[["id"]] <- ls[[service_id]]
    } else {
      ls[["id"]] <- ls[[id_name]]
    }
  }

  ls[[service_id]] <- NULL
  ls[[id_name]] <- NULL
  return(ls)
}

#' Format the date request
#'
#' Users will want to give either start/end dates or
#' period requests.
#'
#' @param datetime character, Date, or POSIX
#' @param date logical, whether to return Date or POSIX
#'
#' @noRd
#' @return character vector with a length of either 1 or 2.
#' @examples
#'
#' start_end <- c("2021-01-01", "2022-01-01")
#' dataRetrieval:::format_api_dates(start_end, date = TRUE)
#' dataRetrieval:::format_api_dates(start_end, date = FALSE)
#'
#' start_end <- c("", "")
#' dataRetrieval:::format_api_dates(start_end)
#'
#' period <- "P7D"
#' dataRetrieval:::format_api_dates(period)
#'
#' start <- c("2021-01-01", NA)
#' dataRetrieval:::format_api_dates(start)
#' dataRetrieval:::format_api_dates(start, TRUE)
#'
#' end <- c(NA, "2021-01-01")
#' dataRetrieval:::format_api_dates(end)
#' dataRetrieval:::format_api_dates(end, TRUE)
#'
#' end <- as.POSIXct(c(NA, "2021-01-01 12:15:00"))
#' dataRetrieval:::format_api_dates(end)
#'
#' start_end <- as.POSIXct(c("2021-01-01 12:15:00",
#'                           "2022-01-01 16:45"))
#' dataRetrieval:::format_api_dates(start_end)
#'
#' start_end <- as.POSIXct(c("2021-01-01 12:15:00",
#'                           "2022-01-01 16:45"),
#'                           tz = "America/New_York")
#'
#' dataRetrieval:::format_api_dates(start_end)
#'
#' # If you don't specify a timezone, it will assume UTC
#' start_end2 <- c("2021-01-01 12:15:00", "")
#' dataRetrieval:::format_api_dates(start_end2)
#'
#' # If you do specify a timezone, it should maintain it, but convert to UTC:
#' start_end2 <- c("2021-01-01T12:15:00-0500", "")
#' dataRetrieval:::format_api_dates(start_end2)
#'
#' time = c("2014-05-01T00:00:00Z", "2014-05-01T12:00:00Z")
#' dataRetrieval:::format_api_dates(time)
#'
#' time = c("2014-05-01T00:00Z", "2014-05-01T12:00Z")
#' dataRetrieval:::format_api_dates(time)
#'
#' start <- "2025-10-01"
#' end <- Sys.Date()
#' dataRetrieval:::format_api_dates(c(start, end), date = TRUE)
#'
#' # This is a problem because the first value forces the
#' # vector to be numeric, and then we don't really
#' # know if the 2nd value is a Date (number of days since 1970)
#' # or if it's a date/time (number of seconds..)
#' half_range <- c(NA, as.Date("2025-01-01"))
#' # Will error:
#' #dataRetrieval:::format_api_dates(half_range, date = FALSE)
#' # Better way to do it:
#' better_half <- as.Date(c(NA, "2025-01-01"))
#' dataRetrieval:::format_api_dates(better_half, date = FALSE)
format_api_dates <- function(datetime, date = FALSE) {
  if (is.character(datetime)) {
    datetime[datetime == ""] <- NA
    datetime <- toupper(datetime)
  }

  if (all(is.na(datetime)) | all(is.null(datetime))) {
    return(NA)
  }

  if (length(datetime) > 2) {
    stop("datetime should only include 1-2 values")
  }

  if (is.numeric(datetime)) {
    # Until we can figure out a way to know if the
    # original input was suppose to be Date or Posix
    # We can't determine what the user meant.
    stop(
      "A time query was entered as numeric. This could lead to errors.
Check any time queries that might have been automatically converted to numeric.
This could happen if you use c(NA, as.Date(Sys.Date())) instead of
as.Date(c(NA, Sys.Date()) for example."
    )
  }

  if (length(datetime) == 1) {
    # If the user has "P" or the "/" we assume they know what they are doing
    if (
      grepl("P", datetime, ignore.case = TRUE) |
        grepl("/", datetime)
    ) {
      return(datetime)
    }

    if (date) {
      datetime <- get_Date(datetime)
    } else {
      datetime1 <- get_dateTime(datetime)
      datetime <- lubridate::format_ISO8601(datetime1, usetz = "Z")
    }
  } else if (length(datetime) == 2) {
    if (date) {
      for (i in seq_along(datetime)) {
        datetime[i] <- get_Date(datetime[i])
      }
      datetime <- paste0(datetime, collapse = "/")
    } else {
      datetime1 <- lapply(datetime, get_dateTime)
      datetime <- paste0(
        vapply(datetime1, lubridate::format_ISO8601, character(1), usetz = "Z"),
        collapse = "/"
      )
    }

    datetime <- gsub("NA", "..", datetime)
  }

  return(datetime)
}

get_dateTime <- function(d) {
  temp_date <- tryCatch({
    strptime(d, format = "%Y-%m-%dT%H:%MZ", tz = "UTC")
  })

  if (all(is.na(temp_date))) {
    temp_date <- tryCatch(
      {
        lubridate::as_datetime(d)
      },
      error = function(e) {
        NA
      }
    )
  }

  return(temp_date)
}

get_Date <- function(d) {
  temp_date <- tryCatch(
    {
      as.Date(d)
    },
    error = function(e) {
      "try again"
    }
  )

  if (is.na(temp_date)) {
    return("..")
  } else if (as.character(temp_date) == "try again") {
    temp_date <- tryCatch(
      {
        as.Date(as.numeric(d), origin = "1970-01-01")
      },
      error = function(e) {
        "try again"
      }
    )
  }

  return(as.character(temp_date))
}


#' Turn request list into POST body cql
#'
#' @noRd
#' @return character vector of CQL filters
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#'
#' query_list <- list(monitoring_location_id = c("USGS-01491000",
#'                                               "USGS-01645000"),
#'                    parameter_code = c("00060", "00010"))
#'
#' dataRetrieval:::explode_post(query_list)
#'
#' }
explode_post <- function(ls) {
  ls <- Filter(Negate(anyNA), ls)
  params <- NULL

  if (length(ls) > 0) {
    if (max(lengths(ls)) > 1) {
      for (i in seq_along(ls)) {
        params[names(ls[i])] <- cql2_param(ls[i])
      }

      if (length(params) > 1) {
        params[seq_along(1:(length(params) - 1))] <- paste0(
          params[seq_along(1:(length(params) - 1))],
          ","
        )
      }
    }
  }
  return(params)
}

#' Create CQL parameters
#'
#' Helps to give more informative messages on some errors.
#'
#' @param parameter named vector
#' @noRd
#' @return list
#' @examples
#'
#' parameter <- list("monitoring_location_id" = c("USGS-02238500",
#'                                                "USGS-01491000"))
#' dataRetrieval:::cql2_param(parameter)
#'
cql2_param <- function(parameter) {
  # Wildcards:
  if (names(parameter) %in% c("hydrologic_unit_code")) {
    template_path <- system.file(
      "templates/param.CQL2.like",
      package = "dataRetrieval"
    )
    template <- readChar(template_path, file.info(template_path)$size)

    params <- c()
    for (i in parameter[[1]]) {
      parameter_list <- list("property" = names(parameter), "parameter" = i)
      params <- c(params, whisker::whisker.render(template, parameter_list))
    }
    template_path_or <- system.file(
      "templates/post.CQL2.or",
      package = "dataRetrieval"
    )
    template_or <- readChar(template_path_or, file.info(template_path_or)$size)

    post_params <- list(
      "params" = paste0(params, collapse = ", ")
    )
    cql_text <- whisker::whisker.render(template_or, post_params)
  } else {
    # INs
    parameters <- paste0(unlist(parameter), collapse = '", "')
    parameters <- paste0('"', parameters, '"')
    parameter_list <- list(
      "property" = names(parameter),
      "parameter" = parameters
    )

    template_path <- system.file(
      "templates/param.CQL2",
      package = "dataRetrieval"
    )
    template <- readChar(template_path, file.info(template_path)$size)

    cql_text <- whisker::whisker.render(template, parameter_list)
  }

  return(cql_text)
}


#' Custom Error Messages
#'
#' Helps to give more informative messages on some errors.
#'
#' @param resp httr2 response
#' @return list
#' @noRd
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' check_collections <- dataRetrieval:::base_url() |>
#'   httr2::req_url_path_append("openapi") |>
#'   httr2::req_url_query(f = "html#/server/getCollections")
#'
#' collect_request <- dataRetrieval:::basic_request(check_collections)
#' query_ret <- httr2::req_perform(collect_request)
#' dataRetrieval:::error_body(query_ret)
#' }
#'
error_body <- function(resp) {
  status <- httr2::resp_status(resp)
  if (status == 429) {
    x <- httr2::resp_body_json(resp)$error
    return(x[["message"]])
  } else if (status == 403) {
    return(
      "Query request denied. Possible reasons include query exceeding server limits."
    )
  }
}


#' Basic request to API services
#'
#' Automatically includes json format, gzip encoding, dataRetrieval
#' user agents, and the X-Api-Key token if available.
#'
#' @param url_base httr2 request
#' @return list
#' @noRd
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' check_collections <- dataRetrieval:::base_url() |>
#'   httr2::req_url_path_append("openapi") |>
#'   httr2::req_url_query(f = "html#/server/getCollections")
#' collect_request <- dataRetrieval:::basic_request(check_collections)
#' collect_request
#' }
#'
basic_request <- function(url_base, format = "json") {
  req <- url_base |>
    httr2::req_user_agent(default_ua()) |>
    httr2::req_headers(`Accept-Encoding` = c("compress", "gzip")) |>
    httr2::req_url_query(f = format, lang = "en-US") |>
    httr2::req_error(body = error_body) |>
    httr2::req_retry(max_tries = 3, retry_on_failure = TRUE) |>
    httr2::req_timeout(seconds = 180)

  req <- add_api_token(req)

  return(req)
}

add_api_token <- function(req) {
  token <- Sys.getenv("API_USGS_PAT")
  if (token != "") {
    req <- req |>
      httr2::req_headers_redacted(`X-Api-Key` = token)
  }
  req
}

# Treat these columns as time:
time_periods <- c(
  "last_modified",
  "datetime",
  "time",
  "begin",
  "end",
  "begin_utc",
  "end_utc"
)

#' Site Data Import from Water Quality Portal
#'
#' Returns a list of sites from the Water Quality Portal web service. This function
#' gets the data from: <https://www.waterqualitydata.us>.
#' Arguments to the function should be based on
#' <https://www.waterqualitydata.us/webservices_documentation>. The return from
#' this function returns the basic metadata on WQP sites. It is
#' generally faster than the [whatWQPdata()] function, but does
#' not return information on what data was collected at the site.
#'
#' @param \dots see <https://www.waterqualitydata.us/webservices_documentation>
#' for a complete list of options. A list of arguments can also be supplied.
#' One way to figure out how to construct a WQP query is to go to the "Advanced"
#' form in the Water Quality Portal:
#' <https://www.waterqualitydata.us/#mimeType=csv&providers=NWIS&providers=STORET>
#' Use the form to discover what parameters are available. Once the query is
#' set in the form, scroll down to the "Query URL". You will see the parameters
#' after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
#' in the Characteristic Group dropdown, you will see characteristicType=Nutrient
#' in the Query URL. The corresponding argument for dataRetrieval is
#' characteristicType = "Nutrient". dataRetrieval users do not need to include
#' mimeType,  and providers is optional (these arguments are picked automatically).
#' @param legacy Logical. If TRUE, uses legacy WQP services. Default is TRUE.
#' Setting legacy = FALSE uses WQX3.0 WQP services, which are in-development, use with caution.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the
#' function will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character.
#' @keywords data import WQP web service
#' @rdname wqpSpecials
#' @name whatWQPsites
#' @seealso whatWQPdata readWQPsummary
#' @return data frame that includes information on site metadata.
#'
#' @export
#' @seealso whatNWISdata
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#'
#' #site1 <- whatWQPsites(siteid = "USGS-01594440")
#'
#' #type <- "Stream"
#' #sites <- whatWQPsites(
#' #   countycode = "US:55:025",
#' #  characteristicName = "Phosphorus",
#' #   siteType = type
#' #)
#' }
whatWQPsites <- function(..., legacy = TRUE, convertType = TRUE) {
  values <- readWQPdots(..., legacy = legacy)

  values <- values[["values"]]

  if (any(c("tz", "service") %in% names(values))) {
    values <- values[!(names(values) %in% c("tz", "service"))]
  }
  POST = FALSE

  if (legacy) {
    baseURL <- httr2::request(pkg.env[["Station"]])

    if ("siteid" %in% names(values)) {
      if (length(values[["siteid"]]) > 1) {
        sites <- values[["siteid"]]
        POST <- nchar(paste0(sites, collapse = "")) > 2048

        baseURL <- get_or_post(
          baseURL,
          POST = POST,
          siteid = sites,
          .multi = function(x) paste0(x, collapse = ";")
        )

        values <- values[names(values) != "siteid"]
      }
    }
  } else {
    baseURL <- httr2::request(pkg.env[["StationWQX3"]])
  }

  baseURL <- get_or_post(baseURL, POST = POST, !!!values, .multi = "explode")

  retval <- importWQP(baseURL, convertType = convertType)

  if (!is.null(retval)) {
    attr(retval, "queryTime") <- Sys.time()
    attr(retval, "url") <- baseURL$url
  }

  return(retval)
}


#' Summary of Data Available from Water Quality Portal
#'
#' Returns a list of sites with year-by-year information on what data is available.
#' The function gets the data from: <https://www.waterqualitydata.us>.
#' Arguments to the function should be based on
#' <https://www.waterqualitydata.us/webservices_documentation>.
#' The information returned from this function describes the
#' available data at the WQP sites, and some metadata on the sites themselves.
#'
#' @param \dots see <https://www.waterqualitydata.us/webservices_documentation>
#'  for a complete list of options. A list of arguments can also be supplied.
#' One way to figure out how to construct a WQP query is to go to the "Advanced"
#' form in the Water Quality Portal:
#' <https://www.waterqualitydata.us/#mimeType=csv&providers=NWIS&providers=STORET>
#' Use the form to discover what parameters are available. Once the query is
#' set in the form, scroll down to the "Query URL". You will see the parameters
#' after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
#' in the Characteristic Group dropdown, you will see characteristicType=Nutrient
#' in the Query URL. The corresponding argument for dataRetrieval is
#' characteristicType = "Nutrient". dataRetrieval users do not need to include
#' mimeType, and providers is optional (these arguments are picked automatically).
#' @return A data frame from the data returned from the Water Quality Portal
#' about the data available for the query parameters.
#' @export
#' @seealso whatWQPsites whatWQPdata
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' # Summary of a single site for the last 5 years:
#' #site_5 <- readWQPsummary(
#' #   siteid = "USGS-07144100",
#' #   summaryYears = 5
#'#  )
#'
#' # Summary of a single site for the full period of record:
#' #site_all <- readWQPsummary(
#' #   siteid = "USGS-07144100",
#' #   summaryYears = "all"
#' # )
#'
#' # Summary of the data available from streams in a single county:
#' # dane_county_data <- readWQPsummary(
#' #   countycode = "US:55:025",
#' #   summaryYears = 5,
#' #   siteType = "Stream"
#' # )
#'
#' # Summary of the data all available from lakes in a single county:
#' # lake_sites <- readWQPsummary(
#' #   siteType = "Lake, Reservoir, Impoundment",
#' #   countycode = "US:55:025"
#' # )
#'
#' # Summary of the data available for the last 5 years in New Jersey:
#' #state1 <- readWQPsummary(
#' #   statecode = "NJ",
#' #   summaryYears = 5,
#' #    siteType = "Stream"
#' # )
#' }
readWQPsummary <- function(...) {
  wqp_message()

  values <- readWQPdots(...)

  values <- values[["values"]]

  if (any(c("tz", "service") %in% names(values))) {
    values <- values[!(names(values) %in% c("tz", "service"))]
  }

  if (!"dataProfile" %in% names(values)) {
    values[["dataProfile"]] <- "periodOfRecord"
  }

  baseURL <- httr2::request(pkg.env[["SiteSummary"]])

  if (length(values[["siteid"]]) > 1) {
    sites <- values[["siteid"]]
    baseURL <- httr2::req_url_query(
      baseURL,
      siteid = sites,
      .multi = function(x) paste0(x, collapse = ";")
    )
    values <- values[names(values) != "siteid"]
  }

  baseURL <- httr2::req_url_query(baseURL, !!!values, .multi = "explode")

  withCallingHandlers(
    {
      retval <- importWQP(baseURL, csv = TRUE)
    },
    warning = function(w) {
      if (any(grepl("Number of rows returned not matched in header", w))) {
        invokeRestart("muffleWarning")
      }
    }
  )

  if (!is.null(retval)) {
    attr(retval, "queryTime") <- Sys.time()
    attr(retval, "url") <- baseURL$url
  }

  return(retval)
}

#' Function to return data from the National Ground Water Monitoring Network waterML2 format
#'
#' This function accepts a url parameter for a WaterML2 getObservation. This function is still under development,
#' but the general functionality is correct.
#'
#' @param input character or raw, containing the url for the retrieval or a path to the data file, or raw XML.
#' @param asDateTime logical, if `TRUE` returns date and time as POSIXct, if `FALSE`, character
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the data's provided time zone offset.
#' Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
#' "America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  `OlsonNames()`
#' for more information on time zones.
#' @return mergedDF a data frame source, time, value, uom, uomTitle, comment, gmlID
#' @export
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#'
#' params <- list(request = "GetObservation",
#'                service = "SOS",
#'                version = "2.0.0",
#'                observedProperty = "urn:ogc:def:property:OGC:GroundWaterLevel",
#'                responseFormat = "text/xml",
#'                featureOfInterest = "VW_GWDP_GEOSERVER.USGS.403836085374401")
#'
#' obs_url <- httr2::request("https://cida.usgs.gov") |>
#'  httr2::req_url_path_append("ngwmn_cache") |>
#'  httr2::req_url_path_append("sos") |>
#'  httr2::req_url_query(!!!params)
#'
#' #data_returned <- importNGWMN(obs_url)
#' }
#'
importNGWMN <- function(input, asDateTime = FALSE, tz = "UTC") {
  if (tz != "") {
    tz <- match.arg(tz, OlsonNames())
  } else {
    tz <- "UTC"
  }

  raw <- FALSE

  if (inherits(input, "httr2_request")) {
    returnedDoc <- getWebServiceData(input)
    if (is.null(returnedDoc)) {
      return(invisible(NULL))
    }
    returnedDoc <- xml2::xml_root(returnedDoc)
  } else if (is.character(input) && file.exists(input)) {
    returnedDoc <- xml2::read_xml(input)
  } else if (
    is.character(input) &&
      grepl("(https)://[^ /$.?#].[^\\s]*", input)
  ) {
    input <- httr2::request(input)
    returnedDoc <- getWebServiceData(input)
    if (is.null(returnedDoc)) {
      return(invisible(NULL))
    }
  } else if (is.raw(input)) {
    returnedDoc <- xml2::read_xml(input)
    raw <- TRUE
  }

  response <- xml2::xml_name(returnedDoc)
  if (response == "GetObservationResponse") {
    timeSeries <- xml2::xml_find_all(
      returnedDoc,
      "//wml2:MeasurementTimeseries"
    ) # each parameter/site combo

    if (0 == length(timeSeries)) {
      df <- data.frame()
      if (!raw) {
        attr(df, "url") <- input
      }
      return(df)
    }

    mergedDF <- NULL

    for (t in timeSeries) {
      df <- importWaterML2(t, asDateTime, tz)

      if (is.null(mergedDF)) {
        mergedDF <- df
      } else {
        similarNames <- intersect(colnames(mergedDF), colnames(df))
        mergedDF <- merge(
          x = mergedDF,
          y = df,
          by = similarNames,
          all = TRUE
        )
      }
    }

    if (!raw) {
      url <- input
      attr(mergedDF, "url") <- url
    }
    if (asDateTime) {
      mergedDF$date <- as.Date(mergedDF$date)
    }
    nonDateCols <- grep("date", names(mergedDF), value = TRUE, invert = TRUE)

    if (nrow(mergedDF) > 0) {
      mergedDF[nonDateCols][
        mergedDF[nonDateCols] == "" | mergedDF[nonDateCols] == -999999.0
      ] <- NA
    }
    attr(mergedDF, "gml:identifier") <- xml2::xml_text(xml2::xml_find_all(
      returnedDoc,
      ".//gml:identifier"
    ))
    attr(mergedDF, "generationDate") <- xml2::xml_text(xml2::xml_find_all(
      returnedDoc,
      ".//wml2:generationDate"
    ))
    meta <- xml2::xml_find_all(returnedDoc, ".//gmd:contact")
    attr(mergedDF, "contact") <- xml2::xml_attr(meta, "href")
    attr(mergedDF, "responsibleParty") <- xml2::xml_text(xml2::xml_find_all(
      meta,
      ".//gco:CharacterString"
    ))
  } else if (response == "GetFeatureOfInterestResponse") {
    featureMembers <- xml2::xml_find_all(returnedDoc, ".//sos:featureMember")
    site <- xml2::xml_text(xml2::xml_find_all(
      featureMembers,
      ".//gml:identifier"
    ))
    site <- substring(site, 8)

    # some sites don't have a description
    siteDesc <- xml2::xml_text(xml2::xml_find_first(
      featureMembers,
      ".//gml:description"
    ))

    siteLocs <- strsplit(
      xml2::xml_text(xml2::xml_find_all(featureMembers, ".//gml:pos")),
      " "
    )
    siteLocs <- data.frame(
      matrix(unlist(siteLocs), nrow = length(siteLocs), byrow = TRUE),
      stringsAsFactors = FALSE
    )
    names(siteLocs) <- c("dec_lat_va", "dec_lon_va")

    siteLocs$dec_lat_va <- as.numeric(siteLocs$dec_lat_va)
    siteLocs$dec_lon_va <- as.numeric(siteLocs$dec_lon_va)
    mergedDF <- cbind.data.frame(
      site,
      description = siteDesc,
      siteLocs,
      stringsAsFactors = FALSE
    )
  } else if (response == "ExceptionReport") {
    return(data.frame())
  } else {
    stop("Unrecognized response from the web service")
    return(data.frame())
  }
  return(mergedDF)
}

#' Parse the WaterML2 timeseries portion of a waterML2 file
#'
#' Returns data frame columns of all information with each time series measurement;
#' Anything defined as a default, is returned as an attribute of that data frame.
#'
#' @param input XML with only the wml2:MeasurementTimeseries node and children
#' @param asDateTime logical, if `TRUE` returns date and time as POSIXct, if `FALSE`, character
#' @param tz character to set timezone attribute of datetime. Default is an empty quote, which converts the
#' datetimes to UTC (properly accounting for daylight savings times based on the data's provided time zone offset).
#' Possible values are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
#' "America/Anchorage", "America/Honolulu", "America/Jamaica", "America/Managua",
#' "America/Phoenix", and "America/Metlakatla"
#' @export
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' baseURL <- httr2::request("https://waterservices.usgs.gov/nwis/dv")
#' baseURL <- httr2::req_url_query(baseURL,
#'                                 format = "waterml,2.0",
#'                                 sites = "01646500",
#'                                 startDT = "2014-09-01",
#'                                 endDT = "2014-09-08",
#'                                 statCd = "00003",
#'                                 parameterCd = "00060" )
#'
#' timeseries <- importWaterML2(baseURL, asDateTime = TRUE, tz = "UTC")
#' }
importWaterML2 <- function(input, asDateTime = FALSE, tz = "UTC") {
  returnedDoc <- check_if_xml(input)

  gmlID <- xml2::xml_attr(returnedDoc, "id") # TODO: make this an attribute
  TVP <- xml2::xml_find_all(returnedDoc, ".//wml2:MeasurementTVP") # time-value pairs
  if (length(TVP) == 0) {
    # empty nodes on some sites
    return(data.frame(
      site = character(0),
      source = character(0),
      date = character(0),
      time = character(0),
      dateTime = character(0),
      value = numeric(0),
      uom = character(0),
      comment = character(0),
      stringsAsFactors = FALSE
    ))
  }
  rawTime <- xml2::xml_text(xml2::xml_find_all(
    returnedDoc,
    ".//wml2:MeasurementTVP/wml2:time"
  ))

  valueNodes <- xml2::xml_find_all(
    returnedDoc,
    ".//wml2:MeasurementTVP/wml2:value"
  )
  charValues <- xml2::xml_text(valueNodes)
  nilValues <- as.logical(xml2::xml_attr(valueNodes, "nil"))
  charValues[nilValues] <- NA
  values <- as.numeric(charValues)
  nVals <- length(values)

  # df of date, time, dateTime
  oneCol <- rep(NA, nVals)
  timeDF <- data.frame(date = oneCol, time = oneCol, dateTime = oneCol)
  splitTime <- data.frame(
    matrix(unlist(strsplit(rawTime, "T")), nrow = nVals, byrow = TRUE),
    stringsAsFactors = FALSE
  )
  if (ncol(splitTime) > 1) {
    # some sites only have a date
    names(splitTime) <- c("date", "time")
  } else {
    names(splitTime) <- "date"
    splitTime$time <- NA
  }

  timeDF <- splitTime
  timeDF$dateTime <- NA
  logicVec <- nchar(rawTime) > 19
  if (!all(!logicVec)) {
    # otherwise sets it to char <NA>
    timeDF$dateTime[logicVec] <- rawTime[logicVec]
  }
  if (asDateTime) {
    timeDF$dateTime <- lubridate::parse_date_time(
      timeDF$dateTime,
      c(
        "%Y",
        "%Y-%m-%d",
        "%Y-%m-%dT%H:%M",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M:%OS",
        "%Y-%m-%dT%H:%M:%OS%z"
      ),
      exact = TRUE
    )
    # ^^setting tz in as.POSIXct just sets the attribute, does not convert the time!
    attr(timeDF$dateTime, "tzone") <- tz
  }

  uom <- xml2::xml_attr(valueNodes, "uom", default = NA)

  source <- xml2::xml_attr(
    xml2::xml_find_all(
      returnedDoc,
      "./wml2:point/wml2:MeasurementTVP/wml2:metadata/wml2:source"
    ),
    "title"
  )
  comment <- xml2::xml_text(xml2::xml_find_all(
    returnedDoc,
    "./wml2:point/wml2:MeasurementTVP/wml2:metadata/wml2:comment"
  ))

  defaultMeta <- xml2::xml_find_all(
    returnedDoc,
    ".//wml2:DefaultTVPMeasurementMetadata"
  )
  defaultQuals <- xml2::xml_text(xml2::xml_find_all(
    defaultMeta,
    ".//swe:description"
  ))
  defaultUOM <- xml2::xml_attr(
    xml2::xml_find_all(defaultMeta, ".//wml2:uom"),
    "title",
    default = NA
  )

  df_vars <- list(
    source = source,
    timeDF,
    value = values,
    uom = uom,
    comment = comment
  )
  df_use <- df_vars[sapply(df_vars, function(x) {
    length(x) > 0 && !all(is.na(x))
  })]
  df <- data.frame(df_use, stringsAsFactors = FALSE)
  if (!"value" %in% names(df)) {
    df$value <- NA_real_
  }
  # from the default metadata section
  # append to existing attributes if they aren't empty
  mdAttribs <- list(
    defaultQualifier = defaultQuals,
    defaultUOM = defaultUOM,
    gmlID = gmlID
  ) # all attributes must have names
  mdAttribs_use <- mdAttribs[sapply(mdAttribs, function(x) length(x) > 0)]
  attributes(df) <- append(attributes(df), mdAttribs_use)
  return(df)
}

context("importRDB")
test_that("CRAN-friendly importRDB test", {
  filePath <- system.file("extdata", package = "dataRetrieval")
  fileName <- "RDB1Example.txt"
  fullPath <- file.path(filePath, fileName)
  importUserRDB <- importRDB1(fullPath)

  # default is to turn dates to characters
  expect_is(importUserRDB$datetime, "Date")
})

context("importWaterML1")
test_that("CRAN-friendly importWaterML1 test", {
  filePath <- system.file("extdata", package = "dataRetrieval")
  fileName <- "WaterML1Example.xml"
  fullPath <- file.path(filePath, fileName)
  importUserWML1 <- importWaterML1(fullPath, asDateTime = TRUE)
  # default is to turn dates to characters
  expect_is(importUserWML1$dateTime, "POSIXct")
})

context("importWaterML2")

test_that("importWaterML2 internal test", {
  filePath <- system.file("extdata", package = "dataRetrieval")
  fileName <- "WaterML2Example.xml"
  xml_full <- xml2::read_xml(file.path(filePath, fileName))
  measurementTS <- xml2::xml_find_all(xml_full, "//wml2:MeasurementTimeseries")
  UserData <- importWaterML2(measurementTS)
  expect_is(UserData$value, "numeric")
})


context("importWQP_noCRAN")

test_that("External WQP tests", {
  testthat::skip_on_cran()

  url2 <- constructWQPURL("USGS-01594440", "01075", "", "", legacy = TRUE)
  rawSample2 <- suppressWarnings(importWQP(url2))
  expect_is(rawSample2$ActivityStartDateTime, "POSIXct")
})

#' General Data Import from Water Quality Portal
#'
#' Imports data from Water Quality Portal web service. This function gets the data from here:
#' <https://www.waterqualitydata.us>.
#'
#' This function uses \dots as a query input, which can be very flexible, but also
#' has a steeper learning curve. For a quick overview, scroll down to the Examples
#' in this help file to see many query options.
#'
#' @details
#'
#' There are currently 10 legacy options
#' for data provided by the Water Quality Portal:
#'
#'
#' Legacy:
#' \tabular{lll}{
#' WQP Radio Button \tab service argument \tab Base URL  \cr
#' Sample Results \tab Result \tab /data/Result/search \cr
#' Site Data Only \tab Station \tab /data/Station/search \cr
#' Sampling Activity \tab Activity \tab /data/Activity/search \cr
#' Sampling Activity Metrics \tab ActivityMetric \tab /data/ActivityMetric/search \cr
#' Site Summary (not advertised on WQP)  \tab SiteSummary \tab /data/summary/monitoringLocation/search \cr
#' Project Data \tab Project \tab /data/Project/search \cr
#' Project Monitoring Location Weighting Data \tab ProjectMonitoringLocationWeighting \tab /data/ProjectMonitoringLocationWeighting/search \cr
#' Result Detection Quantitation Limit Data \tab ResultDetectionQuantitationLimit \tab /data/ResultDetectionQuantitationLimit/search \cr
#' Biological Habitat Metrics \tab BiologicalMetric \tab /data/BiologicalMetric/search \cr
#' Organization Data \tab Organization \tab /data/Organization/search \cr
#' }
#'
#' There are 4 WQX3 options. These are still in-development,
#' and should be used with caution.
#'
#' \tabular{llll}{
#' WQP Radio Button \tab service argument \tab Base URL \tab dataProfile \cr
#' Monitoring Locations \tab StationWQX3 \tab /wqx3/Station/search \tab \cr
#' Full Physical Chemical \tab ResultWQX3 \tab /wqx3/Result/search \tab fullPhysChem \cr
#' Narrow \tab ResultWQX3 \tab /wqx3/Result/search \tab narrow \cr
#' Basic Physical Chemical \tab ResultWQX3 \tab /wqx3/Result/search \tab basicPhysChem \cr
#' Sampling Activity \tab ActivityWQX3 \tab /wqx3/Activity/search \cr
#' }
#'
#'
#' @param \dots see <https://www.waterqualitydata.us/webservices_documentation> for a complete list of options.
#' A list of arguments can also be supplied. For more information see the above
#' description for this help file. One way to figure out how to construct a WQP query is to go to the "Advanced"
#' form in the Water Quality Portal. Use the form to discover what parameters are available. Once the query is
#' set in the form, scroll down to the "Query URL". You will see the parameters
#' after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
#' in the Characteristic Group dropdown, you will see characteristicType=Nutrient
#' in the Query URL. The corresponding argument for dataRetrieval is
#' characteristicType = "Nutrient". dataRetrieval users do not need to include
#' mimeType, and providers is optional (these arguments are picked automatically).
#' @param service character. See Details for more information.
#' @param querySummary logical to only return the number of records and unique sites that
#' will be returned from this query. Choosing TRUE is deprecated, readWQPsummary
#' is recommended instead.
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the
#' data's provided tz_cd column.
#' Possible values to provide are "America/New_York","America/Chicago",
#' "America/Denver","America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings
#' time: "America/Honolulu",
#' "America/Jamaica","America/Managua","America/Phoenix", and "America/Metlakatla".
#' See also  `OlsonNames()`
#' for more information on time zones.
#' @param ignore_attributes logical to choose to ignore fetching site and status
#' attributes. Default is `FALSE`.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character.
#' @keywords data import WQP web service
#' @return A data frame, the specific columns will depend on the "service" and/or "dataProfile".
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' headerInfo \tab data.frame \tab A data frame returned from the WQP status service \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#' @export
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#'
#' # Legacy:
#' #nameToUse <- "pH"
#' #pHData <- readWQPdata(siteid = "USGS-04024315",
#' #                       characteristicName = nameToUse)
#' #ncol(pHData)
#' #attr(pHData, "siteInfo")
#' #attr(pHData, "queryTime")
#' #attr(pHData, "url")
#'
#' # WQX3:
#' #pHData_wqx3 <- readWQPdata(siteid = "USGS-04024315",
#' #                            characteristicName = nameToUse,
#' #                            service = "ResultWQX3",
#' #                            dataProfile = "basicPhysChem")
#' # attr(pHData_wqx3, "url")
#'
#' # More examples:
#' # querying by county
#' #DeWitt <- readWQPdata(
#' #   statecode = "Illinois",
#' #   countycode = "DeWitt",
#' #   characteristicName = "Nitrogen"
#' # )
#'
#' #attr(DeWitt, "url")
#'
#' #DeWitt_wqx3 <- readWQPdata(
#' #    statecode = "Illinois",
#' #    countycode = "DeWitt",
#' #    characteristicName = "Nitrogen",
#' #    service = "ResultWQX3",
#' #    dataProfile = "basicPhysChem",
#' #    ignore_attributes = TRUE)
#'
#' #attr(DeWitt_wqx3, "url")
#'
#' # Data profile: "Sampling Activity"
#' #activity <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   service = "Activity"
#' #)
#' #attr(activity, "url")
#'
#' # activity_wqx3 <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   service = "ActivityWQX3"
#' # )
#' # attr(activity_wqx3, "url")
#'
#' #Dane_activity <- readWQPdata(
#' #   statecode = "Wisconsin",
#' #   countycode = "Dane",
#' #   startDateLo = "2023-01-01",
#' #   startDateHi = "2023-12-31",
#' #   service = "Activity"
#' # )
#' #attr(Dane_activity, "url")
#'
#' # Dane_activity_wqx3 <- readWQPdata(
#' #   statecode = "Wisconsin",
#' #   countycode = "Dane",
#' #   startDateLo = "2023-01-01",
#' #   startDateHi = "2023-12-31",
#' #   service = "ActivityWQX3"
#' # )
#' # attr(Dane_activity_wqx3, "url")
#'
#' ########################################################
#' # Additional examples:
#'
#'
#' # Data profiles: "Organization Data"
#' #org_data <- readWQPdata(
#' #   statecode = "WI",
#' #   countycode = "Dane",
#' #   service = "Organization"
#' # )
#'
#' # Data profiles: "Project Data"
#' #project_data <- readWQPdata(
#' #   statecode = "WI",
#' #   countycode = "Dane",
#' #   service = "Project"
#' # )
#'
#' # Data profiles: "Project Monitoring Location Weighting Data"
#' #proj_mlwd <- readWQPdata(
#' #   statecode = "WI",
#' #   countycode = "Dane",
#' #   service = "ProjectMonitoringLocationWeighting"
#' # )
#'
#' # Data profiles: "Sample Results (physical/chemical metadata)"
#' # samp_data <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   dataProfile = "resultPhysChem",
#' #   service = "Result"
#' # )
#'
#' # Data profiles: "Sample Results (biological metadata)"
#' #samp_bio <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   dataProfile = "biological",
#' #   service = "Result"
#' # )
#'
#'
#' # Data profiles: "Sample Results (narrow)"
#' # samp_narrow <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   service = "Result",
#' #   dataProfile = "narrowResult"
#' # )
#'
#' # samp_narrow_wqx3 <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   service = "ResultWQX3",
#' #   dataProfile = "narrow"
#' # )
#'
#'
#' # Data profiles: "Sampling Activity"
#' #samp_activity <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   dataProfile = "activityAll",
#' #   service = "Activity"
#' #  )
#'
#' # Data profile: "Sampling Activity Metrics"
#' # act_metrics <- readWQPdata(
#' #   statecode = "WI",
#' #   countycode = "Dane",
#' #   service = "ActivityMetric"
#' # )
#'
#' # Data profile: "Result Detection Quantitation Limit Data"
#' # dl_data <- readWQPdata(
#' #   siteid = "USGS-04024315",
#' #   service = "ResultDetectionQuantitationLimit"
#' # )
#'
#' # other options:
#' # Phosphorus <- readWQPdata(
#' #   statecode = "WI", countycode = "Dane",
#' #   characteristicName = "Phosphorus",
#' #   startDateLo = "2023-01-01",
#' #   ignore_attributes = TRUE,
#' #   convertType = FALSE
#' # )
#'
#' #rawPHsites_legacy <- readWQPdata(siteid = c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#' #                         characteristicName = "pH",
#' #                         service = "Result",
#' #                         dataProfile = "narrowResult" )
#'
#' # rawPHsites <- readWQPdata(siteid = c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#' #                           characteristicName = "pH",
#' #                           service = "ResultWQX3",
#' #                           dataProfile = "narrow" )
#'
#' }
readWQPdata <- function(
  ...,
  service = "Result",
  querySummary = FALSE,
  tz = "UTC",
  ignore_attributes = FALSE,
  convertType = TRUE
) {
  tz <- match.arg(tz, OlsonNames())

  service <- match.arg(
    service,
    c(
      "Result",
      "Station",
      "Activity",
      "ActivityMetric",
      "SiteSummary",
      "Project",
      "ProjectMonitoringLocationWeighting",
      "ResultDetectionQuantitationLimit",
      "BiologicalMetric",
      "Organization",
      "ResultWQX3",
      "StationWQX3",
      "ActivityWQX3"
    ),
    several.ok = FALSE
  )

  legacy <- is_legacy(service)

  valuesList <- readWQPdots(..., legacy = legacy)
  values <- valuesList[["values"]]

  baseURL <- httr2::request(pkg.env[[service]])
  POST <- FALSE

  if (!legacy) {
    if (service == "ResultWQX3" & !"dataProfile" %in% names(values)) {
      baseURL <- httr2::req_url_query(baseURL, dataProfile = "fullPhysChem")
    }
    baseURL <- httr2::req_url_query(baseURL, !!!values, .multi = "explode")
  } else {
    if ("siteid" %in% names(values)) {
      if (length(values[["siteid"]]) > 1) {
        sites <- values[["siteid"]]
        POST <- nchar(paste0(sites, collapse = "")) > 2048

        baseURL <- get_or_post(
          baseURL,
          POST = POST,
          siteid = sites,
          .multi = function(x) paste0(x, collapse = ";")
        )
        values <- values[names(values) != "siteid"]
      }
    }
    baseURL <- get_or_post(baseURL, POST = POST, !!!values, .multi = "explode")
  }

  if (querySummary) {
    retquery <- getQuerySummary(baseURL)
    return(retquery)
  } else {
    retval <- importWQP(baseURL, tz = tz, convertType = convertType)

    if (is.null(retval)) {
      return(NULL)
    }

    attr(retval, "legacy") <- legacy

    if (
      !all(is.na(retval)) &&
        !ignore_attributes &&
        !service %in% c("Station", "StationWQX")
    ) {
      params <- convertLists(...)
      params <- params[!names(params) %in% c("dataProfile", "service")]
      retval <- create_WQP_attributes(retval, params)
    }

    attr(retval, "url") <- baseURL$url

    if (legacy) {
      wqp_message()
    } else {
      wqp_message_beta()
      attr(retval, "wqp-request-id") <- attr(
        retval,
        "headerInfo"
      )$`wqp-request-id`
    }

    return(retval)
  }
}


create_WQP_attributes <- function(retval, ...) {
  col_legacy <- c(
    "CharacteristicName", #legacy
    "ResultMeasure.MeasureUnitCode",
    "ResultSampleFractionText"
  )
  col_wqx3 <- c(
    "Result_Characteristic", #wqx3
    "Result_MeasureUnit",
    "Result_SampleFraction"
  )
  if (all(col_legacy %in% names(retval))) {
    retvalVariableInfo <- retval[, col_legacy]
    retvalVariableInfo <- unique(retvalVariableInfo)
    names(retvalVariableInfo) <- c(
      "characteristicName",
      "param_units",
      "valueType"
    )

    attr(retval, "variableInfo") <- retvalVariableInfo
  } else if (all(col_wqx3 %in% names(retval))) {
    retvalVariableInfo <- retval[, col_wqx3]
    retvalVariableInfo <- unique(retvalVariableInfo)
    names(retvalVariableInfo) <- c(
      "characteristicName",
      "param_units",
      "valueType"
    )

    attr(retval, "variableInfo") <- retvalVariableInfo
  }

  siteInfo <- suppressWarnings(whatWQPsites(
    ...,
    legacy = attr(retval, "legacy")
  ))
  attr(retval, "siteInfo") <- siteInfo

  if (!attr(retval, "legacy")) {
    attr(retval, "headerInfo") <- wqp_check_status(
      attr(retval, "headerInfo")$`wqp-request-id`
    )
    attr(retval, "queryTime") <- as.POSIXct(
      attr(retval, "headerInfo")[["requestStartTime"]],
      format = "%Y-%m-%dT%H:%M:%OS",
      tz = "UTC"
    )
  } else {
    attr(retval, "queryTime") <- Sys.time()
  }

  return(retval)
}

#' Get WQP service metadata
#'
#' The information from this request is only available for a
#' limited time after the original query from the WQP. In the
#' readWQPdata and readWQPqw functions, the results from this
#' function will be attached as an attribute to the data.
#'
#' @param wqp_request_id A character returned from the header
#' of a WQP request.
#' @return a list generated from the WQP describing what data
#' was returned.
#' @export
#'
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' #rawPcode <- readWQPqw("USGS-01594440", "01075",
#' #                       ignore_attributes = TRUE, legacy = FALSE)
#' #  headerInfo <- attr(rawPcode, "headerInfo")
#' #wqp_request_id <- headerInfo$`wqp-request-id`
#' #count_info <- wqp_check_status(wqp_request_id)
#' #count_info[["dataProviders"]]
#' }
wqp_check_status <- function(wqp_request_id) {
  id_url <- paste0(pkg.env[["status"]], wqp_request_id)
  counts_list <- get_nldi_sources(id_url)
  return(counts_list)
}

#' Instantaneous value data retrieval from USGS (NWIS)
#'
#' Imports data from NWIS web service.
#' Inputs to this function are USGS site ids, USGS parameter codes,
#' and start and end date. For a more complex query, use [readNWISdata()],
#' including an arguement service="uv".
#' Not all parameter codes are available for all data.
#' Use the function [whatNWISdata()] to discover what data
#' is available for a USGS site. The column data_type_cd with the values "uv"
#' returned from [whatNWISdata()]) are available from this service.
#'
#'
#' @param siteNumbers character USGS site number (or multiple sites).  This is usually an 8 digit number
#' @param parameterCd character USGS parameter code.  This is usually an 5 digit number.
#' @param startDate character starting date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the earliest possible record. Simple date arguments are specified in local time.
#' @param endDate character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the latest possible record. Simple date arguments are specified in local time.
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the data provided tz_cd column.
#' Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
#' "America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  `OlsonNames()`
#' for more information on time zones.
#' @keywords data import USGS web service
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#' dateTime \tab POSIXct \tab The date and time of the value converted to UTC \cr
#' tz_cd \tab character \tab The time zone code for dateTime \cr
#' code \tab character \tab Any codes that qualify the corresponding value\cr
#' value \tab numeric \tab The numeric value for the parameter \cr
#' }
#' Note that code and value are repeated for the parameters requested. The names are of the form:
#' X_D_P_S, where X is literal,
#' D is an option description of the parameter,
#' P is the parameter code,
#' and S is the statistic code (if applicable).
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
#' statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#'
#' @seealso [renameNWISColumns()], [importWaterML1()]
#' @export
#' @examplesIf is_dataRetrieval_user()
#' site_id <- "05114000"
#' parameterCd <- "00060"
#' startDate <- "2014-10-10"
#' endDate <- "2014-10-10"
#' \donttest{
#'
#' #rawData <- readNWISuv(site_id, parameterCd, startDate, endDate)
#'
#' #rawData_today <- readNWISuv(site_id, parameterCd, Sys.Date(), Sys.Date())
#'
#' }
readNWISuv <- function(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  tz = "UTC"
) {
  if (
    as.character(startDate) == "" || (as.Date(startDate) <= Sys.Date() - 120)
  ) {
    service <- "iv"
  } else {
    service <- "iv_recent"
  }

  .Deprecated(
    new = "read_waterdata_continuous",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_continuous."
  )

  url <- constructNWISURL(
    siteNumbers,
    parameterCd,
    startDate,
    endDate,
    service,
    format = "xml"
  )

  data <- importWaterML1(url, asDateTime = TRUE, tz = tz)

  return(data)
}

#' Peak flow data from USGS (NWIS)
#'
#' Reads peak flow from NWISweb.
#'
#' In some cases, the specific date of the peak data is not know. This function
#' will default to
#' converting complete dates to a "Date" object, and converting incomplete dates to
#' "NA". If those incomplete dates are
#' needed, set the `asDateTime` argument to FALSE. No dates will be converted to
#' R Date objects.
#'
#' @param siteNumbers character USGS site number(or multiple sites).  This is usually
#' an 8 digit number.
#' @param startDate character starting date for data retrieval in the form YYYY-MM-DD.
#' Default is "" which indicates
#' retrieval for the earliest possible record.
#' @param endDate character ending date for data retrieval in the form YYYY-MM-DD.
#' Default is "" which indicates
#' retrieval for the latest possible record.
#' @param asDateTime logical default to `TRUE`. When `TRUE`, the peak_dt column is converted
#' to a Date object, and incomplete dates are removed. When `FALSE`, no
#' columns are removed, but no dates are converted.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#' peak_dt \tab Date \tab Date of peak streamflow \cr
#' peak_tm \tab character \tab Time of peak streamflow as character \cr
#' peak_va \tab numeric \tab Annual peak streamflow value in cfs \cr
#' peak_cd \tab character \tab Peak Discharge-Qualification codes (see `comment`
#' for more information) \cr
#' gage_ht \tab numeric \tab Gage height for the associated peak streamflow in feet \cr
#' gage_ht_cd \tab character \tab Gage height qualification codes \cr
#' year_last_pk \tab numeric \tab Peak streamflow reported is the highest since this year \cr
#' ag_dt \tab Date \tab Date of maximum gage-height for water year
#' (if not concurrent with peak) \cr
#' ag_tm \tab character \tab Time of maximum gage-height for water year
#' (if not concurrent with peak) \cr
#' ag_gage_ht \tab numeric \tab maximum Gage height for water year in feet
#' (if not concurrent with peak) \cr
#' ag_gage_ht_cd \tab character \tab maximum Gage height code \cr
#' }
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' comment \tab character \tab Header comments from the RDB file \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' }
#' @seealso [constructNWISURL()], [importRDB1()]
#' @export
#' @examplesIf is_dataRetrieval_user()
#' #site_ids <- c("01594440", "040851325")
#' \donttest{
#' #data <- readNWISpeak(site_ids)
#' #data2 <- readNWISpeak(site_ids, asDateTime = FALSE)
#' #stations <- c("06011000")
#' #peakdata <- readNWISpeak(stations, convertType = FALSE)
#' }
readNWISpeak <- function(
  siteNumbers,
  startDate = "",
  endDate = "",
  asDateTime = TRUE,
  convertType = TRUE
) {
  .Deprecated(
    new = "read_waterdata_peaks",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_peaks."
  )
  # Doesn't seem to be a peak xml service
  url <- constructNWISURL(
    siteNumbers = siteNumbers,
    parameterCd = NA,
    startDate = startDate,
    endDate = endDate,
    service = "peak"
  )

  data <- importRDB1(url, asDateTime = asDateTime, convertType = convertType)

  if (nrow(data) > 0) {
    if (asDateTime && convertType) {
      if ("peak_dt" %in% names(data)) {
        if (
          any(nchar(as.character(data$peak_dt)) <= 7, na.rm = TRUE) ||
            any(grepl("[0-9]*-[0-9]*-00", data$peak_dt), na.rm = TRUE)
        ) {
          stop(
            "Not all dates could be converted to Date object. Use convertType=FALSE to retrieve the raw text"
          )
        } else {
          data$peak_dt <- as.Date(data$peak_dt, format = "%Y-%m-%d")
        }
        if (anyNA(data$peak_dt)) {
          message(
            "Some dates could not be converted to a valid date, and were returned as NA"
          )
        }
      }

      if ("ag_dt" %in% names(data)) {
        data$ag_dt <- as.Date(data$ag_dt, format = "%Y-%m-%d")
      }
    }

    siteInfo <- suppressWarnings(readNWISsite(siteNumbers))
    siteInfo <- merge(
      x = unique(data[, c("agency_cd", "site_no")]),
      y = siteInfo,
      by = c("agency_cd", "site_no"),
      all.x = TRUE
    )

    attr(data, "siteInfo") <- siteInfo
    attr(data, "variableInfo") <- NULL
    attr(data, "statisticInfo") <- NULL
  }

  return(data)
}

#' Rating table for an active USGS streamgage retrieval
#'
#' Reads current rating table for an active USGS streamgage from NWISweb.
#'
#' @param siteNumber character USGS site number.  This is usually an 8 digit number
#' @param type character can be "base", "corr", or "exsa"
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character
#' @return A data frame. If `type` is "base, " then the columns are
#' INDEP, typically the gage height, in feet; DEP, typically the streamflow,
#' in cubic feet per second; and STOR, where "*" indicates that the pair are
#' a fixed point of the rating curve. If `type` is "exsa, " then an
#' additional column, SHIFT, is included that indicates the current shift in
#' the rating for that value of INDEP. If `type` is "corr, " then the
#' columns are INDEP, typically the gage height, in feet; CORR, the correction
#' for that value; and CORRINDEP, the corrected value for CORR.\cr
#' If `type` is "base, " then the data frame has an attribute called "RATING"
#' that describes the rating curve is included.
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' comment \tab character \tab Header comments from the RDB file \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' RATING \tab character \tab Rating information \cr
#' }
#'
#' @note Not all active USGS streamgages have traditional rating curves that
#' relate flow to stage.
#' @seealso [constructNWISURL()], [importRDB1()]
#' @export
#' @examplesIf is_dataRetrieval_user()
#' site_id <- "01594440"
#' \donttest{
#' #data <- readNWISrating(site_id, "base")
#' #attr(data, "RATING")
#' }
readNWISrating <- function(siteNumber, type = "base", convertType = TRUE) {
  .Deprecated(
    new = "read_waterdata_ratings",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_ratings."
  )

  # No rating xml service
  url <- constructNWISURL(siteNumber, service = "rating", ratingType = type)

  data <- importRDB1(url, asDateTime = FALSE, convertType = convertType)

  if ("current_rating_nu" %in% names(data)) {
    data$current_rating_nu <- gsub(" ", "", data$current_rating_nu)
  }

  if (nrow(data) > 0) {
    if (type == "base") {
      Rat <- grep("//RATING ", comment(data), value = TRUE, fixed = TRUE)
      Rat <- sub("# //RATING ", "", Rat)
      Rat <- scan(text = Rat, sep = " ", what = "")
      attr(data, "RATING") <- Rat
    }

    siteInfo <- suppressWarnings(readNWISsite(siteNumbers = siteNumber))

    attr(data, "siteInfo") <- siteInfo
    attr(data, "variableInfo") <- NULL
    attr(data, "statisticInfo") <- NULL
  }

  return(data)
}


#' Site statistics retrieval from USGS (NWIS)
#'
#' Retrieves site statistics from the USGS Statistics Web Service beta.
#'
#' @param siteNumbers character USGS site number (or multiple sites).  This is usually an 8 digit number.
#' @param parameterCd character USGS parameter code.  This is usually a 5 digit number.
#' @param startDate character starting date for data retrieval in the form YYYY, YYYY-MM, or YYYY-MM-DD. Dates cannot
#' be more specific than the statReportType, i.e. startDate for monthly statReportTypes cannot include days, and annual
#' statReportTypes cannot include days or months.  Months and days are optional for the daily statReportType.
#' Default is "" which indicates retrieval for the earliest possible record.  For daily data, this indicates the
#' start of the period the statistics will be computed over.
#' @param endDate character ending date for data retrieval in the form YYYY, YYYY-MM, or YYYY-MM-DD. Default is ""
#' which indicates retrieval for the latest possible record.  For daily data, this
#' indicates the end of the period
#' the statistics will be computed over.  The same restrictions as startDate apply.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function will convert the data to
#' numerics based on a standard algorithm. Years, months, and days (if appliccable) are also returned as numerics
#' in separate columns.  If convertType is false, everything is returned as a character.
#' @param statReportType character time division for statistics: daily, monthly, or annual.  Default is daily.
#' Note that daily provides statistics for each calendar day over the specified range
#' of water years, i.e. no more than 366
#' data points will be returned for each site/parameter.  Use readNWISdata or readNWISdv for daily averages.
#' Also note that 'annual' returns statistics for the calendar year.  Use readNWISdata
#' for water years. Monthly and yearly
#' provide statistics for each month and year within the range indivually.
#' @param statType character type(s) of statistics to output for daily values.
#' Default is mean, which is the only option for monthly and yearly report types.
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#' parameter_cd \tab character \tab The USGS parameter code \cr
#'
#' Other columns will be present depending on statReportType and statType
#' }
#' @seealso [constructNWISURL()], [importRDB1()]
#' @export
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' # x1 <- readNWISstat(
#' #   siteNumbers = c("02319394"),
#' #   parameterCd = c("00060"),
#' #   statReportType = "annual"
#' # )
#'
#' # all the annual mean discharge data for two sites
#' #x2 <- readNWISstat(
#' #   siteNumbers = c("02319394", "02171500"),
#' #   parameterCd = c("00010", "00060"),
#' #   statReportType = "annual"
#' # )
#'
#' # Request p25, p75, and mean values for temperature and discharge for the 2000s
#' # Note that p25 and p75 were not available for temperature, and return NAs
#' #x <- readNWISstat(
#' #   siteNumbers = c("02171500"),
#' #   parameterCd = c("00010", "00060"),
#' #   statReportType = "daily",
#' #   statType = c("mean", "median"),
#' #   startDate = "2000", endDate = "2010"
#' # )
#' }
readNWISstat <- function(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  convertType = TRUE,
  statReportType = "daily",
  statType = "mean"
) {
  .Deprecated(
    new = "read_waterdata_stats_por",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to either read_waterdata_stats_por or read_waterdata_stats_daterange."
  )

  message(new_nwis_message())
  # check for NAs in site numbers
  if (any(is.na(siteNumbers))) {
    siteNumbers <- siteNumbers[!is.na(siteNumbers)]
    if (length(siteNumbers) == 0) {
      stop("siteNumbers was all NAs")
    }
    warning("NAs were passed in siteNumbers; they were ignored")
  }
  url <- constructNWISURL(
    siteNumbers = siteNumbers,
    parameterCd = parameterCd,
    startDate = startDate,
    endDate = endDate,
    service = "stat",
    format = "rdb",
    statType = statType,
    statReportType = statReportType
  )

  data <- importRDB1(
    obs_url = url,
    asDateTime = TRUE,
    convertType = convertType
  )

  siteInfo <- suppressWarnings(readNWISsite(siteNumbers))

  if (nrow(data) > 0) {
    siteInfo <- merge(
      x = unique(data[, c("agency_cd", "site_no")]),
      y = siteInfo,
      by = c("agency_cd", "site_no"),
      all.x = TRUE
    )
  }

  attr(data, "siteInfo") <- siteInfo

  return(data)
}

#' Water use data retrieval from USGS (NWIS)
#'
#' Retrieves water use data from USGS Water Use Data for the Nation.
#'
#' @param stateCd could be character (full name, abbreviation, id), or numeric (id).
#' Only one is accepted per query.
#' @param countyCd could be character (name, with or without "County", or "ALL"),
#' numeric (id), or `NULL`, which will
#' return state or national data depending on the stateCd argument.  "ALL" may
#' also be supplied, which will return data
#' for every county in a state. Can be a vector of counties in the same state.
#' @param years integer Years for data retrieval. Must be years ending in 0 or 5.
#' Default is all available years.
#' @param categories character categories of water use.  Defaults to "ALL".
#' Specific categories must be supplied as two-
#' letter abbreviations as seen in the URL when using the NWIS water use web interface.  Note that
#' there are different codes for national and state level data.
#' @param convertType logical defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to
#' numerics based on a standard algorithm. Years, months, and days (if appliccable) are
#' also returned as numerics
#' in separate columns.  If convertType is false, everything is returned as a character.
#' @param transform logical only intended for use with national data.  Defaults to
#' `FALSE`, with data being returned as
#' presented by the web service.  If `TRUE`, data will be transformed and
#' returned with column names, which will reformat
#' national data to be similar to state data.
#' @return A data frame with at least the year of record, and all available
#' statistics for the given geographic parameters.
#' County and state fields will be included as appropriate.
#'
#' @export
readNWISuse <- function(
  stateCd,
  countyCd,
  years = "ALL",
  categories = "ALL",
  convertType = TRUE,
  transform = FALSE
) {
  .Deprecated(
    new = "read_waterdata_use_data in development",
    package = "dataRetrieval",
    msg = "NWIS servers for water use have been decommission. New functions are being developed."
  )
  return(NULL)
}

#' Set data endpoint
#'
#' access Indicate which dataRetrieval access code
#' you want to use options: `c('public','internal')`
#'
#' @param access code for data access. Options are: "public","internal","cooperator", or "USGS".
#' \itemize{
#' \item{"internal" represents Access=3 ...for a single water science center}
#' \item{"USGS" represents Access=2 ...for all water science centers}
#' \item{"cooperator" represents Access=1}
#' \item{"public" represents Access=0, public access}
#' }
#'
#' @author Luke Winslow, Jordan S Read
#'
#' @examples
#' \donttest{
#' setAccess("internal")
#'
#' setAccess("public")
#' }
#'
#' @export
setAccess <- function(access = "public") {
  access <- match.arg(access, c("public", "internal", "cooperator", "USGS"))

  if (access == "internal") {
    pkg.env$access <- "3"
    message("setting access to internal")
    warning(
      "Internal access is slated for decommision.
Please contact comptools@usgs.gov for more information."
    )
  } else if (access == "cooperator") {
    pkg.env$access <- "1"
    message("setting access to cooperator")
    warning(
      "Cooperator access is slated for decommision.
Please contact comptools@usgs.gov for more information."
    )
  } else if (access == "USGS") {
    pkg.env$access <- "2"
    message("setting access to all USGS Water Science Centers")
    warning(
      "Water Science Center access is slated for decommision.
Please contact comptools@usgs.gov for more information."
    )
  } else {
    pkg.env$access <- NULL
    message("setting access to public")
  }
  # nolint start: line_length_linter
  pkg.env$site <- "https://waterservices.usgs.gov/nwis/site/"
  pkg.env$iv <- "https://nwis.waterservices.usgs.gov/nwis/iv/"
  pkg.env$iv_recent <- "https://waterservices.usgs.gov/nwis/iv/"
  pkg.env$dv <- "https://waterservices.usgs.gov/nwis/dv/"
  pkg.env$peak <- "https://nwis.waterdata.usgs.gov/usa/nwis/peak/"
  pkg.env$rating <- "https://waterdata.usgs.gov/nwisweb/get_ratings/"
  pkg.env$stat <- "https://waterservices.usgs.gov/nwis/stat/"
  pkg.env$pCode <- "https://help.waterdata.usgs.gov/code/parameter_cd_query"
  pkg.env$pCodeSingle <- "https://help.waterdata.usgs.gov/code/parameter_cd_nm_query"

  pkg.env$Result <- "https://www.waterqualitydata.us/data/Result/search"
  pkg.env$Station <- "https://www.waterqualitydata.us/data/Station/search"
  pkg.env$Activity <- "https://www.waterqualitydata.us/data/Activity/search"
  pkg.env$ActivityMetric <- "https://www.waterqualitydata.us/data/ActivityMetric/search"
  pkg.env$SiteSummary <- "https://www.waterqualitydata.us/data/summary/monitoringLocation/search"
  pkg.env$Project <- "https://www.waterqualitydata.us/data/Project/search"
  pkg.env$ProjectMonitoringLocationWeighting <- "https://www.waterqualitydata.us/data/ProjectMonitoringLocationWeighting/search"
  pkg.env$ResultDetectionQuantitationLimit <- "https://www.waterqualitydata.us/data/ResultDetectionQuantitationLimit/search"
  pkg.env$BiologicalMetric <- "https://www.waterqualitydata.us/data/BiologicalMetric/search"
  pkg.env$Organization <- "https://www.waterqualitydata.us/data/Organization/search"
  pkg.env$ResultWQX3 <- "https://www.waterqualitydata.us/wqx3/Result/search"
  pkg.env$StationWQX3 <- "https://www.waterqualitydata.us/wqx3/Station/search"
  pkg.env$ActivityWQX3 <- "https://www.waterqualitydata.us/wqx3/Activity/search"
  pkg.env$samplesData <- "https://api.waterdata.usgs.gov/samples-data/summary"
  pkg.env$status <- "https://www.waterqualitydata.us/wqx3/status/"

  pkg.env$NGWMN <- "https://cida.usgs.gov/ngwmn_cache/sos"
  # nolint end
}

#' renameColumns
#'
#' Rename columns coming back from NWIS data retrievals.  Daily and unit value columns
#' have names derived from their data descriptor, parameter, and statistic codes. This
#' function reads information from the header and the arguments in the call to
#' to rename those columns.
#'
#' @param rawData the daily- or unit-values datset retrieved from NWISweb.
#' @param p00010 the base name for parameter code 00010.
#' @param p00045 the base name for parameter code 00045.
#' @param p00060 the base name for parameter code 00060.
#' @param p00065 the base name for parameter code 00065.
#' @param p00095 the base name for parameter code 00095.
#' @param p00300 the base name for parameter code 00300.
#' @param p00400 the base name for parameter code 00400.
#' @param p62611 the base name for parameter code 62611.
#' @param p63680 the base name for parameter code 63680.
#' @param p72019 the base name for parameter code 72019.
#' @param \dots named arguments for the base name for any other parameter code. The
#' form of the name must be like pXXXXX, where XXXXX is the parameter code.
#' @return A dataset like `data` with selected columns renamed.
#' @note The following statistics codes are converted by `renameNWISColumns`.
#' \describe{
#' \item{00000}{Instantaneous Value, suffix: Inst}
#' \item{00001}{Maximum value, suffix: Max}
#' \item{00002}{Minimum value, suffix: Min}
#' \item{00003}{Mean value, no suffix}
#' \item{00006}{Sum of values, suffix: Sum}
#' \item{00007}{Modal value, suffix: Mode}
#' \item{00008}{Median value, suffix: Median}
#' \item{00012}{Equivalent mean value, suffix: EqMean}
#' \item{00021}{Tidal high-high value, suffix: HiHiTide}
#' \item{00022}{Tidal low-high value, suffix: LoHiTide}
#' \item{00023}{Tidal high-low value, suffix: HiLoTide}
#' \item{00024}{Tidal low-low value, suffix: LoLoTide}
#' }
#' @seealso [readNWISdv()], [readNWISuv()]
#' @keywords manip IO
#' @export
#' @examples
#' siteWithTwo <- "01480015"
#' startDate <- "2012-09-01"
#' endDate <- "2012-10-01"
#' \donttest{
#' twoResults <- readNWISdv(siteWithTwo, "00060", startDate, endDate)
#' names(twoResults)
#' renamedCols <- renameNWISColumns(twoResults)
#' names(renamedCols)
#' # Custom names:
#' newNames <- renameNWISColumns(twoResults, p00060 = "Discharge")
#' names(newNames)
#' }
renameNWISColumns <- function(
  rawData,
  p00010 = "Wtemp",
  p00045 = "Precip",
  p00060 = "Flow",
  p00065 = "GH",
  p00095 = "SpecCond",
  p00300 = "DO",
  p00400 = "pH",
  p62611 = "GWL",
  p63680 = "Turb",
  p72019 = "WLBLS",
  ...
) {
  Cnames <- names(rawData)

  Conv <- list(...)
  Conv$p00010 <- p00010
  Conv$p00060 <- p00060
  Conv$p00045 <- p00045
  Conv$p00065 <- p00065
  Conv$p00095 <- p00095
  Conv$p00300 <- p00300
  Conv$p00400 <- p00400
  Conv$p62611 <- p62611
  Conv$p63680 <- p63680
  Conv$p72019 <- p72019

  Conv$s00000 <- "Inst"
  Conv$s00001 <- "Max"
  Conv$s00002 <- "Min"
  Conv$s00003 <- ""
  Conv$s00006 <- "Sum"
  Conv$s00007 <- "Mode"
  Conv$s00008 <- "Median"
  Conv$s00012 <- "EqMean"
  Conv$s00021 <- "HiHiTide"
  Conv$s00022 <- "LoHiTide"
  Conv$s00023 <- "HiLoTide"
  Conv$s00024 <- "LoLoTide"

  dataColumns <- c(grep("X_", Cnames), grep("X\\d{2}", Cnames))
  dataColumnsChangedParam <- NULL
  for (i in dataColumns) {
    chunks <- strsplit(Cnames[i], "_")[[1]]

    # Pcodes:
    for (j in seq_along(chunks)) {
      if (paste0("p", chunks[j]) %in% names(Conv)) {
        chunks[j] <- as.character(Conv[paste0("p", chunks[j])])
        Cnames[i] <- paste(chunks, collapse = "_")
        dataColumnsChangedParam <- c(dataColumnsChangedParam, i)
        break
      }
    }
    # Stat codes:
    for (j in seq_along(chunks)) {
      if (paste0("s", chunks[j]) %in% names(Conv)) {
        chunks[j] <- as.character(Conv[paste0("s", chunks[j])])
        chunks <- chunks[chunks != ""]
        Cnames[i] <- paste(chunks, collapse = "_")
        break
      }
    }
  }

  Cnames[dataColumnsChangedParam] <- gsub(
    "X_",
    "",
    Cnames[dataColumnsChangedParam]
  )

  names(rawData) <- Cnames

  return(rawData)
}

#' Raw Data Import for Water Quality Portal
#'
#' Imports data from the Water Quality Portal.
#' This function gets the data from here: <https://www.waterqualitydata.us>. There
#' are four required input arguments: siteNumbers, parameterCd, startDate, and endDate.
#' parameterCd can either be a USGS 5-digit code, or a characteristic name. The sites can be
#' either USGS, or other Water Quality Portal offered sites. It is required to use the 'full'
#' site name, such as 'USGS-01234567'.
#'
#' @param siteNumbers character site number. This needs to include the full agency code prefix.
#' @param parameterCd vector of USGS 5-digit parameter code or characteristicNames.
#' Leaving this blank will return all of the measured values during the specified time period.
#' @param startDate character starting date for data retrieval in the form YYYY-MM-DD.
#' Default is "" which indicates
#' retrieval for the earliest possible record. Date arguments are always specified in local time.
#' @param endDate character ending date for data retrieval in the form YYYY-MM-DD.
#' Default is "" which indicates
#' retrieval for the latest possible record. Date arguments are always specified in local time.
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the
#' data provided tz_cd column.
#' Possible values to provide are "America/New_York","America/Chicago",
#' "America/Denver","America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings
#' time: "America/Honolulu",
#' "America/Jamaica","America/Managua","America/Phoenix", and "America/Metlakatla".
#' See also  `OlsonNames()`
#' for more information on time zones.
#' @param querySummary logical to look at number of records and unique sites that
#' will be returned from this query.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character.
#' @param ignore_attributes logical to choose to ignore fetching site and parameter
#' attributes. Default is `FALSE`.
#' @param legacy Logical. If TRUE, uses legacy WQP services. Default is TRUE.
#' Setting legacy = FALSE uses WQX3.0 WQP services, which are in-development, use with caution.
#' @keywords data import USGS web service
#' @return A data frame derived from the default data profile.
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#' @export
#' @seealso [readWQPdata()], [whatWQPsites()],
#' and [importWQP()]
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' #rawPcode <- readWQPqw("USGS-01594440", "01075", "", "")
#'
#' #attr(rawPcode, "siteInfo")
#' #attr(rawPcode, "queryTime")
#' #attr(rawPcode, "url")
#'
#' #rawCharacteristicName <- readWQPqw("WIDNR_WQX-10032762", "Specific conductance", "", "")
#' #pHsites_legacy <- readWQPqw(c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#' #                         "pH", "", "")
#' #ncol(pHsites_legacy)
#' #attr(pHsites_legacy, "url")
#'
#' # pHsites_modern <- readWQPqw(c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#' #                         "pH", "", "", legacy = FALSE)
#' # ncol(pHsites_modern)
#' # attr(pHsites_modern, "url")
#'
#' # nwisEx <- readWQPqw("USGS-04024000", c("34247", "30234", "32104", "34220"), "", "2022-12-20")
#'
#' # DO <- readWQPqw(siteNumbers = "USGS-05288705",
#' #                 parameterCd = "00300",
#' #                 convertType = FALSE)
#' }
readWQPqw <- function(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  tz = "UTC",
  legacy = TRUE,
  querySummary = FALSE,
  ignore_attributes = FALSE,
  convertType = TRUE
) {
  url <- constructWQPURL(siteNumbers, parameterCd, startDate, endDate, legacy)

  if (querySummary) {
    retquery <- getQuerySummary(url)
    return(retquery)
  } else {
    retval <- importWQP(url, tz = tz, convertType = convertType)
    if (is.null(retval)) {
      return(NULL)
    }
    attr(retval, "legacy") <- legacy

    if (legacy) {
      sites <- unique(retval$MonitoringLocationIdentifier)
    } else {
      sites <- unique(retval$Location_Identifier)
    }

    if (!all(is.na(retval)) && !ignore_attributes) {
      retval <- suppressMessages(create_WQP_attributes(retval, siteid = sites))
    }

    if (legacy) {
      wqp_message()
    } else {
      wqp_message_beta()
    }
    attr(retval, "url") <- url$url

    return(retval)
  }
}

#' Get USGS Monitoring Location Data
#'
#' @description `r get_description("monitoring-locations")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("monitoring-locations")$id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param agency_code `r get_ogc_params("monitoring-locations")$agency_code`
#' @param agency_name `r get_ogc_params("monitoring-locations")$agency_name`
#' @param monitoring_location_number `r get_ogc_params("monitoring-locations")$monitoring_location_number`
#' @param monitoring_location_name `r get_ogc_params("monitoring-locations")$monitoring_location_name`
#' @param district_code `r get_ogc_params("monitoring-locations")$district_code`
#' @param state_name `r get_ogc_params("monitoring-locations")$state_name`
#' @param county_code `r get_ogc_params("monitoring-locations")$county_code`
#' @param county_name `r get_ogc_params("monitoring-locations")$county_name`
#' @param country_code `r get_ogc_params("monitoring-locations")$country_code`
#' @param country_name `r get_ogc_params("monitoring-locations")$country_name`
#' @param state_code `r get_ogc_params("monitoring-locations")$state_code`
#' @param minor_civil_division_code `r get_ogc_params("monitoring-locations")$minor_civil_division_code`
#' @param site_type_code `r get_ogc_params("monitoring-locations")$site_type_code`
#' @param site_type `r get_ogc_params("monitoring-locations")$site_type`
#' @param hydrologic_unit_code `r get_ogc_params("monitoring-locations")$hydrologic_unit_code`
#' @param basin_code `r get_ogc_params("monitoring-locations")$basin_code`
#' @param altitude `r get_ogc_params("monitoring-locations")$altitude`
#' @param altitude_accuracy `r get_ogc_params("monitoring-locations")$altitude_accuracy`
#' @param altitude_method_code `r get_ogc_params("monitoring-locations")$altitude_method_code`
#' @param altitude_method_name `r get_ogc_params("monitoring-locations")$altitude_method_name`
#' @param vertical_datum `r get_ogc_params("monitoring-locations")$vertical_datum`
#' @param vertical_datum_name `r get_ogc_params("monitoring-locations")$vertical_datum_name`
#' @param horizontal_positional_accuracy_code `r get_ogc_params("monitoring-locations")$horizontal_positional_accuracy_code`
#' @param horizontal_positional_accuracy `r get_ogc_params("monitoring-locations")$horizontal_positional_accuracy`
#' @param horizontal_position_method_code `r get_ogc_params("monitoring-locations")$horizontal_position_method_code`
#' @param horizontal_position_method_name `r get_ogc_params("monitoring-locations")$horizontal_position_method_name`
#' @param original_horizontal_datum `r get_ogc_params("monitoring-locations")$original_horizontal_datum`
#' @param original_horizontal_datum_name `r get_ogc_params("monitoring-locations")$original_horizontal_datum_name`
#' @param drainage_area `r get_ogc_params("monitoring-locations")$drainage_area`
#' @param contributing_drainage_area `r get_ogc_params("monitoring-locations")$contributing_drainage_area`
#' @param time_zone_abbreviation `r get_ogc_params("monitoring-locations")$time_zone_abbreviation`
#' @param uses_daylight_savings `r get_ogc_params("monitoring-locations")$uses_daylight_savings`
#' @param construction_date `r get_ogc_params("monitoring-locations")$construction_date`
#' @param aquifer_code `r get_ogc_params("monitoring-locations")$aquifer_code`
#' @param national_aquifer_code `r get_ogc_params("monitoring-locations")$national_aquifer_code`
#' @param aquifer_type_code `r get_ogc_params("monitoring-locations")$aquifer_type_code`
#' @param well_constructed_depth `r get_ogc_params("monitoring-locations")$well_constructed_depth`
#' @param hole_constructed_depth `r get_ogc_params("monitoring-locations")$hole_constructed_depth`
#' @param depth_source_code `r get_ogc_params("monitoring-locations")$depth_source_code`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("monitoring-locations", "monitoring_location_id")`.
#' The default (`NA`) will return all columns of the data.
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-02238500"
#' site_info <- read_waterdata_monitoring_location(monitoring_location_id = site)
#'
#' site_slim <- read_waterdata_monitoring_location(
#'                              monitoring_location_id = c("USGS-05114000",
#'                                                         "USGS-09423350"),
#'                              properties = c("monitoring_location_id",
#'                                             "state_name",
#'                                             "county_name",
#'                                             "country_name"))
#'
#' site_slim_no_sf_slim <- read_waterdata_monitoring_location(state_name = "Wisconsin",
#'                                            properties = c("monitoring_location_id",
#'                                                           "state_name",
#'                                                           "county_name",
#'                                                           "country_name",
#'                                                           "site_type"),
#'                                            site_type = "Well",
#'                                            skipGeometry = TRUE)
#'
#' site_info_no_sf <- read_waterdata_monitoring_location(
#'      monitoring_location_id = site_slim_no_sf_slim$monitoring_location_id[1:1000],
#'      skipGeometry = TRUE)
#'
#' bbox_vals = c(-94.00, 35.0, -93.5, 35.5)
#' multi_site <- read_waterdata_monitoring_location(bbox = bbox_vals)
#'
#'
#' }
read_waterdata_monitoring_location <- function(
  monitoring_location_id = NA_character_,
  agency_code = NA_character_,
  agency_name = NA_character_,
  monitoring_location_number = NA_character_,
  monitoring_location_name = NA_character_,
  district_code = NA_character_,
  country_code = NA_character_,
  country_name = NA_character_,
  state_code = NA_character_,
  state_name = NA_character_,
  county_code = NA_character_,
  county_name = NA_character_,
  minor_civil_division_code = NA_character_,
  site_type_code = NA_character_,
  site_type = NA_character_,
  hydrologic_unit_code = NA_character_,
  basin_code = NA_character_,
  altitude = NA_character_,
  altitude_accuracy = NA_character_,
  altitude_method_code = NA_character_,
  altitude_method_name = NA_character_,
  vertical_datum = NA_character_,
  vertical_datum_name = NA_character_,
  horizontal_positional_accuracy_code = NA_character_,
  horizontal_positional_accuracy = NA_character_,
  horizontal_position_method_code = NA_character_,
  horizontal_position_method_name = NA_character_,
  original_horizontal_datum = NA_character_,
  original_horizontal_datum_name = NA_character_,
  drainage_area = NA_character_,
  contributing_drainage_area = NA_character_,
  time_zone_abbreviation = NA_character_,
  uses_daylight_savings = NA_character_,
  construction_date = NA_character_,
  aquifer_code = NA_character_,
  national_aquifer_code = NA_character_,
  aquifer_type_code = NA_character_,
  well_constructed_depth = NA_character_,
  hole_constructed_depth = NA_character_,
  depth_source_code = NA_character_,
  properties = NA_character_,
  bbox = NA,
  skipGeometry = NA,
  ...,
  limit = getOption("dataRetrieval.limit"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "monitoring-locations"
  output_id <- "monitoring_location_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  args[["convertType"]] <- FALSE

  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

.onAttach <- function(libname, pkgname) {
  if (!interactive()) {
    return()
  }
  dataRetrieval_version <- utils::packageVersion("dataRetrieval")
  token_message <- ""
  if (Sys.getenv("API_USGS_PAT") == "") {
    token_message <- "\nConsider adding an API_USGS_PAT for new USGS functions.
See: https://api.waterdata.usgs.gov/signup"
  }
  packageStartupMessage(
    "dataRetrieval ",
    dataRetrieval_version,
    "
Extended Documentation: https://doi-usgs.github.io/dataRetrieval
Learn about the new functions that are replacing NWIS functions here:
https://doi-usgs.github.io/dataRetrieval/articles/read_waterdata_functions.html",
    token_message
  )
}

#' Retrieval functions for USGS and EPA data
#'
#' \tabular{ll}{
#' Package: \tab dataRetrieval\cr
#' Type: \tab Package\cr
#' License: \tab Unlimited for this package, dependencies have more restrictive licensing.\cr
#' Copyright: \tab This software is in the public domain because it contains materials
#' that originally came from the United States Geological Survey, an agency of
#' the United States Department of Interior. \cr
#' LazyLoad: \tab yes\cr
#' }
#'
#' Retrieval functions for USGS and EPA hydrologic and water quality data.
#'
#' Please see <https://doi-usgs.github.io/dataRetrieval/> for more information.
#'
#' @name dataRetrieval
#' @aliases dataRetrieval-package
#' @docType package
#' @author Laura De Cicco \email{ldecicco@@usgs.gov}
#' @keywords internal
"_PACKAGE"

#' List of USGS parameter codes
#'
#' Complete list of USGS parameter codes as of Oct. 24, 2024.
#'
#' @name parameterCdFile
#' @return parameterData data frame with information about USGS parameters.
#'
#' \tabular{lll}{
#' Name \tab Type \tab Description\cr
#' parameter_cd \tab character \tab 5-digit USGS parameter code \cr
#' parameter_group_nm \tab character \tab USGS parameter group name\cr
#' parameter_nm \tab character \tab USGS parameter name\cr
#' casrn \tab character \tab Chemical Abstracts Service (CAS) Registry Number\cr
#' srsname \tab character \tab Substance Registry Services Name\cr
#' parameter_units \tab character \tab Parameter units\cr
#' }
#'
#'
#' @docType data
#' @export parameterCdFile
#' @examples
#' head(parameterCdFile[, 1:2])
NULL


#' Data to convert USGS parameter code to characteristic names
#'
#' Data pulled from Water Quality Portal on December 20, 2021. The data was pulled from
#' <https://www.waterqualitydata.us/Codes/public_srsnames/?mimeType=csv>.
#'
#' @name pCodeToName
#' @return pCodeToName data frame with information about USGS parameters and how they
#' relate to characteristic names (useful for WQP requests).
#'
#' \tabular{lll}{
#' Name \tab Type \tab Description\cr
#' parm_cd \tab character \tab 5-digit USGS parameter code \cr
#' description \tab character \tab Parameter description\cr
#' characteristicname \tab character \tab Characteristic Name \cr
#' measureunitcode \tab character \tab Parameter units\cr
#' resultsamplefraction \tab character \tab Result sample fraction text\cr
#' resulttemperaturebasis \tab character \tab Temperature basis information\cr
#' resultstatisticalbasis \tab character \tab Statistical basis\cr
#' resulttimebasis \tab character \tab Time basis\cr
#' resultweightbasis \tab character \tab Weight basis\cr
#' resultparticlesizebasis \tab character \tab Particle size basis\cr
#' last_rev_dt \tab character \tab Latest revision of information\cr
#' }
#' @docType data
#' @export pCodeToName
#' @keywords internal
#' @examples
#' head(pCodeToName[, 1:2])
NULL

#' US State Code Lookup Table
#'
#' Classic lookup table for states. Has been replaced in functions with
#' `check_waterdata_sample_params("states")`.
#'
#' @name stateCd
#' @return stateCd data frame.
#'
#' \tabular{lll}{
#' Name \tab Type \tab Description\cr
#' STATE \tab character \tab FIPS State Code  \cr
#' STUSAB \tab character \tab Official United States Postal Service (USPS) Code \cr
#' STATE_NAME \tab character \tab State Name \cr
#' STATENS \tab character \tab  Geographic Names Information System Identifier (GNISID)\cr
#' }
#' @docType data
#' @export stateCd
#' @keywords USGS stateCd
#' @examples
#' head(stateCd)
NULL

#' US County Code Lookup Table
#'
#' Classic lookup table for counties. Has been replaced in functions with
#' `check_waterdata_sample_params("counties")`.
#'
#' @name countyCd
#' @return countyCd data frame.
#'
#' \tabular{lll}{
#' Name \tab Type \tab Description\cr
#' STUSAB \tab character \tab State abbreviation \cr
#' STATE \tab character \tab two-digit ANSI code  \cr
#' COUNTY \tab character \tab three-digit county code \cr
#' COUNTY_NAME \tab character \tab County full name \cr
#' COUNTY_ID \tab character \tab County id \cr
#' }
#' @docType data
#' @export countyCd
#' @keywords USGS countyCd
#' @examples
#' head(countyCd)
NULL

## usethis namespace: start
#' @importFrom data.table data.table
#' @importFrom data.table :=
#' @importFrom data.table .SD
#' @importFrom data.table .BY
#' @importFrom data.table .N
#' @importFrom data.table .I
#' @importFrom data.table .GRP
#' @importFrom data.table .NGRP
#' @importFrom data.table .EACHI
## usethis namespace: end
NULL

# nolint start: commented_code_linter
# Here's how to incorporate the state_county.json into the historic
# sysdata.rda. The original data included some IDs that aren't in the json
# We're leaving those in for now because maybe it's helpful,
# even though to my knowledge it's not used in any dataRetrieval query.

# #Step 1: open the json file, parse out what's needed:
# x <- jsonlite::read_json("inst/extdata/state_county.json")
# states <- x[["US"]]
# state_df <- data.frame("STATE" = names(sapply(states$state_cd, function(x) x[[1]])),
#                        "STATE_NAME" = as.character(sapply(states$state_cd, function(x) x[[1]])))
# #Step 2: join it into the original stateCd data frame:
# # This preserves STATE_NAME, which is nice
#
# library(dplyr)
# state_df <- state_df %>%
#   left_join(stateCd, by = c("STATE", "STATE_NAME"))
#
# state_df$STUSAB[state_df$STATE_NAME == "Virgin Islands"] <- "VI"
#
# #Step 3: now get the county names/ids:
# y <- sapply(states$state_cd, function(x) x[["county_cd"]])
#
# county_df <- data.frame()
#
# for(st in state_df$STATE) {
#   county_nums <- names(y[[st]])
#   county_names <- as.character(unlist(y[[st]]))
#   county_df_st <- data.frame(STATE = st,
#                              COUNTY = county_nums,
#                              COUNTY_NAME = county_names)
#   county_df <- dplyr::bind_rows(county_df, county_df_st)
# }
#
# county_df_full <- county_df %>%
#   left_join(select(state_df, STUSAB, STATE), by = "STATE") %>%
#   left_join(select(countyCd, STATE, STUSAB, COUNTY, COUNTY_ID),
#             by = c("STATE", "STUSAB", "COUNTY"))
#
# countyCd <- county_df_full
# stateCd <- state_df
#
# offsetLibrary <- data.frame(
#   offset = c(5, 4, 6, 5, 7, 6, 8, 7, 9, 8, 10, 10, 0, 0, 0, 0),
#   code = c("EST", "EDT", "CST", "CDT", "MST", "MDT", "PST", "PDT",
#            "AKST", "AKDT", "HAST", "HST", "UTC", "", NA, "GMT")
# )
#
# services <- c("daily", "time-series-metadata",
#               "monitoring-locations", "latest-continuous",
#               "field-measurements", "latest-daily",
#               "continuous", "field-measurements-metadata",
#               "combined-metadata", "channel-measurements", "peaks")
#
# property_list <- list()
# for(service in services){
#   property_list[[service]] <- dataRetrieval:::get_properties_for_docs(service)
# }
#
# num_cols <- c("value", "contributing_drainage_area", "drainage_area",
#               "altitude_accuracy", "well_construction_depth",
#               "hole_construction_depth", "channel_flow", "channel_width",
#               "channel_area", "channel_velocity", "channel_location_distance")
#
# save(countyCd, stateCd, parameterCdFile, pCodeToName,
#      offsetLibrary, num_cols, property_list,
#      file = "R/sysdata.rda", compress = "xz")
# nolint end

#' add a water year column
#'
#' Add a column to the dataRetrieval data frame with the water year. WQP
#' queries will return a water year column for the start and end dates
#' of the data.
#'
#' @param rawData the daily- or unit-values datset retrieved from NWISweb. Must
#' have at least one of the following columns to add the new water year columns:
#' `dateTime`, `Date`, `ActivityStartDate`, or `ActivityEndDate`. The date column(s)
#' can be character, POSIXct, Date. They cannot be numeric.
#'
#' @return data.frame with an additional integer column with "WY" appended to the
#' date column name. For WQP, there will be 2 columns: `ActivityStartDateWY` and
#' `ActivityEndDateWY`.
#' @export
#'
#' @examples
#' \donttest{
#' nwisData <- readNWISdv("04085427", "00060", "2022-01-01", "2022-06-30")
#' nwisData <- addWaterYear(nwisData)
#'
#' wqpData <- readWQPqw("USGS-01594440", "01075", "", "")
#' wqpData <- addWaterYear(wqpData)
#' }
addWaterYear <- function(rawData) {
  allowedDateColNames <- c(
    "dateTime",
    "Date",
    "ActivityStartDate",
    "ActivityEndDate",
    "Activity_StartDate",
    "Activity_EndDate"
  )
  allowedWYColNames <- c(
    "waterYear",
    "waterYear",
    "ActivityStartWaterYear",
    "ActivityEndWaterYear",
    "Activity_StartDateWaterYear",
    "Activity_EndDateWaterYear"
  )
  names(allowedWYColNames) <- allowedDateColNames
  # only allow WY to be added if there is an appropriate date column
  if (all(!allowedDateColNames %in% names(rawData))) {
    stop("specified date column does not exist in supplied data frame")
  }

  # set the name of the date column(s) to use for calculating the WY &
  # if the WY column already exists, do not add another (rm that date col
  # from the list that will be looped over)
  dateColNames <- names(rawData)[names(rawData) %in% allowedDateColNames]
  dateColNames <- dateColNames[
    !allowedWYColNames[dateColNames] %in% names(rawData)
  ]

  for (dateCol in dateColNames) {
    dateColWY <- allowedWYColNames[dateCol]

    # calculate WY & add as new column
    rawData[[dateColWY]] <- calcWaterYear(rawData[[dateCol]])

    # move waterYear so that it is always comes right after dateTime
    dateCol_i <- which(names(rawData) == dateCol)
    dateColWY_i <- which(names(rawData) == dateColWY)
    everything_else <- which(!(names(rawData) %in% c(dateCol, dateColWY)))
    everything_else <- everything_else[
      !everything_else %in% c(1:dateCol_i, dateColWY_i)
    ]

    rawData <- rawData[, c(1:dateCol_i, dateColWY_i, everything_else)]
  }

  return(rawData)
}

#' Extract WY from a date
#'
#' Determine the correct water year based on a calendar date.
#'
#' @param dateVec vector of dates as character ("YYYY-DD-MM"),
#' Date, or POSIXct. Numeric does not work.
#'
#' @details This function calculates a water year based on the USGS
#' definition that a water year starts on October 1 of the year before,
#' and ends on September 30. For example, water year 2015 started on
#' 2014-10-01 and ended on 2015-09-30.
#'
#' @return numeric vector indicating the water year
#' @export
#' @examples
#' x <- seq(as.Date("2010-01-01"), as.Date("2010-12-31"), by = "month")
#' calcWaterYear(x)
#'
#' y <- c("2010-01-01", "1994-02", "1980", "2009-11-01", NA)
#' calcWaterYear(y)
calcWaterYear <- function(dateVec) {
  if (is.numeric(dateVec)) {
    message(
      "dateVec is numeric, with insufficient information to determine water year."
    )
    return(rep(NA, length(dateVec)))
  }

  dateTimeVec <- tryCatch(
    {
      as.POSIXlt(dateVec)
    },
    error = function(e) {
      date_vec <- tryCatch(
        {
          as.Date(dateVec)
        },
        error = function(e) {
          return(rep(NA, length(dateVec)))
        }
      )

      if (any(is.na(date_vec))) {
        dateVec <- as.character(dateVec)
        dateVec[grep("^(\\d{4}-\\d{2}$)", dateVec)] <- paste0(
          dateVec[grep("^(\\d{4}-\\d{2}$)", dateVec)],
          "-01"
        )
        dateVec <- as.Date(dateVec)
      }
      dateTimeVec <- as.POSIXlt(dateVec)
      return(dateTimeVec)
    }
  )

  # POSIXlt years start at 100, POSIXlt months start at 0
  calYear <- dateTimeVec$year + 1900
  calMon <- dateTimeVec$mon + 1

  # when the date is NA, it should not try to add 1
  whichPastOct <- calMon >= 10
  whichPastOct[is.na(whichPastOct)] <- FALSE

  # add one to the year if it is in October or after
  waterYear <- calYear
  waterYear[whichPastOct] <- calYear[whichPastOct] + 1

  return(waterYear)
}

context("Unit values")

test_that("Unit value data returns correct types", {
  testthat::skip_on_cran()
  skip_on_ci()
  siteNumber <- "USGS-05114000"
  parameterCd <- "00060"
  startDate <- "2014-10-10"
  endDate <- "2014-10-10"

  rawData <- read_waterdata_continuous(
    monitoring_location_id = siteNumber,
    parameter_code = parameterCd,
    time = c(startDate, endDate)
  )

  spreadOver120 <- read_waterdata_continuous(
    monitoring_location_id = siteNumber,
    parameter_code = parameterCd,
    time = c(as.Date(Sys.Date() - 200), Sys.Date())
  )

  expect_true(min(spreadOver120$time) < as.POSIXct(Sys.Date(), tz = "UTC"))

  recent_uv <- read_waterdata_continuous(
    monitoring_location_id = siteNumber,
    parameter_code = parameterCd,
    time = c(as.Date(Sys.Date() - 10), Sys.Date())
  )
  expect_equal(
    grep(
      x = attr(recent_uv, "request")[["url"]],
      pattern = "https://api.waterdata.usgs.gov/ogcapi/v0/collections/continuous"
    ),
    1
  )

  expect_equal(
    grep(
      x = attr(spreadOver120, "request")[["url"]],
      pattern = "https://api.waterdata.usgs.gov/ogcapi/v0/collections/continuous"
    ),
    1
  )

  # nolint start: line_length_linter
  expect_true(
    grepl(
      x = attr(rawData, "request")[["url"]],
      pattern = "monitoring_location_id=USGS-05114000"
    )
  )

  expect_true(
    grepl(
      x = attr(rawData, "request")[["url"]],
      pattern = "time=2014-10-10T00%3A00%3A00Z%2F2014-10-10T00%3A00%3A00Z"
    )
  )

  expect_true(
    grepl(
      x = attr(rawData, "request")[["url"]],
      pattern = "parameter_code=00060"
    )
  )

  expect_true(
    grepl(
      x = attr(rawData, "request")[["url"]],
      pattern = paste0(
        "https://api.waterdata.usgs.gov/ogcapi/",
        getOption("dataRetrieval.api_version"),
        "/collections/continuous/items"
      )
    )
  )

  # nolint end
  timeZoneChange <- read_waterdata_continuous(
    monitoring_location_id = c("04024430", "04024000"),
    parameter_code = parameterCd,
    time = c("2013-11-03", "2013-11-03")
  )

  expect_is(rawData$time, "POSIXct")
  expect_is(rawData$value, "numeric")
  # nolint start: line_length_linter
  expect_true(
    grepl(
      x = attr(rawData, "request")[["url"]],
      pattern = "time=2014-10-10T00%3A00%3A00Z%2F2014-10-10T00%3A00%3A00Z"
    )
  )
  # nolint end
  site <- "USGS-04087170"
  pCode <- "63680"
  startDate <- "2012-07-10"
  endDate <- "2012-07-17"

  # Example that use to have YSI.6136.UP and regular
  # so now has 2 timeseries ids
  dd_2 <- read_waterdata_continuous(
    monitoring_location_id = site,
    parameter_code = pCode,
    time = c(startDate, endDate)
  )

  expect_true(length(unique(dd_2$time_series_id)) == 2)
})

context("Peak, rating, meas, site")
test_that("peak, rating curves, surface-water measurements", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  siteNumbers <- c("USGS-01594440", "USGS-040851325")
  data <- read_waterdata_peaks(monitoring_location_id = siteNumbers)
  expect_true(ncol(data) > 10)

  # Rating curves:
  siteNumber <- "USGS-01594440"
  data <- read_waterdata_ratings(
    monitoring_location_id = siteNumber,
    file_type = "base"
  )
  expect_gt(length(comment(data[[1]])), 1)

  # Surface meas:
  siteNumbers <- c("USGS-01594440", "USGS-040851325")
  data <- read_waterdata_field_measurements(siteNumbers)
  expect_is(data$monitoring_location_id, "character")

  siteINFO_USGS <- read_waterdata_monitoring_location(
    monitoring_location_id = "USGS-05114000"
  )
  expect_is(siteINFO_USGS$agency_code, "character")
  expect_equal(siteINFO_USGS$monitoring_location_id, "USGS-05114000")

  siteINFOMulti_USGS <- read_waterdata_monitoring_location(
    monitoring_location_id = c("USGS-05114000", "USGS-09423350")
  )
  expect_true(nrow(siteINFOMulti_USGS) == 2)

  Meas07227500.ex <- read_waterdata_field_measurements(
    monitoring_location_id = "USGS-07227500"
  )
  expect_is(Meas07227500.ex$time, "POSIXct")

  expect_equal(
    nrow(read_waterdata_ts_meta(
      monitoring_location_id = "USGS-10312000",
      parameter_code = "50286"
    )),
    0
  )
  # This does come back empty because 50268 isn't at this site

  expect_equal(
    ncol(read_waterdata_ts_meta(
      monitoring_location_id = "USGS-10312000",
      parameter_code = "50286",
      properties = c("geometry", "id", "unit_of_measure", "parameter_name")
    )),
    4
  )

  siteID <- "USGS-263819081585801"
  gwl_1 <- read_waterdata_field_measurements(monitoring_location_id = siteID)
  expect_equal(unique(gwl_1$monitoring_location_id), siteID)

  # No data:
  stations <- "06011000"
  expect_message(readNWISpeak(
    stations,
    startDate = "2024-08-01",
    endDate = "2024-08-31",
    convertType = FALSE
  ))
})

test_that("read_waterdata_daily", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  siteNumber <- "USGS-04085427"
  startDate <- "2012-01-01"
  endDate <- "2012-06-30"
  pCode <- "00060"

  raw_waterdata_daily <- read_waterdata_daily(
    monitoring_location_id = siteNumber,
    parameter_code = pCode,
    time = c(startDate, endDate)
  )
  expect_is(raw_waterdata_daily$time, "Date")

  raw_waterdata_daily_no_start <- read_waterdata_daily(
    monitoring_location_id = siteNumber,
    parameter_code = pCode,
    time = c(NA, endDate)
  )
  expect_equal(
    max(raw_waterdata_daily_no_start$time),
    max(raw_waterdata_daily$time)
  )
  expect_lt(
    min(raw_waterdata_daily_no_start$time),
    min(raw_waterdata_daily$time)
  )

  raw_waterdata_daily_no_end <- read_waterdata_daily(
    monitoring_location_id = siteNumber,
    parameter_code = pCode,
    time = c(startDate, NA)
  )
  expect_gt(max(raw_waterdata_daily_no_end$time), max(raw_waterdata_daily$time))
  expect_equal(
    min(raw_waterdata_daily_no_end$time),
    min(raw_waterdata_daily$time)
  )

  raw_waterdata_TempMeanMax <- read_waterdata_daily(
    monitoring_location_id = siteNumber,
    parameter_code = c("00010", "00060"),
    time = c(startDate, endDate),
    statistic_id = c("00001", "00003")
  )

  expect_true(length(unique(raw_waterdata_TempMeanMax$parameter_code)) == 2)
  expect_true(length(unique(raw_waterdata_TempMeanMax$statistic_id)) == 2)
  expect_true(
    length(unique(raw_waterdata_TempMeanMax$monitoring_location_id)) == 1
  )

  raw_waterdata_MultiSites <- read_waterdata_daily(
    monitoring_location_id = c("USGS-01491000", "USGS-01645000"),
    parameter_code = c("00010", "00060"),
    time = c(startDate, endDate),
    statistic_id = c("00001", "00003")
  )

  expect_true(
    length(unique(raw_waterdata_MultiSites$monitoring_location_id)) == 2
  )

  mixed_qualifiers <- read_waterdata_continuous(
    monitoring_location_id = "USGS-01648010",
    parameter_code = c("00010", "00095", "00400", "00300", "63680"),
    time = "2026-01-01T00:00:00Z/2026-03-13T00:00:00Z"
  )
  expect_all_true(
    unique(mixed_qualifiers$qualifier) %in% c("", "EQUIP", "LESSTHAN")
  )

  # Unclear if these qualifiers will remain persistent:
  # ts_id <- "fa44702f2bd64d9fa493608de9081cbe"
  # multi_q <- read_waterdata_continuous(time_series_id = ts_id)
  # expect_all_true(unique(multi_q$qualifier) %in% c("REGULATED, UNKNOWNREGULATION",
  #                                                  "ESTIMATED, REGULATED, UNKNOWNREGULATION",
  #                                                  "ICE, REGULATED, UNKNOWNREGULATION"))
})

test_that("WQP qw tests", {
  testthat::skip_on_cran()
  skip_on_ci()
  # nameToUse <- "Specific conductance"
  # pcodeToUse <- "00095"
  #
  # INFO_WQP <- readWQPqw(
  #   "USGS-04024315",
  #   pcodeToUse,
  #   startDate = "",
  #   endDate = "",
  #   legacy = FALSE
  # )
  # expect_is(INFO_WQP$Activity_StartDateTime, "POSIXct")
  #
  # INFO2 <- readWQPqw(
  #   "WIDNR_WQX-10032762",
  #   nameToUse,
  #   startDate = "",
  #   endDate = "",
  #   legacy = FALSE
  # )
  # expect_is(INFO2$Activity_StartDateTime, "POSIXct")
  #
  # df <- readWQPqw("USGS-04193500", parameterCd = "00665", legacy = FALSE)
  # expect_true(nrow(df) > 0)
  #
  # df2 <- readWQPqw("USGS-05427718", parameterCd = "all")
  # expect_true(nrow(df2) > 0)
  #
  # #Empty legacy:
  # df3 <- readWQPqw(
  #   siteNumbers = "USGS-385032115220501",
  #   parameterCd = "all",
  #   legacy = TRUE
  # )
  # expect_true(nrow(df3) == 0)
})


context("state tests")
test_that("state county tests", {
  testthat::skip_on_cran()

  fullName <- stateCdLookup("wi", "fullName")
  expect_equal(fullName, "Wisconsin")

  abbriev <- stateCdLookup("Wisconsin", "postal")
  expect_equal(abbriev, "WI")
  id <- stateCdLookup("WI", "id")
  expect_equal(id, 55)
  name <- stateCdLookup(55, "fullName")
  expect_equal(name, "Wisconsin")
  multipleStates <- stateCdLookup(c("West Virginia", "Wisconsin", 55, "MN"))
  expect_equal(multipleStates, c("WV", "WI", "WI", "MN"))

  fip <- countyCdLookup(state = "WI", county = "Dane")
  expect_equal(fip, "US:55:025")
  name <- countyCdLookup(state = "OH", county = 13, output = "fullName")
  expect_equal(name, "Belmont")
  index <- countyCdLookup(
    state = "Pennsylvania",
    county = "ALLEGHENY COUNTY",
    output = "fips"
  )
  expect_equal(index, "US:42:003")
  fromIDs <- countyCdLookup(state = 13, county = 5, output = "fullName")
  expect_equal(fromIDs, "Bacon")
})

context("water year column")

df_test <- data.frame(
  site_no = as.character(1:13),
  dateTime = seq(as.Date("2010-01-01"), as.Date("2011-01-31"), by = "months"),
  result_va = 1:13,
  stringsAsFactors = FALSE
)

test_that("addWaterYear works with Date, POSIXct, character, but breaks with numeric", {
  testthat::skip_on_cran()

  df_date <- df_test
  df_date_wy <- addWaterYear(df_date)
  expect_equal(ncol(df_date_wy), ncol(df_date) + 1)
  df_posixct <- df_test
  df_posixct$dateTime <- as.POSIXct(df_posixct$dateTime)
  df_posixct_wy <- addWaterYear(df_posixct)
  expect_equal(ncol(df_posixct_wy), ncol(df_posixct) + 1)
  df_char <- df_test
  df_char$dateTime <- as.character(df_char$dateTime)
  df_char_wy <- addWaterYear(df_char)
  expect_equal(ncol(df_char_wy), ncol(df_char) + 1)
  df_num <- df_test
  df_num$dateTime <- as.numeric(df_num$dateTime)
  dfnum2 <- addWaterYear(df_num)
  expect_equal(dfnum2$waterYear, rep(NA, 13))
})

test_that("addWaterYear works for each column name", {
  testthat::skip_on_cran()
  nwisqw_style <- df_test
  nwisqw_style_wy <- addWaterYear(nwisqw_style)
  expect_equal(ncol(nwisqw_style_wy), ncol(nwisqw_style) + 1)

  nwisdata_style <- df_test
  names(nwisdata_style)[2] <- "Date"
  nwisdata_style_wy <- addWaterYear(nwisdata_style)
  expect_equal(ncol(nwisdata_style_wy), ncol(nwisdata_style) + 1)

  wqp_style <- df_test
  names(wqp_style)[2] <- "ActivityStartDate"
  wqp_style[["ActivityEndDate"]] <- wqp_style[["ActivityStartDate"]]
  wqp_style_wy <- addWaterYear(wqp_style)
  expect_equal(ncol(wqp_style_wy), ncol(wqp_style) + 2)

  userspecified_style <- df_test
  names(userspecified_style)[2] <- "MyDateCol"
  expect_error(
    addWaterYear(userspecified_style),
    "specified date column does not exist in supplied data frame"
  )
})

test_that("addWaterYear correctly calculates the WY and is numeric", {
  testthat::skip_on_cran()
  df_test_wy <- addWaterYear(df_test)
  expect_is(df_test_wy[["waterYear"]], "numeric")
  expect_true(all(df_test_wy[["waterYear"]][1:9] == 2010))
  expect_true(all(df_test_wy[["waterYear"]][10:13] == 2011))
})

test_that("addWaterYear adds column next to dateTime", {
  testthat::skip_on_cran()
  df_test_wy <- addWaterYear(df_test)
  dateTime_col <- which(names(df_test_wy) == "dateTime")
  expect_equal(names(df_test_wy)[dateTime_col + 1], "waterYear")
})

test_that("addWaterYear can be used with pipes", {
  testthat::skip_on_cran()

  df_test_wy <- df_test
  df_test_wy <- addWaterYear(df_test_wy)
  expect_equal(ncol(df_test_wy), ncol(df_test) + 1)
})

test_that("addWaterYear doesn't add another WY column if it exists", {
  testthat::skip_on_cran()
  df_test_wy <- addWaterYear(df_test)
  expect_equal(ncol(df_test_wy), ncol(df_test) + 1)
  df_test_wy2 <- addWaterYear(df_test_wy)
  expect_equal(ncol(df_test_wy2), ncol(df_test_wy))
})

test_that("calcWaterYear can handle missing values", {
  dateVec <- seq(as.Date("2010-01-01"), as.Date("2011-01-31"), by = "months")
  dateVec[c(3, 7, 12)] <- NA
  wyVec <- dataRetrieval:::calcWaterYear(dateVec)

  expect_is(wyVec, "numeric")
  expect_true(all(is.na(wyVec[c(3, 7, 12)])))
})


context("construct_api_requests")
test_that("Construct USGS urls", {
  testthat::skip_on_cran()

  siteNumber <- "USGS-01594440"
  startDate <- "2024-01-01"
  endDate <- ""
  pCode <- c("00060", "00010")

  url_daily <- construct_api_requests(
    service = "daily",
    output_id = "daily_id",
    monitoring_location_id = siteNumber,
    parameter_code = pCode,
    time = c(startDate, endDate),
    statistic_id = c("00003", "00001"),
    limit = 10000
  )

  # nolint start: line_length_linter
  expect_true(
    grepl(x = url_daily$url, pattern = "parameter_code=00060,00010")
  )

  url_works <- dataRetrieval:::walk_pages(url_daily)
  expect_true(nrow(url_works) > 0)

  url_ts_meta <- construct_api_requests(
    monitoring_location_id = siteNumber,
    output_id = "time_series_id",
    parameter_code = pCode,
    service = "time-series-metadata",
    limit = 10000
  )

  expect_true(
    grepl(
      x = url_ts_meta$url,
      pattern = "collections/time-series-metadata/items"
    )
  )

  url_works_ts <- dataRetrieval:::walk_pages(url_ts_meta)
  expect_true(nrow(url_works_ts) > 0)

  url_ml <- construct_api_requests(
    id = siteNumber,
    output_id = "monitoring_location_id",
    service = "monitoring-locations",
    limit = 50000
  )

  expect_true(
    grepl(x = url_ml$url, pattern = "id=USGS-01594440")
  )

  url_works_ml <- dataRetrieval:::walk_pages(url_ml)
  expect_true(nrow(url_works_ml) > 0)

  # nolint end
})

context("Construct WQP urls")
test_that("Construct WQP urls", {
  testthat::skip_on_cran()

  site_id <- "01594440"
  startDate <- "1985-01-01"
  endDate <- ""
  pCode <- c("00060", "00010")

  url_wqp <- constructWQPURL(
    paste("USGS", site_id, sep = "-"),
    c("01075", "00029", "00453"),
    startDate,
    endDate,
    legacy = FALSE
  )
  # nolint start: line_length_linter
  expect_equal(
    url_wqp$url,
    "https://www.waterqualitydata.us/wqx3/Result/search?siteid=USGS-01594440&pCode=01075&pCode=00029&pCode=00453&startDateLo=01-01-1985&mimeType=csv&dataProfile=basicPhysChem"
  )

  # Multiple characteristicNames
  charNames <- c(
    "Temperature",
    "Temperature, sample",
    "Temperature, water",
    "Temperature, water, deg F"
  )

  obs_url_orig <- constructWQPURL(
    siteNumbers = c("IIDFG-41WSSPAHS", "USGS-02352560"),
    parameterCd = charNames,
    startDate = "",
    endDate = "",
    legacy = FALSE
  )

  expect_equal(
    obs_url_orig$url,
    "https://www.waterqualitydata.us/wqx3/Result/search?siteid=IIDFG-41WSSPAHS&siteid=USGS-02352560&characteristicName=Temperature&characteristicName=Temperature%2C%20sample&characteristicName=Temperature%2C%20water&characteristicName=Temperature%2C%20water%2C%20deg%20F&mimeType=csv&dataProfile=basicPhysChem"
  )

  # nolint end
})

context("Construct WQP urls")
test_that("Construct WQP urls", {
  siteNumber <- "01594440"
  startDate <- "1985-01-01"
  endDate <- ""
  pCode <- c("00060", "00010")
  url_wqp <- constructWQPURL(
    paste("USGS", siteNumber, sep = "-"),
    c("01075", "00029", "00453"),
    startDate,
    endDate,
    legacy = FALSE
  )

  # nolint start: line_length_linter
  expect_equal(
    url_wqp$url,
    "https://www.waterqualitydata.us/wqx3/Result/search?siteid=USGS-01594440&pCode=01075&pCode=00029&pCode=00453&startDateLo=01-01-1985&mimeType=csv&dataProfile=basicPhysChem"
  )

  # nolint end
})


context("pCode Stuff")
test_that("pCode Stuff", {
  testthat::skip_on_cran()

  paramINFO <- read_waterdata_parameter_codes(
    parameter_code = c("00060", "01075", "00931")
  )
  expect_equal(nrow(paramINFO), 3)
  expect_true(all(paramINFO$parameter_code %in% c("00060", "01075", "00931")))

  paramINFO <- read_waterdata_parameter_codes(
    parameter_code = c("00060", "01075", "00931", NA)
  )
  expect_equal(nrow(paramINFO), 3)
  expect_true(all(paramINFO$parameter_code %in% c("00060", "01075", "00931")))

  # pcode 12345 isn't a valid code:
  paramINFO <- read_waterdata_parameter_codes(c("12345"))
  expect_true(nrow(paramINFO) == 0)

  paramINFO <- read_waterdata_parameter_codes()
  expect_true(nrow(paramINFO) > 10000)
})

context("pCode Name Stuff")
test_that("pCode Stuff", {
  testthat::skip_on_cran()

  paramINFO <- pcode_to_name(c("00060", "01075", "00931", NA))
  expect_equal(nrow(paramINFO), 4)
  expect_equal(paramINFO$parm_cd, c("00060", "01075", "00931", NA))

  # pcode 12345 isn't a valid code:
  expect_warning(paramINFO <- pcode_to_name(c("12345")))
  expect_warning(
    paramINFO <- pcode_to_name(c("00060", "01075", "12345", NA_character_))
  )
  expect_equal(nrow(paramINFO), 4)
  expect_equal(paramINFO$parm_cd, c("00060", "01075", "12345", NA_character_))

  expect_equal(paramINFO$description[3:4], c(NA_character_, NA_character_))

  paramINFO <- pcode_to_name("all")
  expect_true(nrow(paramINFO) > 19000)
  expect_equal(
    attr(paramINFO, "url"),
    "https://www.waterqualitydata.us/Codes/public_srsnames/?mimeType=json"
  )
})

context("Smart errors, warnings")
test_that("bad_properties", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  expect_error(read_waterdata_daily(
    monitoring_location_id = "USGS-02238500",
    parameter_code = c("00010"),
    time = c("2021-01-01", "2022-01-01"),
    properties = c("value", "time", "blah")
  ))

  # No paging
  dv_data_quick <- read_waterdata_daily(
    monitoring_location_id = "USGS-02238500",
    parameter_code = "00060",
    no_paging = TRUE
  )

  expect_type(dv_data_quick$parameter_code, "character")
  expect_is(dv_data_quick$time, "Date")
  expect_equal(dv_data_quick$parameter_code[1], "00060")

  # Empty result:
  expect_message(read_waterdata_daily(
    monitoring_location_id = "USGS-02238500",
    parameter_code = c("00010"),
    time = c("2021-01-01", "2022-01-01"),
    no_paging = TRUE
  ))

  empty_return <- read_waterdata_daily(
    monitoring_location_id = "USGS-02238500",
    parameter_code = c("00010"),
    time = c("2021-01-01", "2022-01-01")
  )
  expect_true(nrow(empty_return) == 0)

  empty_return2 <- read_waterdata_daily(
    monitoring_location_id = "USGS-02238500",
    parameter_code = c("00010"),
    time = c("2021-01-01", "2022-01-01"),
    no_paging = TRUE
  )

  expect_true(nrow(empty_return2) == 0)
})

#' Generalized USGS Water Meta Data API retrieval function
#'
#' Function to get metadata from Water Data API. These are useful to get the
#' human readable words and other metadata associated with USGS codes.
#'
#' @export
#' @param collection character, can be any existing collection such
#' as "parameter-codes", "agency-codes", "altitude-datums", "aquifer-codes",
#' "aquifer-types", "coordinate-accuracy-codes", "coordinate-datum-codes",
#' "coordinate-method-codes", "hydrologic-unit-codes", "medium-codes",
#' "national-aquifer-codes", "reliability-codes", "site-types", "statistic-codes",
#' "topographic-codes", "time-zone-codes".
#' @param \dots Optional arguments to pass to the query. Available parameters
#' can be found with the \code{get_ogc_params} function.
#' @param limit The optional limit parameter is used to control the subset of the
#' selected features that should be returned in each page. The maximum allowable
#' limit is 50000. It may be beneficial to set this number lower if your internet
#' connection is spotty. The default (`NA`) will set the limit to the maximum
#' allowable limit for the service.
#' @param attach_request logical, defaults to `r getOption("dataRetrieval.attach_request")`.
#' If set to `TRUE`, the full request sent to the Water Data API is attached
#' as an attribute to the data set.
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' agency_codes <- read_waterdata_metadata("agency-codes")
#' altitude_datums <- read_waterdata_metadata("altitude-datums")
#' aquifer_codes <- read_waterdata_metadata("aquifer-codes")
#' aquifer_types <- read_waterdata_metadata("aquifer-types")
#' counties <- read_waterdata_metadata("counties")
#' countries <- read_waterdata_metadata("countries")
#' us_counties <- read_waterdata_metadata("counties", country_code = "US")
#' coordinate_accuracy_codes <- read_waterdata_metadata("coordinate-accuracy-codes")
#' coordinate_datum_codes <- read_waterdata_metadata("coordinate-datum-codes")
#' coordinate_method_codes <- read_waterdata_metadata("coordinate-method-codes")
#' huc_codes <- read_waterdata_metadata("hydrologic-unit-codes")
#' methods <- read_waterdata_metadata("methods")
#' method_categories <- read_waterdata_metadata("method-categories")
#' method_citations <- read_waterdata_metadata("method-citations")
#' citations <- read_waterdata_metadata("citations")
#' national_aquifer_codes <- read_waterdata_metadata("national-aquifer-codes")
#' parameter_codes <- read_waterdata_metadata("parameter-codes")
#' reliability_codes <- read_waterdata_metadata("reliability-codes")
#' site_types <- read_waterdata_metadata("site-types")
#' states <- read_waterdata_metadata("states")
#' us_states_territories <- read_waterdata_metadata("states", country_code = "US")
#' statistic_codes <- read_waterdata_metadata("statistic-codes")
#' topographic_codes <- read_waterdata_metadata("topographic-codes")
#' time_zone_codes <- read_waterdata_metadata("time-zone-codes")
#' time_zone_limited <- read_waterdata_metadata("time-zone-codes",
#'                         time_zone_description = c("Alaska", "Hawaii", "Pacific North America"))
#' }
read_waterdata_metadata <- function(
  collection,
  ...,
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  match.arg(collection, pkg.env$metadata)

  output_id <- names(pkg.env$metadata)[pkg.env$metadata == collection]

  args <- list(...)

  if (length(args) > 0) {
    possible_args <- get_ogc_params(collection)
    if (!all(names(args) %in% names(possible_args))) {
      wrong_args <- names(args)[!names(args) %in% names(possible_args)]
      wrong_args <- paste0(wrong_args, collapse = ", ")
      stop(paste0("Unknown argument: ", wrong_args))
    }
  }
  args[["attach_request"]] <- attach_request
  args[["limit"]] <- limit
  args[["convertType"]] <- FALSE
  args[["skipGeometry"]] <- NA
  args[["bbox"]] <- NA
  args[["no_paging"]] <- FALSE # drops id if TRUE
  args[["chunk_size"]] <- NA # Chunking doesn't make sense.

  return_list <- get_ogc_data(
    args = args,
    output_id = output_id,
    service = collection
  )

  return(return_list)
}

#' Get USGS Parameter Code Information
#'
#' @description `r get_description("parameter-codes")`
#'
#' @export
#' @param parameter_code `r get_ogc_params("parameter-codes")$id`
#' @param parameter_name `r get_ogc_params("parameter-codes")$parameter_name`
#' @param unit_of_measure `r get_ogc_params("parameter-codes")$unit_of_measure`
#' @param parameter_group_code `r get_ogc_params("parameter-codes")$parameter_group_code`
#' @param parameter_description `r get_ogc_params("parameter-codes")$parameter_description`
#' @param medium `r get_ogc_params("parameter-codes")$medium`
#' @param statistical_basis `r get_ogc_params("parameter-codes")$statistical_basis`
#' @param weight_basis `r get_ogc_params("parameter-codes")$weight_basis`
#' @param sample_fraction `r get_ogc_params("parameter-codes")$sample_fraction`
#' @param temperature_basis `r get_ogc_params("parameter-codes")$temperature_basis`
#' @param epa_equivalence `r get_ogc_params("parameter-codes")$epa_equivalence`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("parameter-codes", "parameter_code_id")`.
#' The default (`NA`) will return all columns of the data.
#' @param \dots Not used. Included to help differentiate official Water Data API arguments
#' from more seldom used, optional dataRetrieval-specific arguments.
#' @param limit The optional limit parameter is used to control the subset of the
#' selected features that should be returned in each page. The maximum allowable
#' limit is 50000. It may be beneficial to set this number lower if your internet
#' connection is spotty. The default (`NA`) will set the limit to the maximum
#' allowable limit for the service.
#' @param attach_request logical, defaults to `r getOption("dataRetrieval.attach_request")`.
#' If set to `TRUE`, the full request sent to the Water Data API is attached
#' as an attribute to the data set.
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' pcode <- "00060"
#' pcode_info <- read_waterdata_parameter_codes(parameter_code = pcode)
#'
#' pcodes <- read_waterdata_parameter_codes(parameter_code = c("00660", "00060"))
#'
#' # equivalent to read_waterdata_metadata("parameter-codes")
#' all_pcodes <- read_waterdata_parameter_codes()
#'
#' total_nutrients <- read_waterdata_parameter_codes(parameter_group_code = "NUT",
#'                                            sample_fraction = "Total")
#'
#' group_of_nutrients <- read_waterdata_parameter_codes(parameter_group_code = "NUT",
#'                                                     unit_of_measure = c("mg/l",
#'                                                     "mg/l as N", "mg/l NO3",
#'                                                     "mg/l asNO2"))
#'
#' }
read_waterdata_parameter_codes <- function(
  parameter_code = NA_character_,
  parameter_name = NA_character_,
  unit_of_measure = NA_character_,
  parameter_group_code = NA_character_,
  parameter_description = NA_character_,
  medium = NA_character_,
  statistical_basis = NA_character_,
  weight_basis = NA_character_,
  sample_fraction = NA_character_,
  temperature_basis = NA_character_,
  epa_equivalence = NA_character_,
  properties = NA_character_,
  ...,
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "parameter-codes"
  output_id <- "parameter_code"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  args[["convertType"]] <- FALSE
  args[["skipGeometry"]] <- NA
  args[["bbox"]] <- NA
  args[["no_paging"]] <- FALSE # drops id if TRUE
  args[["chunk_size"]] <- NA

  return_list <- get_ogc_data(
    args = args,
    output_id = output_id,
    service = service
  )

  return(return_list)
}

#' Construct request for USGS Samples Data
#'
#' This function creates the call for discrete water quality samples data
#' service described at <https://waterdata.usgs.gov/download-samples/>.
#' Note: all possible arguments are included, but it is strongly recommended
#' to only use the NECESSARY arguments. Leave unnecessary arguments as the default
#' NA.
#'
#' See also: <https://api.waterdata.usgs.gov/samples-data/docs>.
#'
#' @param monitoringLocationIdentifier A monitoring location identifier has two parts: the agency code
#' and the location number, separated by a dash (-). Location identifiers should be separated with commas,
#' for example: AZ014-320821110580701, CAX01-15304600, USGS-040851385. Location
#' numbers without an agency prefix are assumed to have the prefix USGS.
#' @param activityMediaName Sample media refers to the environmental medium that
#' was sampled or analyzed. See available options by running
#' `check_waterdata_sample_params("samplemedia")$activityMedia`.
#' @param siteTypeCode Site type code query parameter. See available
#' options by running `check_waterdata_sample_params("sitetype")$typeCode`.
#' @param boundingBox North and South are latitude values; East and West are longitude values.
#' A vector of 4 (west, south, east, north) is expected.
#' An example would be: c(-92.8, 44.2, -88.9, 46.0).
#' @param hydrologicUnit Hydrologic Unit Codes (HUCs) identify physical areas
#' within the US that drain to a certain portion of the stream network.
#' This filter accepts values containing 2, 4, 6, 8, 10 or 12 digits.
#' @param activityMediaName Sample media refers to the environmental medium that
#' was sampled or analyzed.
#' @param activityStartDateLower The service will return records with dates earlier
#' than the value entered for activityStartDateUpper. Can be an R Date object, or
#' a string with format YYYY-MM-DD. The logic is inclusive, i.e. it will also return
#' records that match the date.
#' @param activityStartDateUpper The service will return records with dates later
#' than the value entered for activityStartDateLower. Can be an R Date object, or
#' a string with format YYYY-MM-DD. The logic is inclusive, i.e. it will also return
#' records that match the date.
#' @param characteristicGroup Characteristic group is a broad category describing the sample.
#' See available options by running
#' `check_waterdata_sample_params("characteristicgroup")$characteristicGroup`.
#' @param characteristicUserSupplied Observed property is the USGS term for the
#' constituent sampled and the property name gives a detailed description of what
#' was sampled. Observed property is mapped to characteristicUserSupplied and replaces
#' the parameter name and pcode USGS
#' previously used to describe discrete sample data. Find more information in the
#' Observed Properties and Parameter Codes section of the Code Dictionary found here:
#' <https://waterdata.usgs.gov/code-dictionary/>.
#' @param characteristic Characteristic is a specific category describing the sample.
#' See available options by running
#' `check_waterdata_sample_params("characteristics")$characteristicName`.
#' @param stateFips State query parameter. To get a list of available state fips,
#' run `check_waterdata_sample_params("states")`. The "fips" can be created using the function
#' `stateCdLookup` - for example: `stateCdLookup("WI", "fips")`.
#' FIPs codes for states take the format:
#' CountryAbbrev:StateNumber, like US:55 for Wisconsin.
#' @param countyFips County query parameter. To get a list of available counties,
#' run `check_waterdata_sample_params("counties")`. The "Fips" can be created using the function
#' `countyCdLookup` - for example: `countyCdLookup("WI", "Dane", "fips")`
#' for Dane County, WI.
#' FIPs codes for counties take the format:
#' CountryAbbrev:StateNumber:CountyNumber, like US:55:025 for Dane County, WI.
#' @param countryFips Country query parameter. Do not set redundant parameters.
#' If another query parameter contains the country information, leave this parameter
#' set to the default NA. See available options by running `check_waterdata_sample_params("countries")`,
#' where the "id" field contains the value to use in the countryFips input.
#' @param projectIdentifier Project identifier query parameter. This information
#' would be needed from prior project information.
#' @param recordIdentifierUserSupplied Record identifier, user supplied identifier. This
#' information would be needed from the data supplier.
#' @param siteTypeName Site type name query parameter. See available
#' options by running `check_waterdata_sample_params("sitetype")$typeName`.
#' @param usgsPCode USGS parameter code. See available options by running
#' `check_waterdata_sample_params("characteristics")$parameterCode`.
#' @param pointLocationLatitude Latitude for a point/radius query (decimal degrees). Must be used
#' with pointLocationLongitude and pointLocationWithinMiles.
#' @param pointLocationLongitude Longitude for a point/radius query (decimal degrees). Must be used
#' with pointLocationLatitude and pointLocationWithinMiles.
#' @param pointLocationWithinMiles Radius for a point/radius query. Must be used
#' with pointLocationLatitude and pointLocationLongitude
#' @param dataType Options include: "Results", "Monitoring locations", "Activities",
#' "Projects", and "Organizations".
#' @param dataProfile Profile depends on type. Options for "results" dataType are:
#' "fullphyschem", "basicphyschem", "fullbio", "basicbio", "narrow",
#' "resultdetectionquantitationlimit", "labsampleprep", "count". Options for "locations" are:
#' "site" and "count". Options for "activities" are "sampact", "actmetric", "actgroup",
#' and "count". Options for "projects" are:
#' "project" and "projectmonitoringlocationweight". Options for "organizations" are:
#' "organization" and "count".
#' @export
#' @keywords internal
#' @return data frame returned from web service call.
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' req <- construct_waterdata_sample_request(
#'                monitoringLocationIdentifier = "USGS-04074950",
#'                characteristicUserSupplied = "pH, water, unfiltered, field")
#' rawData <- importWQP(req)
#'
#' }
construct_waterdata_sample_request <- function(
  monitoringLocationIdentifier = NA,
  siteTypeCode = NA,
  boundingBox = NA,
  hydrologicUnit = NA,
  activityMediaName = NA,
  characteristicGroup = NA,
  characteristic = NA,
  characteristicUserSupplied = NA,
  activityStartDateLower = NA,
  activityStartDateUpper = NA,
  countryFips = NA,
  stateFips = NA,
  countyFips = NA,
  projectIdentifier = NA,
  recordIdentifierUserSupplied = NA,
  siteTypeName = NA,
  usgsPCode = NA,
  pointLocationLatitude = NA,
  pointLocationLongitude = NA,
  pointLocationWithinMiles = NA,
  dataType = "results",
  dataProfile = NA
) {
  dataType <- match.arg(
    dataType,
    c("results", "locations", "activities", "projects", "organizations"),
    several.ok = FALSE
  )

  baseURL <- httr2::request("https://api.waterdata.usgs.gov") |>
    httr2::req_url_path_append("samples-data") |>
    httr2::req_url_query(mimeType = "text/csv") |>
    add_api_token()

  switch(
    dataType,
    results = {
      available_profiles <- c(
        "fullphyschem",
        "basicphyschem",
        "fullbio",
        "basicbio",
        "narrow",
        "resultdetectionquantitationlimit",
        "labsampleprep",
        "count"
      )
      baseURL <- httr2::req_url_path_append(baseURL, "results")
    },
    locations = {
      available_profiles <- c("site", "count")

      baseURL <- httr2::req_url_path_append(baseURL, "locations")
    },
    activities = {
      available_profiles <- c("sampact", "actmetric", "actgroup", "count")

      baseURL <- httr2::req_url_path_append(baseURL, "activities")
    },
    projects = {
      available_profiles <- c("project", "projectmonitoringlocationweight")

      baseURL <- httr2::req_url_path_append(baseURL, "projects")
    },
    organizations = {
      available_profiles <- c("organization", "count")

      baseURL <- httr2::req_url_path_append(baseURL, "organizations")
    }
  )

  dataProfile <- check_profile(dataProfile, available_profiles)
  dataProfile <- match.arg(dataProfile, available_profiles, several.ok = FALSE)

  baseURL <- httr2::req_url_path_append(baseURL, dataProfile)

  if (all(!is.na(siteTypeCode))) {
    siteTypeCode <- match.arg(
      siteTypeCode,
      check_waterdata_sample_params("sitetype")$typeCode,
      several.ok = TRUE
    )
  }

  if (all(!is.na(activityMediaName))) {
    activityMediaName <- match.arg(
      activityMediaName,
      check_waterdata_sample_params("samplemedia")$activityMedia,
      several.ok = TRUE
    )
  }

  if (all(!is.na(characteristicGroup))) {
    characteristicGroup <- match.arg(
      characteristicGroup,
      check_waterdata_sample_params("characteristicgroup")$characteristicGroup,
      several.ok = TRUE
    )
  }

  if (all(!is.na(countryFips))) {
    countryFips <- match.arg(
      countryFips,
      check_waterdata_sample_params("countries")$countryCode,
      several.ok = TRUE
    )
  }

  if (all(!is.na(siteTypeName))) {
    siteTypeName <- match.arg(
      siteTypeName,
      check_waterdata_sample_params("sitetype")$typeLongName,
      several.ok = TRUE
    )
  }

  if (all(!is.na(stateFips))) {
    states <- check_waterdata_sample_params("states")
    state_codes <- paste(states$countryCode, states$fipsCode, sep = ":")
    stateFips <- match.arg(stateFips, state_codes, several.ok = TRUE)
  }

  if (all(!is.na(countyFips))) {
    states <- check_waterdata_sample_params("states")
    state_codes <- paste(states$countryCode, states$fipsCode, sep = ":")
    counties <- check_waterdata_sample_params("counties")
    state_cd <- stats::setNames(states$fipsCode, states$stateAbbrev)
    county_codes <- paste(
      counties$countryCode,
      state_cd[counties$stateAbbrev],
      counties$countyCode,
      sep = ":"
    )

    countyFips <- match.arg(countyFips, county_codes, several.ok = TRUE)
  }

  check_radius <- sum(is.na(c(
    pointLocationLatitude,
    pointLocationLongitude,
    pointLocationWithinMiles
  )))

  if (!check_radius %in% c(3, 0)) {
    stop(
      "pointLocationLatitude, pointLocationLongitude, and pointLocationWithinMiles
           must all be defined, or none defined."
    )
  }

  baseURL <- explode_query(
    baseURL,
    POST = FALSE,
    list(
      hydrologicUnit = hydrologicUnit,
      projectIdentifier = projectIdentifier,
      recordIdentifierUserSupplied = recordIdentifierUserSupplied,
      monitoringLocationIdentifier = monitoringLocationIdentifier,
      siteTypeCode = siteTypeCode,
      activityMediaName = activityMediaName,
      characteristicGroup = characteristicGroup,
      characteristic = characteristic,
      characteristicUserSupplied = characteristicUserSupplied,
      countryFips = countryFips,
      siteTypeName = siteTypeName,
      usgsPCode = usgsPCode,
      stateFips = stateFips,
      countyFips = countyFips,
      pointLocationLatitude = pointLocationLatitude,
      pointLocationLongitude = pointLocationLongitude,
      pointLocationWithinMiles = pointLocationWithinMiles
    )
  )

  if (all(!is.na(activityStartDateLower))) {
    start <- checkWQPdates(list(
      activityStartDateLower = activityStartDateLower
    ))
    start <- as.character(as.Date(
      start$activityStartDateLower,
      format = "%m-%d-%Y"
    ))
    baseURL <- httr2::req_url_query(baseURL, activityStartDateLower = start)
  }

  if (all(!is.na(activityStartDateUpper))) {
    end <- checkWQPdates(list(activityStartDateUpper = activityStartDateUpper))
    end <- as.character(as.Date(
      end$activityStartDateUpper,
      format = "%m-%d-%Y"
    ))
    baseURL <- httr2::req_url_query(baseURL, activityStartDateUpper = end)
  }

  if (all(!is.na(boundingBox))) {
    baseURL <- httr2::req_url_query(
      baseURL,
      boundingBox = boundingBox,
      .multi = "comma"
    )
  }

  return(baseURL)
}

check_profile <- function(dataProfile, profile_convert) {
  if (is.na(dataProfile)) {
    message(paste0(
      "No profile specified, defaulting to '",
      profile_convert[1],
      "'\n",
      "Possible values are: \n",
      paste0(profile_convert, collapse = ", ")
    ))
    dataProfile <- profile_convert[1]
  } else {
    dataProfile <- match.arg(dataProfile, profile_convert, several.ok = FALSE)
  }
  return(dataProfile)
}

explode_query <- function(baseURL, POST = FALSE, x, multi = "explode") {
  if (!is.list(x)) {
    return(baseURL)
  }

  if (any(!is.na(x))) {
    x <- Filter(Negate(anyNA), x)
    if (POST) {
      baseURL <- httr2::req_body_json(req = baseURL, data = x)
    } else {
      baseURL <- httr2::req_url_query(baseURL, !!!x, .multi = multi)
    }
  }
  return(baseURL)
}


#' Check values from codeservice
#'
#' Call a service to check on values from:
#' <https://api.waterdata.usgs.gov/samples-data/codeservice/docs>.
#'
#' @param service Options are: "characteristicgroup", "states", "counties",
#' "countries", "sitetype", "samplemedia", "characteristics", "observedproperty"
#' @param ... Optional additional query arguments. Only "characteristics" and
#' "observedproperty" have additional parameters options.
#' @return List, structure depends on service.
#' @export
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' groups <- check_waterdata_sample_params("characteristicgroup")
#' states <- check_waterdata_sample_params("states")
#' countries <- check_waterdata_sample_params("countries")
#' counties <- check_waterdata_sample_params("counties")
#' sitetypes <- check_waterdata_sample_params("sitetype")
#' samplemedia <- check_waterdata_sample_params("samplemedia")
#' characteristics <- check_waterdata_sample_params("characteristics",
#'                                group = "Biological")
#' observedProperties <- check_waterdata_sample_params("observedproperty",
#'                                   text = "phosphorus")
#' ref_list <- check_waterdata_sample_params("reference-list")
#' }
check_waterdata_sample_params <- function(
  service = "characteristicgroup",
  ...
) {
  service_options <- c(
    "characteristicgroup",
    "states",
    "counties",
    "countries",
    "sitetype",
    "samplemedia",
    "characteristics",
    "observedproperty",
    "reference-list"
  )

  match.arg(service, choices = service_options, several.ok = FALSE)

  check_group_req <- httr2::request("https://api.waterdata.usgs.gov") |>
    httr2::req_url_path_append("samples-data") |>
    add_api_token()

  if (service != "reference-list") {
    check_group_req <- check_group_req |>
      httr2::req_url_path_append("codeservice")
  }

  check_group_req <- check_group_req |>
    httr2::req_url_path_append(service) |>
    httr2::req_user_agent(default_ua()) |>
    httr2::req_url_query(mimeType = "application/json")

  if (length(list(...)) > 0) {
    params <- list(...)
    extra_params <- c("group", "pageNumber", "pageSize", "text")
    match.arg(names(params), extra_params, several.ok = TRUE)
    check_group_req <- httr2::req_url_query(check_group_req, !!!params)
  }

  message("GET: ", check_group_req$url)

  check_group <- httr2::req_perform(check_group_req) |>
    httr2::resp_body_string() |>
    jsonlite::fromJSON()

  return(check_group$data)
}

#' USGS Samples Data
#'
#' This function creates the call and gets the data for discrete water quality samples data
#' service described at <https://waterdata.usgs.gov/download-samples/>.
#'
#' @inheritParams construct_waterdata_sample_request
#' @param tz character to set timezone attribute of datetime. Default is UTC
#' (properly accounting for daylight savings times based on the data's provided tz_cd column).
#' Possible values include "America/New_York","America/Chicago", "America/Denver","America/Los_Angeles",
#' "America/Anchorage","America/Honolulu","America/Jamaica","America/Managua",
#' "America/Phoenix", and "America/Metlakatla"
#' @param convertType logical, defaults to \code{TRUE}. If \code{TRUE}, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character.
#' @export
#' @rdname read_waterdata_samples
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' ph_data <- read_waterdata_samples(
#'                monitoringLocationIdentifier = "USGS-04074950",
#'                characteristicUserSupplied = "pH, water, unfiltered, field",
#'                activityStartDateUpper = "2000-01-01",
#'                dataProfile = "narrow")
#'
#' nameToUse <- "pH"
#' pHData <- read_waterdata_samples(monitoringLocationIdentifier = "USGS-04024315",
#'                          characteristic = nameToUse)
#' ncol(pHData)
#' attr(pHData, "url")
#' attr(pHData, "queryTime")
#'
#' summary_data <- read_waterdata_samples(monitoringLocationIdentifier = "USGS-04024315",
#'                                dataType = "projects")
#'
#' }
read_waterdata_samples <- function(
  monitoringLocationIdentifier = NA,
  siteTypeCode = NA,
  boundingBox = NA,
  hydrologicUnit = NA,
  activityMediaName = NA,
  characteristicGroup = NA,
  characteristic = NA,
  characteristicUserSupplied = NA,
  activityStartDateLower = NA,
  activityStartDateUpper = NA,
  countryFips = NA,
  stateFips = NA,
  countyFips = NA,
  projectIdentifier = NA,
  recordIdentifierUserSupplied = NA,
  siteTypeName = NA,
  usgsPCode = NA,
  pointLocationLatitude = NA,
  pointLocationLongitude = NA,
  pointLocationWithinMiles = NA,
  dataType = "results",
  dataProfile = NA,
  tz = "UTC",
  convertType = TRUE
) {
  request_url <- construct_waterdata_sample_request(
    monitoringLocationIdentifier = monitoringLocationIdentifier,
    siteTypeCode = siteTypeCode,
    boundingBox = boundingBox,
    hydrologicUnit = hydrologicUnit,
    activityMediaName = activityMediaName,
    characteristicGroup = characteristicGroup,
    characteristic = characteristic,
    characteristicUserSupplied = characteristicUserSupplied,
    activityStartDateLower = activityStartDateLower,
    activityStartDateUpper = activityStartDateUpper,
    countryFips = countryFips,
    stateFips = stateFips,
    countyFips = countyFips,
    projectIdentifier = projectIdentifier,
    recordIdentifierUserSupplied = recordIdentifierUserSupplied,
    siteTypeName = siteTypeName,
    usgsPCode = usgsPCode,
    pointLocationLatitude = pointLocationLatitude,
    pointLocationLongitude = pointLocationLongitude,
    pointLocationWithinMiles = pointLocationWithinMiles,
    dataType = dataType,
    dataProfile = dataProfile
  )

  df <- importWQP(request_url, tz = tz, convertType = convertType)
  attr(df, "url") <- request_url$url
  attr(df, "queryTime") <- Sys.time()
  return(df)
}


#' USGS Samples Summary Data
#'
#' This function creates the call and gets the data for discrete water quality samples summary data
#' service described at <https://api.waterdata.usgs.gov/samples-data/docs>.
#'
#' @param monitoringLocationIdentifier A single monitoring location identifier
#' with two parts, separated by a dash (-): the agency code and the location
#' number. Examples: USGS-040851385, AZ014-320821110580701, CAX01-15304600.
#' The summary service accepts only one site at a time; supplying a vector of
#' length > 1 raises an error. The agency prefix is required: bare location
#' numbers (e.g. "040851385") are accepted by the service but return an empty
#' result.
#' @export
#' @return data frame with summary of data available based on the monitoringLocationIdentifier
#' @rdname summarize_waterdata_samples
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' monitoringLocationIdentifier <- "USGS-04074950"
#'
#' what_data <- summarize_waterdata_samples(monitoringLocationIdentifier)
#'
#' }
summarize_waterdata_samples <- function(monitoringLocationIdentifier) {
  if (length(monitoringLocationIdentifier) > 1) {
    stop("Summary service only available for one site at a time.")
  }

  baseURL <- httr2::request("https://api.waterdata.usgs.gov") |>
    httr2::req_url_path_append("samples-data") |>
    httr2::req_url_path_append("summary", monitoringLocationIdentifier) |>
    httr2::req_url_query(mimeType = "text/csv") |>
    add_api_token()

  df <- importWQP(baseURL)

  if (all(c("firstActivity", "mostRecentActivity") %in% names(df))) {
    df$firstActivity <- as.Date(df$firstActivity)
    df$mostRecentActivity <- as.Date(df$mostRecentActivity)
  }

  attr(df, "url") <- baseURL$url
  attr(df, "queryTime") <- Sys.time()

  return(df)
}


#' @title Deprecated: Use \code{read_waterdata_samples} instead
#' @description This function has been renamed to \code{\link{read_waterdata_samples}}.
#' @param ... Arguments passed to \code{\link{read_waterdata_samples}}.
#' @return data frame returned from web service call.
#' @export
#' @keywords internal
read_USGS_samples <- function(...) {
  .Deprecated(
    new = "read_waterdata_samples",
    package = "dataRetrieval",
    msg = "Function has been renamed. Please begin to migrate to read_waterdata_samples"
  )

  read_waterdata_samples(...)
}


#' @rdname summarize_waterdata_samples
#' @export
summarize_USGS_samples <- function(monitoringLocationIdentifier) {
  .Deprecated(
    new = "summarize_waterdata_samples",
    package = "dataRetrieval",
    msg = "Function has been renamed. Please begin to migrate to summarize_waterdata_samples"
  )

  summarize_waterdata_samples(monitoringLocationIdentifier)
}

context("General functions")


test_that("General samples-data retrievals work using WQP tests", {
  testthat::skip_on_cran()
  nameToUse <- "pH"
  pHData <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-04024315",
    characteristic = nameToUse
  )
  expect_is(pHData$Activity_StartDateTime, "POSIXct")
  expect_type(pHData$USGSpcode, "character")
  expect_type(pHData$Result_Measure, "double")
  expect_type(pHData$SampleCollectionMethod_Identifier, "character")

  # testing lists:
  startDate <- as.Date("2022-01-01")
  secchi_ops <- check_waterdata_sample_params(
    "observedproperty",
    text = "secchi"
  )

  state_fips <- paste0("US:", stateCdLookup("WI", "id"))
  lakeData <- read_waterdata_samples(
    activityStartDateLower = startDate,
    activityStartDateUpper = "2024-01-01",
    stateFips = "US:55",
    characteristicUserSupplied = secchi_ops$observedProperty,
    dataProfile = "narrow"
  )

  expect_true(nrow(lakeData) > 0)

  lakeSites <- read_waterdata_samples(
    monitoringLocationIdentifier = unique(lakeData$Location_Identifier),
    dataType = "locations",
    dataProfile = "site"
  )
  expect_type(lakeSites, "list")

  rawPcode <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-01594440",
    usgsPCode = "01075"
  )

  expect_true(all(
    c("url", "queryTime", "headerInfo") %in%
      names(attributes(rawPcode))
  ))

  pHData <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-04024315",
    characteristic = "pH",
    dataProfile = "narrow"
  )

  expect_true(all(
    c("url", "queryTime", "headerInfo") %in%
      names(attributes(pHData))
  ))
})


context("samples-data samples")
test_that("samples-data activities working", {
  testthat::skip_on_cran()

  activityInfo <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-01594440",
    dataType = "activities"
  )
  expect_true(nrow(activityInfo) > 0)
})

context("samples-data project")
test_that("samples-data project working", {
  testthat::skip_on_cran()
  type <- "Stream"
  county_fips <- countyCdLookup("WI", "Dane")
  projectInfo <- read_waterdata_samples(
    countyFips = county_fips,
    siteTypeName = type,
    dataType = "projects"
  )
  expect_true(ncol(projectInfo) >= 0)
})

context("summary_waterdata_samples")
test_that("summary_waterdata_samples working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  site1 <- summarize_waterdata_samples(
    monitoringLocationIdentifier = "USGS-01594440"
  )
  expect_is(site1, "data.frame")
})

test_that("profiles", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  # Data profiles: "Organization Data"
  org_data <- read_waterdata_samples(
    countyFips = countyCdLookup("WI", "Dane"),
    dataType = "organizations"
  )

  # Data profiles: "Site Data Only"
  site_data <- read_waterdata_samples(
    countyFips = countyCdLookup("WI", "Dane"),
    dataType = "locations"
  )

  expect_true(all(
    c("ProviderName", "Location_Identifier") %in% names(site_data)
  ))

  # Data profiles: "Project Data"
  project_data <- read_waterdata_samples(
    countyFips = countyCdLookup("WI", "Dane"),
    dataType = "projects"
  )

  expect_true(all(
    c(
      "Org_Identifier",
      "Org_FormalName"
    ) %in%
      names(project_data)
  ))

  # Data profiles: "Project Monitoring Location Weighting Data"
  proj_mlwd <- read_waterdata_samples(
    countyFips = countyCdLookup("WI", "Dane"),
    dataType = "projects",
    dataProfile = "projectmonitoringlocationweight"
  )

  expect_true(all(
    c(
      "Org_Identifier",
      "Org_FormalName"
    ) %in%
      names(proj_mlwd)
  ))

  # Data profiles: "Sample Results (biological metadata)"
  samp_bio <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-04024315",
    dataProfile = "basicbio",
    dataType = "results"
  )

  expect_true(all(
    c(
      "Org_Identifier",
      "Org_FormalName"
    ) %in%
      names(samp_bio)
  ))

  # Data profiles: "Sample Results (narrow)"
  samp_narrow <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-04024315",
    dataProfile = "narrow",
    dataType = "results"
  )

  expect_true(all(
    c(
      "Org_Identifier",
      "Org_FormalName"
    ) %in%
      names(samp_narrow)
  ))

  # Data profiles: "Sampling Activity"
  samp_activity <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-04024315",
    dataProfile = "sampact", # Sampling Activities
    dataType = "activities"
  )

  expect_true(all(
    c(
      "Org_Identifier",
      "Org_FormalName"
    ) %in%
      names(samp_activity)
  ))

  # Data profile: "Result Detection Quantitation Limit Data"
  dl_data <- read_waterdata_samples(
    monitoringLocationIdentifier = "USGS-04024315",
    dataType = "results",
    dataProfile = "resultdetectionquantitationlimit"
  )

  expect_true(all(
    c(
      "Org_Identifier",
      "Org_FormalName"
    ) %in%
      names(dl_data)
  ))
})

#' Basic Water Quality Portal Data parser
#'
#' Imports data from the Water Quality Portal based on a specified url.
#'
#' @param obs_url character URL to Water Quality Portal#' @keywords data import USGS web service
#' @param tz character to set timezone attribute of datetime. Default is UTC
#' (properly accounting for daylight savings times based on the data's provided tz_cd column).
#' Possible values include "America/New_York","America/Chicago", "America/Denver","America/Los_Angeles",
#' "America/Anchorage","America/Honolulu","America/Jamaica","America/Managua",
#' "America/Phoenix", and "America/Metlakatla"
#' @param csv logical. Is the data coming back with a csv or tsv format. Default is `FALSE`.
#' Currently, the summary service does not support tsv, for other services tsv is the safer choice.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character.
#' @return retval dataframe raw data returned from the Water Quality Portal. Additionally,
#' a POSIXct dateTime column is supplied for
#' start and end times, and converted to UTC. See
#' <https://www.waterqualitydata.us/portal_userguide/> for more information.
#' @export
#' @seealso [readWQPdata()], [readWQPqw()], [whatWQPsites()]
#' @examplesIf is_dataRetrieval_user()
#' # These examples require an internet connection to run
#'
#' ## Examples take longer than 5 seconds:
#' \donttest{
#' rawSampleURL <- constructWQPURL("USGS-01594440", "01075", "", "")
#'
#' rawSample <- importWQP(rawSampleURL)
#'
#'
#' STORETex <- constructWQPURL("WIDNR_WQX-10032762", "Specific conductance", "", "")
#'
#' STORETdata <- importWQP(STORETex)
#'
#' STORETdata_char <- importWQP(STORETex, convertType = FALSE)
#' }
#'
importWQP <- function(obs_url, tz = "UTC", csv = TRUE, convertType = TRUE) {
  if (tz != "") {
    tz <- match.arg(tz, OlsonNames())
  } else {
    tz <- "UTC"
  }

  if (
    is.character(obs_url) &&
      grepl("(https)://[^ /$.?#].[^\\s]*", obs_url)
  ) {
    obs_url <- httr2::request(obs_url)
  }

  if (inherits(obs_url, "httr2_request")) {
    doc <- getWebServiceData(obs_url)
    if (is.null(doc)) {
      return(invisible(NULL))
    }
    headerInfo <- attr(doc, "headerInfo")
  } else {
    doc <- obs_url
  }

  last_chars <- as.character(substr(doc, nchar(doc) - 1, nchar(doc)))

  retval <- data.table::fread(
    text = doc,
    data.table = FALSE,
    sep = ifelse(csv, ",", "\t"),
    fill = TRUE,
    colClasses = "character",
    quote = ifelse(csv, '\"', "")
  )

  attr(retval, 'spec') <- NULL

  if (any(grepl("ERROR: INCOMPLETE DATA", retval[[1]]))) {
    stop(retval[[1]][grepl("ERROR: INCOMPLETE DATA", retval[[1]])])
  }

  # this is only needed for legacy
  names(retval)[grep("/", names(retval))] <- gsub(
    "/",
    ".",
    names(retval)[grep("/", names(retval))]
  )

  if (convertType) {
    retval <- parse_WQP(retval, tz)
  }
  if (inherits(obs_url, "httr2_request")) {
    attr(retval, "headerInfo") <- headerInfo
  }
  return(retval)
}

#' Convert WQP columns to correct types
#'
#' Takes the character results and converts to numeric and dates.
#'
#' @param retval Data frame from WQP
#' @param tz character to set timezone attribute of datetime. Default is UTC
#' (properly accounting for daylight savings times based on the associated "TimeZone" column).
#' Possible values include "America/New_York","America/Chicago", "America/Denver","America/Los_Angeles",
#' "America/Anchorage","America/Honolulu","America/Jamaica","America/Managua",
#' "America/Phoenix", and "America/Metlakatla"
#'
#' @export
#' @return data frame retval with converted columns
#'
#' @examplesIf is_dataRetrieval_user()
#' # These examples require an internet connection to run
#' rawSampleURL <- constructWQPURL("USGS-01594440", "01075", "", "")
#'
#' ## Examples take longer than 5 seconds:
#'
#' \donttest{
#'
#' rawSample <- importWQP(rawSampleURL, convertType = FALSE)
#' convertedSample <- parse_WQP(rawSample)
#'
#' }
#'
parse_WQP <- function(retval, tz = "UTC") {
  # this is legacy:
  valueCols <- names(retval)[grep("Value", names(retval))]
  countCols <- names(retval)[grep("Count", names(retval))]

  # this is new:
  latCols <- names(retval)[grep("Latitude", names(retval))]
  lonCols <- names(retval)[grep("Longitude", names(retval))]

  countCols <- countCols[!grepl("Country", countCols)]
  countCols <- countCols[!grepl("County", countCols)]

  measureCols <- names(retval)[grep("Measure", names(retval))]
  yearCols <- names(retval)[grep("Year", names(retval))]

  numberColumns <- unique(c(
    valueCols,
    countCols,
    yearCols,
    latCols,
    lonCols,
    measureCols
  ))
  numberColumns <- numberColumns[!grepl("Code", numberColumns)]
  numberColumns <- numberColumns[!grepl("Unit", numberColumns)]
  numberColumns <- numberColumns[!grepl("Identifier", numberColumns)]
  numberColumns <- numberColumns[!grepl("Type", numberColumns)]

  for (numberCol in numberColumns) {
    suppressWarnings({
      val <- tryCatch(as.numeric(retval[[numberCol]]), warning = function(w) w)
      # we don't want to convert it to numeric if there are non-numeric chars
      # If we leave it to the user, it will probably break a lot of code
      if (!"warning" %in% class(val)) {
        retval[[numberCol]] <- val
      }
    })
  }

  dateCols <- names(retval)[grep("Date", names(retval))]

  if (length(dateCols) > 0) {
    dateCols_to_convert <- NA
    for (date_col in dateCols) {
      # Legacy:
      time_col <- gsub("Date", "Time", date_col)
      tz_col <- paste0(time_col, ".TimeZoneCode")
      time_col <- paste0(time_col, ".Time")

      # WQX3:
      time_col_wqx3 <- gsub("Date", "Time", date_col)
      tz_col_wqx3 <- gsub("Date", "TimeZone", date_col)

      if (all(c(date_col, time_col, tz_col) %in% names(retval))) {
        # Legacy
        if (!all(is.na(retval[[date_col]]))) {
          retval <- create_dateTime(
            retval,
            date_col = date_col,
            time_col = time_col,
            tz_col = tz_col,
            tz = tz
          )
        }
      } else if (
        all(c(date_col, time_col_wqx3, tz_col_wqx3) %in% names(retval))
      ) {
        # WQX3
        retval <- create_dateTime(
          retval,
          date_col = date_col,
          time_col = time_col_wqx3,
          tz_col = tz_col_wqx3,
          tz = tz
        )
      } else {
        # This is the legacy pattern:
        time_col <- gsub("Date", "Time.Time", date_col)
        tz_col <- gsub("Date", "Time.TimeZoneCode", date_col)

        if (all(c(date_col, time_col, tz_col) %in% names(retval))) {
          if (!all(is.na(retval[[date_col]]))) {
            retval <- create_dateTime(
              retval,
              date_col = date_col,
              time_col = time_col,
              tz_col = tz_col,
              tz = tz
            )
          }
        } else {
          dateCols_to_convert <- c(dateCols_to_convert, date_col)
        }
      }
    }

    dateCols_to_convert <- dateCols_to_convert[!is.na(dateCols_to_convert)]
    if (length(dateCols_to_convert) > 0) {
      for (i in dateCols_to_convert) {
        if (i %in% names(retval)) {
          retval[, i] <- suppressWarnings(as.Date(lubridate::parse_date_time(
            retval[[i]],
            c("Ymd", "mdY")
          )))
        }
      }
    }

    if ("Activity_StartDateTime" %in% names(retval)) {
      #WQX 3
      retval <- retval[order(retval$Activity_StartDateTime), ]
    } else if ("ActivityStartDateTime" %in% names(retval)) {
      #legacy
      retval <- retval[order(retval$ActivityStartDateTime), ]
    } else {
      retval <- retval[order(retval[[dateCols[1]]]), ]
    }
  }

  return(retval)
}

create_dateTime <- function(df, date_col, time_col, tz_col, tz) {
  # Difference in behavior between NWIS and WQP
  offsetLibrary$offset[is.na(offsetLibrary$code)] <- NA
  original_order <- names(df)

  df <- merge(
    x = df,
    y = offsetLibrary,
    by.x = tz_col,
    by.y = "code",
    all.x = TRUE
  )

  df$dateTime <- paste(df[[date_col]], df[[time_col]])
  df$dateTime <- lubridate::fast_strptime(
    df$dateTime,
    "%Y-%m-%d %H:%M:%S"
  ) +
    60 * 60 * df$offset

  attr(df$dateTime, "tzone") <- tz

  df[[date_col]] <- suppressWarnings(as.Date(lubridate::parse_date_time(
    df[[date_col]],
    c("Ymd", "mdY")
  )))

  df <- df[, c(original_order, "offset", "dateTime")]

  names(df)[names(df) == "offset"] <- paste0(tz_col, "_offset")
  names(df)[names(df) == "dateTime"] <- paste0(date_col, "Time")

  return(df)
}

#'
#' Format and organize WQP arguments that are passed in as `...`.
#'
#' @keywords internal
readWQPdots <- function(..., legacy = TRUE) {
  if (length(list(...)) == 0) {
    stop("No arguments supplied")
  }

  matchReturn <- convertLists(...)

  if ("service" %in% names(matchReturn)) {
    service <- matchReturn$service
    matchReturn$service <- NULL
  } else {
    if (legacy) {
      service <- "Result"
    } else {
      service <- "ResultWQX3"
    }
  }

  if ("dataProfile" %in% names(matchReturn)) {
    profile <- matchReturn$dataProfile
    if (profile == "activityAll") {
      service <- "ActivityWQX3"
      matchReturn$service <- NULL
    } else if (profile %in% c("resultPhysChem", "biological", "narrowResult")) {
      service <- "Result"
      matchReturn$service <- NULL
    } else if (profile %in% c("fullPhysChem", "narrow", "basicPhysChem")) {
      service <- "ResultWQX3"
      matchReturn$service <- NULL
    }
  }

  match.arg(
    service,
    c(
      "Result",
      "Station",
      "Activity",
      "Organization",
      "ActivityMetric",
      "SiteSummary",
      "Project",
      "ProjectMonitoringLocationWeighting",
      "ResultDetectionQuantitationLimit",
      "BiologicalMetric",
      "ResultWQX3",
      "StationWQX3",
      "ActivityWQX3"
    )
  )

  names(matchReturn)[names(matchReturn) == "bbox"] <- "bBox"

  bbox <- "bBox" %in% names(matchReturn)
  if (bbox) {
    matchReturn["bBox"] <- sapply(matchReturn["bBox"], function(x) {
      as.character(paste0(eval(x), collapse = ","))
    })
  }

  values <- matchReturn
  values <- checkWQPdates(values)

  names(values)[names(values) == "siteNumber"] <- "siteid"
  names(values)[names(values) == "siteNumbers"] <- "siteid"
  names(values)[names(values) == "parameterCd"] <- "pCode"
  names(values)[names(values) == "USGSPCode"] <- "pCode"

  names(values)[names(values) == "stateCd"] <- "statecode"
  if ("statecode" %in% names(values)) {
    stCd <- values["statecode"]
    stCdPrefix <- "US:"
    if (!grepl(stCdPrefix, stCd)) {
      values["statecode"] <- stateCdLookup(stCd, "fips")
    }
  }

  names(values)[names(values) == "countyCd"] <- "countycode"
  if (all(c("countycode", "statecode") %in% names(values))) {
    stCd <- gsub("US:", "", values["statecode"])
    # This will error if more than 1 state is requested
    # It's possible that someone could request more than one state
    # in WQP, but if they also then request county codes,
    # it gets really confusing, and the WQP developers don't recommend.
    values["countycode"] <- countyCdLookup(stCd, values["countycode"], "fips")
    # WQP doesn't need or want both state and county code:
    values["statecode"] <- NULL
  }

  if (!"mimeType" %in% names(values)) {
    values["mimeType"] <- "csv"
  }

  if (legacy & !("count" %in% names(values))) {
    values["count"] <- "no"
  }

  return_list <- list()
  return_list["values"] <- list(values)
  return_list["service"] <- service
  return(return_list)
}

#' Site Data Import from NWIS
#'
#' Returns a list of sites from the NWIS web service. This function gets the data from:
#' <https://waterservices.usgs.gov/docs/site-service/>.
#' Mapper format is used
#'
#' @param \dots see <https://waterservices.usgs.gov/docs/site-service/>
#' for a complete list of options. A list (or lists) can also be supplied.
#'
#' @return A data frame with at least the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#' station_nm \tab character \tab Station name \cr
#' site_tp_cd \tab character \tab Site type code \cr
#' dec_lat_va \tab numeric \tab Decimal latitude \cr
#' dec_long_va \tab numeric \tab Decimal longitude \cr
#' queryTime \tab POSIXct \tab Query time \cr
#' }
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#' @export
#' @seealso [read_waterdata_monitoring_location()]
#'
#' @examples
#'
#' # see ?read_waterdata_monitoring_location
#' #siteListPhos <- whatNWISsites(stateCd = "OH", parameterCd = "00665")
#' #oneSite <- whatNWISsites(sites = "05114000")
#'
whatNWISsites <- function(...) {
  .Deprecated(
    new = "read_waterdata_monitoring_location",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_monitoring_location"
  )

  matchReturn <- convertLists(...)
  if ("service" %in% names(matchReturn)) {
    service <- matchReturn$service
    matchReturn$service <- NULL
  } else {
    service <- NULL
  }

  valuesList <- readNWISdots(matchReturn)

  values <- valuesList[["values"]]
  values <- values[names(values) != "format"]

  #################
  # temporary gwlevels fixes
  values <- values[
    !names(values) %in%
      c(
        "date_format",
        "TZoutput",
        "rdb_inventory_output",
        "list_of_search_criteria"
      )
  ]

  names(values)[names(values) == "state_cd"] <- "stateCd"
  ##################

  if (!is.null(service)) {
    service[service == "peak"] <- "pk"
    service[service == "uv"] <- "id"

    values[["hasDataTypeCd"]] <- service
  }

  POST <- nchar(paste0(unlist(values), collapse = "")) > 2048

  urlCall <- httr2::request(pkg.env[["site"]])

  urlCall <- get_or_post(urlCall, POST = POST, !!!values, .multi = "comma")

  urlCall <- get_or_post(urlCall, POST = POST, format = "mapper")

  rawData <- getWebServiceData(urlCall)
  if (is.null(rawData)) {
    return(invisible(NULL))
  }
  doc <- xml2::xml_root(rawData)
  siteCategories <- xml2::xml_children(doc)
  retVal <- NULL
  for (sc in siteCategories) {
    sites <- xml2::xml_children(sc)
    site_no <- xml2::xml_attr(sites, "sno")
    station_nm <- xml2::xml_attr(sites, "sna")
    site_tp_cd <- xml2::xml_attr(sites, "cat")
    dec_lat_va <- as.numeric(xml2::xml_attr(sites, "lat"))
    dec_long_va <- as.numeric(xml2::xml_attr(sites, "lng"))
    agency_cd <- xml2::xml_attr(sites, "agc")

    colocated <- isTRUE(xml2::xml_name(sc) == "colocated_sites")

    df <- data.frame(
      agency_cd,
      site_no,
      station_nm,
      site_tp_cd,
      dec_lat_va,
      dec_long_va,
      colocated,
      stringsAsFactors = FALSE
    )

    if (is.null(retVal)) {
      retVal <- df
    } else {
      retVal <- r_bind_dr(retVal, df)
    }
  }

  retVal <- retVal[!duplicated(retVal), ]

  attr(retVal, "url") <- urlCall$url

  timenow <- Sys.time()

  attr(retVal, "queryTime") <- timenow
  # Backwards compatible, might remove later:
  retVal$queryTime <- timenow

  return(retVal)
}

#' Walk through the pages
#'
#' @param req httr2 initial request
#'
#' @noRd
#' @return data.frame with attributes
walk_pages <- function(req) {
  resps <- httr2::req_perform_iterative(
    req,
    next_req = next_req_url,
    max_reqs = Inf,
    on_error = "stop"
  )

  failures <- resps |> httr2::resps_failures()

  return_list <- resps |>
    httr2::resps_successes() |>
    httr2::resps_data(\(resp) get_resp_data(resp))

  return(return_list)
}

#' Get single response data frame
#'
#' Depending on skipGeometry to decide to use sf or not.
#'
#' @noRd
#'
#' @param resp httr2 response from last request
#'
#' @return data.frame
#'
get_resp_data <- function(resp) {
  body <- httr2::resp_body_json(resp)
  use_sf <- !grepl("skipGeometry=true", resp$url, ignore.case = TRUE)

  if (isTRUE(body[["numberReturned"]] == 0)) {
    return(data.frame())
  }

  return_df <- sf::read_sf(httr2::resp_body_string(resp))

  return_df <- coerce_num_cols(return_df, is_sf = TRUE)

  if ("qualifier" %in% names(return_df)) {
    return_df$qualifier <- as.character(vapply(
      X = return_df$qualifier,
      FUN = function(x) {
        x[is.na(x)] <- ""
        paste(x, collapse = ", ")
      },
      FUN.VALUE = c(NA_character_)
    ))
  }

  if (!use_sf) {
    return_df <- sf::st_drop_geometry(return_df)
    if ("AsGeoJSON(geometry)" %in% names(return_df)) {
      return_df <- return_df[, !names(return_df) %in% "AsGeoJSON(geometry)"]
    }
  }

  return(return_df)
}

#' Next request URL
#'
#' Custom function to find the "next" URL from the API response.
#' @seealso [httr2::req_perform_iterative]
#'
#' @param resp httr2 response from last request
#' @param req httr2 request from last time
#'
#' @noRd
#' @return the url for the next request
#'
next_req_url <- function(resp, req) {
  body <- httr2::resp_body_json(resp)

  if (isTRUE(body[["code"]] == "InvalidQuery")) {
    message(body[["description"]])
    return(NULL)
  }

  if (isTRUE(body[["numberReturned"]] == 0)) {
    return(NULL)
  }

  log_rate_limit(resp)
  if ("links" %in% names(body)) {
    links <- body$links
    if (any(sapply(links, function(x) x$rel) == "next")) {
      next_index <- which(sapply(links, function(x) x$rel) == "next")

      next_url <- links[[next_index]][["href"]]

      return(httr2::req_url(req = req, url = next_url))
    }
  } else if (!is.null(body[["next"]])) {
    return(httr2::req_url_query(req, next_token = body[["next"]]))
  } else {
    return(NULL)
  }
}


get_csv <- function(req, limit) {
  skip_geo <- grepl("skipGeometry=true", req$url, ignore.case = TRUE)
  resp <- httr2::req_perform(req)

  log_rate_limit(resp)

  if (httr2::resp_has_body(resp)) {
    return_list <- httr2::resp_body_string(resp)
    df <- data.table::fread(
      input = return_list,
      data.table = FALSE,
      colClasses = "character"
    )

    df <- coerce_num_cols(df)

    if (skip_geo) {
      df <- df[, names(df)[!names(df) %in% c("x", "y")]]
    } else {
      if (all(c("x", "y") %in% names(df))) {
        df <- sf::st_as_sf(df, coords = c("x", "y"))
        sf::st_crs(df) <- 4269
      }
    }

    if (nrow(df) == limit) {
      warning(
        "Missing data is probable. Use no_paging = FALSE to 
ensure all requested data is returned."
      )
    }
  } else {
    df <- data.frame()
  }

  return(df)
}

coerce_num_cols <- function(df, is_sf = FALSE) {
  included_num_cols <- names(df)[names(df) %in% num_cols]
  if (length(included_num_cols) == 0) {
    return(df)
  }

  check_df <- if (is_sf) {
    sf::st_drop_geometry(df[, included_num_cols, drop = FALSE])
  } else {
    df[, included_num_cols, drop = FALSE]
  }

  if (!all(vapply(check_df, is.numeric, logical(1)))) {
    df[, included_num_cols] <- lapply(check_df, as.numeric)
  }
  df
}

log_rate_limit <- function(resp) {
  if (Sys.getenv("API_USGS_PAT") != "") {
    header_info <- httr2::resp_headers(resp)
    message(
      "Remaining requests this hour:",
      header_info$`x-ratelimit-remaining`,
      " "
    )
  }
}

#' Daily Value USGS NWIS Data Retrieval
#'
#' Imports data from NWIS daily web service. This function gets the data from here:
#' <https://waterservices.usgs.gov/docs/dv-service/daily-values-service-details/>
#' Inputs to this function are just USGS site ids, USGS parameter codes,
#' USGS statistic codes, and start and end date. For a more complex query, use [readNWISdata()],
#' with an argument service = "dv".
#' Data coming the daily web services are aggregates of the instantaneous
#' (sensor) web services. Not all statistical codes are available for all data.
#' Use the function [whatNWISdata()] to discover what data
#' is available for a USGS site. The column data_type_cd with the values "dv"
#' returned from [whatNWISdata()]) are available from this service.
#'
#' More information on the web service can be found here:
#' <https://waterservices.usgs.gov/test-tools>, choosing the
#' "Daily Value Service".
#'
#' @param siteNumbers character USGS site number.  This is usually an 8 digit number.
#' Multiple sites can be requested with a character vector.
#' @param parameterCd character of USGS parameter code(s).  This is usually an 5 digit number.
#' @param startDate character starting date for data retrieval in the form YYYY-MM-DD.
#' Default is "" which indicates
#' retrieval for the earliest possible record. Date arguments are always specified in local time.
#' @param endDate character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the latest possible record. Date arguments are always specified in local time.
#' @param statCd character USGS statistic code. This is usually 5 digits.  Daily mean (00003) is the default.
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency \tab character \tab The NWIS code for the agency reporting the data\cr
#' site \tab character \tab The USGS site number \cr
#' Date \tab Date \tab The date of the value \cr
#' code \tab character \tab Any codes that qualify the corresponding value\cr
#' value \tab numeric \tab The numeric value for the parameter \cr
#' }
#' Note that code and value are repeated for the parameters requested. The names are of the form
#' X_D_P_S, where X is literal,
#' D is an option description of the parameter,
#' P is the parameter code,
#' and S is the statistic code (if applicable).
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
#' statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#'
#' @seealso [read_waterdata_daily()]
#' @export
#' @keywords data import USGS web service
#' @examples
#'
#' # see ?read_waterdata_daily
#'
#' #site_id <- "04085427"
#' #startDate <- "2012-01-01"
#' #endDate <- "2012-06-30"
#' #pCode <- "00060"
#' #
#' #rawDailyQ <- readNWISdv(site_id, pCode, startDate, endDate)
#'
readNWISdv <- function(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  statCd = "00003"
) {
  .Deprecated(
    new = "read_waterdata_daily",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_daily."
  )

  url <- constructNWISURL(
    siteNumbers = siteNumbers,
    parameterCd = parameterCd,
    startDate = startDate,
    endDate = endDate,
    service = "dv",
    statCd = statCd
  )

  data <- importWaterML1(url, asDateTime = FALSE)

  if (nrow(data) > 0) {
    data$dateTime <- as.Date(data$dateTime)
    data$tz_cd <- NULL

    names(data)[names(data) == "dateTime"] <- "Date"
  }

  return(data)
}

#' USGS Site File Data Retrieval
#'
#' Imports data from USGS site file site. This function gets data from here: <https://waterservices.usgs.gov/>
#'
#' @param siteNumbers character USGS site number (or multiple sites).  This is usually an 8 digit number
#' @keywords data import USGS web service
#' @return A data frame with at least the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#'  station_nm \tab character \tab Site name \cr
#'  site_tp_cd \tab character \tab Site type \cr
#'  lat_va \tab numeric \tab DMS latitude \cr
#'  long_va \tab numeric \tab DMS longitude \cr
#'  dec_lat_va \tab numeric \tab Decimal latitude \cr
#'  dec_long_va \tab numeric \tab Decimal longitude \cr
#'  coord_meth_cd  \tab character \tab Latitude-longitude method \cr
#'  coord_acy_cd \tab character \tab Latitude-longitude accuracy \cr
#'  coord_datum_cd \tab character \tab Latitude-longitude datum \cr
#'  dec_coord_datum_cd \tab character \tab Decimal Latitude-longitude datum \cr
#'  district_cd \tab character \tab District code \cr
#'  state_cd \tab character \tab State code \cr
#'  county_cd \tab character \tab County code \cr
#'  country_cd \tab character \tab Country code \cr
#'  land_net_ds \tab character \tab Land net location description \cr
#'  map_nm \tab character \tab Name of location map \cr
#'  map_scale_fc \tab character \tab Scale of location map \cr
#'  alt_va \tab numeric \tab Altitude of Gage/land surface \cr
#'  alt_meth_cd \tab character \tab Method altitude determined \cr
#'  alt_acy_va  \tab numeric \tab Altitude accuracy \cr
#'  alt_datum_cd \tab character \tab Altitude datum \cr
#'  huc_cd \tab character \tab Hydrologic unit code \cr
#'  basin_cd \tab character \tab Drainage basin code \cr
#'  topo_cd \tab character \tab Topographic setting code \cr
#'  instruments_cd \tab character \tab Flags for instruments at site \cr
#'  construction_dt \tab character \tab Date of first construction \cr
#'  inventory_dt \tab character \tab Date site established or inventoried \cr
#'  drain_area_va \tab numeric \tab Drainage area \cr
#'  contrib_drain_area_va \tab numeric \tab Contributing drainage area \cr
#'  tz_cd  \tab character \tab Time Zone abbreviation \cr
#'  local_time_fg \tab character \tab Site honors Daylight Savings Time \cr
#'  reliability_cd \tab character \tab Data reliability code \cr
#'  gw_file_cd \tab character \tab Data-other GW files \cr
#'  nat_aqfr_cd \tab character \tab National aquifer code \cr
#'  aqfr_cd \tab character \tab Local aquifer code \cr
#'  aqfr_type_cd \tab character \tab Local aquifer type code \cr
#'  well_depth_va \tab numeric \tab Well depth \cr
#'  hole_depth_va \tab numeric \tab Hole depth \cr
#'  depth_src_cd \tab character \tab Source of depth data \cr
#'  project_no \tab character \tab Project number \cr
#' }
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' comment \tab character \tab Header comments from the RDB file \cr
#' }
#' @export
#' @seealso [read_waterdata_monitoring_location()]
#' @examples
#'
#' # see ?read_waterdata_monitoring_location
#' # siteINFOMulti <- readNWISsite(c("05114000", "09423350"))
#'
readNWISsite <- function(siteNumbers) {
  .Deprecated(
    new = "read_waterdata_monitoring_location",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_monitoring_location"
  )

  baseURL <- httr2::request(pkg.env[["site"]])
  urlSitefile <- httr2::req_url_query(
    baseURL,
    siteOutput = "Expanded",
    format = "rdb"
  )

  POST <- nchar(paste0(siteNumbers, collapse = "")) > 2048

  urlSitefile <- get_or_post(
    urlSitefile,
    POST = POST,
    site = siteNumbers,
    .multi = "comma"
  )

  data <- importRDB1(urlSitefile, asDateTime = FALSE)
  # readr needs multiple lines to convert to anything but characters:
  data[grep("_va", names(data))][data[grep("_va", names(data))] == "."] <- NA

  data[, grep("_va", names(data))] <- sapply(
    data[, grep("_va", names(data))],
    as.numeric
  )

  return(data)
}

#' Get USGS Channel Measurements
#'
#' @description `r get_description("channel-measurements")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("channel-measurements")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param last_modified `r get_ogc_params("channel-measurements")$last_modified`
#'                                    field_visit_id = NA_character_,
#' @param measurement_number `r get_ogc_params("channel-measurements")$measurement_number`
#' @param field_visit_id `r get_ogc_params("channel-measurements")$field_visit_id`
#' @param time `r get_ogc_params("channel-measurements")$time`
#' @param channel_name `r get_ogc_params("channel-measurements")$channel_name`
#' @param channel_flow `r get_ogc_params("channel-measurements")$channel_flow`
#' @param channel_flow_unit `r get_ogc_params("channel-measurements")$channel_flow_unit`
#' @param channel_width `r get_ogc_params("channel-measurements")$channel_width`
#' @param channel_width_unit `r get_ogc_params("channel-measurements")$channel_width_unit`
#' @param channel_area `r get_ogc_params("channel-measurements")$channel_area`
#' @param channel_area_unit `r get_ogc_params("channel-measurements")$channel_area_unit`
#' @param channel_velocity `r get_ogc_params("channel-measurements")$channel_velocity`
#' @param channel_velocity_unit `r get_ogc_params("channel-measurements")$channel_velocity_unit`
#' @param channel_location_distance `r get_ogc_params("channel-measurements")$channel_location_distance`
#' @param channel_location_distance_unit `r get_ogc_params("channel-measurements")$channel_location_distance_unit`
#' @param channel_location_direction `r get_ogc_params("channel-measurements")$channel_location_direction`
#' @param channel_stability `r get_ogc_params("channel-measurements")$channel_stability`
#' @param channel_material `r get_ogc_params("channel-measurements")$channel_material`
#' @param channel_evenness `r get_ogc_params("channel-measurements")$channel_evenness`
#' @param horizontal_velocity_description `r get_ogc_params("channel-measurements")$horizontal_velocity_description`
#' @param vertical_velocity_description `r get_ogc_params("channel-measurements")$vertical_velocity_description`
#' @param longitudinal_velocity_description `r get_ogc_params("channel-measurements")$longitudinal_velocity_description`
#' @param measurement_type `r get_ogc_params("channel-measurements")$measurement_type`
#' @param channel_measurement_type `r get_ogc_params("channel-measurements")$channel_measurement_type`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("channel-measurements", "channel_measurements_id")`.
#' The default (`NA`) will return all columns of the data.
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @inherit read_waterdata_continuous details
#'
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-02238500"
#' df <- read_waterdata_channel(monitoring_location_id = site)
#'
#' }
read_waterdata_channel <- function(
  monitoring_location_id = NA_character_,
  field_visit_id = NA_character_,
  measurement_number = NA_character_,
  time = NA_character_,
  channel_name = NA_character_,
  channel_flow = NA_character_,
  channel_flow_unit = NA_character_,
  channel_width = NA_character_,
  channel_width_unit = NA_character_,
  channel_area = NA_character_,
  channel_area_unit = NA_character_,
  channel_velocity = NA_character_,
  channel_velocity_unit = NA_character_,
  channel_location_distance = NA_character_,
  channel_location_distance_unit = NA_character_,
  channel_location_direction = NA_character_,
  channel_stability = NA_character_,
  channel_material = NA_character_,
  channel_evenness = NA_character_,
  horizontal_velocity_description = NA_character_,
  vertical_velocity_description = NA_character_,
  longitudinal_velocity_description = NA_character_,
  measurement_type = NA_character_,
  last_modified = NA_character_,
  channel_measurement_type = NA_character_,
  properties = NA_character_,
  skipGeometry = NA,
  bbox = NA,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  limit = getOption("dataRetrieval.limit"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_data"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "channel-measurements"
  output_id <- "channel_measurements_id"

  rlang::check_dots_empty()
  args <- mget(names(formals()))
  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

#' Get USGS Peak Data
#'
#' @description `r get_description("peaks")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("peaks")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param parameter_code `r get_ogc_params("peaks")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param time `r get_ogc_params("peaks")$time`
#'
#' See also Details below for more information.
#' @param value `r get_ogc_params("peaks")$value`
#' @param unit_of_measure `r get_ogc_params("peaks")$unit_of_measure`
#' @param time_series_id `r get_ogc_params("peaks")$time_series_id`
#' @param last_modified `r get_ogc_params("peaks")$last_modified`
#'
#' See also Details below for more information.
#' @param water_year `r get_ogc_params("peaks")$water_year`
#' @param year `r get_ogc_params("peaks")$year`
#' @param month `r get_ogc_params("peaks")$month`
#' @param day `r get_ogc_params("peaks")$day`
#' @param time_of_day `r get_ogc_params("peaks")$time_of_day`
#' @param peak_since `r get_ogc_params("peaks")$peak_since`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("peaks", "peak_id")`.
#' The default (`NA`) will return all columns of the data.
#'
#' @param allow_incomplete_dates Specifically in the peaks data, exact peak dates
#' are not always known. Sometimes peaks are known just for the year, sometimes
#' they are known to the year and month, and and sometimes to the exact date.
#' This argument determines if incomplete dates + uncertain month/day values are
#' allowed in the "time" column so that it can be a complete Date object (`TRUE`),
#' or whether to set those dates to `NA` (`FALSE`). Peaks with uncertain days
#' are stored on the first of the month, and those with uncertain
#' month stored on January 1. Default is `FALSE`.
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @inherit read_waterdata_continuous details
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' wi_peaks <- read_waterdata_combined_meta(
#'                state_name = "Wisconsin",
#'                data_type = "Peaks",
#'                parameter_code = "00060")
#'
#'
#' dv_data_sf <- read_waterdata_peaks(
#'                monitoring_location_id = wi_peaks$monitoring_location_id[1],
#'                parameter_code = "00060")
#'
#' incomplete_dates_not_allowed <- read_waterdata_peaks(
#'                monitoring_location_id = "USGS-06334330",
#'                parameter_code = "00060")
#' incomplete_dates_not_allowed$time
#' incomplete_dates_allowed <- read_waterdata_peaks(
#'                monitoring_location_id = "USGS-06334330",
#'                parameter_code = "00060",
#'                allow_incomplete_dates = TRUE)
#' incomplete_dates_allowed$time
#'
#' }
read_waterdata_peaks <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  properties = NA_character_,
  time_series_id = NA_character_,
  unit_of_measure = NA_character_,
  value = NA,
  last_modified = NA_character_,
  water_year = NA_character_,
  year = NA_character_,
  month = NA_character_,
  day = NA_character_,
  time_of_day = NA_character_,
  peak_since = NA_character_,
  skipGeometry = NA,
  time = NA_character_,
  bbox = NA,
  ...,
  allow_incomplete_dates = FALSE,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "peaks"
  output_id <- "peak_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  args[["allow_incomplete_dates"]] <- NULL
  return_list <- get_ogc_data(args, output_id, service)

  if (anyNA(return_list[, c("year", "month", "day")])) {
    if (allow_incomplete_dates) {
      warning("Incomplete dates are included in time column.")
    } else {
      return_list$time[is.na(return_list$month)] <- NA
      return_list$time[is.na(return_list$day)] <- NA
    }
  }

  return(return_list)
}

pkg.env <- new.env()

.onLoad <- function(libname, pkgname) {
  suppressMessages(setAccess("public"))
  pkg.env$local_sf <- requireNamespace("sf", quietly = TRUE)
  options("dataRetrieval.nldi_base" = "https://api.water.usgs.gov/nldi/linked-data/")
  options("dataRetrieval.api_version" = "v0")
  options("dataRetrieval.api_version_stat" = "v0")
  options("dataRetrieval.attach_request" = TRUE)
  options("dataRetrieval.convertType" = TRUE)
  options("dataRetrieval.no_paging" = FALSE)
  options("dataRetrieval.site_chunk_size_meta" = 250)
  options("dataRetrieval.site_chunk_size_data" = 10)
  options("dataRetrieval.limit" = 50000)

  services <- c(
    "server",
    "daily",
    "time-series-metadata",
    "monitoring-locations",
    "latest-continuous",
    "field-measurements",
    "latest-daily",
    "continuous",
    "field-measurements-metadata",
    "combined-metadata",
    "channel-measurements",
    "peaks"
  )
  collections <- c(
    "parameter-codes",
    "agency-codes",
    "altitude-datums",
    "aquifer-codes",
    "aquifer-types",
    "coordinate-accuracy-codes",
    "coordinate-datum-codes",
    "coordinate-method-codes",
    "medium-codes",
    "counties",
    "countries",
    "hydrologic-unit-codes",
    "methods",
    "method-categories",
    "method-citations",
    "citations",
    "states",
    "national-aquifer-codes",
    "reliability-codes",
    "site-types",
    "statistic-codes",
    "topographic-codes",
    "time-zone-codes"
  )
  names(collections) <- c(
    "parameter_code",
    "agency_code",
    "altitude_datum",
    "aquifer_code",
    "aquifer_type",
    "coordinate_accuracy_code",
    "coordinate_datum_code",
    "coordinate_method_code",
    "medium_code",
    "county",
    "country",
    "hydrologic_unit_code",
    "methods",
    "method_categories",
    "method_citations",
    "citations",
    "state",
    "national_aquifer_code",
    "reliability_code",
    "site_types",
    "statistic_code",
    "topographic_code",
    "time_zone_code"
  )

  pkg.env$api_endpoints <- services
  pkg.env$metadata <- collections
}


#' Is this a dataRetrieval user
#'
#' Reveals if this is a user or not
#' @export
#' @examples
#' is_dataRetrieval_user()
is_dataRetrieval_user <- function() {
  interactive() ||
    !identical(Sys.getenv("CI"), "") ||
    identical(Sys.getenv("NOT_CRAN"), "true")
}

wqp_message <- function() {
  message(
    "NEWS: Data does not include USGS data newer than March 11, 2024. More details:
https://doi-usgs.github.io/dataRetrieval/articles/Status.html"
  )
}

wqp_message_beta <- function() {
  message("WQX3 services are in-development, use with caution.")
}

is_legacy <- function(service) {
  legacy <- service %in%
    c(
      "Result",
      "Station",
      "Activity",
      "Organization",
      "ActivityMetric",
      "SiteSummary",
      "Project",
      "ProjectMonitoringLocationWeighting",
      "ResultDetectionQuantitationLimit",
      "BiologicalMetric"
    )
  return(legacy)
}

nwis_message <- function() {
  return(
    "WARNING: whatNWISdata does not include
discrete water quality data newer than March 11, 2024.
For additional details, see:
https://doi-usgs.github.io/dataRetrieval/articles/Status.html"
  )
}

new_nwis_message <- function() {
  return(
    "ALERT: All NWIS services are slated for decommission
and new dataRetrieval functions will be added.
For up-to-date information, see: 
https://doi-usgs.github.io/dataRetrieval/articles/Status.html"
  )
}

#' Pad string with leading zeros
#'
#' Function to pad a string with leading zeros. Useful for parameter codes and USGS site IDs.
#'
#' @param x character
#' @param padTo number Final desired length of the character
#' @keywords data import USGS web service
#' @return x character returned with leading zeros
#' @export
#' @examples
#' pCode <- "10"
#' correctPCode <- zeroPad(pCode, 5)
#' pCodes <- c("100", "1000", "0", "12345", "1565465465465465")
#' correctPCodes <- zeroPad(pCodes, 5)
#' pCodeNA <- c(1, 2, NA)
#' padPCodeNA <- zeroPad(pCodeNA, 4)
zeroPad <- function(x, padTo) {
  if (padTo <= 1) {
    return(x)
  }

  numDigits <- nchar(x, keepNA = TRUE)
  padding <- padTo - numDigits

  if (anyNA(padding)) {
    padding[is.na(padding)] <- 0
  }

  paddingZeros <- vapply(
    X = padding[padding > 0],
    FUN = function(y) paste0(rep("0", y), collapse = ""),
    FUN.VALUE = ""
  )

  x[padding > 0] <- paste0(paddingZeros, x[padding > 0])
  return(x)
}

#' USGS data availability
#'
#' Imports a table of available parameters, period of record, and count. See
#' <https://waterservices.usgs.gov/docs/site-service/>
#' for more information.
#'
#' @param \dots see <https://waterservices.usgs.gov/docs/site-service/>
#' for a complete list of options.  A list of arguments can also be supplied.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function will
#' convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character
#' @keywords data import USGS web service
#'
#' @details This function requires users to create their own arguments
#' based on the NWIS web services. It is a more complicated function to use
#' compared to other NWIS functions such as [readNWISdv()], [readNWISuv()],
#' etc. However, this function adds a lot of
#' flexibility to the possible queries. If the "service" argument is included,
#' the results will be filtered to the proper data_type_cd. This is a great
#' function to use before a large data set, by filtering down the number
#' of sites that have useful data.
#'
#'
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#' station_nm \tab character \tab Site name \cr
#' site_tp_cd \tab character \tab Site type \cr
#' dec_lat_va \tab numeric \tab Decimal latitude\cr
#' dec_long_va \tab numeric \tab Decimal longitude \cr
#' coord_acy_cd \tab character \tab Latitude-longitude accuracy \cr
#' dec_coord_datum_cd \tab character \tab Decimal Latitude-longitude datum \cr
#' alt_va \tab character \tab Altitude of Gage or land surface \cr
#' alt_acy_va \tab character \tab Altitude accuracy \cr
#' alt_datum_cd \tab character \tab Altitude datum \cr
#' huc_cd \tab character \tab Hydrologic unit code \cr
#' data_type_cd \tab character \tab Data type \cr
#' parm_cd \tab character \tab Parameter code \cr
#' stat_cd \tab character \tab Statistical code \cr
#' dd_nu \tab character \tab Internal database key \cr
#' loc_web_ds \tab character \tab Additional measurement description \cr
#' medium_grp_cd \tab character \tab Medium group code \cr
#' parm_grp_cd \tab character \tab Parameter group code \cr
#' srs_id \tab character \tab SRS ID \cr
#' access_cd \tab character \tab Access code \cr
#' begin_date \tab Date \tab Begin date \cr
#' end_date \tab Date \tab End date \cr
#' count_nu \tab integer \tab Record count\cr
#' parameter_group_nm \tab character \tab Parameter group name \cr
#' parameter_nm \tab character \tab Parameter name \cr
#' casrn \tab character \tab Chemical Abstracts Service (CAS) Registry Number \cr
#' srsname \tab character \tab Substance Registry Services \cr
#' parameter_units \tab character \tab Parameter units \cr
#' }
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' comment \tab character \tab Header comments from the RDB file \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#' @export
#' @seealso [read_waterdata_ts_meta()]
#' @examples
#'
#' # see ?read_waterdata_ts_meta
#'
#' #site1 <- whatWQPsamples(siteid = "USGS-01594440")
#'
#' #type <- "Stream"
#'
#' #sites <- whatWQPsamples(countycode = "US:55:025", siteType = type)
#'
#' #lakeSites_samples <- whatWQPsamples(siteType = "Lake, Reservoir, Impoundment",
#' #                                    countycode = "US:55:025")
#'
#'
whatNWISdata <- function(..., convertType = TRUE) {
  matchReturn <- convertLists(...)

  prewarned <- FALSE

  .Deprecated(
    new = "read_waterdata_ts_meta",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_ts_meta"
  )

  if ("service" %in% names(matchReturn)) {
    service <- matchReturn$service

    if (any(service %in% c("qw", "qwdata"))) {
      .Deprecated(
        old = "whatNWISdata",
        package = "dataRetrieval",
        new = "whatWQPdata",
        msg = nwis_message()
      )
      prewarned <- TRUE
    }
  } else {
    service <- "all"
  }

  if (any(service == "site")) {
    service <- "all"
  } else if (any(service == "iv")) {
    service[service == "iv"] <- "uv"
  } else if (any(service == "peak")) {
    service[service == "peak"] <- "pk"
  }

  if ("statCd" %in% names(matchReturn)) {
    statCd <- matchReturn$statCd
    matchReturn <- matchReturn[names(matchReturn) != "statCd"]
  } else {
    statCd <- "all"
  }

  if ("parameterCd" %in% names(matchReturn)) {
    parameterCd <- matchReturn$parameterCd
    matchReturn[["parameterCd"]] <- NULL
  } else {
    parameterCd <- "all"
  }

  if (
    "startDate" %in% names(matchReturn) && !"endDate" %in% names(matchReturn)
  ) {
    matchReturn[["endDate"]] <- matchReturn[["startDate"]]
  }

  matchReturn$service <- "site"

  valuesList <- readNWISdots(matchReturn)

  values <- valuesList[["values"]]
  values <- values[names(values) != "format"]

  POST <- nchar(paste0(unlist(values), collapse = "")) > 2048

  urlSitefile <- httr2::request(pkg.env[["site"]])
  urlSitefile <- get_or_post(
    urlSitefile,
    POST = POST,
    seriesCatalogOutput = "true"
  )
  urlSitefile <- get_or_post(
    urlSitefile,
    POST = POST,
    !!!values,
    .multi = "comma"
  )

  SiteFile <- importRDB1(
    urlSitefile,
    asDateTime = FALSE,
    convertType = convertType
  )

  if (!("all" %in% service)) {
    SiteFile <- SiteFile[SiteFile$data_type_cd %in% service, ]
  }
  if (!("all" %in% statCd)) {
    SiteFile <- SiteFile[SiteFile$stat_cd %in% c(statCd, NA), ]
  }
  if (!("all" %in% parameterCd)) {
    SiteFile <- SiteFile[SiteFile$parm_cd %in% parameterCd, ]
  }

  if (nrow(SiteFile) > 0 && convertType) {
    SiteFile$begin_date <- as.Date(suppressWarnings(lubridate::parse_date_time(
      SiteFile$begin_date,
      c("Ymd", "mdY", "Y!")
    )))
    SiteFile$end_date <- as.Date(suppressWarnings(lubridate::parse_date_time(
      SiteFile$end_date,
      c("Ymd", "mdY", "Y!")
    )))
  }

  if (any(SiteFile$data_type_cd == "qw")) {
    if (!prewarned) {
      message(nwis_message())
    }
  }
  return(SiteFile)
}

#' Parameter code to characteristic name
#'
#' This function is useful to fine what characteristic name, result sample
#' fraction, unit code, and other parameters are mapped with USGS parameter
#' codes. This information is useful for converting workflows from a more
#' traditional NWIS water quality retrieval to a Water Quality Portal retrieval.
#'
#' @export
#' @param parameterCd character that contains the code for a character vector
#' of 5-digit parameter codes. Default is "all" which will return a complete
#' list of parameter codes that have been mapped to a characteristic name.
#' @return a data frame with columns  "parm_cd", "description",
#' "characteristicname", "measureunitcode", "resultsamplefraction",
#' "resulttemperaturebasis", "resultstatisticalbasis",  "resulttimebasis",
#' "resultweightbasis", "resultparticlesizebasis", "last_rev_dt"
#' @examples
#' pcodes <- c("00070", "00075", "00430", "52642")
#' \donttest{
#'
#' all <- pcode_to_name()
#' some <- pcode_to_name(pcodes)
#'
#' }
pcode_to_name <- function(parameterCd = "all") {
  parameterCd.orig <- parameterCd
  parameterCd <- parameterCd[!is.na(parameterCd)]

  url_all <- "https://www.waterqualitydata.us/Codes/public_srsnames/?mimeType=json"
  doc <- get_nldi_sources(url_all)
  retval <- doc[["pcodes"]]

  if (all(tolower(parameterCd) != "all")) {
    retval <- retval[retval$parm_cd %in% parameterCd, ]
  }

  attr(retval, "url") <- url_all

  if (any(parameterCd != "all")) {
    if (nrow(retval) != length(unique(parameterCd))) {
      badPcode <- parameterCd[!(parameterCd %in% retval$parm_cd)]
      warning(
        "The following pCodes seem mistyped, and no information was returned: ",
        paste(badPcode, collapse = ", ")
      )
    }
  }

  if (nrow(retval) != sum(is.na(parameterCd.orig))) {
    na.params <- data.frame(matrix(
      ncol = ncol(retval),
      nrow = sum(is.na(parameterCd.orig))
    ))
    names(na.params) <- names(retval)
    retval <- rbind(retval, na.params)
  }

  # order by parameterCd.orig
  if (!isTRUE(parameterCd.orig == "all")) {
    retval <- retval[match(parameterCd.orig, retval$parm_cd), ]
    retval$parm_cd <- parameterCd.orig
  }

  return(retval)
}

#' USGS Parameter Data Retrieval
#'
#' Imports data from NWIS about measured parameter based on user-supplied parameter code or codes.
#'
#' @param parameterCd character of USGS parameter codes (or multiple parameter codes).
#' These are 5 digit number codes. To get a complete list of all current parameter
#' codes in the USGS, use "all" as the input.
#' @keywords data import USGS web service
#' @return parameterData data frame with the following information:
#' \tabular{lll}{
#'   Name \tab Type \tab Description\cr
#'   parameter_cd \tab character \tab 5-digit USGS parameter code \cr
#'   parameter_group_nm \tab character \tab USGS parameter group name\cr
#'   parameter_nm \tab character \tab USGS parameter name\cr
#'   casrn \tab character \tab Chemical Abstracts Service (CAS) Registry Number\cr
#'   srsname \tab character \tab Substance Registry Services Name\cr
#'   parameter_units \tab character \tab Parameter units\cr
#' }
#'
#' @export
#' @seealso [importRDB1()]
readNWISpCode <- function(parameterCd) {
  .Deprecated(
    new = "read_waterdata_parameter_codes",
    package = "dataRetrieval",
    msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_parameter_codes"
  )

  parameterCd.orig <- parameterCd
  parameterCd <- parameterCd[!is.na(parameterCd)]
  baseURL <- httr2::request(pkg.env[["pCode"]])
  fullURL <- httr2::req_url_query(baseURL, fmt = "rdb", group_cd = "%")

  if (any(parameterCd == "all")) {
    temp_df <- importRDB1(fullURL, asDateTime = FALSE)
    parameterData <- data.frame(
      parameter_cd = temp_df$parm_cd,
      parameter_group_nm = temp_df$group,
      parameter_nm = temp_df$parm_nm,
      casrn = temp_df$CASRN,
      srsname = temp_df$SRSName,
      parameter_units = temp_df$parm_unit,
      stringsAsFactors = FALSE
    )
    attr(parameterData, "url") <- fullURL$url
  } else {
    parameterData <- parameterCdFile[
      parameterCdFile$parameter_cd %in% parameterCd,
    ]

    if (nrow(parameterData) != length(parameterCd)) {
      parameterCd_lookup <- parameterCd[
        !parameterCd %in% unique(parameterData$parameter_cd)
      ]

      if (length(parameterCd_lookup) == 1) {
        baseURL <- httr2::request(pkg.env[["pCodeSingle"]])
        baseURL <- httr2::req_url_query(baseURL, fmt = "rdb")
        baseURL <- httr2::req_url_query(
          baseURL,
          parm_nm_cd = parameterCd_lookup
        )

        temp_df <- importRDB1(baseURL, asDateTime = FALSE)

        temp_df <- data.frame(
          parameter_cd = temp_df$parameter_cd,
          parameter_group_nm = temp_df$group,
          parameter_nm = temp_df$parm_nm,
          casrn = temp_df$CASRN,
          srsname = temp_df$SRSName,
          parameter_units = temp_df$parm_unit,
          stringsAsFactors = FALSE
        )

        if (nrow(temp_df) > 0) {
          parameterData <- rbind(parameterData, temp_df)
        }

        attr(parameterData, "url") <- baseURL$url
      } else {
        temp_df <- importRDB1(fullURL, asDateTime = FALSE)
        trim_df <- data.frame(
          parameter_cd = temp_df$parm_cd,
          parameter_group_nm = temp_df$group,
          parameter_nm = temp_df$parm_nm,
          casrn = temp_df$CASRN,
          srsname = temp_df$SRSName,
          parameter_units = temp_df$parm_unit,
          stringsAsFactors = FALSE
        )
        parameterData <- trim_df[trim_df$parameter_cd %in% parameterCd, ]
        attr(parameterData, "url") <- fullURL$url
      }

      if (nrow(parameterData) != length(parameterCd)) {
        badPcode <- parameterCd[!(parameterCd %in% parameterData$parameter_cd)]
        warning(
          "The following pCodes seem mistyped, and no information was returned: ",
          paste(badPcode, collapse = ", ")
        )
      }
    }
  }

  if (nrow(parameterData) != sum(is.na(parameterCd.orig))) {
    na.params <- data.frame(matrix(
      ncol = ncol(parameterData),
      nrow = sum(is.na(parameterCd.orig))
    ))
    names(na.params) <- names(parameterData)
    parameterData <- rbind(parameterData, na.params)
  }
  # order by parameterCd.orig
  if (!isTRUE(parameterCd.orig == "all")) {
    parameterData <- parameterData[
      match(parameterCd.orig, parameterData$parameter_cd),
    ]
    parameterData$parameter_cd <- parameterCd.orig
  }
  return(parameterData)
}

#' Get Continuous USGS Water Data
#'
#' @description `r get_description("continuous")`
#'
#' Currently, the services only allow up to 3 years of data to be requested with
#' a single request. If no "time" is specified, the service will return the
#' last single year of data. If this is a bottleneck, please check back
#' for new direct download functions that are expected to be available sometime
#' in 2026.
#'
#' Geometry output is not supported in the continuous data API endpoint.
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("continuous")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param parameter_code `r get_ogc_params("continuous")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param time `r get_ogc_params("continuous")$time`
#'
#' See also Details below for more information.
#'
#' @param value `r get_ogc_params("continuous")$value`
#' @param unit_of_measure `r get_ogc_params("continuous")$unit_of_measure`
#' @param approval_status `r get_ogc_params("continuous")$approval_status`
#' @param last_modified `r get_ogc_params("continuous")$last_modified`
#'
#' See also Details below for more information.
#' @param time_series_id `r get_ogc_params("continuous")$time_series_id`
#' Multiple time_series_ids can be requested as a character vector.
#' @param qualifier `r get_ogc_params("continuous")$qualifier`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("continuous", "continuous_id")`.
#' The default (`NA`) will return all columns of the data.
#' @param \dots Not used. Included to help differentiate official Water Data API arguments
#' from more seldom used, optional dataRetrieval-specific arguments.
#' @inheritParams check_arguments_non_api
#'
#' @details
#' You can also use a vector of length 2 for any time queries (such as time
#' or last_modified). The first value is the starting date (or datetime),
#' the second value is the ending date(or datetime).
#' NA's within the vector indicate a half-bound date.
#' For example, \code{time = c("2024-01-01", NA)} will return all data starting
#' at 2024-01-01.
#' \code{time = c(NA, "2024-01-01")} will return all data from the beginning of
#' the timeseries until 2024-01-01.
#' By default, time is assumed UTC, although time zone attributes
#' will be accommodated. As an example, setting \code{time = as.POSIXct(c("2021-01-01 12:00:00",
#' "2021-01-01 14:00"), tz = "America/New_York")} will request data that between
#' noon and 2pm eastern time on 2021-01-01.
#' All time values RETURNED from the service are UTC with the exception of
#' daily data, which returns time values in local dates.
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-451605097071701"
#' pcode <- "72019"
#'
#' uv_data_trim <- read_waterdata_continuous(monitoring_location_id = site,
#'                            parameter_code = pcode,
#'                            properties = c("value", "time"),
#'                            time = as.POSIXct(c("2026-02-07 12:00",
#'                                                "2026-02-08 12:00"),
#'                                               tz = "America/Chicago"))
#'
#' uv_data <- read_waterdata_continuous(monitoring_location_id = site,
#'                            parameter_code = pcode,
#'                            time = "P2D")
#'
#'
#' # Only return data that has been modified in last 7 days
#' multi_site2 <- read_waterdata_continuous(monitoring_location_id =  c("USGS-451605097071701",
#'                                                                      "USGS-14181500"),
#'                                                parameter_code = c("00060", "72019"),
#'                                                last_modified = "P7D")
#'
#' # how to split up request into roughly 3 year chunks
#'
#' site <- "USGS-0208458892"
#' pcode <- "00095" # Specific conductance
#' start <- as.Date("2013-01-01")
#' end <- as.Date("2025-12-31")
#'
#' n_days <- difftime(end, start, units = "days")
#'
#' # create a vector of dates that are about 3 years apart:
#' time_chunks <- seq(from = start,
#'                    to = end,
#'                    length.out = ceiling(n_days/(3*365.25)) + 1)
#'
#'
#' # create a list where each element starts at the beginning
#' # of a chunk, and ends the day before the next chunk:
#' time_df <- data.frame(start = time_chunks[-length(time_chunks)],
#'                       end = time_chunks[-1]-1)
#'
#' #all_data <- data.frame()
#' #for(i in seq_along(time_df$start)){
#' #   sub_df <- read_waterdata_continuous(monitoring_location_id = site,
#' #                                       parameter_code = pcode,
#' #                                      time = c(time_df$start[i],
#' #                                                time_df$end[i]))
#' #   all_data <- rbind(all_data, sub_df)
#' # }
#'
#' # Set the time to Eastern:
#' # all_data$time <- lubridate::force_tz(all_data$time, "America/New_York")
#'
#' }
read_waterdata_continuous <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  properties = NA_character_,
  time_series_id = NA_character_,
  approval_status = NA_character_,
  unit_of_measure = NA_character_,
  qualifier = NA_character_,
  value = NA,
  last_modified = NA_character_,
  time = NA_character_,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  limit = getOption("dataRetrieval.limit"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_data"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "continuous"
  output_id <- "continuous_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  args[["skipGeometry"]] <- TRUE

  return_list <- get_ogc_data(args, output_id, service)

  attr(return_list$time, "tzone") <- "UTC"
  return(return_list)
}

#' General Data Import from NWIS
#'
#' Returns data from the NWIS web service.
#' Arguments to the function should be based on <https://waterservices.usgs.gov> service calls.
#' See examples below for ideas of constructing queries.
#'
#' @param asDateTime logical, if `TRUE` returns date and time as POSIXct, if `FALSE`, Date
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the
#' function will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the data's provided tz_cd column.
#' Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
#' "America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  `OlsonNames()`
#' for more information on time zones.
#' @param \dots see <https://waterservices.usgs.gov/docs/site-service/> for
#' a complete list of options.  A
#' list of arguments can also be supplied. One important argument to include is
#' "service". Possible values are "iv"
#' (for instantaneous),
#' "dv" (for daily values),
#' "site" (for site service), and "stat" (for
#' statistics service).
#' The statistics service has a limited selection of arguments
#' (see <https://waterservices.usgs.gov/docs/site-service/>).
#'
#' @details This function requires users to create their own arguments
#' based on the NWIS web services. It is a more complicated function to use
#' compared to other NWIS functions such as [readNWISdv()], [readNWISuv()],
#' etc. However, this function adds a lot of
#' flexibility to the possible queries. This function will also behave exactly
#' as NWIS when it comes to date queries. NWIS by default will only return the latest
#' value for the daily and instantaneous services. So if you do not provide
#' a starting date, you will only get back the latest value. If you want
#' the full period of record, you can use "startDate = '1900-01-01'". Other options
#' for dates are periods, such as "period = 'P7D'" which translates to a period
#' of 7 days. For period, use only a positive ISO-8601 duration format, which should
#' not be expressed in periods of less than a day, or in increments of months M or years Y.
#' period returns data for a site generally from now to a time in the past.
#' Note that when period is used all data up to the most recent value are returned.
#'
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency \tab character \tab The NWIS code for the agency reporting the data\cr
#' site \tab character \tab The USGS site number \cr
#' dateTime \tab POSIXct \tab The date and time (if applicable) of the measurement,
#'           converted to UTC for unit value data. R only allows one time zone attribute per column. For unit data
#'           spanning a time zone change, converting the data to UTC solves this problem. For daily data,
#'           the time zone attribute is the time zone of the first returned measurement.
#'            \cr
#' tz_cd \tab character \tab The time zone code for dateTime column\cr
#' code \tab character \tab Any codes that qualify the corresponding value\cr
#' value \tab numeric \tab The numeric value for the parameter \cr
#' }
#' Note that code and value are repeated for the parameters requested. The names are of the form
#' X_D_P_S, where X is literal,
#' D is an option description of the parameter,
#' P is the parameter code,
#' and S is the statistic code (if applicable).
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
#' statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#'
#' @seealso [read_waterdata()]
#' @export
readNWISdata <- function(
  ...,
  asDateTime = TRUE,
  convertType = TRUE,
  tz = "UTC"
) {
  tz <- match.arg(tz, OlsonNames())

  valuesList <- readNWISdots(...)

  values <- valuesList[["values"]]
  values <- values[names(values) != "format"]
  format <- valuesList[["values"]][["format"]]

  service <- valuesList$service
  if (length(service) > 1) {
    warning(
      "Only one service value is allowed. Service: ",
      service[1],
      " will be used."
    )
    service <- service[1]
  }

  if (any(service %in% c("qw", "qwdata"))) {
    .Deprecated(
      old = "readNWISdata",
      package = "dataRetrieval",
      new = "readWQPdata",
      msg = "NWIS qw web services are being retired.
Please see vignette('qwdata_changes', package = 'dataRetrieval')
for more information.
https://cran.r-project.org/web/packages/dataRetrieval/vignettes/qwdata_changes.html"
    )
  } else if (service == "dv") {
    .Deprecated(
      new = "read_waterdata_daily",
      package = "dataRetrieval",
      msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_daily."
    )
  } else if (service == "site") {
    .Deprecated(
      new = "read_waterdata_monitoring_location",
      package = "dataRetrieval",
      msg = "NWIS servers are slated for decommission. Please begin to migrate to read_waterdata_monitoring_location"
    )
  } else {
    message(new_nwis_message())
  }

  baseURL <- httr2::request(pkg.env[[service]])

  POST <- nchar(paste0(unlist(values), collapse = "")) > 2048

  baseURL <- get_or_post(baseURL, POST = POST, !!!values, .multi = "comma")

  if (service != "rating") {
    baseURL <- get_or_post(baseURL, POST = POST, format = format)
  }

  if (length(grep("rdb", format)) > 0) {
    retval <- importRDB1(
      baseURL,
      tz = tz,
      asDateTime = asDateTime,
      convertType = convertType
    )
  } else {
    retval <- importWaterML1(baseURL, tz = tz, asDateTime = asDateTime)
  }

  if ("dv" == service) {
    tzLib <- stats::setNames(
      c(
        "America/New_York",
        "America/New_York",
        "America/Chicago",
        "America/Chicago",
        "America/Denver",
        "America/Denver",
        "America/Los_Angeles",
        "America/Los_Angeles",
        "America/Anchorage",
        "America/Anchorage",
        "America/Honolulu",
        "America/Honolulu",
        "UTC"
      ),
      c(
        "EST",
        "EDT",
        "CST",
        "CDT",
        "MST",
        "MDT",
        "PST",
        "PDT",
        "AKST",
        "AKDT",
        "HAST",
        "HST",
        "UTC"
      )
    )
    # TODO: Think about dates that cross a time zone boundary.
    if (format == "waterml,1.1" && nrow(retval) > 0) {
      retval$dateTime <- as.POSIXct(
        retval$dateTime,
        tzLib[tz = retval$tz_cd[1]]
      )
    }
  }

  if ("iv" == service || "iv_recent" == service) {
    if (tz == "") {
      retval$tz_cd <- rep("UTC", nrow(retval))
    } else {
      retval$tz_cd <- rep(tz, nrow(retval))
    }
  }

  return(retval)
}

#' State code look up
#'
#' Function to simplify finding state and state code definitions. Used in `readNWISdata`
#' and `readWQPdata`.
#'
#' @param input could be character (full name, abbreviation, id), or numeric (id)
#' @param country description
#' @param outputType character can be "postal", "fullName", "tableIndex", or "id".
#' @export
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' fullName <- stateCdLookup("wi", "fullName")
#' abbriev <- stateCdLookup("Wisconsin", "postal")
#' id <- stateCdLookup("WI", "id")
#' name <- stateCdLookup(55, "fullName")
#' fips <- stateCdLookup("WI", "fips")
#' canada_st <- stateCdLookup(13, "fullName", country = "CA")
#' mexico_st <- stateCdLookup(13, "fullName", country = "MX")
#' stateCdLookup(c("West Virginia", "Wisconsin", 200, 55, "MN"))
#' }
stateCdLookup <- function(input, outputType = "postal", country = "US") {
  outputType <- match.arg(outputType, c("postal", "fullName", "id", "fips"))

  states <- check_waterdata_sample_params("states")
  country <- match.arg(
    country,
    choices = unique(states$countryCode),
    several.ok = FALSE
  )
  states <- states[states$countryCode == country, ]
  retVal <- rep(NA, length(input))
  index <- 1
  for (i in input) {
    if (is.numeric(i) || !is.na(suppressWarnings(as.numeric(i)))) {
      out_i <- which(as.numeric(i) == as.numeric(states$fipsCode))
    } else if (nchar(i) == 2) {
      out_i <- which(tolower(i) == tolower(states$stateAbbrev))
    } else {
      out_i <- which(tolower(i) == tolower(states$stateName))
    }

    if (length(out_i) > 0) {
      output <- switch(
        outputType,
        postal = states$stateAbbrev[out_i],
        fullName = states$stateName[out_i],
        id = as.integer(states$fipsCode[out_i]),
        fips = paste(
          states$countryCode[out_i],
          states$fipsCode[out_i],
          sep = ":"
        )
      )

      retVal[index] <- output
    }

    index <- index + 1
  }

  if (length(retVal[-1]) == 0) {
    paste(
      "Could not find",
      input,
      "in the state lookup table. See `stateCd` for complete list."
    )
  }

  return(retVal)
}

#' US county code look up
#'
#' Function to simplify finding county and county code definitions. Used in `readNWISdata`
#' and `readNWISuse`. Currently only has US counties.
#'
#' @param state could be character (full name, abbreviation, id), or numeric (id)
#' @param county could be character (name, with or without "County") or numeric (id)
#' @param outputType character can be "fullName", "tableIndex", "id", or "fullEntry".
#' @export
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#'
#' fips <- countyCdLookup(state = "WI", county = "Dane")
#' id <- countyCdLookup(state = "WI", county = "Dane", outputType = "id")
#' name <- countyCdLookup(state = "OH", county = 13, output = "fullName")
#' entry <- countyCdLookup(state = "Pennsylvania", county = "ALLEGHENY COUNTY", output = "fullEntry")
#' fromIDs <- countyCdLookup(state = 13, county = 5, output = "fullName")
#' }
countyCdLookup <- function(state, county, outputType = "fips") {
  outputType <- match.arg(outputType, c("fullName", "fullEntry", "fips", "id"))

  if (missing(state)) {
    stop("No state code provided")
  }

  if (missing(county)) {
    stop("No county code provided")
  }

  if (length(state) > 1) {
    stop("Only one state allowed in countyCdLookup.")
  }

  counties <- check_waterdata_sample_params("counties")

  # first turn state into stateCd postal name
  state_postal <- stateCdLookup(
    state,
    country = unique(counties$countryCode),
    outputType = "postal"
  )
  state_counties <- counties[counties$stateAbbrev == state_postal, ]

  if (is.numeric(county) || !is.na(suppressWarnings(as.numeric(county)))) {
    county_i <- which(
      as.numeric(county) == as.numeric(state_counties$countyCode)
    )
  } else {
    county_i <- grep(
      tolower(gsub(" ", "", county)),
      tolower(gsub(" ", "", state_counties$countyName))
    ) # takes care of questionable spaces...DeWitt vs De Witt
    if (length(county_i) == 0) {
      county <- gsub(" county", "", county, ignore.case = TRUE)
      county_i <- grep(
        tolower(gsub(" ", "", county)),
        tolower(gsub(" ", "", state_counties$countyName))
      )
    }
  }

  if (length(county_i) == 0) {
    stop(paste(
      "Could not find",
      county,
      "(county), ",
      state_postal,
      "(state) in the county lookup table. See `countyCd` for complete list."
    ))
  } else if (length(county_i) > 1) {
    stop(paste(
      county,
      "(county), ",
      state_postal,
      "(state) matched more than one county. See `countyCd` for complete list."
    ))
  }

  fips_val <- paste(
    stateCdLookup(state_counties$stateAbbrev[county_i], outputType = "fips"),
    state_counties$countyCode[county_i],
    sep = ":"
  )
  retVal <- switch(
    outputType,
    fullName = state_counties$countyName[county_i],
    fullEntry = state_counties[county_i, ],
    fips = fips_val,
    id = state_counties$countyCode[county_i]
  )

  return(retVal)
}

#'
#' Format and organize NWIS arguments that are passed in as `...`.
#'
#' @keywords internal
readNWISdots <- function(...) {
  if (length(list(...)) == 0) {
    stop("No arguments supplied")
  }

  matchReturn <- convertLists(...)

  if (anyNA(unlist(matchReturn))) {
    stop("NA's are not allowed in query")
  }

  if ("service" %in% names(matchReturn)) {
    service <- matchReturn$service
    matchReturn$service <- NULL
  } else {
    service <- "dv"
  }

  match.arg(
    service,
    c(
      "dv",
      "iv",
      "iv_recent",
      "uv",
      "site",
      "qwdata",
      "stat",
      "rating",
      "peak"
    )
  )

  if (service == "uv") {
    service <- "iv"
  }

  if (length(service) > 1) {
    stop("Only one service call allowed.")
  }

  values <- matchReturn

  names(values)[names(values) == "startDate"] <- "startDT"
  names(values)[names(values) == "endDate"] <- "endDT"
  names(values)[names(values) == "siteNumber"] <- "sites"
  names(values)[names(values) == "siteNumbers"] <- "sites"

  format.default <- "waterml,1.1"

  if (service == "iv" && "startDT" %in% names(values)) {
    if (as.Date(values[["startDT"]]) >= Sys.Date() - 120) {
      service <- "iv_recent"
    }
  }

  names(values)[names(values) == "statecode"] <- "stateCd"
  if ("stateCd" %in% names(values)) {
    state <- stateCdLookup(values["stateCd"], "postal")

    if (length(state) > 1) {
      stop("Multiple states are not allowed in NWIS queries.")
    }

    values["stateCd"] <- state
    values[["stateCd"]][values["stateCd"] == "AS"] <- "AQ" # Leaving the correct abb in stateCd
    if (values["stateCd"] == "UM") {
      stop("NWIS does not include U.S. Minor Outlying Islands")
    }
  }

  if ("parameterCd" %in% names(matchReturn)) {
    pcodeCheck <- (nchar(matchReturn$parameterCd) == 5) &
      !is.na(suppressWarnings(as.numeric(matchReturn$parameterCd)))
    if (!all(pcodeCheck)) {
      badPcode <- matchReturn$parameterCd[which(!pcodeCheck)]
      stop(
        "The following pCodes appear mistyped:",
        paste(badPcode, collapse = ", ")
      )
    }
  }

  names(values)[names(values) == "countycode"] <- "countyCd"
  if ("countyCd" %in% names(values)) {
    if ("stateCd" %in% names(values)) {
      values["countyCd"] <- paste0(
        zeroPad(stateCdLookup(values["stateCd"], "id"), padTo = 2),
        countyCdLookup(values["stateCd"], values["countyCd"], "id")
      )
      values <- values[names(values) != "stateCd"]
    }
  }

  if (service %in% c("peak")) {
    format.default <- "rdb"

    names(values)[names(values) == "startDT"] <- "begin_date"
    names(values)[names(values) == "endDT"] <- "end_date"
    names(values)[names(values) == "sites"] <- "site_no"

    if ("bBox" %in% names(values)) {
      values["nw_longitude_va"] <- as.character(matchReturn$bBox[1])
      values["nw_latitude_va"] <- as.character(matchReturn$bBox[2])
      values["se_longitude_va"] <- as.character(matchReturn$bBox[3])
      values["se_latitude_va"] <- as.character(matchReturn$bBox[4])
      values["coordinate_format"] <- "decimal_degrees"
      values <- values[-which("bBox" %in% names(values))]
    }

    values["date_format"] <- "YYYY-MM-DD"
    values["rdb_inventory_output"] <- "file"
    values["TZoutput"] <- "0"

    if (all(c("begin_date", "end_date") %in% names(values))) {
      values["range_selection"] <- "date_range"
    }
  }

  if ("bbox" %in% names(values)) {
    values[["bbox"]] <- paste0(values[["bbox"]], collapse = ",")
  }

  if (service %in% c("peak") && "stateCd" %in% names(values)) {
    names(values)[names(values) == "stateCd"] <- "state_cd"
    values["list_of_search_criteria"] <- "state_cd"
  }

  if (service %in% c("peak") && "huc2_cd" %in% names(values)) {
    values["list_of_search_criteria"] <- "huc2_cd"
  }

  if (service %in% c("peak") && "bBox" %in% names(values)) {
    values["list_of_search_criteria"] <- "lat_long_bounding_box"
  }

  if (service %in% c("site", "stat", "rating", "peak")) {
    format.default <- "rdb"
  }

  if (service == "stat") {
    message(
      "Please be aware the NWIS data service feeding this function is in BETA.\n
            Data formatting could be changed at any time, and is not guaranteed"
    )
  }

  if (!("format" %in% names(values))) {
    values["format"] <- format.default
  }
  return_list <- list()
  return_list["values"] <- list(values)
  return_list["service"] <- service
  return(return_list)
}

#' convert variables in dots to usable format
#'
#' @keywords internal
convertLists <- function(...) {
  matchReturn <- c(
    do.call("c", list(...)[sapply(list(...), class) == "list"]), # get the list parts
    list(...)[sapply(list(...), class) != "list"]
  ) # get the non-list parts
  return(matchReturn)
}

context("General functions")

test_that("General USGS retrievals working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  cql <- '{
  "op": "and",
  "args": [
    {
      "op": "in",
      "args": [
        { "property": "parameter_code" },
        [ "00060", "00065" ]
      ]
    },
   {
      "op": "in",
      "args": [
        { "property": "monitoring_location_id" },
        [ "USGS-07367300", "USGS-03277200" ]
      ]
    }
  ]
  }'

  dv_data <- read_waterdata(
    service = "daily",
    CQL = cql,
    time = c("2023-01-01", "2024-01-01")
  )

  expect_equal(as.Date(c("2023-01-01", "2024-01-01")), range(dv_data$time))
  expect_true(all(
    unique(dv_data$monitoring_location_id) %in%
      c("USGS-07367300", "USGS-03277200")
  ))

  cql_not_active <- '{
  "op": "and",
  "args": [
    {
      "op": "in",
      "args": [
        { "property": "parameter_code" },
        [ "00060", "00065" ]
      ]
    },
   {
      "op": "in",
      "args": [
        { "property": "monitoring_location_id" },
        [ "USGS-05212700"]
      ]
    }
  ]
  }'

  notActiveUSGS <- read_waterdata(
    CQL = cql_not_active,
    service = "daily",
    time = c("2014-01-01", "2014-01-07")
  )
  expect_true(nrow(notActiveUSGS) == 0)
})

test_that("General NWIS retrievals working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  multiSite <- read_waterdata_continuous(
    monitoring_location_id = c("USGS-04025500", "USGS-040263491"),
    parameter_code = "00060",
    time = c("2020-11-01", "2020-11-02")
  )

  expect_is(multiSite$time, "POSIXct")

  expect_error(readNWISdata(), "No arguments supplied")
  expect_error(readNWISdata(siteNumber = NA), "NA's are not allowed in query")

  bBox_inventory <- read_waterdata_ts_meta(
    bbox = c(-83, 38, -82.5, 38.5),
    parameter_code = "00010"
  )

  expect_true(length(unique(bBox_inventory$monitoring_location_id)) > 1)

  siteInfo <- read_waterdata_monitoring_location(state_name = "Wisconsin")

  timeseriesInfo <- read_waterdata_ts_meta(
    bbox = sf::st_bbox(siteInfo),
    parameter_code = "00010",
    computation_period_identifier = "Points"
  )

  expect_is(timeseriesInfo$begin, "POSIXct")

  # nolint start: line_length_linter

  # Test list:
  args <- list(
    monitoring_location_id = "USGS-05114000",
    parameter_code = "00060",
    time = c("2014-05-01T00:00Z", "2014-05-01T12:00Z")
  )

  instData <- do.call(read_waterdata_continuous, args)

  args2 <- list(
    monitoring_location_id = "USGS-05114000",
    parameter_code = "00060",
    time = c("2014-05-01", "2014-05-01")
  )

  daily_USGS <- do.call(read_waterdata_daily, args2)
  expect_lt(nrow(daily_USGS), nrow(instData))

  ohio <- read_waterdata_monitoring_location(
    state_name = "Ohio",
    site_type_code = "ST"
  )
  bbox <- sf::st_bbox(ohio)
  what_sites <- read_waterdata_ts_meta(parameter_code = "00665", bbox = bbox)
  expect_true(all(
    c("monitoring_location_id", "begin", "end", "parameter_name") %in%
      names(what_sites)
  ))

  huc <- read_waterdata_monitoring_location(hydrologic_unit_code = "02080202")
  expect_true(nrow(huc) > 0)

  # Test counties:

  county_code_stafford <- countyCdLookup(
    state = "Virginia",
    county = "Stafford",
    outputType = "id"
  )
  state_code_va <- stateCdLookup(input = "Virginia", outputType = "id")
  stafford <- read_waterdata_monitoring_location(
    county_code = county_code_stafford,
    state_code = state_code_va
  )
  stafford_bbox <- sf::st_bbox(stafford)

  dailyStaffordVA <- read_waterdata_daily(
    bbox = stafford_bbox,
    parameter_code = "00060",
    time = c("2015-01-01", "2015-01-30")
  )
  expect_gt(nrow(dailyStaffordVA), 1)

  # America Samoa?
  AS <- read_waterdata_monitoring_location(state_name = "American Samoa")
  expect_gt(nrow(AS), 0)

  site_id <- "USGS-01594440"
  rating_curve <- read_waterdata_ratings(
    monitoring_location_id = site_id,
    file_type = "base"
  )

  expect_equal(names(rating_curve), "USGS-01594440.base.rdb")

  state_rating_list <- read_waterdata_ratings(
    datetime = c(Sys.Date() - 1, NA),
    download_and_parse = FALSE
  )

  expect_true(length(state_rating_list) > 0)

  multi_hucs <- c("07130007", "07130011")
  multi_huc_sites <- read_waterdata_monitoring_location(
    hydrologic_unit_code = multi_hucs
  )

  multi_huc <- read_waterdata_daily(
    bbox = sf::st_bbox(multi_huc_sites),
    parameter_code = "63680",
    statistic_id = "00003",
    time = c("2015-06-18", "2015-06-18")
  )
  expect_equal(4, length(unique(multi_huc$monitoring_location_id)))

  peak_data <- readNWISdata(
    service = "peak",
    state_cd = "PA"
  )
  expect_lt(nrow(peak_data), 100000)

  peak_data <- readNWISdata(
    service = "peak",
    huc2_cd = "20"
  )
  expect_lt(nrow(peak_data), 100000)
})

test_that("read_waterdata_ts_meta", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  # no service specified:
  availableData <- read_waterdata_ts_meta(
    monitoring_location_id = "USGS-05114000"
  )
  expect_gte(ncol(availableData), 18)

  uvData <- read_waterdata_ts_meta(
    monitoring_location_id = "USGS-05114000",
    computation_period_identifier = c("Points")
  )
  expect_equal(unique(uvData$computation_period_identifier), "Points")

  # multiple services
  uvDataMulti <- read_waterdata_ts_meta(
    monitoring_location_id = c("USGS-05114000", "USGS-09423350"),
    computation_period_identifier = c("Daily", "Points")
  )

  expect_true(all(
    unique(uvDataMulti$computation_period_identifier) %in% c("Daily", "Points")
  ))

  # state codes:
  wi_sites <- read_waterdata_monitoring_location(state_name = "Wisconsin")
  flow_and_temp <- read_waterdata_ts_meta(
    bbox = sf::st_bbox(wi_sites),
    parameter_code = c("00060", "00010"),
    statistic_id = "00003",
    computation_period_identifier = c("Daily", "Points")
  )

  expect_true(all(
    unique(flow_and_temp$computation_period_identifier) %in%
      c("Daily", "Points")
  ))
  expect_true(all(
    unique(flow_and_temp$parameter_code) %in% c("00060", "00010")
  ))
  expect_true(all(unique(flow_and_temp$statistic_id) %in% c("00003")))
})

test_that("General WQP retrievals working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  # nameToUse <- "pH"
  # pHData <- readWQPdata(
  #   siteid = "USGS-04024315",
  #   characteristicName = nameToUse,
  #   service = "ResultWQX3"
  # )
  # expect_is(pHData$Activity_StartDateTime, "POSIXct")
  # expect_type(pHData$USGSpcode, "character")
  # expect_type(pHData$Result_Measure, "double")
  # expect_type(pHData$SampleCollectionMethod_Identifier, "character")
  #
  # # testing lists:
  startDate <- as.Date("2022-01-01")
  secchi.names <- c(
    "Depth, Secchi disk depth",
    "Secchi depth",
    "Water transparency, Secchi disc",
    "Depth, Secchi disk depth (choice list)"
  )
  # "Transparency, Secchi tube with disk",
  # "Secchi Reading Condition (choice list)",
  # "Depth, Secchi disk visible at bottom (Y/N) (choice list)")

  args_2 <- list(
    "startDateLo" = startDate,
    "startDateHi" = "2024-01-01",
    statecode = "WI",
    characteristicName = secchi.names
  )

  # # Testing multiple lists:
  arg_3 <- list(
    "startDateLo" = startDate,
    "startDateHi" = "2023-12-31"
  )
  arg_4 <- list(
    statecode = "WI",
    characteristicName = secchi.names
  )

  # wqp.summary_no_atts <- readWQPdata(
  #   siteid = "USGS-04024315",
  #   characteristicName = nameToUse,
  #   ignore_attributes = TRUE,
  #   service = "ResultWQX3"
  # )
  # expect_true(
  #   !all(
  #     c("siteInfo", "variableInfo") %in% names(attributes(wqp.summary_no_atts))
  #   )
  # )
  #
  # rawPcode <- readWQPqw("USGS-01594440", "01075", "", "", legacy = FALSE)
  # expect_true(all(
  #   c("url", "queryTime", "siteInfo") %in%
  #     names(attributes(rawPcode))
  # ))
  #
  # # This means wqp_check_status was called:
  # expect_true("dataProviders" %in% names(attr(rawPcode, "headerInfo")))
  #
  # rawPcode2 <- readWQPqw(
  #   "USGS-01594440",
  #   "01075",
  #   "",
  #   "",
  #   ignore_attributes = TRUE
  # )
  # expect_true(all(
  #   !c("queryTime", "siteInfo") %in%
  #     names(attributes(rawPcode2))
  # ))
  #
  # # This means wqp_check_status wasn't called:
  # expect_false("dataProviders" %in% names(attr(rawPcode2, "headerInfo")))
  #
  # pHData <- readWQPdata(
  #   siteid = "USGS-04024315",
  #   characteristicName = "pH",
  #   service = "ResultWQX3"
  # )
  # expect_true(all(
  #   c("url", "queryTime", "siteInfo", "headerInfo") %in%
  #     names(attributes(pHData))
  # ))
  #
  # # # This means wqp_check_status was called:
  # expect_true("dataProviders" %in% names(attr(pHData, "headerInfo")))
  #
  # pHData2 <- readWQPdata(
  #   siteid = "USGS-04024315",
  #   characteristicName = "pH",
  #   ignore_attributes = TRUE,
  #   service = "ResultWQX3"
  # )
  # expect_true(all(
  #   !c("queryTime", "siteInfo") %in%
  #     names(attributes(pHData2))
  # ))
  #
  # # # This means wqp_check_status was called:
  # expect_false("dataProviders" %in% names(attr(pHData2, "headerInfo")))
  #
  # rawPcode <- readWQPqw(
  #   "USGS-01594440",
  #   "01075",
  #   ignore_attributes = TRUE,
  #   legacy = FALSE
  # )
  # headerInfo <- attr(rawPcode, "headerInfo")
  # wqp_request_id <- headerInfo$`wqp-request-id`
  # count_info <- wqp_check_status(wqp_request_id)
  #
  # expect_true("dataProviders" %in% names(count_info))
})


test_that("zeroPad handles NAs", {
  toPad <- c(1, 5, 55, NA)
  padded <- zeroPad(toPad, 3)
  expect_true(identical(c("001", "005", "055", NA), padded))
})

test_that("Dates with no days can be handled", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  empty_df <- read_waterdata_field_measurements(
    monitoring_location_id = "USGS-425957088141001",
    time = c("1980-01-01", NA)
  )
  expect_true(nrow(empty_df) > 0)
})

context("whatWQPsamples")
test_that("whatWQPsamples working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  # The warning is caused by a confirmed bug in WQP
  siteInfo <- whatWQPsamples(siteid = "USGS-01594440")
  expect_true(nrow(siteInfo) > 0)
})

context("whatWQPmetrics")
test_that("whatWQPmetrics working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  type <- "Stream"
  siteInfo <- whatWQPmetrics(countycode = "US:55:025", siteType = type)
  expect_true(ncol(siteInfo) >= 21)
})

context("whatWQPdata")
test_that("whatWQPdata working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  site1 <- whatWQPdata(siteid = "USGS-01594440")
  expect_is(site1, "data.frame")
  expect_equal(1, nrow(site1))

  type <- "Stream"
  sites <- whatWQPdata(countycode = "US:55:025", siteType = type)
  expect_gt(nrow(sites), 1)

  lakeSites <- whatWQPdata(
    siteType = "Lake, Reservoir, Impoundment",
    statecode = "US:55"
  )
  expect_is(lakeSites$activityCount, "numeric")
})

context("read_waterdata_ts_meta")
test_that("read_waterdata_ts_meta working", {
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  siteListOhio <- read_waterdata_monitoring_location(state_name = "Ohio")
  siteListPhos <- read_waterdata_ts_meta(
    bbox = sf::st_bbox(siteListOhio),
    parameter_code = "00665"
  )
  expect_true(nrow(siteListPhos) > 0)
  expect_is(siteListPhos$begin, "POSIXct")

  bboxSites <- read_waterdata_ts_meta(
    bbox = c(-92.5, 45.4, -87, 47),
    parameter_code = "00060"
  )
  expect_true(nrow(bboxSites) > 0)
})

context("readWQPdots")
test_that("readWQPdots working", {
  testthat::skip_on_cran()

  # bbox vector turned into single string with coords separated by semicolons
  formArgs_bbox <- dataRetrieval:::readWQPdots(bbox = c(-92.5, 45.4, -87, 47))
  expect_true(length(formArgs_bbox) == 2)
  expect_true(length(formArgs_bbox$values$bBox) == 1)

  # NWIS names (siteNumber) converted to WQP expected names (siteid)
  formArgs_site <- dataRetrieval:::readWQPdots(siteNumber = "04010301")
  expect_true(length(formArgs_site) == 2)
  expect_true("siteid" %in% names(formArgs_site$values))
  expect_false("siteNumber" %in% names(formArgs_site$values))

  # NWIS names (stateCd) converted to WQP expected names (statecode)
  formArgs <- dataRetrieval:::readWQPdots(stateCd = "OH", parameterCd = "00665")
  expect_true(length(formArgs$values) == 4)
  expect_true("statecode" %in% names(formArgs$values))
  expect_false("stateCd" %in% names(formArgs$values))

  bbox <- c(-86.97361, 34.48827, -86.61349, 34.65623)
  what_bbox <- whatWQPdata(bBox = bbox)
  expect_true(nrow(what_bbox) > 0)
  x <- whatWQPsites(bBox = bbox)
  expect_true(nrow(x) > 0)
  df <- readWQPdata(
    bBox = bbox,
    characteristicName = "Total Coliform",
    startDateLo = "2023-01-01",
    startDateHi = "2023-12-31",
    service = "Result",
    dataProfile = "narrowResult"
  )
  expect_true(nrow(df) > 0)
  df_legacy <- readWQPdata(
    bBox = bbox,
    characteristicName = "Total Coliform",
    startDateLo = "2023-01-01",
    startDateHi = "2023-12-31",
    service = "Result",
    dataProfile = "narrowResult"
  )
  expect_true(nrow(df_legacy) > 0)
})


test_that("internal functions", {
  # get empty_col type
  expect_equal(dataRetrieval:::empty_col("numeric"), numeric())
  expect_equal(dataRetrieval:::empty_col("character"), character())
  expect_equal(
    dataRetrieval:::empty_col("Date"),
    as.Date(numeric(), origin = "1970-01-01")
  )

  # add empty columns
  df1 <- data.frame(
    a = 1:2,
    b = c("a", "b"),
    c = as.Date(c("2010-01-01", "2010-01-02")),
    stringsAsFactors = FALSE
  )
  df2 <- data.frame(
    d = 1:2,
    b = c("a", "b"),
    e = factor(c("c", "d")),
    stringsAsFactors = FALSE
  )

  df2_ret <- dataRetrieval:::add_empty_col(df2, df1, c("a", "c"))
  df1_ret <- dataRetrieval:::add_empty_col(df1, df2, c("d", "e"))

  expect_true(ncol(df2_ret) == 5)
  expect_true(ncol(df1_ret) == 5)
  expect_true(all(is.na(df2_ret[, c("a", "c")])))
  expect_true(all(is.na(df1_ret[, c("d", "e")])))

  bound_df <- dataRetrieval:::r_bind_dr(df1, df2)

  expect_true(nrow(bound_df) == 4)
  expect_true(class(bound_df$c) == "Date")
  expect_true(class(bound_df$a) == "integer")
  expect_true(class(bound_df$b) == "character")
})


test_that("importWQP convertType", {
  testthat::skip_on_cran()

  SC <- readWQPqw(
    siteNumbers = "USGS-05288705",
    parameterCd = "00300",
    convertType = TRUE,
    legacy = TRUE
  )
  expect_in("ActivityStartDateTime", names(SC))

  SC2 <- readWQPqw(
    siteNumbers = "USGS-05288705",
    parameterCd = "00300",
    convertType = FALSE,
    legacy = TRUE
  )
  expect_true(!"ActivityStartDateTime" %in% names(SC2))

  lakeSites_chars <- whatWQPdata(
    siteType = "Lake, Reservoir, Impoundment",
    statecode = "US:55",
    convertType = FALSE
  )
  expect_is(lakeSites_chars$lat, "character")
})

test_that("format_dates", {
  start_end <- c("2021-01-01", "2022-01-01")
  expect_equal(
    dataRetrieval:::format_api_dates(start_end, date = TRUE),
    "2021-01-01/2022-01-01"
  )
  expect_equal(
    dataRetrieval:::format_api_dates(start_end, date = FALSE),
    "2021-01-01T00:00:00Z/2022-01-01T00:00:00Z"
  )

  start_end <- c("", "")
  expect_equal(dataRetrieval:::format_api_dates(start_end), NA)

  period <- "P7D"

  expect_equal(dataRetrieval:::format_api_dates(period), "P7D")

  start <- c("2021-01-01", NA)

  expect_equal(
    dataRetrieval:::format_api_dates(start),
    "2021-01-01T00:00:00Z/.."
  )
  expect_equal(dataRetrieval:::format_api_dates(start, TRUE), "2021-01-01/..")

  end <- c(NA, "2021-01-01")

  expect_equal(dataRetrieval:::format_api_dates(end), "../2021-01-01T00:00:00Z")
  expect_equal(dataRetrieval:::format_api_dates(end, TRUE), "../2021-01-01")
  #
  # Bad test because it assumes a local timezone
  # This worked locally, but failed on a CI job
  # # end <- c(NA, as.POSIXct("2021-01-01 12:15:00"))
  # # expect_equal(dataRetrieval:::format_api_dates(end),
  # #              "../2021-01-01T18:15:00Z")
  #
  # start_end <- as.POSIXct(c("2021-01-01 12:15:00",
  #                           "2022-01-01 16:45"))
  # expect_equal(dataRetrieval:::format_api_dates(start_end),
  #              "2021-01-01T18:15:00Z/2022-01-01T22:45:00Z")

  start_end <- as.POSIXct(
    c("2021-01-01 12:15:00", "2022-01-01 16:45"),
    tz = "America/New_York"
  )

  expect_equal(
    dataRetrieval:::format_api_dates(start_end),
    "2021-01-01T17:15:00Z/2022-01-01T21:45:00Z"
  )

  # If you don't specify a timezone, it will assume UTC
  start_end2 <- c("2021-01-01 12:15:00", "")
  expect_equal(
    dataRetrieval:::format_api_dates(start_end2),
    "2021-01-01T12:15:00Z/.."
  )

  # If you do specify a timezone, it should maintain it, but convert to UTC:
  start_end2 <- c("2021-01-01T12:15:00-0500", "")
  expect_equal(
    dataRetrieval:::format_api_dates(start_end2),
    "2021-01-01T17:15:00Z/.."
  )

  time = c("2014-05-01T00:00:00Z", "2014-05-01T12:00:00Z")
  expect_equal(
    dataRetrieval:::format_api_dates(time),
    "2014-05-01T00:00:00Z/2014-05-01T12:00:00Z"
  )

  time = c("2014-05-01T00:00Z", "2014-05-01T12:00Z")
  expect_equal(
    dataRetrieval:::format_api_dates(time),
    "2014-05-01T00:00:00Z/2014-05-01T12:00:00Z"
  )

  expect_equal(
    dataRetrieval:::format_api_dates(c("2010-01-01T00:00Z", NA)),
    "2010-01-01T00:00:00Z/.."
  )

  expect_equal(
    dataRetrieval:::format_api_dates(c(
      "2010-01-01T00:00Z",
      "2010-01-01T05:00Z"
    )),
    "2010-01-01T00:00:00Z/2010-01-01T05:00:00Z"
  )

  start <- "2025-10-01"
  end <- as.Date("2026-02-02")

  expect_equal(
    dataRetrieval:::format_api_dates(c(start, end), date = TRUE),
    "2025-10-01/2026-02-02"
  )
})

#' Create service descriptions dynamically
#'
#' This function populates the parameter descriptions.
#'
#' @param service Character, can be any of the endpoints
#' @return list
#' @noRd
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' ml_desc <- dataRetrieval:::get_description("monitoring-locations")
#' ml_desc
#' }
#'
get_description <- function(service) {
  query_ret <- get_collection()

  tags <- query_ret[["tags"]]

  service_index <- which(sapply(tags, function(x) {
    x$name == service
  }))

  tags[[service_index]][["description"]]
}

#' Get collection response
#'
#'
#' @return httr2 response
#' @noRd
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' collection <- dataRetrieval:::get_collection()
#' collection
#' }
#'
get_collection <- function() {
  check_collections <- base_url() |>
    httr2::req_url_path_append("openapi") |>
    httr2::req_url_query(f = "html#/server/getCollections")

  check_endpoints_req <- basic_request(check_collections)

  query_ret <- httr2::req_perform(check_endpoints_req) |>
    httr2::resp_body_json()

  return(query_ret)
}

#' Get parameter descriptions
#'
#' This function returns a list of properties available from an endpoint. When
#' available, it will also contain a description.
#'
#' @param service Character, can be any of the USGS Waterdata API endpoints or collections.
#' @return list
#' @export
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' ml <- dataRetrieval::get_ogc_params("monitoring-locations")
#' ml$national_aquifer_code
#' }
#'
get_ogc_params <- function(service) {
  check_queryables_req <- base_url() |>
    httr2::req_url_path_append("collections") |>
    httr2::req_url_path_append(service) |>
    httr2::req_url_path_append("schema") |>
    basic_request()

  query_ret <- httr2::req_perform(check_queryables_req) |>
    httr2::resp_body_json()

  params <- sapply(query_ret$properties, function(x) x[["description"]])
}


#' Get property list
#'
#' This function gets a list of available properties, and
#' renames the id column to what is used in dataRetrieval.
#'
#' @param service Character, can be any of the endpoints
#' @param output_id Character, dataRetrieval output name
#' @return list
#' @noRd
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' dataRetrieval:::get_properties_for_docs("monitoring-locations",
#'                                         "monitoring_location_id")
#'
#' }
#'
get_properties_for_docs <- function(service, output_id = NA) {
  schema <- check_OGC_requests(endpoint = service, type = "schema")
  properties <- names(schema$properties)
  if (!is.na(output_id)) {
    properties[properties == "id"] <- output_id
  }

  return(paste(properties, collapse = ", "))
}

#' Check OGC requests
#'
#' @param endpoint Character, can be any existing collection
#' @param type Character, can be "queryables", "schema"
#' @export
#' @keywords internal
#' @return list
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#'
#' dv_queryables <- check_OGC_requests(endpoint = "daily",
#'                                 type = "queryables")
#' dv_schema <- check_OGC_requests(endpoint = "daily",
#'                             type = "schema")
#' ts_meta_queryables <- check_OGC_requests(endpoint = "time-series-metadata",
#'                                 type = "queryables")
#' ts_meta_schema <- check_OGC_requests(endpoint = "time-series-metadata",
#'                                 type = "schema")
#' }
check_OGC_requests <- function(endpoint = "daily", type = "queryables") {
  match.arg(type, c("queryables", "schema"))

  match.arg(endpoint, c(pkg.env$api_endpoints, pkg.env$metadata))

  req <- base_url() |>
    httr2::req_url_path_append("collections") |>
    httr2::req_url_path_append(endpoint) |>
    httr2::req_url_path_append(type) |>
    basic_request()

  query_ret <- req |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  return(query_ret)
}

#' Construct NWIS url for data retrieval
#'
#' Using USGS water web services to construct urls.
#'
#' @param siteNumbers string or vector of strings USGS site number.  This is usually an 8 digit number
#' @param parameterCd string or vector of USGS parameter code.  This is usually an 5 digit number.
#' @param startDate character starting date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the earliest possible record.
#' @param endDate character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the latest possible record.
#' @param statCd string or vector USGS statistic code only used for daily value service.
#' This is usually 5 digits.  Daily mean (00003) is the default.
#' @param service string USGS service to call. Possible values are "dv" (daily values),
#' "uv" (unit/instantaneous values), and "rating" (rating curve),
#' "peak", "stat" (statistics web service BETA).
#' @param format string, can be "tsv" or "xml", and is only applicable for daily
#' and unit value requests.  "tsv" returns results faster, but there is a possibility
#' that an incomplete file is returned without warning. XML is slower,
#' but will offer a warning if the file was incomplete (for example, if there was a
#' momentary problem with the internet connection). It is possible to safely use the "tsv" option,
#' but the user must carefully check the results to see if the data returns matches
#' what is expected. The default is therefore "xml".
#' @param expanded logical defaults to `TRUE`. If `TRUE`, retrieves additional
#' information, only applicable for qw data.
#' @param ratingType can be "base", "corr", or "exsa". Only applies to rating curve data.
#' @param statReportType character Only used for statistics service requests.  Time
#' division for statistics: daily, monthly, or annual.  Default is daily.
#' Note that daily provides statistics for each calendar day over the specified
#' range of water years, i.e. no more than 366 data points will be returned for
#' each site/parameter.  Use `readNWISdata` or `readNWISdv` for daily averages.
#' Also note that "annual" returns statistics for the calendar year.  Use
#' `readNWISdata` for water years. Monthly and yearly
#' provide statistics for each month and year within the range individually.
#' @param statType character Only used for statistics service requests. Type(s)
#' of statistics to output for daily values.  Default is mean, which is the only
#' option for monthly and yearly report types.
#' @keywords data import USGS web service
#' @return url string
#' @export
#' @examples
#' site_id <- "01594440"
#' startDate <- "1985-01-01"
#' endDate <- ""
#' pCode <- c("00060", "00010")
#' url_daily <- constructNWISURL(site_id, pCode,
#'   startDate, endDate, "dv",
#'   statCd = c("00003", "00001")
#' )
#' url_unit <- constructNWISURL(site_id, pCode, "2012-06-28", "2012-06-30", "iv")
#'
#' url_daily_tsv <- constructNWISURL(site_id, pCode, startDate, endDate, "dv",
#'   statCd = c("00003", "00001"), format = "tsv"
#' )
#' url_rating <- constructNWISURL(site_id, service = "rating", ratingType = "base")
#' url_peak <- constructNWISURL(site_id, service = "peak")
constructNWISURL <- function(
  siteNumbers,
  parameterCd = "00060",
  startDate = "",
  endDate = "",
  service,
  statCd = "00003",
  format = "xml",
  expanded = TRUE,
  ratingType = "base",
  statReportType = "daily",
  statType = "mean"
) {
  service <- match.arg(
    service,
    c(
      "dv",
      "uv",
      "iv",
      "iv_recent",
      "rating",
      "peak",
      "stat"
    )
  )

  service[service == "meas"] <- "measurements"
  service[service == "uv"] <- "iv"

  POST <- nchar(paste0(siteNumbers, parameterCd, collapse = "")) > 2048

  baseURL <- httr2::request(pkg.env[[service]])

  if (!is.null(pkg.env$access)) {
    baseURL <- httr2::req_url_query(baseURL, Access = pkg.env$access)
  }

  if (any(!is.na(parameterCd) & parameterCd != "all")) {
    pcodeCheck <- all(nchar(parameterCd) == 5) &
      all(!is.na(suppressWarnings(as.numeric(parameterCd))))

    if (!pcodeCheck) {
      badIndex <- which(
        nchar(parameterCd) != 5 |
          is.na(suppressWarnings(as.numeric(parameterCd)))
      )
      stop(
        "The following pCodes appear mistyped:",
        paste(parameterCd[badIndex], collapse = ", ")
      )
    }

    if (length(parameterCd) > 200) {
      stop(
        "Maximum parameter codes allowed is 200, please adjust data request."
      )
    }
  }

  switch(
    service,
    rating = {
      ratingType <- match.arg(ratingType, c("base", "corr", "exsa"))
      url <- get_or_post(
        baseURL,
        POST = POST,
        site_no = siteNumbers,
        file_type = ratingType
      )
    },
    peak = {
      url <- get_or_post(
        baseURL,
        POST = POST,
        range_selection = "date_range",
        format = "rdb"
      )
      url <- get_or_post(
        url,
        POST = POST,
        site_no = siteNumbers,
        .multi = "comma"
      )

      if (nzchar(startDate)) {
        url <- get_or_post(url, POST = POST, begin_date = startDate)
      }
      if (nzchar(endDate)) {
        url <- get_or_post(url, POST = POST, end_date = endDate)
      }
    },
    stat = {
      # for statistics service

      message(
        "Please be aware the NWIS data service feeding this function is in BETA.\n
          Data formatting could be changed at any time, and is not guaranteed"
      )

      # make sure only statTypes allowed for the statReportType are being requested
      if (
        !grepl("(?i)daily", statReportType) &&
          !all(grepl("(?i)mean", statType)) &&
          !all(grepl("(?i)all", statType))
      ) {
        stop("Monthly and annual report types can only provide means")
      }

      # make sure dates aren't too specific for statReportType
      if (
        grepl("(?i)monthly", statReportType) &&
          (length(unlist(gregexpr("-", startDate))) > 1 ||
            length(unlist(gregexpr("-", endDate))) > 1)
      ) {
        stop(
          "Start and end dates for monthly statReportType can only include months and years"
        )
      }
      if (
        grepl("(?i)annual", statReportType) &&
          (grepl("-", startDate) || grepl("-", endDate))
      ) {
        stop(
          "Start and end dates for annual statReportType can only include years"
        )
      }

      url <- get_or_post(
        baseURL,
        POST = POST,
        sites = siteNumbers,
        .multi = "comma"
      )
      url <- get_or_post(
        url,
        POST = POST,
        statReportType = statReportType,
        .multi = "comma"
      )
      url <- get_or_post(
        url,
        POST = POST,
        statType = statType,
        .multi = "comma"
      )
      url <- get_or_post(
        url,
        POST = POST,
        parameterCd = parameterCd,
        .multi = "comma"
      )

      if (nzchar(startDate)) {
        url <- get_or_post(url, POST = POST, startDT = startDate)
      }
      if (nzchar(endDate)) {
        url <- get_or_post(url, POST = POST, endDT = endDate)
      }
      if (!grepl("(?i)daily", statReportType)) {
        url <- get_or_post(url, POST = POST, missingData = "off")
      }
    },
    {
      # this will be either dv, uv

      format <- match.arg(format, c("xml", "tsv", "wml1", "wml2", "rdb"))

      formatURL <- switch(
        format,
        xml = "waterml,1.1",
        rdb = "rdb,1.0",
        tsv = "rdb,1.0",
        wml2 = "waterml,2.0",
        wml1 = "waterml,1.1"
      )

      url <- get_or_post(
        baseURL,
        POST = POST,
        site = siteNumbers,
        .multi = "comma"
      )
      url <- get_or_post(url, POST = POST, format = formatURL)

      if (!all(is.na(parameterCd))) {
        url <- get_or_post(
          url,
          POST = POST,
          ParameterCd = parameterCd,
          .multi = "comma"
        )
      }

      if ("dv" == service) {
        url <- get_or_post(url, POST = POST, StatCd = statCd, .multi = "comma")
      }

      if (nzchar(startDate)) {
        url <- get_or_post(url, POST = POST, startDT = startDate)
      } else {
        startorgin <- "1851-01-01"
        if ("iv" == service) {
          startorgin <- "1900-01-01"
        }
        url <- get_or_post(url, POST = POST, startDT = startorgin)
      }

      if (nzchar(endDate)) {
        url <- get_or_post(url, POST = POST, endDT = endDate)
      }
    }
  )

  url <- httr2::req_headers(
    url,
    `Accept-Encoding` = c("compress", "gzip", "deflate")
  )

  return(url)
}


#' Construct WQP url for data retrieval
#'
#' Construct WQP url for data retrieval. This function gets the data from here: <https://www.waterqualitydata.us>
#'
#' @param siteNumbers string or vector of strings USGS site number.
#' @param parameterCd string or vector of USGS parameter code.  This is usually an 5 digit number.
#' @param startDate character starting date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the earliest possible record.
#' @param endDate character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
#' retrieval for the latest possible record.
#' @param legacy Logical. If TRUE, uses legacy WQP services. Default is TRUE.
#' Setting legacy = FALSE uses WQX3.0 WQP services, which are in-development, use with caution.
#' @keywords data import WQP web service
#' @return url string
#' @export
#' @examples
#' site_ids <- c("USGS-02292010", "USGS-02276877")
#' startDate <- "2020-01-01"
#' endDate <- ""
#' pCode <- c("80154", "00613")
#' url_wqp <- constructWQPURL(
#'   site_ids,
#'   pCode,
#'   startDate, endDate
#' )
#' url_wqp
#' charNames <- c(
#'   "Temperature",
#'   "Temperature, sample",
#'   "Temperature, water",
#'   "Temperature, water, deg F"
#' )
#' obs_url_orig <- constructWQPURL(
#'   siteNumbers = c(
#'     "IIDFG-41WSSPAHS",
#'     "USGS-02352560"
#'   ),
#'   parameterCd = charNames,
#'   startDate, ""
#' )
#' obs_url_orig
constructWQPURL <- function(
  siteNumbers,
  parameterCd,
  startDate,
  endDate,
  legacy = TRUE
) {
  allPCode <- any(toupper(parameterCd) == "ALL")

  pCodeLogic <- TRUE

  POST = nchar(paste0(siteNumbers, collapse = "")) > 2048

  if (!allPCode) {
    multiplePcodes <- length(parameterCd) > 1
    if (all(nchar(parameterCd) == 5)) {
      suppressWarnings(pCodeLogic <- all(!is.na(as.numeric(parameterCd))))
    } else {
      pCodeLogic <- FALSE
    }
  }

  if (legacy) {
    baseURL <- httr2::request(pkg.env[["Result"]])
    baseURL <- get_or_post(
      baseURL,
      POST = POST,
      siteid = siteNumbers,
      .multi = function(x) paste0(x, collapse = ";")
    )
    baseURL <- get_or_post(baseURL, POST = POST, count = "no")
  } else {
    baseURL <- httr2::request(pkg.env[["ResultWQX3"]])
    baseURL <- httr2::req_url_query(
      baseURL,
      siteid = siteNumbers,
      .multi = "explode"
    )
  }

  if (legacy & !allPCode) {
    if (pCodeLogic) {
      baseURL <- get_or_post(
        baseURL,
        POST = POST,
        pCode = parameterCd,
        .multi = function(x) paste0(x, collapse = ";")
      )
    } else {
      baseURL <- httr2::req_url_query(
        baseURL,
        characteristicName = parameterCd,
        .multi = function(x) paste0(x, collapse = ";")
      )
    }
  } else if (!legacy & !allPCode) {
    if (pCodeLogic) {
      baseURL <- httr2::req_url_query(
        baseURL,
        pCode = parameterCd,
        .multi = "explode"
      )
    } else {
      baseURL <- httr2::req_url_query(
        baseURL,
        characteristicName = parameterCd,
        .multi = "explode"
      )
    }
  }

  if (nzchar(startDate)) {
    startDate <- format(as.Date(startDate), format = "%m-%d-%Y")
    baseURL <- get_or_post(baseURL, POST = POST, startDateLo = startDate)
  }

  if (nzchar(endDate)) {
    endDate <- format(as.Date(endDate), format = "%m-%d-%Y")
    baseURL <- get_or_post(baseURL, POST = POST, startDateHi = endDate)
  }

  baseURL <- httr2::req_url_query(baseURL, mimeType = "csv")
  if (!legacy) {
    baseURL <- httr2::req_url_query(baseURL, dataProfile = "basicPhysChem")
  }

  baseURL <- httr2::req_headers(
    baseURL,
    `Accept-Encoding` = c("compress", "gzip", "deflate")
  )

  return(baseURL)
}

#' Function to return data from web services
#'
#' This function accepts a url parameter, and returns the raw data.
#'
#' To add a custom user agent, create an environmental variable: CUSTOM_DR_UA
#'
#' @param obs_url character containing the url for the retrieval
#' @param \dots information to pass to header request
#' @export
#' @return Returns xml, json, or text depending on the requested data.
#' @examplesIf is_dataRetrieval_user()
#' siteNumber <- "02177000"
#' startDate <- "2012-09-01"
#' endDate <- "2012-10-01"
#' offering <- "00003"
#' property <- "00060"
#' obs_url <- constructNWISURL(siteNumber, property, startDate, endDate, "dv")
#' \donttest{
#' rawData <- getWebServiceData(obs_url)
#' }
getWebServiceData <- function(obs_url, ...) {
  if (!has_internet_2(obs_url)) {
    message("No internet connection.")
    return(invisible(NULL))
  }

  if (is.character(obs_url)) {
    obs_url <- httr2::request(obs_url)
  }

  obs_url <- httr2::req_user_agent(obs_url, default_ua())
  obs_url <- httr2::req_throttle(obs_url, rate = 30 / 60)
  obs_url <- httr2::req_retry(
    obs_url,
    max_tries = 3,
    retry_on_failure = TRUE
  )
  obs_url <- httr2::req_headers(
    obs_url,
    `Accept-Encoding` = c("compress", "gzip")
  )

  url_method <- "GET"
  if (!is.null(obs_url$body)) {
    url_method <- "POST"
  }
  message(url_method, ": ", obs_url$url)
  returnedList <- httr2::req_perform(obs_url)

  good <- check_non_200s(returnedList)

  return_readLines <- c("text/html", "text/html;charset=utf-8")

  return_content <- c(
    "text/tab-separated-values;charset=utf-8",
    "text/csv;charset=utf-8",
    "text/csv",
    "text/plain",
    "text/plain;charset=utf-8"
  )

  return_json <- c("application/vnd.geo+json;charset=utf-8")

  if (good) {
    headerInfo <- httr2::resp_headers(returnedList)
    content <- gsub(" ", "", tolower(headerInfo$`content-type`))
    if (content %in% return_content) {
      returnedDoc <- httr2::resp_body_string(returnedList)
    } else if (content %in% return_readLines) {
      returnedList <- httr2::resp_body_string(returnedList)
      message(returnedList)
      returnedDoc <- returnedList
    } else if (content %in% return_json) {
      returnedDoc <- httr2::resp_body_json(returnedList)
    } else {
      returnedDoc <- httr2::resp_body_xml(returnedList, encoding = "UTF-8")
      if (
        all(grepl(
          "No sites/data found using the selection criteria specified",
          returnedDoc
        ))
      ) {
        message(returnedDoc)
      }
    }

    attr(returnedDoc, "headerInfo") <- headerInfo

    return(returnedDoc)
  } else {
    return(NULL)
  }
}

check_non_200s <- function(returnedList) {
  status <- httr2::resp_status(returnedList)

  return(status == 200)
}

#' Create user agent
#'
#' @keywords internal
default_ua <- function() {
  if (!is.null(pkg.env$ua)) {
    return(pkg.env$ua)
  }

  versions <- c(
    libcurl = curl::curl_version()$version,
    httr2 = as.character(utils::packageVersion("httr2")),
    dataRetrieval = as.character(utils::packageVersion("dataRetrieval"))
  )

  ua <- paste0(names(versions), "/", versions, collapse = " ")

  if (Sys.getenv("CUSTOM_DR_UA") != "") {
    ua <- paste0(ua, "/", Sys.getenv("CUSTOM_DR_UA"))
  }

  pkg.env$ua <- ua
  return(ua)
}

#' has_internet_2
#'
#' Function to check for internet even if the user
#' is behind a proxy
#'
#' If this is giving you problems, override the false negative:
#' Sys.setenv("OVERRIDE_INTERNET_TEST" = TRUE)
#'
#' @keywords internal
#' @param obs_url character obs_url to check
has_internet_2 <- function(obs_url) {
  if (nzchar(Sys.getenv("OVERRIDE_INTERNET_TEST"))) {
    if (Sys.getenv("OVERRIDE_INTERNET_TEST") == "FALSE") {
      return(FALSE)
    } else {
      return(TRUE)
    }
  }

  if ("url" %in% names(obs_url)) {
    url <- obs_url$url
  } else {
    url <- obs_url
  }

  host <- gsub("^https://(?:www[.])?([^/]*).*$", "\\1", url)

  !is.null(curl::nslookup(host, error = FALSE))
}

#' getting header information from a WQP query
#'
#' @param url the query url
getQuerySummary <- function(url) {
  wqp_message()

  queryHEAD <- httr2::req_method(req = url, method = "HEAD")
  queryHEAD <- httr2::req_perform(queryHEAD)
  headerInfo <- httr2::resp_headers(queryHEAD)
  retquery <- data.frame(t(unlist(headerInfo)))
  names(retquery) <- gsub("\\.", "-", names(retquery))
  retquery[, grep("-count", names(retquery))] <- as.numeric(retquery[grep(
    "-count",
    names(retquery)
  )])

  if ("date" %in% names(retquery)) {
    retquery$date <- as.Date(retquery$date, format = "%a, %d %b %Y %H:%M:%S")
  }

  return(retquery)
}

#' Import data from the National Groundwater Monitoring Network.
#'
#' Only water level data and site locations and names are currently available through the web service.
#' @param service char Service for the request - "observation" and "featureOfInterest" are implemented.
#' @param \dots Other parameters to supply, namely `siteNumbers` or `bbox`
#' @param asDateTime logical if `TRUE`, will convert times to POSIXct format.  Currently defaults to
#' `FALSE` since time zone information is not included.
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the data's provided time zone offset.
#' Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
#' "America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  `OlsonNames()`
#' for more information on time zones.
#' @export
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' # one site
#' site <- "USGS.430427089284901"
#' #oneSite <- readNGWMNdata(siteNumbers = site, service = "observation")
#'
#' # multiple sites
#' sites <- c("USGS.272838082142201", "USGS.404159100494601", "USGS.401216080362703")
#' # Very slow:
#' # multiSiteData <- readNGWMNdata(siteNumbers = sites, service = "observation")
#' # attributes(multiSiteData)
#'
#' # non-USGS site
#' # accepts colon or period between agency and ID
#' site <- "MBMG:702934"
#' # data <- readNGWMNdata(siteNumbers = site, service = "featureOfInterest")
#'
#' # bounding box
#' # bboxSites <- readNGWMNdata(service = "featureOfInterest", bbox = c(30, -102, 31, 99))
#' # retrieve  sites.  Set asDateTime to false since one site has an invalid date
#' # Very slow:
#' # bboxData <- readNGWMNdata(service = "observation", siteNumbers = bboxSites$site[1:3],
#' #                           asDateTime = FALSE)
#' }
#'
readNGWMNdata <- function(service, ..., asDateTime = TRUE, tz = "UTC") {
  .Deprecated(
    "",
    msg = "read_ngwmn_data coming soon. Check back at
              https://doi-usgs.github.io/dataRetrieval/articles/Status.html
              for more information"
  )

  dots <- convertLists(...)

  match.arg(service, c("observation", "featureOfInterest"))

  if (service == "observation") {
    # these attributes are pulled out and saved when doing binds to be reattached
    attrs <- c(
      "url",
      "gml:identifier",
      "generationDate",
      "responsibleParty",
      "contact"
    )
    featureID <- stats::na.omit(gsub(":", ".", dots[["siteNumbers"]]))

    obs_list <- vector("list", length(featureID))
    attr_list <- vector("list", length(featureID))
    for (idx in seq_along(featureID)) {
      obsFID <- retrieveObservation(
        featureID = featureID[idx], asDateTime, attrs, tz = tz
      )
      attr_list[[idx]] <- saveAttrs(attrs, obsFID)
      obs_list[[idx]] <- removeAttrs(attrs, obsFID)
    }
    allObs <- Reduce(r_bind_dr, obs_list, init = data.frame())
    allAttrs <- Reduce(r_bind_dr, attr_list, init = data.frame())

    allSites <- tryCatch(
      {
        retrieveFeatureOfInterest(featureID = featureID)
      },
      error = function(cond) {
        return(NULL)
      }
    )

    if (!is.null(allSites)) {
      attr(allObs, "siteInfo") <- allSites
    }

    attr(allObs, "other") <- allAttrs
    returnData <- allObs
  } else if (service == "featureOfInterest") {
    if ("siteNumbers" %in% names(dots)) {
      featureID <- stats::na.omit(gsub(":", ".", dots[["siteNumbers"]]))
      allSites <- tryCatch({
        retrieveFeatureOfInterest(featureID = featureID)
      })
    }

    if ("bbox" %in% names(dots)) {
      allSites <- tryCatch({
        retrieveFeatureOfInterest(bbox = dots[["bbox"]])
      })
    }
    returnData <- allSites
  }

  return(returnData)
}

#' Retrieve groundwater levels from the National Ground Water Monitoring Network.
#'
#' @param siteNumbers character Vector of feature IDs formatted with agency code and site number
#' separated by a period or semicolon, e.g. `USGS.404159100494601`.
#' @param asDateTime logical Should dates and times be converted to date/time objects,
#' or returned as character?  Defaults to `TRUE`.  Must be set to `FALSE` if a site
#' contains non-standard dates.
#' @param tz character to set timezone attribute of dateTime. Default is "UTC", and converts the
#' date times to UTC, properly accounting for daylight savings times based on the data's provided time zone offset.
#' Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
#' "America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
#' "America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  `OlsonNames()`
#' for more information on time zones.
#' @export
#'
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' # one site
#' site <- "USGS.430427089284901"
#' # oneSite <- readNGWMNlevels(siteNumbers = site)
#'
#' # multiple sites
#' sites <- c("USGS:272838082142201", "USGS:404159100494601", "USGS:401216080362703")
#' # multiSiteData <- readNGWMNlevels(sites)
#'
#' # non-USGS site
#' site <- "MBMG.103306"
#' # data <- readNGWMNlevels(siteNumbers = site, asDateTime = FALSE)
#'
#' # site with no data returns empty data frame
#' noDataSite <- "UTGS.401544112060301"
#' # noDataSite <- readNGWMNlevels(siteNumbers = noDataSite)
#' }
readNGWMNlevels <- function(siteNumbers, asDateTime = TRUE, tz = "UTC") {
  data <- readNGWMNdata(
    siteNumbers = siteNumbers,
    service = "observation",
    asDateTime = asDateTime,
    tz = tz
  )
  return(data)
}

#' Retrieve site data from the National Ground Water Monitoring Network.
#'
#' @param siteNumbers character Vector of feature IDs formatted with agency code and site number
#' separated by a period or semicolon, e.g. `USGS.404159100494601`.
#'
#' @export
#' @return A data frame the following columns:
#' #' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' site \tab char \tab Site FID \cr
#' description \tab char \tab Site description \cr
#' dec_lat_va, dec_lon_va \tab numeric \tab Site latitude and longitude \cr
#' }
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' # one site
#' site <- "USGS.430427089284901"
#' #oneSite <- readNGWMNsites(siteNumbers = site)
#'
#' # non-USGS site
#' site <- "MBMG.103306"
#' #siteInfo <- readNGWMNsites(siteNumbers = site)
#' }
readNGWMNsites <- function(siteNumbers) {
  sites <- readNGWMNdata(
    siteNumbers = siteNumbers,
    service = "featureOfInterest"
  )
  return(sites)
}

retrieveObservation <- function(featureID, asDateTime, attrs, tz) {
  baseURL <- httr2::request(pkg.env[["NGWMN"]])
  baseURL <- httr2::req_url_query(
    baseURL,
    request = "GetObservation",
    service = "SOS",
    version = "2.0.0",
    observedProperty = "urn:ogc:def:property:OGC:GroundWaterLevel",
    responseFormat = "text/xml",
    featureOfInterest = paste("VW_GWDP_GEOSERVER", featureID, sep = ".")
  )

  returnData <- importNGWMN(baseURL, asDateTime = asDateTime, tz = tz)
  if (nrow(returnData) == 0) {
    # need to add NA attributes, so they aren't messed up when stored as DFs
    attr(returnData, "gml:identifier") <- NA
    attr(returnData, "generationDate") <- NA
  }

  # mutate removes the attributes, need to save and append
  attribs <- saveAttrs(attrs, returnData)
  if (nrow(returnData) > 0) {
    # tack on site number
    siteNum <- rep(sub(".*\\.", "", featureID), nrow(returnData))
    returnData$site <- siteNum
    numCol <- ncol(returnData)
    returnData <- returnData[, c(numCol, 1:(numCol - 1))] # move siteNum to the left
  }
  attributes(returnData) <- c(attributes(returnData), as.list(attribs))

  return(returnData)
}

# retrieve feature of interest
# could allow pass through srsName - needs to be worked in higher-up in dots
retrieveFeatureOfInterest <- function(
  ...,
  asDateTime,
  srsName = "urn:ogc:def:crs:EPSG::4269"
) {
  values <- convertLists(...)

  baseURL <- httr2::request(pkg.env[["NGWMN"]])
  baseURL <- httr2::req_url_query(
    baseURL,
    request = "GetFeatureOfInterest",
    service = "SOS",
    version = "2.0.0",
    responseFormat = "text/xml"
  )

  if ("featureID" %in% names(values)) {
    features <- paste("VW_GWDP_GEOSERVER", values[["featureID"]], sep = ".")

    baseURL <- httr2::req_url_query(
      baseURL,
      featureOfInterest = features,
      .multi = "comma"
    )
  } else if ("bbox" %in% names(values)) {
    baseURL <- httr2::req_url_query(
      baseURL,
      bbox = paste(values[["bbox"]], collapse = ","),
      srsName = srsName
    )
  } else {
    stop("Geographical filter not specified. Please use siteNumbers or bbox")
  }

  siteDF <- importNGWMN(baseURL, asDateTime, tz = "")
  attr(siteDF, "url") <- baseURL$url
  attr(siteDF, "queryTime") <- Sys.time()
  return(siteDF)
}


# save specified attributes from a data frame
saveAttrs <- function(attrs, df) {
  attribs <- sapply(attrs, function(x) attr(df, x))
  if (is.vector(attribs)) {
    toReturn <- as.data.frame(t(attribs), stringsAsFactors = FALSE)
  } else {
    # don't need to transpose
    toReturn <- as.data.frame(attribs, stringsAsFactors = FALSE)
  }
  return(toReturn)
}

# strip specified attributes from a data frame
removeAttrs <- function(attrs, df) {
  for (a in attrs) {
    attr(df, a) <- NULL
  }
  return(df)
}

#' @name whatWQPsamples
#' @rdname wqpSpecials
#' @export
#' @return A data frame with information on the sampling activity
#' available from the Water Quality Portal for the query parameters.
#' @examples
#' \donttest{
#'
#' #site1 <- whatWQPsamples(siteid = "USGS-01594440")
#'
#' #type <- "Stream"
#'
#' #sites <- whatWQPsamples(countycode = "US:55:025", siteType = type)
#'
#' #lakeSites_samples <- whatWQPsamples(siteType = "Lake, Reservoir, Impoundment",
#' #                                   countycode = "US:55:025")
#' }
whatWQPsamples <- function(..., convertType = TRUE, legacy = TRUE) {
  values <- readWQPdots(..., legacy = legacy)

  values <- values[["values"]]

  if ("tz" %in% names(values)) {
    values <- values[!(names(values) %in% "tz")]
  }

  if ("service" %in% names(values)) {
    values <- values[!(names(values) %in% "service")]
  }

  if (legacy) {
    baseURL <- httr2::request(pkg.env[["Activity"]])
  } else {
    baseURL <- httr2::request(pkg.env[["ActivityWQX3"]])
  }
  POST = FALSE
  if (!legacy) {
    baseURL <- httr2::req_url_query(baseURL, !!!values, .multi = "explode")
  } else {
    if ("siteid" %in% names(values)) {
      if (length(values[["siteid"]]) > 1) {
        sites <- values[["siteid"]]
        POST <- nchar(paste0(sites, collapse = "")) > 2048

        baseURL <- get_or_post(
          baseURL,
          POST = POST,
          siteid = sites,
          .multi = function(x) paste0(x, collapse = ";")
        )

        values <- values[names(values) != "siteid"]
      }
    }
    baseURL <- get_or_post(baseURL, POST = POST, !!!values, .multi = "explode")
  }

  retval <- importWQP(baseURL, convertType = convertType)
  if (!is.null(retval)) {
    attr(retval, "legacy") <- legacy
    attr(retval, "queryTime") <- Sys.time()
    attr(retval, "url") <- baseURL$url

    if (legacy) {
      wqp_message()
    } else {
      wqp_message_beta()
      attr(retval, "wqp-request-id") <- attr(
        retval,
        "headerInfo"
      )$`wqp-request-id`
    }
  }

  return(retval)
}

#' @name whatWQPmetrics
#' @rdname wqpSpecials
#' @export
#' @examples
#' \donttest{
#'
#' #type <- "Stream"
#'
#' #sites <- whatWQPmetrics(countycode = "US:55:025", siteType = type)
#' #lakeSites_metrics <- whatWQPmetrics(siteType = "Lake, Reservoir, Impoundment",
#' #                                   countycode = "US:55:025")
#' }
whatWQPmetrics <- function(..., convertType = TRUE) {
  values <- readWQPdots(..., legacy = TRUE)

  values <- values[["values"]]

  if ("tz" %in% names(values)) {
    values <- values[!(names(values) %in% "tz")]
  }

  if ("service" %in% names(values)) {
    values <- values[!(names(values) %in% "service")]
  }
  POST <- FALSE
  baseURL <- httr2::request(pkg.env[["ActivityMetric"]])

  if ("siteid" %in% names(values)) {
    if (length(values[["siteid"]]) > 1) {
      sites <- values[["siteid"]]
      POST <- nchar(paste0(sites, collapse = "")) > 2048

      baseURL <- get_or_post(
        baseURL,
        POST = POST,
        siteid = sites,
        .multi = function(x) paste0(x, collapse = ";")
      )

      values <- values[names(values) != "siteid"]
    }
  }

  baseURL <- get_or_post(baseURL, POST = POST, !!!values, .multi = "explode")

  withCallingHandlers(
    {
      retval <- importWQP(baseURL, convertType = convertType)
    },
    warning = function(w) {
      if (any(grepl("Number of rows returned not matched in header", w))) {
        invokeRestart("muffleWarning")
      }
    }
  )

  if (is.null(retval)) {
    return(NULL)
  } else {
    wqp_message()
    attr(retval, "queryTime") <- Sys.time()
    attr(retval, "url") <- baseURL$url

    return(retval)
  }
}


#' Data Available from Water Quality Portal
#'
#' Returns a list of sites from the Water Quality Portal web service. This function gets
#' the data from: <https://www.waterqualitydata.us>.
#' Arguments to the function should be based on
#' <https://www.waterqualitydata.us/webservices_documentation>.
#' The information returned from whatWQPdata describes the
#' available data at the WQP sites, and some metadata on the sites themselves.
#' For example, a row is returned for each individual site that fulfills this
#' query. In that we can learn how many sampling activities and results
#' are available for the query. It does not break those results down by any finer
#' grain. For example, if you ask for "Nutrients" (characteristicGroup), you will
#' not learn what specific nutrients are available at that site. For that
#' kind of data discovery see `readWQPsummary`.
#'
#' @param \dots see <https://www.waterqualitydata.us/webservices_documentation> for
#' a complete list of options. A list of arguments can also be supplied.
#' One way to figure out how to construct a WQP query is to go to the "Advanced"
#' form in the Water Quality Portal:
#' <https://www.waterqualitydata.us/#mimeType=csv&providers=NWIS&providers=STORET>
#' Use the form to discover what parameters are available. Once the query is
#' set in the form, scroll down to the "Query URL". You will see the parameters
#' after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
#' in the Characteristic Group dropdown, you will see characteristicType=Nutrient
#' in the Query URL. The corresponding argument for dataRetrieval is
#' characteristicType = "Nutrient". dataRetrieval users do not need to include
#' mimeType, and providers is optional (these arguments are picked automatically).
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character.
#' @keywords data import WQP web service
#' @return A data frame that returns basic data availability such as
#' sites, number of results, and number of sampling activities from the
#' query parameters for the Water Quality Portal.
#'
#' @export
#' @seealso whatWQPsites readWQPsummary readWQPdata
#' @examplesIf is_dataRetrieval_user()
#' \donttest{
#' #site1 <- whatWQPdata(siteid = "USGS-01594440")
#'
#' #type <- "Stream"
#' #sites <- whatWQPdata(countycode = "US:55:025", siteType = type)
#'
#' #lakeSites <- whatWQPdata(siteType = "Lake, Reservoir, Impoundment",
#' #                        countycode = "US:55:025")
#' #lakeSites_chars <- whatWQPdata(
#' #   siteType = "Lake, Reservoir, Impoundment",
#' #   countycode = "US:55:025", convertType = FALSE)
#'
#'
#' #bbox <- c(-86.9736, 34.4883, -86.6135, 34.6562)
#' #what_bb <- whatWQPdata(bBox = bbox)
#' }
whatWQPdata <- function(..., convertType = TRUE) {
  args <- convertLists(...)
  if ("legacy" %in% names(args)) {
    if (!args$legacy) {
      stop(
        "There is not an equivalent 'whatWQPdata' offering for legacy=FALSE."
      )
    }
    args <- args[names(args) != "legacy"]
  }

  values <- readWQPdots(args, legacy = TRUE)
  service <- values[["service"]]
  values <- values[["values"]]

  # Not sure if there's a geojson option with WQX3
  wqp_message()

  if (grepl("WQX", service)) {
    stop("There is not an equivalent 'whatWQPdata' offering for WQX3 services.")
  } else if (service != "Result") {
    message("service argument is not used in whatWQPdata and will be ignored")
  }

  if (any(c("tz", "mimeType") %in% names(values))) {
    values <- values[!(names(values) %in% c("tz", "mimeType"))]
  }
  POST <- FALSE

  baseURL <- httr2::request(pkg.env[["Station"]])

  if ("siteid" %in% names(values)) {
    if (length(values[["siteid"]]) > 1) {
      sites <- values[["siteid"]]

      POST <- nchar(paste0(sites, collapse = "")) > 2048

      baseURL <- get_or_post(
        baseURL,
        POST = POST,
        siteid = sites,
        .multi = function(x) paste0(x, collapse = ";")
      )

      values <- values[names(values) != "siteid"]
    }
  }

  baseURL <- get_or_post(
    req = baseURL,
    POST = POST,
    !!!values,
    .multi = "explode"
  )

  baseURL <- get_or_post(baseURL, POST = POST, mimeType = "geojson")

  doc <- getWebServiceData(baseURL)

  if (is.null(doc)) {
    return(invisible(NULL))
  }

  headerInfo <- attr(doc, "headerInfo")

  if (headerInfo$`total-site-count` == 0) {
    y <- data.frame(
      total_type = character(),
      lat = numeric(),
      lon = numeric(),
      ProviderName = character(),
      OrganizationIdentifier = character(),
      OrganizationFormalName = character(),
      MonitoringLocationIdentifier = character(),
      MonitoringLocationName = character(),
      MonitoringLocationTypeName = character(),
      ResolvedMonitoringLocationTypeName = character(),
      HUCEightDigitCode = character(),
      siteUrl = character(),
      activityCount = numeric(),
      resultCount = numeric(),
      StateName = character(),
      CountyName = character(),
      stringsAsFactors = FALSE
    )
    if (!convertType) {
      y <- data.frame(lapply(y, as.character), stringsAsFactors = FALSE)
    }
  } else {
    features <- doc[["features"]]
    y <- data.frame(matrix(NA, nrow = length(features), ncol = 15))
    names(y) <- c(names(features[[1]][["properties"]]), "lat", "lon")
    for (i in seq_along(features)) {
      single_feature <- features[[i]][["properties"]]
      single_feature[["lat"]] <- unlist(features[[i]][["geometry"]][[
        "coordinates"
      ]][2])
      single_feature[["lon"]] <- unlist(features[[i]][["geometry"]][[
        "coordinates"
      ]][1])
      y[i, ] <- single_feature
    }

    if (convertType) {
      y[, grep("Count$", names(y))] <- sapply(
        y[, grep("Count$", names(y))],
        as.numeric
      )
    }

    names(y)[names(y) == "type"] <- paste(
      "type",
      letters[seq_along(names(y)[names(y) == "type"])],
      sep = "_"
    )

    if (all(c("type_a", "type_b", "features.type") %in% names(y))) {
      y$total_type <- paste(y$type_a, y$features.type, y$type_b)
      y$type_a <- NULL
      if (all(y$type_b == "Point")) {
        y$lon <- sapply(y$coordinates, function(x) x[[1]][1])
        y$lat <- sapply(y$coordinates, function(x) x[[2]][1])
      }
      y$type_b <- NULL
      y$coordinates <- NULL
      y$features.type <- NULL
      y <- y[, c(
        "total_type",
        "lat",
        "lon",
        names(y)[!(names(y) %in% c("total_type", "lat", "lon"))]
      )]
    }
  }

  if (!convertType) {
    y <- data.frame(lapply(y, as.character), stringsAsFactors = FALSE)
  }

  attr(y, "queryTime") <- Sys.time()
  attr(y, "url") <- baseURL
  return(y)
}

get_or_post <- function(req, POST = FALSE, ...) {
  if (POST) {
    req <- httr2::req_body_form(req, ...)
  } else {
    req <- httr2::req_url_query(req, ...)
  }
  return(req)
}

#' Get USGS Field Measurement Water Data
#'
#' @description `r get_description("field-measurements")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("field-measurements")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param parameter_code `r get_ogc_params("field-measurements")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param observing_procedure_code `r get_ogc_params("field-measurements")$observing_procedure_code`
#' @param observing_procedure `r get_ogc_params("field-measurements")$observing_procedure`
#' @param time `r get_ogc_params("field-measurements")$time`
#'
#' See also Details below for more information.
#' @param value `r get_ogc_params("field-measurements")$value`
#' @param unit_of_measure `r get_ogc_params("field-measurements")$unit_of_measure`
#' @param approval_status `r get_ogc_params("field-measurements")$approval_status`
#' @param last_modified `r get_ogc_params("field-measurements")$last_modified`
#'
#' See also Details below for more information.
#' @param qualifier `r get_ogc_params("field-measurements")$qualifier`
#' @param field_visit_id `r get_ogc_params("field-measurements")$field_visit_id`
#' @param field_measurements_series_id A unique identifier representing a single
#' collection series. This corresponds to the `field_measurements_series_id` field in the
#' `read_waterdata_field_meta` endpoint. Collection series are defined as the
#' set of field measurements at a given monitoring location for a single parameter
#' code using a single reading type.
#'
#' @param vertical_datum `r get_ogc_params("field-measurements")$vertical_datum`
#' @param measuring_agency `r get_ogc_params("field-measurements")$measuring_agency`
#' @param control_condition `r get_ogc_params("field-measurements")$control_condition`
#' What and where the control of flow is for the gage pool.
#' @param measurement_rated  `r get_ogc_params("field-measurements")$measurement_rated`
#' Rated measurement based on the hydrologic/hydraulic conditions in which the measurement was made
#' (excellent (2 percent), good (5 percent), fair (8 percent), or poor (more than 8 percent).                                                        percent)
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("field-measurements", "field_measurement_id")`.
#' The default (`NA`) will return all columns of the data.
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @inherit read_waterdata_continuous details
#'
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-02238500"
#' field_data_sf <- read_waterdata_field_measurements(monitoring_location_id = site)
#'
#' groundwater <- read_waterdata_field_measurements(monitoring_location_id = "USGS-375907091432201")
#'
#' field_data <- read_waterdata_field_measurements(monitoring_location_id = "USGS-02238500",
#'                            parameter_code = "00060",
#'                            time = as.POSIXct(c("2024-02-26 15:00:00",
#'                                                "2025-08-27 12:00:00"),
#'                                              tz = "America/Chicago"),
#'                            skipGeometry = TRUE)
#'
#' gwl_data_period <- read_waterdata_field_measurements(
#'                                   monitoring_location_id = "USGS-375907091432201",
#'                                   parameter_code = "72019",
#'                                   time = "P20Y")
#'
#' multi_site <- read_waterdata_field_measurements(
#'                               monitoring_location_id =  c("USGS-451605097071701",
#'                                                           "USGS-263819081585801"),
#'                               parameter_code = c("62611", "72019"))
#'
#' old_df <- read_waterdata_field_measurements(monitoring_location_id = "USGS-425957088141001",
#'                                               time = c("1980-01-01", NA))
#'
#' new_df <- read_waterdata_field_measurements(monitoring_location_id = "USGS-425957088141001",
#'                                               time = c(NA, "2020-01-01"))
#'
#' surface_water <- read_waterdata_field_measurements(
#'                          monitoring_location_id = c("USGS-07069000",
#'                                                     "USGS-07064000",
#'                                                     "USGS-07068000"),
#'                          time = "2024-07-01T00:00:00Z/..",
#'                          parameter_code = "00060")
#'
#'
#' }
read_waterdata_field_measurements <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  observing_procedure_code = NA_character_,
  properties = NA_character_,
  field_visit_id = NA_character_,
  field_measurements_series_id = NA_character_,
  approval_status = NA_character_,
  unit_of_measure = NA_character_,
  qualifier = NA_character_,
  value = NA,
  last_modified = NA_character_,
  observing_procedure = NA_character_,
  vertical_datum = NA_character_,
  measuring_agency = NA_character_,
  control_condition = NA_character_,
  measurement_rated = NA_character_,
  skipGeometry = NA,
  time = NA_character_,
  bbox = NA,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  limit = getOption("dataRetrieval.limit"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_data"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "field-measurements"
  output_id <- "field_measurement_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

#' Generalized USGS Water Data API retrieval function
#'
#' Function that allows complex CQL queries.
#' See <https://api.waterdata.usgs.gov/docs/ogcapi/complex-queries/>
#' for more information.
#'
#' @export
#' @param service character, can be any existing collection.
#' @param CQL A string in a Common Query Language format.
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the function
#' will convert the data to dates and qualifier to string vector.
#' @param \dots Additional arguments to send to the request.
#' @inheritParams check_arguments_non_api
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' cql <- '{
#' "op": "and",
#' "args": [
#'   {
#'     "op": "in",
#'     "args": [
#'       { "property": "parameter_code" },
#'       [ "00060", "00065" ]
#'     ]
#'   },
#'  {
#'     "op": "in",
#'     "args": [
#'       { "property": "monitoring_location_id" },
#'       [ "USGS-07367300", "USGS-03277200" ]
#'     ]
#'   }
#' ]
#' }'
#'
#' dv_data <- read_waterdata(service = "daily",
#'                           CQL = cql,
#'                           time = c("2023-01-01", "2024-01-01"))
#'
#' # A wildcard in CQL2 is %
#' # Here's how to get HUCs that fall within 02070010
#' cql_huc_wildcard <- '{
#' "op": "like",
#' "args": [
#'   { "property": "hydrologic_unit_code" },
#'   "02070010%"
#' ]
#' }'
#'
#' what_huc_sites <- read_waterdata(service = "monitoring-locations",
#'                                  CQL = cql_huc_wildcard)
#'
#' }
read_waterdata <- function(
  service,
  CQL,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  match.arg(service, pkg.env$api_endpoints)

  args <- list(...)

  args[["convertType"]] <- convertType
  args[["limit"]] <- limit
  args[["attach_request"]] <- attach_request
  args[["bbox"]] <- NA
  args[["no_paging"]] <- FALSE # drops id if TRUE
  args[["chunk_size"]] <- NA # Chunking doesn't make sense.

  args[["output_id"]] <- switch(
    service,
    "daily" = "daily_id",
    "latest-daily" = "latest_daily_id",
    "time-series-metadata" = "time_series_id",
    "monitoring-locations" = "monitoring_location_id",
    "latest-continuous" = "latest_continuous_id",
    "continuous" = "continuous_id",
    "field-measurements" = "field_measurement_id",
    service
  )

  if (!"properties" %in% names(args)) {
    args[["properties"]] <- NA_character_
  }

  args[["service"]] <- service

  data_req <- suppressWarnings(do.call(construct_api_requests, args))

  data_req <- data_req |>
    httr2::req_headers(`Content-Type` = "application/query-cql-json") |>
    httr2::req_body_raw(CQL)

  return_list <- walk_pages(data_req)

  return_list <- deal_with_empty(
    return_list,
    args[["properties"]],
    service,
    isTRUE(args[["skipGeometry"]]),
    convertType
  )

  return_list <- rejigger_cols(
    return_list,
    args[["properties"]],
    args[["output_id"]]
  )

  if (convertType) {
    return_list <- cleanup_cols(return_list, service)
    return_list <- order_results(return_list)
    return_list <- move_id_col(return_list, args[["output_id"]])
  }

  return(return_list)
}

#' Get USGS Daily Data
#'
#' @description `r get_description("daily")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("daily")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param parameter_code `r get_ogc_params("daily")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param statistic_id `r get_ogc_params("daily")$statistic_id`
#' Multiple statistic_ids can be requested as a character vector.
#' @param time `r get_ogc_params("daily")$time`
#'
#' See also Details below for more information.
#' @param value `r get_ogc_params("daily")$value`
#' @param unit_of_measure `r get_ogc_params("daily")$unit_of_measure`
#' @param approval_status `r get_ogc_params("daily")$approval_status`
#' @param last_modified `r get_ogc_params("daily")$last_modified`
#'
#' See also Details below for more information.
#' @param time_series_id `r get_ogc_params("daily")$time_series_id`
#' Multiple time_series_ids can be requested as a character vector.
#' @param qualifier `r get_ogc_params("daily")$qualifier`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("daily", "daily_id")`.
#' The default (`NA`) will return all columns of the data.
#'
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @inherit read_waterdata_continuous details
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-02238500"
#' dv_data_sf <- read_waterdata_daily(monitoring_location_id = site,
#'                               parameter_code = "00060",
#'                               time = c("2021-01-01", "2022-01-01"))
#'
#' dv_data_last_modified <- read_waterdata_daily(monitoring_location_id = site,
#'                               parameter_code = "00060",
#'                               last_modified = "P7D")
#'
#' dv_data_trim <- read_waterdata_daily(monitoring_location_id = site,
#'                           parameter_code = "00060",
#'                           properties = c("value",
#'                                          "time"),
#'                           time = c("2021-01-01", "2022-01-01"))
#'
#' dv_data <- read_waterdata_daily(monitoring_location_id = site,
#'                            parameter_code = "00060",
#'                            skipGeometry = TRUE)
#'
#' dv_data_period <- read_waterdata_daily(monitoring_location_id = site,
#'                                   parameter_code = "00060",
#'                                   time = "P7D")
#'
#' multi_site <- read_waterdata_daily(monitoring_location_id =  c("USGS-01491000",
#'                                                           "USGS-01645000"),
#'                               parameter_code = c("00060", "00010"),
#'                               time = c("2023-01-01", "2024-01-01"))
#'
#' dv_data_quick <- read_waterdata_daily(monitoring_location_id = site,
#'                                    parameter_code = "00060",
#'                                    no_paging = TRUE)
#'
#' dv_post <- read_waterdata_daily(monitoring_location_id = site,
#'                                 approval_status = c("Approved", "Provisional"))
#' # Don't attach "request" attribute:
#' options("dataRetrieval.attach_request" = FALSE)
#' dv_data_no_request <- read_waterdata_daily(monitoring_location_id = site,
#'                               parameter_code = "00060",
#'                               time = c("2021-01-01", "2022-01-01"))
#' }
read_waterdata_daily <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  statistic_id = NA_character_,
  properties = NA_character_,
  time_series_id = NA_character_,
  approval_status = NA_character_,
  unit_of_measure = NA_character_,
  qualifier = NA_character_,
  value = NA,
  last_modified = NA_character_,
  skipGeometry = NA,
  time = NA_character_,
  bbox = NA,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_data"),
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "daily"
  output_id <- "daily_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

#' Function to return data from the NWIS RDB 1.0 format
#'
#' This function accepts a url parameter that already contains the desired
#' NWIS site, parameter code, statistic, startdate and enddate. It is not
#' recommended to use the RDB format for importing multi-site data.
#'
#' @param obs_url character containing the url for the retrieval or a file path to the data file.
#' @param asDateTime logical, if `TRUE` returns date and time as POSIXct, if `FALSE`, Date
#' @param tz character to set timezone attribute of datetime. Default converts the datetimes to UTC
#' (properly accounting for daylight savings times based on the data's provided tz_cd column).
#' Recommended US values include "UTC", "America/New_York", "America/Chicago", "America/Denver",
#' "America/Los_Angeles",  "America/Anchorage", "America/Honolulu", "America/Jamaica", "America/Managua",
#' "America/Phoenix", and "America/Metlakatla".
#' For a complete list, see <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>
#' @param convertType logical, defaults to `TRUE`. If `TRUE`, the
#' function will convert the data to dates, datetimes,
#' numerics based on a standard algorithm. If false, everything is returned as a character
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#' datetime \tab POSIXct \tab The date and time of the value converted to
#' UTC (if asDateTime = `TRUE`), \cr
#' \tab character \tab or raw character string (if asDateTime = FALSE) \cr
#' tz_cd \tab character \tab The time zone code for datetime \cr
#' code \tab character \tab Any codes that qualify the corresponding value\cr
#' value \tab numeric \tab The numeric value for the parameter \cr
#' tz_cd_reported \tab The originally reported time zone \cr
#' }
#' Note that code and value are repeated for the parameters requested. The names are of the form
#' XD_P_S, where X is literal,
#' D is an option description of the parameter,
#' P is the parameter code,
#' and S is the statistic code (if applicable).
#' If a date/time (dt) column contained incomplete date and times, a new column
#' of dates and time was inserted. This could happen
#' when older data was reported as dates, and newer data was reported as a date/time.
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' comment \tab character \tab Header comments from the RDB file \cr
#' }
#' @export
#' @examplesIf is_dataRetrieval_user()
#' site_id <- "02177000"
#' startDate <- "2012-09-01"
#' endDate <- "2012-10-01"
#' offering <- "00003"
#' property <- "00060"
#'
#' obs_url <- constructNWISURL(site_id, property,
#'   startDate, endDate, "dv",
#'   format = "tsv"
#' )
#' \donttest{
#' data <- importRDB1(obs_url)
#'
#'
#' urlMultiPcodes <- constructNWISURL("04085427", c("00060", "00010"),
#'   startDate, endDate, "dv",
#'   statCd = c("00003", "00001"), "tsv"
#' )
#'
#' multiData <- importRDB1(urlMultiPcodes)
#'
#' unitDataURL <- constructNWISURL(site_id, property,
#'   "2020-10-30", "2020-11-01", "uv",
#'   format = "tsv"
#' ) # includes timezone switch
#'
#' unitData <- importRDB1(unitDataURL, asDateTime = TRUE)
#'
#' iceSite <- "04024000"
#' start <- "2015-11-09"
#' end <- "2015-11-24"
#' urlIce <- constructNWISURL(iceSite, "00060", start, end, "uv", format = "tsv")
#'
#' ice <- importRDB1(urlIce, asDateTime = TRUE)
#' iceNoConvert <- importRDB1(urlIce, convertType = FALSE)
#' }
#' # User file:
#' filePath <- system.file("extdata", package = "dataRetrieval")
#' fileName <- "RDB1Example.txt"
#' fullPath <- file.path(filePath, fileName)
#' importUserRDB <- importRDB1(fullPath)
#'
importRDB1 <- function(
  obs_url,
  asDateTime = TRUE,
  convertType = TRUE,
  tz = "UTC"
) {
  if (tz == "") {
    tz <- "UTC"
  }

  tz <- match.arg(tz, OlsonNames())

  if (
    is.character(obs_url) &&
      grepl("(https)://[^ /$.?#].[^\\s]*", obs_url)
  ) {
    obs_url <- httr2::request(obs_url)
  }

  if (inherits(obs_url, "httr2_request")) {
    doc <- getWebServiceData(obs_url)

    if (is.null(doc)) {
      return(invisible(NULL))
    }
  } else {
    if (!file.exists(obs_url)) {
      warning("Unknown Input")
      return(NULL)
    }
    doc <- obs_url
  }
  if (file.exists(doc)) {
    readr.total <- readr::read_lines(doc)
  } else {
    readr.total <- readr::read_lines(I(doc))
  }

  if (readr.total[length(readr.total)] == "") {
    readr.total <- readr.total[-length(readr.total)]
  }
  total.rows <- length(readr.total)

  readr.meta <- readr.total[grep("^#", readr.total)]
  meta.rows <- length(readr.meta)
  header.names <- strsplit(readr.total[meta.rows + 1], "\t")[[1]]

  data.rows <- total.rows - meta.rows - 2

  char.names <- c(
    header.names[grep("_cd", header.names)],
    header.names[grep("_id", header.names)],
    header.names[grep("_tx", header.names)],
    header.names[grep("_tm", header.names)],
    header.names[header.names == "site_no"],
    header.names[header.names == "project_no"]
  )

  if (data.rows > 0) {
    args_list <- list(
      file = doc,
      delim = "\t",
      quote = "",
      skip = meta.rows + 2,
      col_names = FALSE
    )

    args_list[["show_col_types"]] <- FALSE

    if (convertType) {
      args_list[["guess_max"]] <- data.rows
      args_list[["col_types"]] <- readr::cols()
    } else {
      args_list[["col_types"]] <- readr::cols(.default = "c")
    }

    readr.data <- suppressWarnings(do.call(readr::read_delim, args = args_list))

    readr.data <- as.data.frame(readr.data)

    if (nrow(readr.data) > 0) {
      names(readr.data) <- header.names

      readr.data <- as.data.frame(readr.data)

      if (length(char.names) > 0) {
        char.names.true <- char.names[sapply(
          readr.data[, char.names],
          is.character
        )]
        char.names <- char.names[!(char.names %in% char.names.true)]
      }

      if (length(char.names) > 0) {
        args_list[["col_types"]] <- readr::cols(.default = "c")
        readr.data.char <- do.call(readr::read_delim, args = args_list)
        names(readr.data.char) <- header.names

        for (j in char.names) {
          readr.data[, j] <- readr.data.char[[j]]
        }
      }

      if (convertType) {
        if (length(grep("_va", names(readr.data))) > 0) {
          # note... if we simply convert any _va to numeric...we lose some QW censoring information from some formats
          vaCols <- grep("_va", names(readr.data))

          for (i in vaCols) {
            readr.data[[i]] <- tryCatch(
              {
                as.numeric(readr.data[[i]])
              },
              warning = function(cond) {
                message(paste(
                  "Column",
                  i,
                  "contains characters that cannot be automatically converted to numeric."
                ))
                return(readr.data[[i]])
              }
            )
          }
        }

        if (asDateTime && convertType) {
          header.suffix <- sapply(strsplit(header.names, "_"), function(x) {
            x[length(x)]
          })
          header.base <- substr(header.names, 1, nchar(header.names) - 3)

          dt_cols <- unique(header.base[header.suffix %in% c("dt", "tm")])

          if (all(c("sample", "sample_end") %in% dt_cols)) {
            if ("sample_start_time_datum_cd" %in% header.names) {
              readr.data[, "tz_cd"] <- readr.data[,
                "sample_start_time_datum_cd"
              ]

              readr.data[,
                "sample_start_time_datum_cd_reported"
              ] <- readr.data[, "sample_start_time_datum_cd"]
              readr.data[, "sample_end_time_datum_cd_reported"] <- readr.data[,
                "sample_start_time_datum_cd"
              ]
              readr.data <- readr.data[, names(readr.data)[
                names(readr.data) != "sample_start_time_datum_cd"
              ]]
            }
          }

          for (i in dt_cols) {
            if (all(c(paste0(i, "_dt"), paste0(i, "_tm")) %in% header.names)) {
              varname <- paste0(i, "_dateTime")

              varval <- suppressWarnings(lubridate::parse_date_time(
                paste(
                  readr.data[, paste0(i, "_dt")],
                  readr.data[, paste0(i, "_tm")]
                ),
                c(
                  "%Y-%m-%d %H:%M:%S",
                  "%Y-%m-%d %H:%M"
                ),
                tz = "UTC"
              ))

              if (!all(is.na(varval))) {
                readr.data[, varname] <- varval
                if (paste0(i, "_tz_cd") %in% names(readr.data)) {
                  tz.name <- paste0(i, "_tz_cd")
                } else {
                  tz.name <- paste0(i, "_time_datum_cd")
                }

                if (tz.name %in% names(readr.data)) {
                  readr.data <- convertTZ(readr.data, tz.name, varname, tz)
                }
              }
            }
          }

          if ("tz_cd" %in% names(readr.data)) {
            date.time.cols <- which(sapply(readr.data, function(x) {
              inherits(x, "POSIXct")
            }))
            if (length(date.time.cols) > 0) {
              readr.data <- convertTZ(
                readr.data,
                "tz_cd",
                date.time.cols,
                tz,
                flip.cols = FALSE
              )
            }
          }

          if ("DATE" %in% header.names) {
            readr.data[, "DATE"] <- lubridate::parse_date_time(
              readr.data[, "DATE"],
              "Ymd"
            )
          }

          if (all(c("DATE", "TIME", "TZCD") %in% header.names)) {
            varname <- "DATETIME"
            varval <- as.POSIXct(lubridate::fast_strptime(
              paste(
                readr.data[, "DATE"],
                readr.data[, "TIME"]
              ),
              "%Y-%m-%d %H%M%S",
              tz = "UTC"
            ))
            readr.data[, varname] <- varval
            readr.data <- convertTZ(
              readr.data,
              "TZCD",
              varname,
              tz,
              flip.cols = TRUE
            )
          }

          if ("sample_dateTime" %in% names(readr.data)) {
            names(readr.data)[
              names(readr.data) == "sample_dateTime"
            ] <- "startDateTime"
          }
        }
        row.names(readr.data) <- NULL
      }
    }
  } else {
    readr.data <- data.frame(
      matrix(
        vector(),
        0,
        length(header.names),
        dimnames = list(c(), header.names)
      ),
      stringsAsFactors = FALSE
    )
  }

  names(readr.data) <- make.names(names(readr.data))

  # People get confused having the tz_cd_reported next to the POSIXs:

  if ("tz_cd_reported" %in% names(readr.data)) {
    new_order <- names(readr.data)
    new_order <- c(
      new_order[!new_order %in% c("tz_cd_reported", "tz_cd")],
      "tz_cd"
    )
    readr.data <- readr.data[, new_order]
  }

  attr(readr.data, "queryTime") <- Sys.time()
  if (inherits(obs_url, "httr2_request")) {
    attr(readr.data, "url") <- obs_url$url
    attr(readr.data, "headerInfo") <- attr(doc, "headerInfo")
  }

  if ("spec" %in% names(attributes(readr.data))) {
    attr(readr.data, "spec") <- NULL
  }

  attr(readr.data, "comment") <- readr.meta

  return(readr.data)
}

convertTZ <- function(df, tz.name, date.time.cols, tz, flip.cols = TRUE) {
  offset <- offsetLibrary$offset[match(df[, tz.name], offsetLibrary$code)]

  df[, paste0(tz.name, "_reported")] <- df[, tz.name, drop = FALSE]

  df[, date.time.cols] <- df[, date.time.cols] + offset * 60 * 60

  for (i in date.time.cols) {
    df[, i] <- as.POSIXct(df[, i])
    if (tz != "") {
      attr(df[, i], "tzone") <- tz
      df[, tz.name] <- tz
    } else {
      attr(df[, i], "tzone") <- "UTC"
      df[!is.na(df[, i]), tz.name] <- "UTC"
    }
  }

  if (flip.cols) {
    reported.col <- which(names(df) %in% paste0(tz.name, "_reported"))
    orig.col <- which(names(df) %in% tz.name)

    new.order <- seq_len(ncol(df))
    new.order[orig.col] <- reported.col
    new.order[reported.col] <- orig.col

    df <- df[, new.order]
  }

  if (all(is.na(df[, date.time.cols]))) {
    df[, date.time.cols] <- NULL
  }

  return(df)
}

#' Rejigger and rename
#'
#' Reorder columns based on users property input.
#' Add "service" prefix to returned "id" column.
#' This allows better integration with other endpoints.
#'
#' @param df data frame returned from walk_pages
#' @param properties A vector of requested columns
#' @param service character, can be any existing collection such
#' as "daily", "monitoring-locations", "time-series-metadata"
#'
#' @return data.frame
#' @noRd
#' @examples
#'
#' df <- dataRetrieval:::deal_with_empty(data.frame(NULL),
#'                                       properties = c("state_code", "county_code", "id"),
#'                                       service = "monitoring-locations")
#' df2 <- dataRetrieval:::rejigger_cols(df,
#'                                      properties = c("state_code", "id", "county_code"),
#'                                      output_id = "monitoring_location_id")
#'
#' df3 <- dataRetrieval:::rejigger_cols(df,
#'                                      properties = c("state_code", "monitoring_location_id", "county_code"),
#'                                      output_id = "monitoring_location_id")
#'
rejigger_cols <- function(df, properties, output_id) {
  if (!all(is.na(properties))) {
    if (!"id" %in% properties) {
      if (output_id %in% properties) {
        names(df)[(names(df) == "id")] <- output_id
      }
    }
    df <- df[, properties]
  } else {
    names(df)[(names(df) == "id")] <- output_id
  }
  df
}


#' Convert columns if needed
#'
#' These are columns that have caused problems in testing.
#' Mostly if the columns are empty on 1 page, but not the next.
#' The qualifier column also comes back as a list column. This
#' is fine for many, but others prefer a character column.
#'
#'
#' @param df data frame returned from walk_pages
#' @param service character, can be any existing collection such
#' as "daily"
#' @return data.frame
#' @noRd
#' @examples
#'
#' df <- dataRetrieval:::deal_with_empty(data.frame(NULL),
#'                                       properties = c("time", "value", "id", "qualifier"),
#'                                       service = "daily")
#' df2 <- dataRetrieval:::rejigger_cols(df,
#'                                      properties = c("value", "id", "time", "qualifier"),
#'                                      service = "daily")
#' df3 <- dataRetrieval:::cleanup_cols(df2)
#'
cleanup_cols <- function(df, service) {
  if ("time" %in% names(df)) {
    if (service %in% c("daily", "peaks")) {
      df$time <- as.Date(df$time)
    } else {
      attr(df$time, "tzone") <- "UTC"
    }
  }

  for (time_period_columns in time_periods) {
    if (time_period_columns %in% names(df)) {
      attr(df[[time_period_columns]], "tzone") <- "UTC"
    }
  }

  df
}

#' Get USGS Field Measurement Metadata
#'
#' @description `r get_description("field-measurements-metadata")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("field-measurements-metadata")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param parameter_code `r get_ogc_params("field-measurements-metadata")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param parameter_name `r get_ogc_params("field-measurements-metadata")$parameter_name`
#' Multiple parameter_names can be requested as a character vector.
#' @param parameter_description `r get_ogc_params("field-measurements-metadata")$parameter_description`
#' Multiple parameter_descriptions can be requested as a character vector.
#' @param last_modified `r get_ogc_params("field-measurements-metadata")$last_modified`
#'
#' See also Details below for more information.
#' @param begin `r get_ogc_params("field-measurements-metadata")$begin`
#'
#' See also Details below for more information.
#' @param end `r get_ogc_params("field-measurements-metadata")$end`
#'
#' See also Details below for more information.
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("field-measurements-metadata", "field_measurement_id")`.
#' The default (`NA`) will return all columns of the data.
#'
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @inherit read_waterdata_continuous details
#'
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-02238500"
#' field_data_sf <- read_waterdata_field_meta(monitoring_location_id = site)
#'
#' groundwater <- read_waterdata_field_meta(monitoring_location_id = "USGS-375907091432201")
#'
#' gwl_data <- read_waterdata_field_meta(monitoring_location_id = "USGS-02238500",
#'                            parameter_code = "00060",
#'                            begin = as.POSIXct(c(NA,
#'                                                "2025-08-27 12:00:00"),
#'                                              tz = "America/Chicago"),
#'                            skipGeometry = TRUE)
#'
#' gwl_data_period <- read_waterdata_field_meta(
#'                                   monitoring_location_id = "USGS-375907091432201",
#'                                   parameter_code = "72019",
#'                                   last_modified = "P1Y")
#'
#' multi_site <- read_waterdata_field_meta(
#'                               monitoring_location_id =  c("USGS-451605097071701",
#'                                                           "USGS-263819081585801"),
#'                               parameter_code = c("62611", "72019"))
#'
#'
#' surface_water <- read_waterdata_field_meta(
#'                          monitoring_location_id = c("USGS-07069000",
#'                                                     "USGS-07064000",
#'                                                     "USGS-07068000"),
#'                          end = "2024-07-01T00:00:00Z/..",
#'                          parameter_code = "00060")
#'
#'
#' }
read_waterdata_field_meta <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  parameter_name = NA_character_,
  parameter_description = NA_character_,
  begin = NA_character_,
  end = NA_character_,
  last_modified = NA_character_,
  properties = NA_character_,
  skipGeometry = NA,
  bbox = NA,
  limit = NA,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "field-measurements-metadata"
  output_id <- "field_measurements_series_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

#' Get USGS Rating Curve Data
#'
#' Reads current rating table for an active USGS streamgages. More information
#' can be found at https://api.waterdata.usgs.gov/docs/stac/.
#'
#' @param monitoring_location_id A unique identifier representing a single
#' monitoring location. Monitoring location IDs are created by combining the
#' agency code of the agency responsible for the monitoring location (e.g. USGS)
#' with the ID number of the monitoring location (e.g. 02238500), separated by
#' a hyphen (e.g. USGS-02238500).
#' @param file_type Rating file time. Could be any of "exsa", "corr", or "base".
#' If `file_type` is "base" then the columns are
#' INDEP, typically the gage height, in feet; DEP, typically the streamflow,
#' in cubic feet per second; and STOR, where "*" indicates that the pair are
#' a fixed point of the rating curve. If `file_type` is "exsa" then an
#' additional column, SHIFT, is included that indicates the current shift in
#' the rating for that value of INDEP. If `file_type` is "corr" then the
#' columns are INDEP, typically the gage height, in feet; CORR, the correction
#' for that value; and CORRINDEP, the corrected value for CORR.
#' @param file_path Path to save the rating curve rdb files. The
#' default is `tempdir()`, which will wipe out the files.
#' @param datetime Only return items that have a temporal property that
#' intersects this value. Either a date-time or an interval, open or closed.
#' See Details below.
#' @param bbox Only features that have a geometry that intersects the bounding
#' box are selected.The bounding box is provided as four or six numbers, depending
#' on whether the coordinate reference system includes a vertical axis (height or
#' depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
#' vector structured: c(xmin,ymin,xmax,ymax).
#' Another way to think of it is c(Western-most longitude,
#' Southern-most latitude, Eastern-most longitude, Northern-most longitude).
#' @param \dots Not used.
#' @param limit Limits the number of results that are included in each page of
#' the response (capped at the default 10,000).
#' @param download_and_parse Logical to define whether or not to download, parse,
#' and return a list of data frames with rating curve data (`TRUE`), or to return
#' just a list of available rating curve files (`FALSE`). Default is `TRUE`.
#' @export
#' @inherit read_waterdata_continuous details
#'
#' @return List of data frames which contain the requested rating curves.
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#'
#' monitoring_location_id <- c("USGS-01104475", "USGS-01104460")
#' ratings_exsa <- read_waterdata_ratings(
#'       monitoring_location_id = monitoring_location_id,
#'       file_type = "exsa")
#'
#' head(ratings_exsa[["USGS-01104475.exsa.rdb"]])
#' comment(ratings_exsa[["USGS-01104475.exsa.rdb"]])[1:15]
#'
#' ratings_corr <- read_waterdata_ratings(
#'       monitoring_location_id = monitoring_location_id,
#'       file_type = "corr")
#'
#' head(ratings_corr[["USGS-01104460.corr.rdb"]])
#' comment(ratings_corr[["USGS-01104460.corr.rdb"]])[1:15]
#'
#' rating_2 <- read_waterdata_ratings(
#'       monitoring_location_id = monitoring_location_id,
#'       file_type = c("corr", "exsa"))
#' names(rating_2)
#'
#' bbox <- c(-95.00, 40.0, -92.0, 42)
#'
#' bbox_query <- read_waterdata_ratings(bbox = bbox,
#'                                      download_and_parse = FALSE)
#' length(bbox_query)
#' recent_query <- read_waterdata_ratings(bbox = bbox,
#'                                        datetime = c(Sys.Date()-7, NA),
#'                                        download_and_parse = FALSE)
#' length(recent_query)
#'}
read_waterdata_ratings <- function(
  monitoring_location_id = NA_character_,
  file_type = c("exsa", "base", "corr"),
  file_path = tempdir(),
  bbox = NA,
  datetime = NA_character_,
  ...,
  limit = 10000,
  download_and_parse = TRUE
) {
  match.arg(
    arg = file_type,
    choices = c("exsa", "base", "corr"),
    several.ok = TRUE
  )
  rlang::check_dots_empty()

  request <- httr2::request("https://api.waterdata.usgs.gov/stac/v0/") |>
    httr2::req_url_path_append("search")

  filter <- NA_character_

  if (!all(is.na(monitoring_location_id))) {
    if (length(monitoring_location_id) > 1) {
      monitoring_location_id <- paste0(
        monitoring_location_id,
        collapse = "', '"
      )
    }

    filter <- sprintf(
      "monitoring_location_id IN ('%s')",
      monitoring_location_id
    )
  }

  if (length(file_type) == 1) {
    filter <- sprintf("%s AND file_type = '%s'", filter, file_type)
  }

  if (!is.na(filter)) {
    if (substr(filter, 1, 3) == "AND") {
      filter <- substr(filter, 4, nchar(filter))
    }

    request <- request |>
      httr2::req_url_query(filter = filter)
  }

  if (!all(is.na(datetime))) {
    if (any(grepl("P", datetime))) {
      stop(
        "Periods are not supported in datetime argument in the rating curve service."
      )
    }
    datetime <- format_api_dates(datetime, date = FALSE)

    request <- request |>
      httr2::req_url_query(datetime = datetime)
  }

  if (all(!is.na(bbox))) {
    request <- httr2::req_url_query(
      request,
      bbox = as.numeric(bbox),
      .multi = "comma"
    )
  }

  request <- request |>
    httr2::req_url_query(limit = limit) |>
    basic_request()

  resp <- httr2::req_perform(request)
  log_rate_limit(resp)

  features <- httr2::resp_body_json(resp)[["features"]]

  if (download_and_parse) {
    return_list <- list()
    for (feature in features) {
      id <- feature$id
      df <- download_convert(feature, file_path, file_type)
      if (!is.null(df)) {
        return_list[[id]] <- df
      }
    }

    return(return_list)
  } else {
    return(features)
  }
}

download_convert <- function(feature, file_path, file_type) {
  links <- feature$links
  id <- feature$id
  url <- feature$assets$data$href

  req <- httr2::request(url) |>
    basic_request()

  if (any(sapply(file_type, function(x) grepl(x, url)))) {
    full_file_path <- file.path(file_path, id)
    message("Requesting: \n", url)
    resp <- httr2::req_perform(req, path = full_file_path)
    rating <- importRDB1(full_file_path)
    return(rating)
  }

  return(NULL)
}

#' Function to return data from the NWISWeb WaterML1.1 service
#'
#' This function accepts a url parameter that already contains the desired
#' NWIS site, parameter code, statistic, startdate and enddate.
#'
#' @param obs_url character or raw, containing the url for the retrieval or a
#' file path to the data file, or raw XML.
#' @param asDateTime logical, if `TRUE` returns date and time as POSIXct, if `FALSE`, Date
#' @param tz character to set timezone attribute of datetime. Default converts the datetimes to UTC
#' (properly accounting for daylight savings times based on the data's provided tz_cd column).
#' Recommended US values include "UTC", "America/New_York", "America/Chicago", "America/Denver",
#' "America/Los_Angeles", "America/Anchorage", "America/Honolulu", "America/Jamaica", "America/Managua",
#' "America/Phoenix", and "America/Metlakatla".
#' For a complete list, see <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>
#' @return A data frame with the following columns:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
#' site_no \tab character \tab The USGS site number \cr
#'  \tab POSIXct \tab The date and time of the value converted to UTC (if asDateTime = TRUE), \cr
#' \tab character \tab or raw character string (if asDateTime = FALSE) \cr
#' tz_cd \tab character \tab The time zone code for  \cr
#' code \tab character \tab Any codes that qualify the corresponding value\cr
#' value \tab numeric \tab The numeric value for the parameter \cr
#' }
#' Note that code and value are repeated for the parameters requested. The names are of the form
#' X_D_P_S, where X is literal,
#' D is an option description of the parameter,
#' P is the parameter code,
#' and S is the statistic code (if applicable).
#'
#' There are also several useful attributes attached to the data frame:
#' \tabular{lll}{
#' Name \tab Type \tab Description \cr
#' url \tab character \tab The url used to generate the data \cr
#' siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
#' variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
#' statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
#' queryTime \tab POSIXct \tab The time the data was returned \cr
#' }
#'
#' @seealso [renameNWISColumns()]
#' @export
#' @examplesIf is_dataRetrieval_user()
#' site_id <- "02177000"
#' startDate <- "2012-09-01"
#' endDate <- "2012-10-01"
#' offering <- "00003"
#' property <- "00060"
#' obs_url <- constructNWISURL(site_id, property, startDate, endDate, "dv")
#' \donttest{
#' data <- importWaterML1(obs_url, asDateTime = TRUE)
#'
#'
#' unitDataURL <- constructNWISURL(
#'   site_id, property,
#'   "2013-11-03", "2013-11-03", "uv"
#' )
#' unitData <- importWaterML1(unitDataURL, TRUE)
#'
#' # Two sites, two pcodes, one site has two data descriptors:
#' siteNumber <- c("01480015", "04085427")
#' obs_url <- constructNWISURL(
#'   siteNumber, c("00060", "00010"),
#'   startDate, endDate, "dv"
#' )
#' data <- importWaterML1(obs_url)
#' data$dateTime <- as.Date(data$dateTime)
#' data <- renameNWISColumns(data)
#' names(attributes(data))
#' attr(data, "url")
#' attr(data, "disclaimer")
#'
#' inactiveSite <- "05212700"
#' inactiveSite <- constructNWISURL(inactiveSite, "00060",
#'                                  "2014-01-01", "2014-01-10", "dv")
#' inactiveSite <- importWaterML1(inactiveSite)
#'
#' inactiveAndAcitive <- c("07334200", "05212700")
#' inactiveAndAcitive <- constructNWISURL(inactiveAndAcitive,
#'                          "00060", "2014-01-01", "2014-01-10", "dv")
#' inactiveAndAcitive <- importWaterML1(inactiveAndAcitive)
#'
#' # Timezone change with specified local timezone:
#' tzURL <- constructNWISURL("04027000", c("00300", "63680"),
#'                           "2011-11-05", "2011-11-07", "uv")
#' tzIssue <- importWaterML1(tzURL,
#'   asDateTime = TRUE, tz = "America/Chicago"
#' )
#'
#' }
#' filePath <- system.file("extdata", package = "dataRetrieval")
#' fileName <- "WaterML1Example.xml"
#' fullPath <- file.path(filePath, fileName)
#' importFile <- importWaterML1(fullPath, TRUE)
#'
importWaterML1 <- function(obs_url, asDateTime = FALSE, tz = "UTC") {
  # note: obs_url is a dated name, does not have to be a url/path

  returnedDoc <- check_if_xml(obs_url)
  raw <- !is.character(obs_url) & !("httr2_request" %in% class(obs_url))

  if (tz == "") {
    # check tz is valid if supplied
    tz <- "UTC"
  }

  tz <- match.arg(tz, OlsonNames())

  timeSeries <- xml2::xml_find_all(returnedDoc, ".//ns1:timeSeries") # each parameter/site combo

  # some initial attributes
  queryNodes <- xml2::xml_children(xml2::xml_find_all(
    returnedDoc,
    ".//ns1:queryInfo"
  ))
  notes <- queryNodes[xml2::xml_name(queryNodes) == "note"]
  noteTitles <- xml2::xml_attrs(notes)
  noteText <- xml2::xml_text(notes)
  noteList <- as.list(noteText)
  names(noteList) <- noteTitles

  if (0 == length(timeSeries)) {
    df <- data.frame(
      agency_cd = character(),
      site_no = character(),
      dateTime = as.POSIXct(character()),
      tz_cd = character()
    )
    attr(df, "queryInfo") <- noteList
    if (!raw) {
      attr(df, "url") <- obs_url$url
    }
    return(df)
  }

  mergedDF <- NULL

  for (t in timeSeries) {
    # check if there are multiple time series (ie with different descriptors)
    # descriptor will be appended to col name if so
    valParents <- xml2::xml_find_all(t, ".//ns1:values")
    obsDF <- NULL
    useMethodDesc <- FALSE
    if (length(valParents) > 1) {
      useMethodDesc <- TRUE
    } # append the method description to colnames later

    sourceInfo <- xml2::xml_children(xml2::xml_find_all(t, ".//ns1:sourceInfo"))
    variable <- xml2::xml_children(xml2::xml_find_all(t, ".//ns1:variable"))
    agency_cd <- xml2::xml_attr(
      sourceInfo[xml2::xml_name(sourceInfo) == "siteCode"],
      "agencyCode"
    )
    pCode <- xml2::xml_text(variable[
      xml2::xml_name(variable) == "variableCode"
    ])
    statCode <- xml2::xml_attr(
      xml2::xml_find_all(variable, ".//ns1:option"),
      "optionCode"
    )

    # site info
    srsNode <- xml2::xml_find_all(sourceInfo, ".//ns1:geogLocation")
    srs <- xml2::xml_attr(srsNode, "srs")
    locNodes <- xml2::xml_children(srsNode)
    locNames <- xml2::xml_name(locNodes)
    locText <- as.numeric(xml2::xml_text(locNodes))
    names(locText) <- sub(
      "longitude",
      "dec_lon_va",
      sub("latitude", "dec_lat_va", locNames)
    )
    sitePropNodes <- sourceInfo[xml2::xml_name(sourceInfo) == "siteProperty"]
    siteProp <- xml2::xml_text(sitePropNodes)
    names(siteProp) <- xml2::xml_attr(sitePropNodes, "name")
    tzInfo <- unlist(xml2::xml_attrs(xml2::xml_find_all(
      sourceInfo,
      "ns1:defaultTimeZone"
    )))
    siteName <- xml2::xml_text(sourceInfo[
      xml2::xml_name(sourceInfo) == "siteName"
    ])
    siteCodeNode <- sourceInfo[xml2::xml_name(sourceInfo) == "siteCode"]
    site_no <- xml2::xml_text(siteCodeNode)
    siteCodeAtt <- unlist(xml2::xml_attrs(siteCodeNode))
    siteDF <- cbind.data.frame(
      t(locText),
      t(tzInfo),
      station_nm = siteName,
      t(siteCodeAtt),
      srs,
      t(siteProp),
      site_no,
      stringsAsFactors = FALSE
    )
    defaultTZ <- xml2::xml_attr(
      xml2::xml_find_all(sourceInfo, ".//ns1:defaultTimeZone"),
      "zoneAbbreviation"
    )

    for (v in valParents) {
      obsColName <- paste(pCode, statCode, sep = "_")
      obs <- xml2::xml_find_all(v, ".//ns1:value")
      values <- as.numeric(xml2::xml_text(obs)) # actual observations

      nObs <- length(values)
      qual <- xml2::xml_attr(obs, "qualifiers")

      dateTime <- xml2::xml_attr(obs, "dateTime")
      if (asDateTime) {
        numChar <- nchar(dateTime)
        dateTime <- lubridate::parse_date_time(
          dateTime,
          c(
            "%Y",
            "%Y-%m-%d",
            "%Y-%m-%dT%H:%M",
            "%Y-%m-%dT%H:%M:%S",
            "%Y-%m-%dT%H:%M:%OS",
            "%Y-%m-%dT%H:%M:%OS%z"
          ),
          exact = TRUE
        )
        if (any(numChar < 20) && any(numChar > 16)) {
          # not sure there is still a case for this (no offset on times)?
          dateTime[numChar < 20 & numChar > 16] <- dateTime[
            numChar < 20 & numChar > 16
          ] +
            offsetLibrary[offsetLibrary$code == defaultTZ, "offset"] * 60 * 60
          warning(paste(
            "site",
            site_no[1],
            "had data without time zone offsets, so DST could not be accounted for"
          ))
        }

        # ^^setting tz in as.POSIXct just sets the attribute, does not convert the time!
        # Set to UTC now, convert the whole thing to tz later.
        attr(dateTime, "tzone") <- "UTC"
        tzCol <- rep(tz, nObs)
      } else {
        tzCol <- rep(defaultTZ, nObs)
      }
      # create column names, addressing if methodDesc is needed
      if (useMethodDesc) {
        methodDesc <- xml2::xml_text(xml2::xml_find_all(
          v,
          ".//ns1:methodDescription"
        ))
        # this keeps column names consistent with old version
        methodDesc <- gsub("\\[|\\]| |\\(|\\)", ".", methodDesc)

        # sometimes methodDesc is empty
        if (nchar(methodDesc) > 0) {
          obsColName <- paste("X", methodDesc, obsColName, sep = "_")
        } else {
          obsColName <- paste("X", obsColName, sep = "_")
        }
      } else {
        obsColName <- paste("X", obsColName, sep = "_")
      }
      qualColName <- paste(obsColName, "cd", sep = "_")

      valParentDF <- cbind.data.frame(
        dateTime,
        values,
        qual,
        tzCol,
        stringsAsFactors = FALSE
      )
      names(valParentDF) <- c("dateTime", obsColName, qualColName, "tz_cd")
      # delete qual column if all NA
      if (all(is.na(valParentDF[, eval(qualColName)]))) {
        valParentDF <- subset(
          valParentDF,
          select = c("dateTime", eval(obsColName), "tz_cd")
        )
      }
      if (nrow(valParentDF) > 0) {
        if (is.null(obsDF)) {
          obsDF <- valParentDF
        } else {
          obsDF <- merge(
            x = obsDF,
            y = valParentDF,
            by = c("dateTime", "tz_cd"),
            all = TRUE
          )
        }
      } else {
        # need column names for joining later
        # but don't overwrite:
        if (is.null(obsDF)) {
          obsDF <- data.frame(
            dateTime = character(0),
            tz_cd = character(0),
            stringsAsFactors = FALSE
          )
          if (asDateTime) {
            obsDF$dateTime <- as.POSIXct(obsDF$dateTime)
            attr(obsDF$dateTime, "tzone") <- "UTC"
          }
        }
      }
    }

    if (is.null(obsDF)) {
      mergedSite <- data.frame()
      next
    }
    nObs <- nrow(obsDF)

    # statistic info
    options <- xml2::xml_find_all(variable, "ns1:option")
    stat <- options[xml2::xml_attr(options, "name") == "Statistic"]
    stat_nm <- xml2::xml_text(options[
      xml2::xml_attr(stat, "name") == "Statistic"
    ])
    statCd <- xml2::xml_attr(stat, "optionCode")
    statDF <- cbind.data.frame(
      statisticCd = statCd,
      statisticName = stat_nm,
      stringsAsFactors = FALSE
    )

    # variable info
    varText <- as.data.frame(
      t(xml2::xml_text(variable)),
      stringsAsFactors = FALSE
    )
    varNames <- xml2::xml_name(variable)
    varNames <- sub("unit", "param_unit", varNames) # rename to stay consistent with orig importWaterMl1
    names(varText) <- varNames

    # replace no data vals with NA, change attribute df
    noDataVal <- as.numeric(varText$noDataValue)
    if (nObs > 0 && obsColName %in% names(obsDF)) {
      obsDF[[obsColName]][obsDF[[obsColName]] == noDataVal] <- NA
    }
    varText$noDataValue <- NA

    # rep site no & agency, combine into DF
    obsDFrows <- nrow(obsDF)
    df <- cbind.data.frame(
      agency_cd = rep(agency_cd, obsDFrows),
      site_no = rep(site_no, obsDFrows),
      obsDF,
      stringsAsFactors = FALSE
    )

    # join by site no
    # append siteInfo, stat, and variable if they don't match a previous one
    if (is.null(mergedDF)) {
      mergedDF <- df
      mergedSite <- siteDF
      mergedVar <- varText
      mergedStat <- statDF
    } else {
      if (nrow(df) > 0) {
        # merge separately with any same site nos, then recombine
        sameSite <- mergedDF[mergedDF$site_no == site_no, ]
        if (nrow(sameSite) > 0) {
          diffSite <- mergedDF[mergedDF$site_no != site_no, ]
          # first need to delete the obs and qual columns if they have already been filled with NA
          deleteCols <- grepl(obsColName, colnames(sameSite))
          sameSite <- sameSite[, !deleteCols]
          sameSite_simNames <- intersect(colnames(sameSite), colnames(df))
          sameSite <- merge(
            x = sameSite,
            y = df,
            by = sameSite_simNames,
            all = TRUE
          )
          na.omit.unique <- function(x) {
            if (all(is.na(x))) NA else stats::na.omit(x)
          }

          if (
            any(duplicated(sameSite[, c(
              "agency_cd",
              "site_no",
              "dateTime",
              "tz_cd"
            )]))
          ) {
            types <- lapply(sameSite, class)
            sameSite <- stats::aggregate(
              . ~ agency_cd + site_no + dateTime + tz_cd,
              data = sameSite,
              FUN = na.omit.unique,
              na.action = NULL
            )
            sameSite[, which(types == "numeric")] <- sapply(
              sameSite[, which(types == "numeric")],
              as.numeric
            )
            sameSite[, which(types == "character")] <- sapply(
              sameSite[, which(types == "character")],
              as.character
            )
          }

          sameSite <- sameSite[order(as.Date(sameSite$dateTime)), ]

          mergedDF <- r_bind_dr(sameSite, diffSite)
        } else {
          similarNames <- intersect(colnames(mergedDF), colnames(df))
          mergedDF <- merge(
            x = mergedDF,
            y = df,
            by = similarNames,
            all = TRUE
          )
        }
      }
      mergedSite <- merge(
        x = mergedSite,
        y = siteDF,
        by = colnames(mergedSite),
        all = TRUE
      )
      mergedVar <- merge(
        x = mergedVar,
        y = varText,
        by = colnames(mergedVar),
        all = TRUE
      )
      mergedStat <- merge(
        x = mergedStat,
        y = statDF,
        by = colnames(mergedStat),
        all = TRUE
      )
    }
  }

  if (!is.null(mergedSite)) {
    # keep attribute df names the same as old version
    names(mergedSite) <- c(
      "dec_lat_va",
      "dec_lon_va",
      "timeZoneOffset",
      "timeZoneAbbreviation",
      "station_nm",
      "network",
      "agency_cd",
      "srs",
      "siteTypeCd",
      "hucCd",
      "stateCd",
      "countyCd",
      "site_no"
    )
    mergedSite <- mergedSite[c(
      "station_nm",
      "site_no",
      "agency_cd",
      "timeZoneOffset",
      "timeZoneAbbreviation",
      "dec_lat_va",
      "dec_lon_va",
      "srs",
      "siteTypeCd",
      "hucCd",
      "stateCd",
      "countyCd",
      "network"
    )]
  }

  # move tz column to far right and sort by increasing site number to be consistent with old version
  mergedNames <- names(mergedDF)
  tzLoc <- grep("tz_cd", names(mergedDF))
  attr(mergedDF$dateTime, "tzone") <- tz
  mergedDF <- mergedDF[c(mergedNames[-tzLoc], mergedNames[tzLoc])]

  mergedDF <- mergedDF[order(mergedDF$site_no, mergedDF$dateTime), ]
  ###############################################################
  names(mergedDF) <- make.names(names(mergedDF))

  # attach other site info etc as attributes of mergedDF
  if ("httr2_request" %in% class(obs_url)) {
    attr(mergedDF, "url") <- obs_url$url
  }
  attr(mergedDF, "siteInfo") <- mergedSite
  attr(mergedDF, "variableInfo") <- mergedVar
  attr(mergedDF, "disclaimer") <- noteText[noteTitles == "disclaimer"]
  attr(mergedDF, "statisticInfo") <- mergedStat
  attr(mergedDF, "queryTime") <- Sys.time()

  return(mergedDF)
}

r_bind_dr <- function(df1, df2) {
  # Note...this function doesn't retain factors/levels
  # That is not a problem with any dataRetrieval function
  # but, if this function gets used else-where,
  # that should be addressed.
  df1 <- add_empty_col(df1, df2, setdiff(names(df2), names(df1)))
  df2 <- add_empty_col(df2, df1, setdiff(names(df1), names(df2)))

  df3 <- rbind(df1, df2)

  return(df3)
}

add_empty_col <- function(df, df_ref, col_names) {
  if (length(col_names) > 0) {
    if (nrow(df) > 0) {
      df[, col_names] <- NA
    } else {
      for (i in col_names) {
        column_type <- class(df_ref[[i]])
        df[i] <- empty_col(column_type)
      }
    }
  }
  return(df)
}

empty_col <- function(column_type) {
  if (all(column_type %in% c("POSIXct", "POSIXt"))) {
    column_type <- "POSIXct"
  }

  col_return <- switch(
    column_type,
    "numeric" = as.numeric(),
    "factor" = as.factor(),
    "list" = list(),
    "integer" = as.integer(),
    "Date" = as.Date(numeric(), origin = "1970-01-01"),
    "POSIXct" = as.POSIXct(numeric(), origin = "1970-01-01"),
    "POSIXlt" = as.POSIXlt(numeric(), origin = "1970-01-01"),
    "character" = as.character()
  )

  return(col_return)
}

check_if_xml <- function(obs_url) {
  if (is.character(obs_url) && file.exists(obs_url)) {
    returnedDoc <- xml2::read_xml(obs_url)
  } else if (is.raw(obs_url)) {
    returnedDoc <- xml2::read_xml(obs_url)
  } else if (inherits(obs_url, c("xml_node", "xml_nodeset"))) {
    returnedDoc <- obs_url
  } else {
    doc <- getWebServiceData(obs_url)
    if (is.null(doc)) {
      return(invisible(NULL))
    }
    returnedDoc <- xml2::xml_root(doc)
  }
  return(returnedDoc)
}

#' Get USGS Time Series Metadata
#'
#' @description `r get_description("time-series-metadata")`
#'
#' @export
#' @param monitoring_location_id `r get_ogc_params("time-series-metadata")$monitoring_location_id`
#' Multiple monitoring_location_ids can be requested as a character vector.
#' @param parameter_code `r get_ogc_params("time-series-metadata")$parameter_code`
#' Multiple parameter_codes can be requested as a character vector.
#' @param parameter_name `r get_ogc_params("time-series-metadata")$parameter_name`
#' @param statistic_id `r get_ogc_params("time-series-metadata")$statistic_id`
#' Multiple statistic_ids can be requested as a character vector.
#' @param computation_identifier `r get_ogc_params("time-series-metadata")$computation_identifier`
#' Multiple computation_identifiers can be requested as a character vector.
#' @param computation_period_identifier `r get_ogc_params("time-series-metadata")$computation_period_identifier`
#' Multiple computation_period_identifiers can be requested as a character vector.
#' @param sublocation_identifier `r get_ogc_params("time-series-metadata")$sublocation_identifier`
#' @param last_modified `r get_ogc_params("time-series-metadata")$last_modified`
#'
#' See also Details below for more information.
#' @param begin_utc `r get_ogc_params("time-series-metadata")$begin_utc`
#' #'
#' See also Details below for more information.
#' @param end_utc `r get_ogc_params("time-series-metadata")$end_utc`
#'
#' See also Details below for more information.
#' @param hydrologic_unit_code `r get_ogc_params("time-series-metadata")$hydrologic_unit_code`
#' @param state_name `r get_ogc_params("time-series-metadata")$state_name`
#' @param thresholds `r get_ogc_params("time-series-metadata")$thresholds`
#' @param unit_of_measure `r get_ogc_params("time-series-metadata")$unit_of_measure`
#' @param primary
#' A flag identifying if the time series is a "primary" time series. "Primary" time
#' series (which have this flag) are standard observations which undergo Bureau
#' review and approval processes. Non-primary time series, which will have missing
#' values for "primary", are provisional datasets made available to meet the need
#' for timely best science and to assist with daily operations which need
#' real-time information. Non-primary time series data are only retained by
#' this system for 120 days.
#' @param parent_time_series_id `r get_ogc_params("time-series-metadata")$parent_time_series_id`
#' @param web_description `r get_ogc_params("time-series-metadata")$web_description`
#' @param begin `r get_ogc_params("time-series-metadata")$begin`
#' @param end `r get_ogc_params("time-series-metadata")$end`
#' @param properties A vector of requested columns to be returned from the query.
#' Available options are:
#' `r dataRetrieval:::get_properties_for_docs("time-series-metadata", "time_series_id")`.
#' The default (`NA`) will return all columns of the data.
#' @param time_series_id `r get_ogc_params("time-series-metadata")$id`
#' @inheritParams check_arguments_api
#' @inheritParams check_arguments_non_api
#'
#' @inherit read_waterdata_continuous details
#'
#' @examplesIf is_dataRetrieval_user()
#'
#' \donttest{
#' site <- "USGS-02238500"
#' meta_1 <- read_waterdata_ts_meta(monitoring_location_id = site)
#'
#' meta_multi <- read_waterdata_ts_meta(monitoring_location_id =  c("USGS-01491000",
#'                                                                  "USGS-01645000"),
#'                             parameter_code = c("00060", "00010"),
#'                             properties = c("monitoring_location_id",
#'                                            "parameter_code",
#'                                            "begin_utc",
#'                                            "end_utc",
#'                                            "time_series_id"),
#'                             skipGeometry = TRUE)
#'
#' meta_wi <- read_waterdata_ts_meta(state_name = "Wisconsin")
#' }
read_waterdata_ts_meta <- function(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  parameter_name = NA_character_,
  properties = NA_character_,
  statistic_id = NA_character_,
  last_modified = NA_character_,
  begin_utc = NA_character_,
  end_utc = NA_character_,
  hydrologic_unit_code = NA_character_,
  state_name = NA_character_,
  unit_of_measure = NA_character_,
  computation_period_identifier = NA_character_,
  computation_identifier = NA_character_,
  thresholds = NA,
  sublocation_identifier = NA_character_,
  primary = NA_character_,
  parent_time_series_id = NA_character_,
  time_series_id = NA_character_,
  web_description = NA_character_,
  skipGeometry = NA,
  bbox = NA,
  begin = NA_character_,
  end = NA_character_,
  ...,
  limit = getOption("dataRetrieval.limit"),
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  attach_request = getOption("dataRetrieval.attach_request")
) {
  service <- "time-series-metadata"
  output_id <- "time_series_id"
  rlang::check_dots_empty()

  args <- mget(names(formals()))
  return_list <- get_ogc_data(args, output_id, service)

  return(return_list)
}

---
title: "Introduction to the dataRetrieval package"
date: "`r format(Sys.time(), '%d %B, %Y')`"
author: Laura A. DeCicco and Robert M. Hirsch
output: 
  rmarkdown::html_vignette:
    toc: true
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Introduction to the dataRetrieval package}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---


```{r setup, include=FALSE, message=FALSE}
library(knitr)
library(dataRetrieval)

options(continue = " ")
options(width = 60)
knitr::opts_chunk$set(
  echo = TRUE,
  warning = FALSE,
  message = FALSE,
  fig.height = 7,
  fig.width = 7
)
```

The `dataRetrieval` package was created to simplify the process of loading hydrologic data into the R environment. It is designed to retrieve the major data types of U.S. Geological Survey (USGS) hydrologic data that are available on the Web, as well as data from the Water Quality Portal (WQP), which currently houses water quality data from the Environmental Protection Agency (EPA), U.S. Department of Agriculture (USDA), and USGS. Direct USGS data is obtained from a service called the USGS Water Data API. 

For information on getting started in R and installing the package, see [Getting Started](#getting-started-in-r). Any use of trade, firm, or product names is for descriptive purposes only and does not imply endorsement by the U.S. Government.

A quick workflow using `dataRetrieval` functions:

```{r workflow, echo=TRUE,eval=FALSE}
library(dataRetrieval)
# Choptank River near Greensboro, MD
siteNumber <- "USGS-01491000"
ChoptankInfo <- read_waterdata_monitoring_location(siteNumber)
parameterCd <- "00060"

# Raw daily data:
rawDailyData <- read_waterdata_daily(monitoring_location_id = siteNumber,
                                     parameter_code = parameterCd,
                                     time = c("1980-01-01", "2010-01-01"))


pCode <- read_waterdata_parameter_codes(parameter_code = parameterCd)
```

## USGS Data API Access Tokens

USGS data are made available through the USGS water data APIs. You can register an API key for use with USGS water data APIs. There are now limits on how many queries can be requested per IP address per hour. If you find yourself running into limits, you can request an API token here: <https://api.waterdata.usgs.gov/signup/>

Then save your token in your .Renviron file like this:

```
API_USGS_PAT = "my_super_secret_token"
```

You can use `usethis::edit_r_environ()` to edit find and open your .Renviron file. You will need to restart R for that variable to be recognized. 

Table 1 describes the functions available in the `dataRetrieval` package.

```{r echo=FALSE}
Functions <- c(
  "read_waterdata",
  "read_waterdata_daily",
  "read_waterdata_continuous",
  "readNWISrating",
  "read_waterdata_field_measurements",
  "readNWISpeak",
  "read_waterdata_field_measurements",
  "readNWISuse",
  "read_waterdata_stats_por, read_waterdata_stats_daterange",
  "read_waterdata_parameter_codes",
  "read_waterdata_monitoring_location",
  "read_waterdata_samples",
  "summarize_waterdata_samples",
  "whatNWISsites",
  "read_waterdata_ts_meta",
  "readWQPdata",
  "readWQPqw",
  "whatWQPsites",
  "whatWQPdata",
  "readWQPsummary",
  "whatWQPmetrics",
  "whatWQPsamples"
)

Description <- c(
  "Time series data using user-specified queries", # read_waterdata
  "Daily values", # read_waterdata_daily
  "Instantaneous values", # read_waterdata_continuous
  "Rating table for active streamgage", # readNWISrating
  "Surface-water measurements", # read_waterdata_field_measurements
  "Peak flow", # readNWISpeak
  "Groundwater levels", # read_waterdata_field_measurements
  "Water use", # readNWISuse
  "Statistical service", # readNWISstat
  "Parameter code information", # read_waterdata_parameter_codes
  "Site information", # read_waterdata_monitoring_location
  "Discrete UGSS water quality data", # read_waterdata_samples
  "Discrete USGS water quality summary",
  "Site search using user-specified queries",
  "Data availability",
  "User-specified queries",
  "Water quality data",
  "Site search",
  "Data availability",
  "Summary data",
  "Metric availability",
  "Sample availability"
)
Source <- c("USGS Water Data API",
            "USGS Water Data API",
            "USGS Water Data API",
            "NWIS",
            "USGS Water Data API",
            "NWIS",
            "USGS Water Data API",
            rep("NWIS",2),
            "USGS Water Data API",
            "USGS Water Data API",
            "USGS Samples Data",
            "USGS Samples Data",
            "NWIS",
            "USGS Water Data API",
            rep("WQP", 7))


data.df <- data.frame(
  Name = Functions,
  `Data Returned` = Description,
  Source, stringsAsFactors = FALSE
)

kable(data.df,
  caption = "Table 1: dataRetrieval functions"
)
```

The arguments `startDate` and `endDate` have defaults to request the maximum data.

# USGS Web Retrievals

In this section we'll show how to get raw data into R. This data includes [site information](#site-information), measured [parameter information](#parameter-information), historical [daily values](#daily-data), [unit values](#unit-data) (which include real-time data but can also include other sensor data stored at regular time intervals), [groundwater level data](#groundwater-level-data), [peak flow data](#peak-flow-data), [rating curve data](#rating-curve-data),  [surface-water measurement data](#surface-water-measurement-data), [water use data](#water-use-data), and [statistics data](#statistics-data). The section [Embedded Metadata](#embedded-metadata) shows instructions for getting metadata that is attached to each returned data frame.

The USGS organizes hydrologic data in a standard structure.  Streamgages are located throughout the United States, and each streamgage has a unique ID (referred in this document and throughout the `dataRetrieval` package as `siteNumber`).  Often (but not always), these ID's are 8 digits for surface-water sites and 15 digits for groundwater sites.  The first step to finding data is discovering this `siteNumber`. 

Once the `siteNumber` is known, the next required input for USGS data retrievals is the "parameter code".  This is a 5-digit code that specifies the measured parameter being requested.  For example, parameter code 00631 represents "Nitrate plus nitrite, water, filtered, milligrams per liter as nitrogen", with units of "mg/l as N". 

Not every station will measure all parameters. A short list of commonly measured parameters is shown in Table 2.

```{r tableParameterCodes, echo=FALSE}


pCode <- c("00060", "00065", "00010", "00045", "00400")
shortName <- c(
  "Discharge [ft<sup>3</sup>/s]",
  "Gage height [ft]",
  "Temperature [C]",
  "Precipitation [in]",
  "pH"
)

data.df <- data.frame(pCode, shortName, stringsAsFactors = FALSE)

kable(data.df,
  caption = "Table 2: Common USGS Parameter Codes"
)
```

Two output columns that may not be obvious are "srsname" and "casrn". Srsname stands for "Substance Registry Services". Casrn stands for "Chemical Abstracts Service (CAS) Registry Number". 

For unit values data (sensor data measured at regular time intervals such as 15 minutes or hourly), knowing the parameter code and `siteNumber` is enough to make a request for data.  For most variables that are measured on a continuous basis, the USGS also stores the historical data as daily values.  These daily values are statistical summaries of the continuous data, e.g. maximum, minimum, mean, or median. The different statistics are specified by a 5-digit statistics code.  

Some common codes are shown in Table 3.

```{r tableStatCodes, echo=FALSE}
StatCode <- c("00001", "00002", "00003", "00008")
shortName <- c("Maximum", "Minimum", "Mean", "Median")

data.df <- data.frame(StatCode, shortName, stringsAsFactors = FALSE)

kable(data.df,
  caption = "Table 3: Commonly used USGS Stat Codes"
)
```

Examples for using these site numbers, parameter codes, and statistic codes will be presented in subsequent sections.


## Site Information


### read_waterdata_monitoring_location

Use the `read_waterdata_monitoring_location` function to obtain all of the information available for a particular USGS site (or sites) such as full station name, drainage area, latitude, and longitude. `read_waterdata_monitoring_location` can also access information about multiple sites with a vector input.


```{r getSite, echo=TRUE, eval=FALSE}
siteNumbers <- c("USGS-01491000", "USGS-01645000")
siteINFO <- read_waterdata_monitoring_location(siteNumbers)
```

Site information is obtained from:
<https://api.waterdata.usgs.gov/ogcapi/v0/collections/monitoring-locations>

### read_waterdata_ts_meta

To discover what time series data is available for a particular USGS site, including measured parameters, period of record, and number of samples (count), use the `read_waterdata_ts_meta` function. 

In the following example, we limit the retrieved data to only daily data. 


```{r getSiteExtended, echo=TRUE, eval=FALSE}
# Continuing from the previous example:
# This pulls out just the daily, mean data:

dailyDataAvailable <- read_waterdata_ts_meta(
  monitoring_location_id = siteNumbers, 
  computation_period_identifier = "Daily",
  statistic_id = "00003"
)
```

```{r echo=FALSE, eval=FALSE}

tableData <- dailyDataAvailable[c("monitoring_location_id",
                         "parameter_description",
                         "unit_of_measure",
                         "begin", "end")]

tableData$begin <- as.Date(tableData$begin)
tableData$end <- as.Date(tableData$end)
tableData <- sf::st_drop_geometry(tableData)


knitr::kable(tableData,
             caption = "Table 4: Reformatted version of output from the read_waterdata_ts_meta function for the Choptank River near Greensboro, MD, and from Seneca Creek at Dawsonville, MD from the daily values service [Some columns deleted for space considerations]")


# nolint end
```

Table 4: Reformatted version of output from the read_waterdata_ts_meta function for the Choptank River near Greensboro, MD, and from Seneca Creek at Dawsonville, MD from the daily values service [Some columns deleted for space considerations]

|monitoring_location_id |parameter_description                                                                                                                               |unit_of_measure |begin      |end        |
|:----------------------|:---------------------------------------------------------------------------------------------------------------------------------------------------|:---------------|:----------|:----------|
|USGS-01491000          |Specific conductance, water, unfiltered, microsiemens per centimeter at 25 degrees Celsius                                                          |uS/cm           |2010-10-01 |2012-05-09 |
|USGS-01491000          |Dissolved oxygen, water, unfiltered, milligrams per liter                                                                                           |mg/l            |2023-04-21 |2025-12-02 |
|USGS-01491000          |Discharge, cubic feet per second                                                                                                                    |ft^3/s          |1948-01-01 |2025-12-02 |
|USGS-01645000          |Discharge, cubic feet per second                                                                                                                    |ft^3/s          |1930-09-26 |2025-12-02 |
|USGS-01491000          |Suspended sediment concentration, milligrams per liter                                                                                              |mg/l            |1980-10-01 |1991-09-29 |
|USGS-01491000          |Suspended sediment discharge, short tons per day                                                                                                    |tons/day        |1980-10-01 |1991-09-29 |
|USGS-01491000          |Nitrate plus nitrite, water, in situ, milligrams per liter as nitrogen                                                                              |mg/l            |2023-08-02 |2025-12-02 |
|USGS-01491000          |Temperature, water, degrees Celsius                                                                                                                 |degC            |2023-04-21 |2025-12-02 |
|USGS-01491000          |Turbidity, water, unfiltered, monochrome near infra-red LED light, 780-900 nm, detection angle 90 +-2.5 degrees, formazin nephelometric units (FNU) |_FNU            |2023-04-21 |2025-12-02 |
|USGS-01491000          |Temperature, water, degrees Celsius                                                                                                                 |degC            |2010-10-01 |2012-05-09 |
|USGS-01491000          |Specific conductance, water, unfiltered, microsiemens per centimeter at 25 degrees Celsius                                                          |uS/cm           |2023-04-21 |2025-12-02 |


## Parameter Information

To obtain all of the available information concerning a measured parameter (or multiple parameters), use the `read_waterdata_parameter_codes` function:

```{r label=getPCodeInfo, echo=TRUE, eval=FALSE}
# Using defaults:
parameterCd <- "00618"
parameterINFO <- read_waterdata_parameter_codes(parameter_code = parameterCd)
```


## Daily Data

To obtain daily records of USGS data, use the `read_waterdata_daily` function. 

```{r label=getNWISDaily, echo=TRUE, eval=FALSE}

# Choptank River near Greensboro, MD:
siteNumber <- "USSG-01491000"
parameterCd <- "00060" # Discharge
startDate <- "2009-10-01"
endDate <- "2012-09-30"

discharge <- read_waterdata_daily(monitoring_location_id = siteNumber,
                                  parameter_code = parameterCd, 
                                  time = c(startDate, endDate))
```

The column "time" in the returned data frame is automatically imported as a variable of class "Date" in R. 

Another example would be a request for mean and maximum daily temperature and discharge in early 2012:

```{r label=getNWIStemperature, echo=TRUE, eval=FALSE}
siteNumber <- "USGS-01491000"
parameterCd <- c("00010", "00060") # Temperature and discharge
statCd <- c("00001", "00003") # Mean and maximum
startDate <- "2012-01-01"
endDate <- "2012-05-01"

temperatureAndFlow <- read_waterdata_daily(monitoring_location_id = siteNumber,
                                  parameter_code = parameterCd, 
                                  statistic_id = statCd,
                                  time = c(startDate, endDate))
```

```{r label=getNWIStemperature2, echo=FALSE, eval=TRUE}
filePath <- system.file("extdata", package = "dataRetrieval")
fileName <- "temperatureAndFlow.RData"
fullPath <- file.path(filePath, fileName)
load(fullPath)
```


An example of plotting the above data:

```{r}

temperature <- temperatureAndFlow[temperatureAndFlow$parameter_code == "00010",]
temperature <- temperature[temperature$statistic_id == "00001",]

flow <- temperatureAndFlow[temperatureAndFlow$parameter_code == "00060",]

par(mar = c(5, 5, 5, 5)) # sets the size of the plot window

plot(temperature$time, temperature$value,
  ylab = "Maximum Temperture [C]",
  xlab = ""
)
par(new = TRUE)
plot(flow$time,
  flow$value,
  col = "red", type = "l",
  xaxt = "n", yaxt = "n",
  xlab = "", ylab = "",
  axes = FALSE
)
axis(4, col = "red", col.axis = "red")
mtext("Discharge [ft3/s]", side = 4, line = 3, col = "red")
title("CHOPTANK RIVER NEAR GREENSBORO, MD")
legend("topleft", unique(temperatureAndFlow$unit_of_measure),
  col = c("black", "red"), lty = c(NA, 1),
  pch = c(1, NA)
)
```


## Unit Data

Any data collected at regular time intervals (such as 15-minute or hourly) are known as "unit values". Many of these are delivered on a real time basis and very recent data (even less than an hour old in many cases) are available through the function `read_waterdata_continuous`.  Here is an example of a retrieval of such data.  

```{r label=readNWISuv, eval=FALSE}

parameterCd <- "00060" # Discharge
startDate <- "2012-05-12"
endDate <- "2012-05-13"
dischargeUnit <- read_waterdata_continuous(monitoring_location_id = siteNumber,
                                           parameter_code = parameterCd,
                                           time = c(startDate, endDate))

```

The retrieval produces a data frame that contains 96 rows (one for every 15 minute period in the day).  They include all data collected from the `startDate` through the `endDate` (starting and ending with midnight locally-collected time). The dateTime column is converted to UTC (Coordinated Universal Time), so midnight EST will be 5 hours earlier in the dateTime column (the previous day, at 7pm).


## Groundwater Level Data

Groundwater level measurements can be obtained with the `read_waterdata_field_measurements` function. 

```{r gwlexample, echo=TRUE, eval=FALSE}
siteNumber <- "USGS-434400121275801"
groundWater <- read_waterdata_field_measurements(monitoring_location_id = siteNumber)
```


## Peak Flow Data

Peak flow data are instantaneous discharge or stage data that record the maximum values of these variables during a flood event.  They include the annual peak flood event but can also include records of other peaks that are lower than the annual maximum. Peak discharge measurements can be obtained with the `readNWISpeak` function. Information on the returned data can be found with the `comment` function and attached attributes as described in the [metadata](#embedded-metadata) section.

```{r peakexample, echo=TRUE, eval=FALSE}
siteNumber <- "01594440"
peakData <- readNWISpeak(siteNumber)
```


## Rating Curve Data

Rating curves are the calibration curves that are used to convert measurements of stage to discharge.  Because of changing hydrologic conditions these rating curves change over time. Information on the returned data can be found with the `comment` function and attached attributes as described in the [metadata](#embedded-metadata) section.

Rating curves can be obtained with the `readNWISrating` function.

```{r ratingexample, echo=TRUE, eval=FALSE}
ratingData <- readNWISrating(siteNumber, "base")
attr(ratingData, "RATING")
```


## Surface-Water Measurement Data

These data are the discrete measurements of discharge that are made for the purpose of developing or revising the rating curve.  Surface-water measurement data can be obtained with the `read_waterdata_field_measurements` function.

```{r surfexample, echo=TRUE, eval=FALSE}
surfaceData <- read_waterdata_field_measurements(monitoring_location_id = "USGS-01594440")
```


## Water Use Data

Legacy water use data services have been retired. Check back in for developments on replacement functions.


## Statistics Data
Retrieves site statistics from the [USGS Water Data Statistics API](https://api.waterdata.usgs.gov/statistics/v0/docs). 
The `read_waterdata_stats_por` function retieves day-of-year and month-of-year daily data statistics from the `observationNormals` endpoint.
The `read_waterdata_stats_daterange` function retieves calendar month, calendar year, and water year daily data statistics from the `observationIntervals` endpoint.

```{r eval=FALSE}
discharge_stats_por <- read_waterdata_stats_por(
  monitoring_location_id = "USGS-05428500",
  parameter_code = "00060"
)

discharge_stats_daterange <- read_waterdata_stats_daterange(
  monitoring_location_id = "USGS-05428500",
  parameter_code = "00060"
)
```


# Water Quality Portal Web Retrievals

Water quality data sets available from the [Water Quality Data Portal](https://www.waterqualitydata.us/).  These data sets can be housed in either the STORET database (data from EPA) and NWIS database (data from USGS).  Because only USGS uses parameter codes, a "characteristic name" must be supplied.  The `readWQPqw` function can take either a USGS parameter code, or a more general characteristic name in the parameterCd input argument. The Water Quality Data Portal includes data discovery tools and information on characteristic names. The following example retrieves specific conductance from a DNR site in Wisconsin. 


```{r label=getQWData, echo=TRUE, eval=FALSE}
specificCond <- readWQPqw(
  "WIDNR_WQX-10032762",
  "Specific conductance",
  "2011-05-01", "2011-09-30"
)
```


## WQP 

Just as with NWIS, the Water Quality Portal (WQP) offers a variety of ways to search for sites and request data. The possible Web service arguments for WQP site searches is found [here](https://www.waterqualitydata.us/webservices_documentation).

### Sites: whatWQPsites

To discover available sites in the WQP in New Jersey that have measured Chloride, use the function `whatWQPsites`.

```{r NJChloride, eval=FALSE}

sitesNJ <- whatWQPsites(
  statecode = "US:34",
  characteristicName = "Chloride"
)
```


### Data: readWQPdata

To get data from the WQP using generalized Web service calls, use the function `readWQPdata`. For example, to get all the pH data in Wisconsin:

```{r phData, eval=FALSE}
dataPH <- readWQPdata(
  statecode = "US:55",
  characteristicName = "pH"
)
```


### Availability: whatWQPdata

The function `whatWQPdata` returns a data frame with information on the amount of data collected at a site. For example:

```{r eval=FALSE}
type <- "Stream"
sites <- whatWQPdata(countycode = "US:55:025", siteType = type)
```

This returns a data frame with all of the sites that were measured in streams in Dane County, WI. Also, in that table, there is a measure of `activityCount` (how often the site was sampled), and `resultCount` (how many individual results are available).

### Samples: whatWQPsamples

The function `whatWQPsamples` returns information on the individual samples collected at a site. For example:

```{r eval=FALSE}
site <- whatWQPsamples(siteid = "USGS-01594440")
```

This returns one row for each instance that a sample was collect.

### Metrics: whatWQPmetrics

The function `whatWQPmetrics` provides metric information. This is only currently available for STORET data:

```{r eval=FALSE}
type <- "Stream"
sites <- whatWQPmetrics(countycode = "US:55:025", siteType = type)
```


# Getting Started in R

This section describes the options for downloading and installing the `dataRetrieval` package.

## New to R?
 
If you are new to R, you will need to first install the latest version of R, which can be found [here] (www.R-project.org).

At any time, you can get information about any function in R by typing a question mark before the functions name.  This will open a file (in RStudio, in the Help window) that describes the function, the required arguments, and provides working examples. This will open a help file similar to the image below. To see the raw code for a particular code, type the name of the function, without parentheses.


```
?read_waterdata_parameter_codes
```


![A simple R help file](Rhelp.png)


Additionally, many R packages have vignette files attached (such as this paper). To see the list of vignettes in a package:
```{r seeVignette,eval = FALSE}
# to see all available vignettes:
vignette(package="dataRetrieval")

#to open a specific vignette:
vignette(topic = "qwdata_changes", package = "dataRetrieval")
```


# Citations

## Citing the dataRetrieval package

```{r cite, eval=TRUE}
citation(package = "dataRetrieval")
```

## Citing NWIS data

The requested format is: 

U.S. Geological Survey, 2023, National Water Information System data available on the World Wide Web (USGS Water Data for the Nation), accessed [April 26, 2023], at https://dx.doi.org/10.5066/F7P55KJN

This can be created from the attributes of the data using the `create_NWIS_bib` function:


```{r nwisCite, eval=FALSE}

dv <- read_waterdata_daily(monitoring_location_id = "USGS-04085427",
                           parameter_code = "00060",
                           time = c("2012-01-01", "2012-06-30"))

NWIScitation <- create_NWIS_bib(dv)
NWIScitation
```

```
U.S. Geological Survey (2026). _USGS Water Data for the Nation: U.S. Geological Survey
National Water Information System database_. doi:10.5066/F7P55KJN
<https://doi.org/10.5066/F7P55KJN>, Accessed Mar 09, 2026.
```
Alternative output is available:

```{r show1, eval=FALSE}
print(NWIScitation, style = "Bibtex")
```

```
@Manual{,
  title = {USGS Water Data for the Nation: U.S. Geological Survey National Water Information System database},
  author = {{U.S. Geological Survey}},
  doi = {10.5066/F7P55KJN},
  note = {Accessed Mar 09, 2026},
  year = {2026},
}
```

```{r show2, eval=FALSE}
print(NWIScitation, style = "citation")
```

```
U.S. Geological Survey, 2024, National Water Information System
data available on the World Wide Web (USGS Water Data for the
Nation), accessed Feb 09, 2024, at
https://waterservices.usgs.gov/nwis/dv/?site=09010500&format=waterml,1.1&ParameterCd=00060&StatCd=00003&startDT=1851-01-01,
https://dx.doi.org/10.5066/F7P55KJN

A BibTeX entry for LaTeX users is

  @Manual{,
    title = {National Water Information System data available on the World Wide Web (USGS Water Data for the Nation)},
    author = {{U.S. Geological Survey}},
    doi = {10.5066/F7P55KJN},
    note = {Accessed Feb 09, 2024},
    year = {2024},
    url = {https://waterservices.usgs.gov/nwis/dv/?site=09010500&format=waterml,1.1&ParameterCd=00060&StatCd=00003&startDT=1851-01-01},
  }
```



## Citing WQP data

Citations for specific datasets should use this format:

National Water Quality Monitoring Council, YYYY, Water Quality Portal, accessed mm, dd, yyyy, hyperlink_for_query, https://doi.org/10.5066/P9QRKUVJ.

This can be obtained using the `create_WQP_bib` function:

```{r WQPcite, eval=FALSE}
SC <- readWQPqw(siteNumbers = "USGS-05288705",
                parameterCd = "00300")

WQPcitation <- create_WQP_bib(SC)
WQPcitation
```

```
National Water Quality Monitoring Council (2024). _ Water
Quality Portal_. doi:10.5066/P9QRKUVJ
<https://doi.org/10.5066/P9QRKUVJ>, Accessed Feb 09, 2024,
<https://www.waterqualitydata.us/data/Result/search?siteid=USGS-05288705&pCode=00300&mimeType=tsv&zip=yes>.
```

```{r show3, eval=FALSE}
print(WQPcitation, style = "Bibtex")
```

```
@Manual{,
  title = { Water Quality Portal},
  author = {{National Water Quality Monitoring Council}},
  doi = {10.5066/P9QRKUVJ},
  note = {Accessed Feb 09, 2024},
  year = {2024},
  url = {https://www.waterqualitydata.us/data/Result/search?siteid=USGS-05288705&pCode=00300&mimeType=tsv&zip=yes},
}
```

```{r show4, eval=FALSE}
print(WQPcitation, style = "citation")
```

```
National Water Quality Monitoring Council, 2024, Water Quality
Portal, accessed 02, 09, 2024,
https://www.waterqualitydata.us/data/Result/search?siteid=USGS-05288705&pCode=00300&mimeType=tsv&zip=yes,
https://doi.org/10.5066/P9QRKUVJ.

A BibTeX entry for LaTeX users is

  @Manual{,
    title = { Water Quality Portal},
    author = {{National Water Quality Monitoring Council}},
    doi = {10.5066/P9QRKUVJ},
    note = {Accessed Feb 09, 2024},
    year = {2024},
    url = {https://www.waterqualitydata.us/data/Result/search?siteid=USGS-05288705&pCode=00300&mimeType=tsv&zip=yes},
  }
```

## Citing Water Quality Portal itself

General Water Quality Portal citations should use the following:

Water Quality Portal. Washington (DC): National Water Quality Monitoring Council, United States Geological Survey (USGS), Environmental Protection Agency (EPA); 2021. https://doi.org/10.5066/P9QRKUVJ.


# Disclaimer
This information is preliminary and is subject to revision. It is being provided to meet the need for timely best science. The information is provided on the condition that neither the U.S. Geological Survey nor the U.S. Government may be held liable for any damages resulting from the authorized or unauthorized use of the information.

---
title: "Changes to NWIS QW services"
author: Laura A. DeCicco
editor_options: 
  chunk_output_type: console
output:
  rmarkdown::html_vignette:
    toc: true
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Changes to NWIS QW services}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---


```{r setup, include=FALSE, message=FALSE}
library(knitr)
library(dataRetrieval)
library(dplyr)

options(continue = " ")

knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  fig.height = 7,
  fig.width = 7
)
```

**IMPORTANT** These recommendations have been updated using WQX3.0 profiles. 

# Changes to NWIS water quality services

USGS discrete water samples data are undergoing modernization, and NWIS services will no longer be updated with the latest data starting mid-March 2024, with a full decommission expected 6 months later.

For the latest news on USGS water quality data, see:
[https://doi-usgs.github.io/dataRetrieval/articles/Status.html](https://doi-usgs.github.io/dataRetrieval/articles/Status.html). Learn more about the changes and where to find the new samples data in the [WDFN blog](https://waterdata.usgs.gov/blog/changes-to-sample-data/).

What does this mean for `dataRetrieval` users? Eventually water quality data will ONLY be available from the [Water Quality Portal](https://www.waterqualitydata.us/) rather than the NWIS services. There are 3 major `dataRetrieval` functions that will be affected: `readNWISqw`, `whatNWISdata`, and `readNWISdata`. This vignette will describe the most common workflows conversions needed to update existing scripts. 

# How to find more help

This vignette is being provided in advance of any breaking changes, and more information and guidance will be provided. These changes are big, and initially sound overwhelming. But in the end, they are thoughtful changes that will make understanding USGS water data more intuitive. Please reach out for questions and comments to: gs-w-IOW_PO_team@usgs.gov

# WQX2 -> WQX3

If you have already converted your workflow from NWIS to WQP, much of the hard work has been done! The final step of the process is to make sure workflows are using the most modern WQX3 formats. WQX stands for Water Quality Exchange. WQX2 is the format that has been historically available on the Water Quality Portal, WQX3 is the more modern format that all the data is being converted to.


This is available as of `dataRetrieval` version 2.7.16. Starting with 2.7.16, all WQP functions default to the newer "WQX3" profiles if available. See  [WQX Conversions](#conversion) below to get a table of column names in WQX3 vs WQX2.

# NWIS -> WQX3

## `readNWISqw`

This function was retired as of Oct. 24, 2024.

So...what do you use instead? The function you will need to move to is `readWQPqw`. First, you'll need to convert the numeric USGS site ID's into something that the Water Quality Portal will accept, which requires the agency prefix. For most USGS sites this will mean pasting 'USGS-' before the site number, although it is important to note that there are some USGS sites that begin with a different prefix: it is up to users to determine the agency code..

Here's an example:

```{r eval=FALSE}
wqpData <- readWQPqw(paste0("USGS-", site_ids), parameterCd)
```

Let's say we have a data frame that we got from the retired `readNWISqw` function and we saved it as `nwisData`.
```{r echo=FALSE}
nwisData <- readRDS("nwisData.rds")
wqpData <- readRDS("wqpData.rds")
```

First we compare the number of rows, number of columns, and attributes to each return:


```{r}
nrow(nwisData)
nrow(wqpData)
```

So, same number of rows returned. That's good, since it's the same data.

```{r}
ncol(nwisData)
ncol(wqpData)
```

Different columns!

```{r}
names(attributes(nwisData))
names(attributes(wqpData))
```

Slightly different attributes. You can explore the differences of those attributes:

```{r}
site_NWIS <- attr(nwisData, "siteInfo")
site_WQP <- attr(wqpData, "siteInfo")
```

The next big task is figuring out which columns from the WQP output map to the original columns from the NWIS output. Look at your workflow and determine what columns from the original NWIS output are needed to preserve the integrity of the workflow. 

Let's use the `dplyr` package to pull out the columns are used in this example workflow, and make sure both NWIS and WQP are ordered in the same way.

```{r}
library(dplyr)

nwisData_relevant <- nwisData |> 
  select(
    site_no, startDateTime, parm_cd,
    remark_cd, result_va
  ) |> 
  arrange(startDateTime, parm_cd)

knitr::kable(head(nwisData_relevant))
```

If we explore the output from WQP, we can try to find the columns that include the same relevant information:
```{r}
wqpData_relevant <- wqpData |> 
  select(
    site_no = Location_Identifier,
    startDateTime = Activity_StartDateTime,
    parm_cd = USGSpcode,
    remark_cd = Result_ResultDetectionCondition,
    result_va = Result_Measure,
    detection_level = DetectionLimit_MeasureA
  ) |> 
  arrange(startDateTime, parm_cd)
knitr::kable(head(wqpData_relevant))
```

Now we can start looking at the results and trying to decide how future workflows should be setup. Here are some decisions for this example that we can consider:

### Censored values

The result_va in the NWIS service came back with a value. However, the data is actually censored, meaning we only know it's below the detection limit. With some lazier coding, it might have been really easy to not realize these values are left-censored. So, while we could substitute the detection levels into the measured values if there's an `NA` in the measured value, this might be a great time to update your workflow to handle censored values more robustly. We are probably interested in maintaining the detection level in another column. 

For this theoretical workflow, let's think about what we are trying to find out. Let's say that we want to know if a value is "left-censored" or not. Maybe in this case, what would make the most sense is to have a column that is a logical TRUE/FALSE. For this example, there was only the text "Not Detected" in the "ResultDetectionConditionText" column PLEASE NOTE that other data may include different messages about detection conditions, you will need to examine your data carefully. Here's an example from the `EGRET` package on how to decide if a "ResultDetectionConditionText" should be considered a censored value:


```{r}

censored_text <- c(
  "Not Detected",
  "Non-Detect",
  "Non Detect",
  "Detected Not Quantified",
  "Below Quantification Limit"
)

wqpData_relevant <- wqpData |> 
  mutate(left_censored = grepl(paste(censored_text, collapse = "|"),
    Result_ResultDetectionCondition,
    ignore.case = TRUE
  )) |> 
  select(
    site_no = Location_Identifier,
    startDateTime = Activity_StartDateTime,
    parm_cd = USGSpcode,
    left_censored,
    result_va = Result_Measure,
    detection_level = DetectionLimit_MeasureA,
    dl_units = DetectionLimit_MeasureUnitA
  ) |> 
  arrange(startDateTime, parm_cd)

knitr::kable(head(wqpData_relevant))
```

### NWIS codes

Another difference that is going to require some thoughtful decisions is how to interpret additional NWIS codes. They will now be descriptive text. Columns such as samp_type_cd and medium_cd will now all be reported with descriptive words rather than single letters or numbers. It will be the responsibility of the user to consider the best way to deal with these changes. 

If you use the `readNWISqw` function, you WILL need to adjust your workflows, and you may find there are more codes you will need to account for. Hopefully this section helped get you started. It does not include every scenario, so you may find more columns or codes or other conditions you need to account for.  

# whatNWISdata

This function will continue to work for any service EXCEPT "qw" (water quality discrete data). "qw" results will eventually no longer be returned, and are currently showing values that were frozen in March 2024. 

The function to replace this functionality for discrete water quality data is currently `whatWQPdata`:

```{r whatdata, eval=FALSE}
whatNWIS <- whatNWISdata(
  siteNumber = site_ids,
  service = "qw"
)
```

```
WARNING: NWIS does not deliver
new discrete water quality data or updates to existing data. 
For additional details, see:
https://doi-usgs.github.io/dataRetrieval/articles/Status.html
```

```{r whatdatanew, eval=FALSE}
whatWQP <- whatWQPdata(siteNumber = paste0("USGS-", site_ids))
```

There are some major differences in the output. The NWIS services offers back one row per site/parameter code to learn how many samples are available. This is not *currently* available from the Water Quality Portal, however there are new summary services being developed. When those become available, we will include new documentation on how to get this information.

# readNWISdata

If you get your water quality data from the `readNWISdata` function, no new data will be available and the function will generate a warning message. The other services are working as before. This is not an especially common `dataRetrieval` workflow, so there are not a lot of details here. Please reach out if more information is needed to update your workflows.

See `?readWQPdata` to see all the ways to query data in the Water Quality Portal. Use the suggestions above to convert the output of the `readWQPdata` function to convert the WQP output to what is important for your workflow.

# WQX Conversion {#conversion}

A table is provided on the EPA website that shows the conversions of WQX3 to the WQX2 mappings. This table may change periodically while WQP services are under active development, which is expected to last through Fall 2024.


Here is an example for pulling the EPA table and comparing column names. 

```{r echo=TRUE, eval=TRUE}
schema <- readr::read_csv("https://www.epa.gov/system/files/other-files/2024-07/schema_outbound_wqx3.0.csv")
```

Within that schema table, the column "FieldName3.0" shows the column names for the new WQX3 profile. There are several "FieldName2.0.XXXX" columns that show how the older 2.0 profiles align with the newer columns.

For example, the 2.0 "narrow" dataProfile match up with these new WQP3 columns:
```{r echo=TRUE, eval=TRUE}
sub_schema <- schema |> 
  select(WQX3 = FieldName3.0,
         WQX2 = FieldName2.0.Narrow) |> 
  filter(!is.na(WQX2))

knitr::kable(sub_schema)

```

# Disclaimer

This information is preliminary and is subject to revision. It is being provided to meet the need for timely best science. The information is provided on the condition that neither the U.S. Geological Survey nor the U.S. Government may be held liable for any damages resulting from the authorized or unauthorized use of the information.

# Contributing Guidelines

Contributions to `dataretrieval` are welcome and greatly appreciated, but
please read this document for information on *how* to contribute.

`dataretrieval` broadly follows a ["forking" workflow](https://docs.github.com/en/get-started/quickstart/contributing-to-projects),
however writing code is not the only way to contribute.

---

## Table of Contents

- [Contributing Guidelines](#contributing-guidelines)
  - [Table of Contents](#table-of-contents)
  - [Bugs](#bugs)
    - [Reporting Bugs](#reporting-bugs)
    - [Fixing Bugs](#fixing-bugs)
  - [Code Contributions](#code-contributions)
    - [Implementing Features](#implementing-features)
    - [Pull Request Guidelines](#pull-request-guidelines)
    - [Coding Standards and Style](#coding-standards-and-style)
      - [Style](#style)
      - [Docstrings](#docstrings)
      - [Quotes](#quotes)
  - [Documentation](#documentation)
    - [Contributing to the Documentation](#contributing-to-the-documentation)
    - [Adding Examples to the Documentation](#adding-examples-to-the-documentation)
  - [Feedback and Feature Requests](#feedback-and-feature-requests)
    - [Submitting Feedback](#submitting-feedback)
    - [Feature Requests](#feature-requests)
  - [Acknowledgements](#acknowledgements)

---

## Bugs

### Reporting Bugs

Report bugs at https://github.com/USGS-python/dataretrieval/issues.

When reporting a bug, please include:

* Detailed steps to reproduce the bug
* Your operating system name and version
* The Python version, as well as information about your local Python
  environment, such as the versions of installed packages
* Any additional details about your local setup that might be helpful in
  troubleshooting

### Fixing Bugs

Look through the GitHub [issues](https://github.com/USGS-python/dataretrieval/issues)
for known and unresolved bugs. Any issues labeled "bug" that are unassigned,
are open for resolution. You are welcome to comment in the relevant issue to
state your intention to resolve the bug, which will help ensure there is no
duplication of the same work by multiple contributors.

---

## Code Contributions

Code contributions should be made following a ["forking" workflow](https://docs.github.com/en/get-started/quickstart/contributing-to-projects).
This means that first, one should *fork* the repository, essentially creating a
personal mirror of the project. Next, you will want to create a *feature branch*
in your fork, which you can push code changes to. Once you have completed your
modifications and additions, open a pull request from the *feature branch* in
your fork, to the original upstream repository.

### Implementing Features

Look through the GitHub [issues](https://github.com/USGS-python/dataretrieval/issues)
for outstanding feature requests. Anything tagged with "enhancement"
and "please-help" is open to whomever wants to implement it.

Please do not combine multiple feature enhancements into a single pull request.

### Pull Request Guidelines

Before you submit a pull request, check that it meets these guidelines:

1. If the pull request adds or modifies package functionality, unit tests
   should be written to test the new functionality
2. If the pull request adds or modifies functionality, the documentation should
   be updated. To do so, either add or modify a functions docstring which will
   automatically become part of the API documentation
3. The pull request should work for Python 3.7, 3.8, 3.9, 3.10 - refer to the
   [python-package.yml file](https://github.com/USGS-python/dataretrieval/blob/master/.github/workflows/python-package.yml)
   for the latest versions of Python being tested by the continuous integration
   pipelines. This will be checked automatically by the CI pipelines once the
   pull request is opened.

### Coding Standards and Style

Note that coding standards and style as described below are strong suggestions,
the `dataretrieval` project does not strictly lint or enforce style guidelines
via any automated processes or pipelines.

#### Style

* Attempt to write code following the [PEP8 style guidelines](https://peps.python.org/pep-0008/) as much as possible
* The public interace should emphasize functions over classes; however, classes
  can and should be used internally and in tests
* Functions for downloading data from a specific web portal must be grouped
  within their own submodule
  * For example, all NWIS functions are located at `dataretrieval.nwis`

#### Docstrings
* Docstrings should follow the [numpy standard](https://numpydoc.readthedocs.io/en/v1.5.0/format.html):
  * Example:
    ``` python
    def foo(param1, param2):
    """Example function with types documented in the docstring.

    A more detailed description of the function and its implementation.

    Parameters
    ----------
    param1 : int
        The first parameter.
    param2 : str
        The second parameter.

    Returns
    -------
    bool
        True if successful, False otherwise.

    Examples
    --------
    Examples should be written in doctest format and should demonstrate basic usage.

    >>> foo(1,'bar')
    True

    """
    ```
  * For more details see https://github.com/sphinx-doc/sphinx/blob/master/doc/ext/example_numpy.py

#### Quotes

* Quote conventions follow http://stackoverflow.com/a/56190/5549:

  * Use double quotes around strings that are used for interpolation or that
    are natural language messages
  * Use single quotes for small symbol-like strings (but break the rules if
    the strings contain quotes)
  * Use triple double quotes for docstrings and raw string literals for
    regular expressions even if they aren't needed
  * Example:

    ``` python

        LIGHT_MESSAGES = {
            'English': "There are %(number_of_lights)s lights.",
            'Pirate':  "Arr! Thar be %(number_of_lights)s lights."
        }

        def lights_message(language, number_of_lights):
            """Return a language-appropriate string reporting the light count."""
            return LIGHT_MESSAGES[language] % locals()

        def is_pirate(message):
            """Return True if the given message sounds piratical."""
            return re.search(r"(?i)(arr|avast|yohoho)!", message) is not None
    ```

---

## Documentation

### Contributing to the Documentation

Documentation is built using [sphinx](https://www.sphinx-doc.org/en/master/),
and is located within the `docs/source/` subdirectory in the repository.
Documentation is written using [reStructuredText](https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html).

Contributions to the documentation should be made in a similar fashion to code
contributions - by following a forking workflow. When opening a pull request
please be sure to have tested your documentation modifications locally, and
clearly describe what it is your proposed changes add or fix.

### Adding Examples to the Documentation

A number of examples are provided in the documentation in the form of Jupyter
notebooks. These example notebooks are all contained within the `demos/`
subdirectory of the repository. If you have an example use of the package you
would like to add to the documentation as a run and rendered notebook, you
will need to do the following (in a separate branch of the repository):

1. Add your notebook to the `demos/` subdirectory after clearing all outputs
2. Add a corresponding `.nblink` file to `docs/source/examples/` subdirectory,
   see existing examples for reference, or refer to the [nbsphinx-link](https://nbsphinx-link.readthedocs.io/en/latest/) documentation.
3. Add the example and some text describing it to one of the `.rst` files in
   the examples subdirectory.
4. Run the documentation locally to ensure it renders as you expect, and then
   open a pull request wherein you describe the proposed addition.

---

## Feedback and Feature Requests

### Submitting Feedback

The best way to send feedback is to open an issue at
https://github.com/USGS-python/dataretrieval/issues.

Please be as clear as possible in your feedback, if you are reporting a bug
refer to [Reporting Bugs](#reporting-bugs).


### Feature Requests

To request or propose a new feature, open an issue at
https://github.com/USGS-python/dataretrieval/issues.

Please be sure to:
* Explain in detail how it would work, possibly with pseudo-code or an example
  use-case
* Keep the scope of the proposed feature as narrow as possible
* Make clear whether you would like to implement this feature, you need help
  devising the implementation, or you are unable to implement the feature
  yourself but would like it as a user

---

## Acknowledgements
This document was adapted from the cookiecutter project's [CONTRIBUTING file](https://github.com/audreyr/cookiecutter/blob/master/CONTRIBUTING.rst).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pcode_to_name.R
\name{pcode_to_name}
\alias{pcode_to_name}
\title{Parameter code to characteristic name}
\usage{
pcode_to_name(parameterCd = "all")
}
\arguments{
\item{parameterCd}{character that contains the code for a character vector
of 5-digit parameter codes. Default is "all" which will return a complete
list of parameter codes that have been mapped to a characteristic name.}
}
\value{
a data frame with columns  "parm_cd", "description",
"characteristicname", "measureunitcode", "resultsamplefraction",
"resulttemperaturebasis", "resultstatisticalbasis",  "resulttimebasis",
"resultweightbasis", "resultparticlesizebasis", "last_rev_dt"
}
\description{
This function is useful to fine what characteristic name, result sample
fraction, unit code, and other parameters are mapped with USGS parameter
codes. This information is useful for converting workflows from a more
traditional NWIS water quality retrieval to a Water Quality Portal retrieval.
}
\examples{
pcodes <- c("00070", "00075", "00430", "52642")
\donttest{

all <- pcode_to_name()
some <- pcode_to_name(pcodes)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructNWISURL.R
\name{constructWQPURL}
\alias{constructWQPURL}
\title{Construct WQP url for data retrieval}
\usage{
constructWQPURL(siteNumbers, parameterCd, startDate, endDate, legacy = TRUE)
}
\arguments{
\item{siteNumbers}{string or vector of strings USGS site number.}

\item{parameterCd}{string or vector of USGS parameter code.  This is usually an 5 digit number.}

\item{startDate}{character starting date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the earliest possible record.}

\item{endDate}{character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the latest possible record.}

\item{legacy}{Logical. If TRUE, uses legacy WQP services. Default is TRUE.
Setting legacy = FALSE uses WQX3.0 WQP services, which are in-development, use with caution.}
}
\value{
url string
}
\description{
Construct WQP url for data retrieval. This function gets the data from here: \url{https://www.waterqualitydata.us}
}
\examples{
site_ids <- c("USGS-02292010", "USGS-02276877")
startDate <- "2020-01-01"
endDate <- ""
pCode <- c("80154", "00613")
url_wqp <- constructWQPURL(
  site_ids,
  pCode,
  startDate, endDate
)
url_wqp
charNames <- c(
  "Temperature",
  "Temperature, sample",
  "Temperature, water",
  "Temperature, water, deg F"
)
obs_url_orig <- constructWQPURL(
  siteNumbers = c(
    "IIDFG-41WSSPAHS",
    "USGS-02352560"
  ),
  parameterCd = charNames,
  startDate, ""
)
obs_url_orig
}
\keyword{WQP}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ogc_data.R
\name{check_arguments_non_api}
\alias{check_arguments_non_api}
\title{Check non-API arguments}
\usage{
check_arguments_non_api(
  convertType,
  no_paging,
  limit,
  attach_request,
  chunk_size,
  ...
)
}
\arguments{
\item{convertType}{logical, defaults to TRUE.
If \code{TRUE}, the function will convert the data to dates, any qualifiers to string
vector and reorder the returned data frame.}

\item{no_paging}{logical, defaults to FALSE.
If \code{TRUE}, the data will
be requested from a native csv format. This can be dangerous because the
data will cut off at 50,000 rows without indication that more data
is available. Use \code{TRUE} with caution.}

\item{limit}{numeric, The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50,000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}

\item{chunk_size}{Number of monitoring_location_ids to chunk requests into.
The default for functions that don't generally return long-term data records
is 250, while
the default for time series functions is
10.
Setting to \code{NA} will eliminate site chunking, giving users full control.}

\item{\dots}{Not used. Included to help differentiate official Water Data API arguments
from more seldom used, optional dataRetrieval-specific arguments.}
}
\description{
Function to check types and create parameter descriptions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRetrieval-package.R
\docType{data}
\name{stateCd}
\alias{stateCd}
\title{US State Code Lookup Table}
\value{
stateCd data frame.

\tabular{lll}{
Name \tab Type \tab Description\cr
STATE \tab character \tab FIPS State Code  \cr
STUSAB \tab character \tab Official United States Postal Service (USPS) Code \cr
STATE_NAME \tab character \tab State Name \cr
STATENS \tab character \tab  Geographic Names Information System Identifier (GNISID)\cr
}
}
\description{
Classic lookup table for states. Has been replaced in functions with
\code{check_waterdata_sample_params("states")}.
}
\examples{
head(stateCd)
}
\keyword{USGS}
\keyword{stateCd}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_samples.R
\name{check_waterdata_sample_params}
\alias{check_waterdata_sample_params}
\title{Check values from codeservice}
\usage{
check_waterdata_sample_params(service = "characteristicgroup", ...)
}
\arguments{
\item{service}{Options are: "characteristicgroup", "states", "counties",
"countries", "sitetype", "samplemedia", "characteristics", "observedproperty"}

\item{...}{Optional additional query arguments. Only "characteristics" and
"observedproperty" have additional parameters options.}
}
\value{
List, structure depends on service.
}
\description{
Call a service to check on values from:
\url{https://api.waterdata.usgs.gov/samples-data/codeservice/docs}.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
groups <- check_waterdata_sample_params("characteristicgroup")
states <- check_waterdata_sample_params("states")
countries <- check_waterdata_sample_params("countries")
counties <- check_waterdata_sample_params("counties")
sitetypes <- check_waterdata_sample_params("sitetype")
samplemedia <- check_waterdata_sample_params("samplemedia")
characteristics <- check_waterdata_sample_params("characteristics",
                               group = "Biological")
observedProperties <- check_waterdata_sample_params("observedproperty",
                                  text = "phosphorus")
ref_list <- check_waterdata_sample_params("reference-list")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_ts_meta.R
\name{read_waterdata_ts_meta}
\alias{read_waterdata_ts_meta}
\title{Get USGS Time Series Metadata}
\usage{
read_waterdata_ts_meta(
  monitoring_location_id = NA_character_,
  parameter_code = NA_character_,
  parameter_name = NA_character_,
  properties = NA_character_,
  statistic_id = NA_character_,
  last_modified = NA_character_,
  begin_utc = NA_character_,
  end_utc = NA_character_,
  hydrologic_unit_code = NA_character_,
  state_name = NA_character_,
  unit_of_measure = NA_character_,
  computation_period_identifier = NA_character_,
  computation_identifier = NA_character_,
  thresholds = NA,
  sublocation_identifier = NA_character_,
  primary = NA_character_,
  parent_time_series_id = NA_character_,
  time_series_id = NA_character_,
  web_description = NA_character_,
  skipGeometry = NA,
  bbox = NA,
  begin = NA_character_,
  end = NA_character_,
  ...,
  limit = getOption("dataRetrieval.limit"),
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  attach_request = getOption("dataRetrieval.attach_request")
)
}
\arguments{
\item{monitoring_location_id}{A unique identifier representing a single monitoring location. This corresponds to the \code{id} field in the \code{monitoring-locations} endpoint. Monitoring location IDs are created by combining the agency code of the agency responsible for the monitoring location (e.g. USGS) with the ID number of the monitoring location (e.g. 02238500), separated by a hyphen (e.g. USGS-02238500).

Multiple monitoring_location_ids can be requested as a character vector.}

\item{parameter_code}{Parameter codes are 5-digit codes used to identify the constituent measured and the units of measure. A complete list of parameter codes and associated groupings can be found at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/parameter-codes/items}.

Multiple parameter_codes can be requested as a character vector.}

\item{parameter_name}{A human-understandable name corresponding to \code{parameter_code}.}

\item{properties}{A vector of requested columns to be returned from the query.
Available options are:
geometry, time_series_id, unit_of_measure, parameter_name, parameter_code, statistic_id, hydrologic_unit_code, state_name, last_modified, begin, end, begin_utc, end_utc, computation_period_identifier, computation_identifier, thresholds, sublocation_identifier, primary, monitoring_location_id, web_description, parameter_description, parent_time_series_id.
The default (\code{NA}) will return all columns of the data.}

\item{statistic_id}{A code corresponding to the statistic an observation represents. Example codes include 00001 (max), 00002 (min), and 00003 (mean). A complete list of codes and their descriptions can be found at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/statistic-codes/items}.

Multiple statistic_ids can be requested as a character vector.}

\item{last_modified}{The last time a record was refreshed in our database. This may happen due to regular operational processes and does not necessarily indicate anything about the measurement has changed.
You can query this field using date-times or intervals, adhering to RFC 3339, or using ISO 8601 duration objects. Intervals may be bounded or half-bounded (double-dots at start or end).
Examples:
\itemize{
\item A date-time: "2018-02-12T23:20:50Z"
\item A bounded interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
\item Half-bounded intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"
\item Duration objects: "P1M" for data from the past month or "PT36H" for the last 36 hours
}

Only features that have a \code{last_modified} that intersects the value of datetime are selected.

See also Details below for more information.}

\item{begin_utc}{The datetime of the earliest observation in the time series. Together with \code{end_utc}, this field represents the period of record of a time series. Note that some time series may have large gaps in their collection record.
You can query this field using date-times or intervals, adhering to RFC 3339, or using ISO 8601 duration objects. Intervals may be bounded or half-bounded (double-dots at start or end).
Examples:
\itemize{
\item A date-time: "2018-02-12T23:20:50Z"
\item A bounded interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
\item Half-bounded intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"
\item Duration objects: "P1M" for data from the past month or "PT36H" for the last 36 hours
}

Only features that have a \code{begin_utc} that intersects the value of datetime are selected.

#'
See also Details below for more information.}

\item{end_utc}{The datetime of the most recent observation in the time series. Data returned by this endpoint updates at most once per day, and potentially less frequently than that, and as such there may be more recent observations within a time series than the time series \code{end_utc} value reflects. Together with \code{begin_utc}, this field represents the period of record of a time series. It is additionally used to determine whether a time series is "active".
You can query this field using date-times or intervals, adhering to RFC 3339, or using ISO 8601 duration objects. Intervals may be bounded or half-bounded (double-dots at start or end).
Examples:
\itemize{
\item A date-time: "2018-02-12T23:20:50Z"
\item A bounded interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
\item Half-bounded intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"
\item Duration objects: "P1M" for data from the past month or "PT36H" for the last 36 hours
}

Only features that have a \code{end_utc} that intersects the value of datetime are selected.

See also Details below for more information.}

\item{hydrologic_unit_code}{The United States is divided and sub-divided into successively smaller hydrologic units which are classified into four levels: regions, sub-regions, accounting units, and cataloging units. The hydrologic units are arranged within each other, from the smallest (cataloging units) to the largest (regions). Each hydrologic unit is identified by a unique hydrologic unit code (HUC) consisting of two to eight digits based on the four levels of classification in the hydrologic unit system.}

\item{state_name}{The name of the state or state equivalent in which the monitoring location is located.}

\item{unit_of_measure}{A human-readable description of the units of measurement associated with an observation.}

\item{computation_period_identifier}{Indicates the period of data used for any statistical computations.
Multiple computation_period_identifiers can be requested as a character vector.}

\item{computation_identifier}{Indicates the computation performed to calculate this time series. Values of "Instantaneous" reflect point measurements.
Multiple computation_identifiers can be requested as a character vector.}

\item{thresholds}{Thresholds represent known numeric limits for a time series, for example the historic maximum value for a parameter or a level below which a sensor is non-operative. These thresholds are sometimes used to automatically determine if an observation is erroneous due to sensor error, and therefore shouldn't be included in the time series.}

\item{sublocation_identifier}{An optional human-readable identifier used to specify where measurements are recorded at a monitoring location.}

\item{primary}{A flag identifying if the time series is a "primary" time series. "Primary" time
series (which have this flag) are standard observations which undergo Bureau
review and approval processes. Non-primary time series, which will have missing
values for "primary", are provisional datasets made available to meet the need
for timely best science and to assist with daily operations which need
real-time information. Non-primary time series data are only retained by
this system for 120 days.}

\item{parent_time_series_id}{The unique identifier representing the parent or "upchain" time series that a daily values time series is generated from. Daily values time series have one and only one parent time series.}

\item{time_series_id}{A unique identifier representing a single time series. This corresponds to the "time_series_id" field in other endpoints.}

\item{web_description}{An optional description of the time series. WDFN and other USGS data dissemination products use this field, in combination with sublocation_identifier, to distinguish the differences between multiple time series for the same parameter code, statistic code, and monitoring location.}

\item{skipGeometry}{This parameter can be used to skip response geometries for
each feature. The returning object will be a data frame with no spatial
information. The default \code{NA} will not specify the argument in the request.}

\item{bbox}{Only features that have a geometry that intersects the bounding
box are selected.The bounding box is provided as four or six numbers, depending
on whether the coordinate reference system includes a vertical axis (height or
depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
vector structured: c(xmin,ymin,xmax,ymax).
Another way to think of it is c(Western-most longitude,
Southern-most latitude, Eastern-most longitude, Northern-most longitude).}

\item{begin}{This field contains the same information as "begin_utc", but in the local time of the monitoring location. It is retained for backwards compatibility, but will be removed in V1 of these APIs.}

\item{end}{This field contains the same information as "end_utc", but in the local time of the monitoring location. It is retained for backwards compatibility, but will be removed in V1 of these APIs.}

\item{...}{Not used. Included to help differentiate official Water Data API arguments
from more seldom used, optional dataRetrieval-specific arguments.}

\item{limit}{numeric, The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50,000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{convertType}{logical, defaults to TRUE.
If \code{TRUE}, the function will convert the data to dates, any qualifiers to string
vector and reorder the returned data frame.}

\item{no_paging}{logical, defaults to FALSE.
If \code{TRUE}, the data will
be requested from a native csv format. This can be dangerous because the
data will cut off at 50,000 rows without indication that more data
is available. Use \code{TRUE} with caution.}

\item{chunk_size}{Number of monitoring_location_ids to chunk requests into.
The default for functions that don't generally return long-term data records
is 250, while
the default for time series functions is
10.
Setting to \code{NA} will eliminate site chunking, giving users full control.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}
}
\description{
Daily data and continuous measurements are grouped into time series, which represent a collection of observations of a single parameter, potentially aggregated using a standard statistic, at a single monitoring location. This endpoint provides metadata about those time series, including their operational thresholds, units of measurement, and when the earliest and most recent observations in a time series occurred.
}
\details{
You can also use a vector of length 2 for any time queries (such as time
or last_modified). The first value is the starting date (or datetime),
the second value is the ending date(or datetime).
NA's within the vector indicate a half-bound date.
For example, \code{time = c("2024-01-01", NA)} will return all data starting
at 2024-01-01.
\code{time = c(NA, "2024-01-01")} will return all data from the beginning of
the timeseries until 2024-01-01.
By default, time is assumed UTC, although time zone attributes
will be accommodated. As an example, setting \code{time = as.POSIXct(c("2021-01-01 12:00:00",
"2021-01-01 14:00"), tz = "America/New_York")} will request data that between
noon and 2pm eastern time on 2021-01-01.
All time values RETURNED from the service are UTC with the exception of
daily data, which returns time values in local dates.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
site <- "USGS-02238500"
meta_1 <- read_waterdata_ts_meta(monitoring_location_id = site)

meta_multi <- read_waterdata_ts_meta(monitoring_location_id =  c("USGS-01491000",
                                                                 "USGS-01645000"),
                            parameter_code = c("00060", "00010"),
                            properties = c("monitoring_location_id",
                                           "parameter_code",
                                           "begin_utc",
                                           "end_utc",
                                           "time_series_id"),
                            skipGeometry = TRUE)

meta_wi <- read_waterdata_ts_meta(state_name = "Wisconsin")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ogc_data.R
\name{check_arguments_api}
\alias{check_arguments_api}
\title{Check other arguments}
\usage{
check_arguments_api(bbox, skipGeometry)
}
\arguments{
\item{bbox}{Only features that have a geometry that intersects the bounding
box are selected.The bounding box is provided as four or six numbers, depending
on whether the coordinate reference system includes a vertical axis (height or
depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
vector structured: c(xmin,ymin,xmax,ymax).
Another way to think of it is c(Western-most longitude,
Southern-most latitude, Eastern-most longitude, Northern-most longitude).}

\item{skipGeometry}{This parameter can be used to skip response geometries for
each feature. The returning object will be a data frame with no spatial
information. The default \code{NA} will not specify the argument in the request.}
}
\description{
Additional functions to check types and create parameter descriptions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setAccess.R
\name{setAccess}
\alias{setAccess}
\title{Set data endpoint}
\usage{
setAccess(access = "public")
}
\arguments{
\item{access}{code for data access. Options are: "public","internal","cooperator", or "USGS".
\itemize{
\item{"internal" represents Access=3 ...for a single water science center}
\item{"USGS" represents Access=2 ...for all water science centers}
\item{"cooperator" represents Access=1}
\item{"public" represents Access=0, public access}
}}
}
\description{
access Indicate which dataRetrieval access code
you want to use options: \code{c('public','internal')}
}
\examples{
\donttest{
setAccess("internal")

setAccess("public")
}

}
\author{
Luke Winslow, Jordan S Read
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addWaterYear.R
\name{calcWaterYear}
\alias{calcWaterYear}
\title{Extract WY from a date}
\usage{
calcWaterYear(dateVec)
}
\arguments{
\item{dateVec}{vector of dates as character ("YYYY-DD-MM"),
Date, or POSIXct. Numeric does not work.}
}
\value{
numeric vector indicating the water year
}
\description{
Determine the correct water year based on a calendar date.
}
\details{
This function calculates a water year based on the USGS
definition that a water year starts on October 1 of the year before,
and ends on September 30. For example, water year 2015 started on
2014-10-01 and ended on 2015-09-30.
}
\examples{
x <- seq(as.Date("2010-01-01"), as.Date("2010-12-31"), by = "month")
calcWaterYear(x)

y <- c("2010-01-01", "1994-02", "1980", "2009-11-01", NA)
calcWaterYear(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/findNLDI.R
\name{get_nldi_sources}
\alias{get_nldi_sources}
\title{Get current NLDI offerings}
\usage{
get_nldi_sources(url = getOption("dataRetrieval.nldi_base"))
}
\arguments{
\item{url}{URL for NLDI sources. Default is supplied by package environment.}
}
\value{
data.frame
}
\description{
Used to query the current resources available through the NLDI
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
get_nldi_sources()
}
\dontshow{\}) # examplesIf}
}
\keyword{nldi}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISdata.R
\name{stateCdLookup}
\alias{stateCdLookup}
\title{State code look up}
\usage{
stateCdLookup(input, outputType = "postal", country = "US")
}
\arguments{
\item{input}{could be character (full name, abbreviation, id), or numeric (id)}

\item{outputType}{character can be "postal", "fullName", "tableIndex", or "id".}

\item{country}{description}
}
\description{
Function to simplify finding state and state code definitions. Used in \code{readNWISdata}
and \code{readWQPdata}.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
fullName <- stateCdLookup("wi", "fullName")
abbriev <- stateCdLookup("Wisconsin", "postal")
id <- stateCdLookup("WI", "id")
name <- stateCdLookup(55, "fullName")
fips <- stateCdLookup("WI", "fips")
canada_st <- stateCdLookup(13, "fullName", country = "CA")
mexico_st <- stateCdLookup(13, "fullName", country = "MX")
stateCdLookup(c("West Virginia", "Wisconsin", 200, 55, "MN"))
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_monitoring_location.R
\name{read_waterdata_monitoring_location}
\alias{read_waterdata_monitoring_location}
\title{Get USGS Monitoring Location Data}
\usage{
read_waterdata_monitoring_location(
  monitoring_location_id = NA_character_,
  agency_code = NA_character_,
  agency_name = NA_character_,
  monitoring_location_number = NA_character_,
  monitoring_location_name = NA_character_,
  district_code = NA_character_,
  country_code = NA_character_,
  country_name = NA_character_,
  state_code = NA_character_,
  state_name = NA_character_,
  county_code = NA_character_,
  county_name = NA_character_,
  minor_civil_division_code = NA_character_,
  site_type_code = NA_character_,
  site_type = NA_character_,
  hydrologic_unit_code = NA_character_,
  basin_code = NA_character_,
  altitude = NA_character_,
  altitude_accuracy = NA_character_,
  altitude_method_code = NA_character_,
  altitude_method_name = NA_character_,
  vertical_datum = NA_character_,
  vertical_datum_name = NA_character_,
  horizontal_positional_accuracy_code = NA_character_,
  horizontal_positional_accuracy = NA_character_,
  horizontal_position_method_code = NA_character_,
  horizontal_position_method_name = NA_character_,
  original_horizontal_datum = NA_character_,
  original_horizontal_datum_name = NA_character_,
  drainage_area = NA_character_,
  contributing_drainage_area = NA_character_,
  time_zone_abbreviation = NA_character_,
  uses_daylight_savings = NA_character_,
  construction_date = NA_character_,
  aquifer_code = NA_character_,
  national_aquifer_code = NA_character_,
  aquifer_type_code = NA_character_,
  well_constructed_depth = NA_character_,
  hole_constructed_depth = NA_character_,
  depth_source_code = NA_character_,
  properties = NA_character_,
  bbox = NA,
  skipGeometry = NA,
  ...,
  limit = getOption("dataRetrieval.limit"),
  no_paging = getOption("dataRetrieval.no_paging"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_meta"),
  attach_request = getOption("dataRetrieval.attach_request")
)
}
\arguments{
\item{monitoring_location_id}{A unique identifier representing a single monitoring location. This corresponds to the \code{monitoring_location_id} field in other endpoints. Monitoring location IDs are created by combining the agency code of the agency responsible for the monitoring location (e.g. USGS) with the ID number of the monitoring location (e.g. 02238500), separated by a hyphen (e.g. USGS-02238500).

Multiple monitoring_location_ids can be requested as a character vector.}

\item{agency_code}{The agency that is reporting the data. Agency codes are fixed values assigned by the National Water Information System (NWIS). A list of agency codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/agency-codes/items}.}

\item{agency_name}{The name of the agency that is reporting the data.}

\item{monitoring_location_number}{Each monitoring location in the USGS data base has a unique 8- to 15-digit identification number.}

\item{monitoring_location_name}{This is the official name of the monitoring location in the database. For well information this can be a district-assigned local number.}

\item{district_code}{The Water Science Centers (WSCs) across the United States use the FIPS state code as the district code. In some case, monitoring locations and samples may be managed by a water science center that is adjacent to the state in which the monitoring location actually resides. For example a monitoring location may have a district code of 30 which translates to Montana, but the state code could be 56 for Wyoming because that is where the monitoring location actually is located.}

\item{country_code}{The code for the country in which the monitoring location is located.}

\item{country_name}{The name of the country in which the monitoring location is located.}

\item{state_code}{State code. A \href{https://www2.census.gov/geo/docs/reference/state.txt}{two-digit ANSI code} (formerly FIPS code) as defined by the American National Standards Institute, to define States and equivalents. A three-digit ANSI code is used to define counties and county equivalents. \href{https://www.census.gov/library/reference/code-lists/ansi.html#states}{A lookup table is available.} The only countries with political subdivisions other than the US are Mexico and Canada. The Mexican states have US state codes ranging from 81-86 and Canadian provinces have state codes ranging from 90-98.}

\item{state_name}{The name of the state or state equivalent in which the monitoring location is located.}

\item{county_code}{The code for the county or county equivalent (parish, borough, etc.) in which the monitoring location is located. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/counties/items}.}

\item{county_name}{The name of the county or county equivalent (parish, borough, etc.) in which the monitoring location is located. [A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/counties/items}.}

\item{minor_civil_division_code}{Codes for primary governmental or administrative divisions of the county or county equivalent in which the monitoring location is located.}

\item{site_type_code}{A code describing the hydrologic setting of the monitoring location. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/site-types/items}.}

\item{site_type}{A description of the hydrologic setting of the monitoring location. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/site-types/items}.}

\item{hydrologic_unit_code}{The United States is divided and sub-divided into successively smaller hydrologic units which are classified into four levels: regions, sub-regions, accounting units, and cataloging units. The hydrologic units are arranged within each other, from the smallest (cataloging units) to the largest (regions). Each hydrologic unit is identified by a unique hydrologic unit code (HUC) consisting of two to eight digits based on the four levels of classification in the hydrologic unit system.}

\item{basin_code}{The Basin Code or "drainage basin code" is a two-digit code that further subdivides the 8-digit hydrologic-unit code. The drainage basin code is defined by the USGS State Office where the monitoring location is located.}

\item{altitude}{Altitude of the monitoring location referenced to the specified Vertical Datum.}

\item{altitude_accuracy}{Accuracy of the altitude, in feet. An accuracy of +/- 0.1 foot would be entered as “.1”. Many altitudes are interpolated from the contours on topographic maps; accuracies determined in this way are generally entered as one-half of the contour interval.}

\item{altitude_method_code}{Codes representing the method used to measure altitude.}

\item{altitude_method_name}{The name of the method used to measure altitude.}

\item{vertical_datum}{The datum used to determine altitude and vertical position at the monitoring location. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/altitude-datums/items}.}

\item{vertical_datum_name}{The datum used to determine altitude and vertical position at the monitoring location. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/altitude-datums/items}.}

\item{horizontal_positional_accuracy_code}{Indicates the accuracy of the latitude longitude values. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/coordinate-accuracy-codes/items}.}

\item{horizontal_positional_accuracy}{Indicates the accuracy of the latitude longitude values. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/coordinate-accuracy-codes/items}.}

\item{horizontal_position_method_code}{Indicates the method used to determine latitude longitude values. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/coordinate-method-codes/items}.}

\item{horizontal_position_method_name}{Indicates the method used to determine latitude longitude values. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/coordinate-method-codes/items}.}

\item{original_horizontal_datum}{Coordinates are published in EPSG:4326 / WGS84 / World Geodetic System 1984. This field indicates the original datum used to determine coordinates before they were converted. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/coordinate-datum-codes/items}.}

\item{original_horizontal_datum_name}{Coordinates are published in EPSG:4326 / WGS84 / World Geodetic System 1984. This field indicates the original datum used to determine coordinates before they were converted. A list of codes is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/coordinate-datum-codes/items}.}

\item{drainage_area}{The area enclosed by a topographic divide from which direct surface runoff from precipitation normally drains by gravity into the stream above that point.}

\item{contributing_drainage_area}{The contributing drainage area of a lake, stream, wetland, or estuary monitoring location, in square miles. This item should be present only if the contributing area is different from the total drainage area. This situation can occur when part of the drainage area consists of very porous soil or depressions that either allow all runoff to enter the groundwater or traps the water in ponds so that rainfall does not contribute to runoff. A transbasin diversion can also affect the total drainage area.}

\item{time_zone_abbreviation}{A short code describing the time zone used by a monitoring location.}

\item{uses_daylight_savings}{A flag indicating whether or not a monitoring location uses daylight savings.}

\item{construction_date}{Date the well was completed.}

\item{aquifer_code}{Local aquifers in the USGS water resources data base are identified by a geohydrologic unit code (a three-digit number related to the age of the formation, followed by a 4 or 5 character abbreviation for the geologic unit or aquifer name).}

\item{national_aquifer_code}{National aquifers are the principal aquifers or aquifer systems in the United States, defined as regionally extensive aquifers or aquifer systems that have the potential to be used as a source of potable water. Not all groundwater monitoring locations can be associated with a National Aquifer. Such monitoring locations will not be retrieved using this search criteria. A list of National aquifer codes and names is available at \url{https://api.waterdata.usgs.gov/ogcapi/v0/collections/national-aquifer-codes/items}.}

\item{aquifer_type_code}{Describes the confinement status of an aquifer at the monitoring location. A confined aquifer is an aquifer below the land surface that is saturated with water. A water table--or unconfined--aquifer is an aquifer whose upper water surface (water table) is at atmospheric pressure, and thus is able to rise and fall.}

\item{well_constructed_depth}{The depth of the finished well, in feet below land surface datum. Note: Not all groundwater monitoring locations have information on Well Depth. Such monitoring locations will not be retrieved using this search criteria.}

\item{hole_constructed_depth}{The total depth to which the hole is drilled, in feet below land surface datum. Note: Not all groundwater monitoring locations have information on Hole Depth. Such monitoring locations will not be retrieved using this search criteria.}

\item{depth_source_code}{A code indicating the source of water-level data.}

\item{properties}{A vector of requested columns to be returned from the query.
Available options are:
geometry, monitoring_location_id, agency_code, agency_name, monitoring_location_number, monitoring_location_name, district_code, country_code, country_name, state_code, state_name, county_code, county_name, minor_civil_division_code, site_type_code, site_type, hydrologic_unit_code, basin_code, altitude, altitude_accuracy, altitude_method_code, altitude_method_name, vertical_datum, vertical_datum_name, horizontal_positional_accuracy_code, horizontal_positional_accuracy, horizontal_position_method_code, horizontal_position_method_name, original_horizontal_datum, original_horizontal_datum_name, drainage_area, contributing_drainage_area, time_zone_abbreviation, uses_daylight_savings, construction_date, aquifer_code, national_aquifer_code, aquifer_type_code, well_constructed_depth, hole_constructed_depth, depth_source_code, revision_note, revision_created, revision_modified.
The default (\code{NA}) will return all columns of the data.}

\item{bbox}{Only features that have a geometry that intersects the bounding
box are selected.The bounding box is provided as four or six numbers, depending
on whether the coordinate reference system includes a vertical axis (height or
depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
vector structured: c(xmin,ymin,xmax,ymax).
Another way to think of it is c(Western-most longitude,
Southern-most latitude, Eastern-most longitude, Northern-most longitude).}

\item{skipGeometry}{This parameter can be used to skip response geometries for
each feature. The returning object will be a data frame with no spatial
information. The default \code{NA} will not specify the argument in the request.}

\item{...}{Not used. Included to help differentiate official Water Data API arguments
from more seldom used, optional dataRetrieval-specific arguments.}

\item{limit}{numeric, The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50,000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{no_paging}{logical, defaults to FALSE.
If \code{TRUE}, the data will
be requested from a native csv format. This can be dangerous because the
data will cut off at 50,000 rows without indication that more data
is available. Use \code{TRUE} with caution.}

\item{chunk_size}{Number of monitoring_location_ids to chunk requests into.
The default for functions that don't generally return long-term data records
is 250, while
the default for time series functions is
10.
Setting to \code{NA} will eliminate site chunking, giving users full control.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}
}
\description{
Location information is basic information about the monitoring location including the name, identifier, agency responsible for data collection, and the date the location was established. It also includes information about the type of location, such as stream, lake, or groundwater, and geographic information about the location, such as state, county, latitude and longitude, and hydrologic unit code (HUC).
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
site <- "USGS-02238500"
site_info <- read_waterdata_monitoring_location(monitoring_location_id = site)

site_slim <- read_waterdata_monitoring_location(
                             monitoring_location_id = c("USGS-05114000",
                                                        "USGS-09423350"),
                             properties = c("monitoring_location_id",
                                            "state_name",
                                            "county_name",
                                            "country_name"))

site_slim_no_sf_slim <- read_waterdata_monitoring_location(state_name = "Wisconsin",
                                           properties = c("monitoring_location_id",
                                                          "state_name",
                                                          "county_name",
                                                          "country_name",
                                                          "site_type"),
                                           site_type = "Well",
                                           skipGeometry = TRUE)

site_info_no_sf <- read_waterdata_monitoring_location(
     monitoring_location_id = site_slim_no_sf_slim$monitoring_location_id[1:1000],
     skipGeometry = TRUE)

bbox_vals = c(-94.00, 35.0, -93.5, 35.5)
multi_site <- read_waterdata_monitoring_location(bbox = bbox_vals)


}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRetrieval-package.R
\docType{data}
\name{parameterCdFile}
\alias{parameterCdFile}
\title{List of USGS parameter codes}
\value{
parameterData data frame with information about USGS parameters.

\tabular{lll}{
Name \tab Type \tab Description\cr
parameter_cd \tab character \tab 5-digit USGS parameter code \cr
parameter_group_nm \tab character \tab USGS parameter group name\cr
parameter_nm \tab character \tab USGS parameter name\cr
casrn \tab character \tab Chemical Abstracts Service (CAS) Registry Number\cr
srsname \tab character \tab Substance Registry Services Name\cr
parameter_units \tab character \tab Parameter units\cr
}
}
\description{
Complete list of USGS parameter codes as of Oct. 24, 2024.
}
\examples{
head(parameterCdFile[, 1:2])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISdv.R
\name{readNWISdv}
\alias{readNWISdv}
\title{Daily Value USGS NWIS Data Retrieval}
\usage{
readNWISdv(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  statCd = "00003"
)
}
\arguments{
\item{siteNumbers}{character USGS site number.  This is usually an 8 digit number.
Multiple sites can be requested with a character vector.}

\item{parameterCd}{character of USGS parameter code(s).  This is usually an 5 digit number.}

\item{startDate}{character starting date for data retrieval in the form YYYY-MM-DD.
Default is "" which indicates
retrieval for the earliest possible record. Date arguments are always specified in local time.}

\item{endDate}{character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the latest possible record. Date arguments are always specified in local time.}

\item{statCd}{character USGS statistic code. This is usually 5 digits.  Daily mean (00003) is the default.}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency \tab character \tab The NWIS code for the agency reporting the data\cr
site \tab character \tab The USGS site number \cr
Date \tab Date \tab The date of the value \cr
code \tab character \tab Any codes that qualify the corresponding value\cr
value \tab numeric \tab The numeric value for the parameter \cr
}
Note that code and value are repeated for the parameters requested. The names are of the form
X_D_P_S, where X is literal,
D is an option description of the parameter,
P is the parameter code,
and S is the statistic code (if applicable).

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Imports data from NWIS daily web service. This function gets the data from here:
\url{https://waterservices.usgs.gov/docs/dv-service/daily-values-service-details/}
Inputs to this function are just USGS site ids, USGS parameter codes,
USGS statistic codes, and start and end date. For a more complex query, use \code{\link[=readNWISdata]{readNWISdata()}},
with an argument service = "dv".
Data coming the daily web services are aggregates of the instantaneous
(sensor) web services. Not all statistical codes are available for all data.
Use the function \code{\link[=whatNWISdata]{whatNWISdata()}} to discover what data
is available for a USGS site. The column data_type_cd with the values "dv"
returned from \code{\link[=whatNWISdata]{whatNWISdata()}}) are available from this service.
}
\details{
More information on the web service can be found here:
\url{https://waterservices.usgs.gov/test-tools}, choosing the
"Daily Value Service".
}
\examples{

# see ?read_waterdata_daily

#site_id <- "04085427"
#startDate <- "2012-01-01"
#endDate <- "2012-06-30"
#pCode <- "00060"
#
#rawDailyQ <- readNWISdv(site_id, pCode, startDate, endDate)

}
\seealso{
\code{\link[=read_waterdata_daily]{read_waterdata_daily()}}
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importRDB1.R
\name{importRDB1}
\alias{importRDB1}
\title{Function to return data from the NWIS RDB 1.0 format}
\usage{
importRDB1(obs_url, asDateTime = TRUE, convertType = TRUE, tz = "UTC")
}
\arguments{
\item{obs_url}{character containing the url for the retrieval or a file path to the data file.}

\item{asDateTime}{logical, if \code{TRUE} returns date and time as POSIXct, if \code{FALSE}, Date}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the
function will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character}

\item{tz}{character to set timezone attribute of datetime. Default converts the datetimes to UTC
(properly accounting for daylight savings times based on the data's provided tz_cd column).
Recommended US values include "UTC", "America/New_York", "America/Chicago", "America/Denver",
"America/Los_Angeles",  "America/Anchorage", "America/Honolulu", "America/Jamaica", "America/Managua",
"America/Phoenix", and "America/Metlakatla".
For a complete list, see \url{https://en.wikipedia.org/wiki/List_of_tz_database_time_zones}}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
datetime \tab POSIXct \tab The date and time of the value converted to
UTC (if asDateTime = \code{TRUE}), \cr
\tab character \tab or raw character string (if asDateTime = FALSE) \cr
tz_cd \tab character \tab The time zone code for datetime \cr
code \tab character \tab Any codes that qualify the corresponding value\cr
value \tab numeric \tab The numeric value for the parameter \cr
tz_cd_reported \tab The originally reported time zone \cr
}
Note that code and value are repeated for the parameters requested. The names are of the form
XD_P_S, where X is literal,
D is an option description of the parameter,
P is the parameter code,
and S is the statistic code (if applicable).
If a date/time (dt) column contained incomplete date and times, a new column
of dates and time was inserted. This could happen
when older data was reported as dates, and newer data was reported as a date/time.

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
comment \tab character \tab Header comments from the RDB file \cr
}
}
\description{
This function accepts a url parameter that already contains the desired
NWIS site, parameter code, statistic, startdate and enddate. It is not
recommended to use the RDB format for importing multi-site data.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
site_id <- "02177000"
startDate <- "2012-09-01"
endDate <- "2012-10-01"
offering <- "00003"
property <- "00060"

obs_url <- constructNWISURL(site_id, property,
  startDate, endDate, "dv",
  format = "tsv"
)
\donttest{
data <- importRDB1(obs_url)


urlMultiPcodes <- constructNWISURL("04085427", c("00060", "00010"),
  startDate, endDate, "dv",
  statCd = c("00003", "00001"), "tsv"
)

multiData <- importRDB1(urlMultiPcodes)

unitDataURL <- constructNWISURL(site_id, property,
  "2020-10-30", "2020-11-01", "uv",
  format = "tsv"
) # includes timezone switch

unitData <- importRDB1(unitDataURL, asDateTime = TRUE)

iceSite <- "04024000"
start <- "2015-11-09"
end <- "2015-11-24"
urlIce <- constructNWISURL(iceSite, "00060", start, end, "uv", format = "tsv")

ice <- importRDB1(urlIce, asDateTime = TRUE)
iceNoConvert <- importRDB1(urlIce, convertType = FALSE)
}
# User file:
filePath <- system.file("extdata", package = "dataRetrieval")
fileName <- "RDB1Example.txt"
fullPath <- file.path(filePath, fileName)
importUserRDB <- importRDB1(fullPath)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AAA.R
\name{is_dataRetrieval_user}
\alias{is_dataRetrieval_user}
\title{Is this a dataRetrieval user}
\usage{
is_dataRetrieval_user()
}
\description{
Reveals if this is a user or not
}
\examples{
is_dataRetrieval_user()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_stats.R
\name{read_waterdata_stats_por}
\alias{read_waterdata_stats_por}
\alias{read_waterdata_stats_daterange}
\title{Get USGS daily data statistics}
\usage{
read_waterdata_stats_por(
  approval_status = NA,
  computation_type = NA_character_,
  country_code = NA_character_,
  state_code = NA_character_,
  county_code = NA_character_,
  start_date = NA_character_,
  end_date = NA_character_,
  normal_type = NA_character_,
  monitoring_location_id = NA_character_,
  parent_time_series_id = NA_character_,
  site_type_code = NA_character_,
  site_type_name = NA_character_,
  parameter_code = NA_character_,
  page_size = NA
)

read_waterdata_stats_daterange(
  approval_status = NA,
  computation_type = NA_character_,
  country_code = NA_character_,
  state_code = NA_character_,
  county_code = NA_character_,
  start_date = NA_character_,
  end_date = NA_character_,
  interval_type = NA_character_,
  monitoring_location_id = NA_character_,
  parent_time_series_id = NA_character_,
  site_type_code = NA_character_,
  site_type_name = NA_character_,
  parameter_code = NA_character_,
  page_size = NA
)
}
\arguments{
\item{approval_status}{Whether to include approved and/or provisional
observations. At this time, only approved observations are returned.}

\item{computation_type}{Desired statistical computation method. Available
values: "arithmetic_mean", "maximum", "median", "minimum", "percentile".}

\item{country_code}{Country Query Parameter. Accepts multiple values (see
examples). If one of country, county, or state code is supplied then the
other two arguments do not need to be specified.}

\item{state_code}{State Query Parameter. Accepts multiple values in a
character vector.}

\item{county_code}{County Query Parameter. Accepts multiple values in a
character vector.}

\item{start_date}{Start Date Query Parameter. The logic is inclusive i.e., it
will also return records that match the date. If an end date is supplied,
but no start date is supplied, then statistics will be supplied for the
entire period of record ending with the end date. If an end date is not
supplied, but a start date is supplied then statistics will be supplied for
the period of record following the start date. If no start or end date are
supplied then statistics will be supplied for the entire period of record.}

\item{end_date}{End Date Query Parameter. The logic is inclusive i.e., it will
also return records that match the date.}

\item{normal_type}{Normal Type Query Parameter. If unspecified, all matching data
will be returned. Otherwise, it will filter the results to one of the following
normals: day-of-year, month-of-year. Available values: "DOY", "MOY".}

\item{monitoring_location_id}{Each monitoring location has been assigned a
unique station number that places them in downstream order. Accepts
multiple values in a character vector.}

\item{parent_time_series_id}{The parent_time_series_id returns statistics
tied to a particular database entry. Accepts multiple values in a character
vector. If no parent time series identifier is supplied, then all records
matching the rest of the provided criteria will be returned.}

\item{site_type_code}{Site Type Code Query Parameter. Accepts multiple values
in a character vector. If no Site Type code is specified, statistics of all
site types with the matching Monitoring Location Identifier will be
returned.}

\item{site_type_name}{Site Type Name Query Parameter. If no Site Type name is
specified, statistics of all site types with the matching Monitoring
Location Identifier will be returned.}

\item{parameter_code}{USGS Parameter Code Query Parameter. Accepts multiple
values in a character vector. If no USGS parameter code is specified, but a
Monitoring Location Identifier is supplied, then all statistics and their
parameter codes with a matching monitoring location identifier will be
returned. All statistics within the period of record will be returned if no
parameter code or monitoring location identifier are specified.}

\item{page_size}{Return a defined number of results (default: 1000).}

\item{interval_type}{Interval Type Query Parameter. If unspecified, all matching
data will be returned. Otherwise, it will filter the results to one or more of
the following intervals: month, calendar year, water year.
Available values: "M", "CY", "WY".}
}
\description{
This service provides endpoints for access to computations on the historical
record regarding water conditions. For more information regarding the
calculation of statistics and other details, please visit the Statistics
documentation page.

Note: The /statistics API is under active beta development and subject to change.
Improved handling of significant figures will be addressed in a future
release.

\code{read_waterdata_stats_por} Returns day-of-year and month-of-year
statistics matching your query.

\code{read_waterdata_stats_daterange} Returns monthly and annual statistics matching
your query.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
# All day-of-year and month-of-year statistics for two sites
x1 <- read_waterdata_stats_por(
  monitoring_location_id = c("USGS-02319394", "USGS-02171500")
)

# Request only month-of-year statistics using normal_type arg
x1 <- read_waterdata_stats_por(
  monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
  normal_type = "MOY"
)

# Request temperature percentiles for specific month-day range
# Returns:
# - Day-of-year temperature percentiles for each day between June 1 through June 15.
# - Month-of-year percentiles for June, computed using
#    all June data (not just June 1 through June 15).
# Note: the month-of-year percentiles are returned only when the month-day range includes
#  the beginning of the month (e.g., "06-01")
x2 <- read_waterdata_stats_por(
  monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
  parameter_code = "00010",
  start_date = "06-01", end_date = "06-15",
  computation_type = "percentile"
)

# All calendar month, calendar year, and water year statistics for two sites
x3 <- read_waterdata_stats_daterange(
  monitoring_location_id = c("USGS-02319394", "USGS-02171500")
)

# Request only calendar year statistics
x3 <- read_waterdata_stats_daterange(
  monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
  interval_type = "CY"
)

# Request specific gage height and discharge summaries for a limited date range
# Returns:
# - calendar month summaries for each month between January, 2010 through December, 2011
# - calendar year summaries for 2010 and 2011
# - water year summaries for WY2010, WY2011, and WY2012
x4 <- read_waterdata_stats_daterange(
  monitoring_location_id = c("USGS-02319394", "USGS-02171500"),
  parameter_code = c("00065", "00060"),
  start_date = "2010-01-01", end_date = "2011-12-31",
  computation_type = c("minimum", "median", "maximum")
)

}
\dontshow{\}) # examplesIf}
}
\seealso{
\url{https://api.waterdata.usgs.gov/statistics/v0/docs}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whatWQPdata.R, R/whatWQPsites.R
\name{whatWQPsamples}
\alias{whatWQPsamples}
\alias{whatWQPmetrics}
\alias{whatWQPsites}
\title{Site Data Import from Water Quality Portal}
\usage{
whatWQPsamples(..., convertType = TRUE, legacy = TRUE)

whatWQPmetrics(..., convertType = TRUE)

whatWQPsites(..., legacy = TRUE, convertType = TRUE)
}
\arguments{
\item{\dots}{see \url{https://www.waterqualitydata.us/webservices_documentation}
for a complete list of options. A list of arguments can also be supplied.
One way to figure out how to construct a WQP query is to go to the "Advanced"
form in the Water Quality Portal:
\url{https://www.waterqualitydata.us/#mimeType=csv&providers=NWIS&providers=STORET}
Use the form to discover what parameters are available. Once the query is
set in the form, scroll down to the "Query URL". You will see the parameters
after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
in the Characteristic Group dropdown, you will see characteristicType=Nutrient
in the Query URL. The corresponding argument for dataRetrieval is
characteristicType = "Nutrient". dataRetrieval users do not need to include
mimeType,  and providers is optional (these arguments are picked automatically).}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the
function will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character.}

\item{legacy}{Logical. If TRUE, uses legacy WQP services. Default is TRUE.
Setting legacy = FALSE uses WQX3.0 WQP services, which are in-development, use with caution.}
}
\value{
A data frame with information on the sampling activity
available from the Water Quality Portal for the query parameters.

data frame that includes information on site metadata.
}
\description{
Returns a list of sites from the Water Quality Portal web service. This function
gets the data from: \url{https://www.waterqualitydata.us}.
Arguments to the function should be based on
\url{https://www.waterqualitydata.us/webservices_documentation}. The return from
this function returns the basic metadata on WQP sites. It is
generally faster than the \code{\link[=whatWQPdata]{whatWQPdata()}} function, but does
not return information on what data was collected at the site.
}
\examples{
\donttest{

#site1 <- whatWQPsamples(siteid = "USGS-01594440")

#type <- "Stream"

#sites <- whatWQPsamples(countycode = "US:55:025", siteType = type)

#lakeSites_samples <- whatWQPsamples(siteType = "Lake, Reservoir, Impoundment",
#                                   countycode = "US:55:025")
}
\donttest{

#type <- "Stream"

#sites <- whatWQPmetrics(countycode = "US:55:025", siteType = type)
#lakeSites_metrics <- whatWQPmetrics(siteType = "Lake, Reservoir, Impoundment",
#                                   countycode = "US:55:025")
}
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{

#site1 <- whatWQPsites(siteid = "USGS-01594440")

#type <- "Stream"
#sites <- whatWQPsites(
#   countycode = "US:55:025",
#  characteristicName = "Phosphorus",
#   siteType = type
#)
}
\dontshow{\}) # examplesIf}
}
\seealso{
whatWQPdata readWQPsummary

whatNWISdata
}
\keyword{WQP}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISunit.R
\name{readNWISpeak}
\alias{readNWISpeak}
\title{Peak flow data from USGS (NWIS)}
\usage{
readNWISpeak(
  siteNumbers,
  startDate = "",
  endDate = "",
  asDateTime = TRUE,
  convertType = TRUE
)
}
\arguments{
\item{siteNumbers}{character USGS site number(or multiple sites).  This is usually
an 8 digit number.}

\item{startDate}{character starting date for data retrieval in the form YYYY-MM-DD.
Default is "" which indicates
retrieval for the earliest possible record.}

\item{endDate}{character ending date for data retrieval in the form YYYY-MM-DD.
Default is "" which indicates
retrieval for the latest possible record.}

\item{asDateTime}{logical default to \code{TRUE}. When \code{TRUE}, the peak_dt column is converted
to a Date object, and incomplete dates are removed. When \code{FALSE}, no
columns are removed, but no dates are converted.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
peak_dt \tab Date \tab Date of peak streamflow \cr
peak_tm \tab character \tab Time of peak streamflow as character \cr
peak_va \tab numeric \tab Annual peak streamflow value in cfs \cr
peak_cd \tab character \tab Peak Discharge-Qualification codes (see \code{comment}
for more information) \cr
gage_ht \tab numeric \tab Gage height for the associated peak streamflow in feet \cr
gage_ht_cd \tab character \tab Gage height qualification codes \cr
year_last_pk \tab numeric \tab Peak streamflow reported is the highest since this year \cr
ag_dt \tab Date \tab Date of maximum gage-height for water year
(if not concurrent with peak) \cr
ag_tm \tab character \tab Time of maximum gage-height for water year
(if not concurrent with peak) \cr
ag_gage_ht \tab numeric \tab maximum Gage height for water year in feet
(if not concurrent with peak) \cr
ag_gage_ht_cd \tab character \tab maximum Gage height code \cr
}

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
comment \tab character \tab Header comments from the RDB file \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
}
}
\description{
Reads peak flow from NWISweb.
}
\details{
In some cases, the specific date of the peak data is not know. This function
will default to
converting complete dates to a "Date" object, and converting incomplete dates to
"NA". If those incomplete dates are
needed, set the \code{asDateTime} argument to FALSE. No dates will be converted to
R Date objects.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
#site_ids <- c("01594440", "040851325")
\donttest{
#data <- readNWISpeak(site_ids)
#data2 <- readNWISpeak(site_ids, asDateTime = FALSE)
#stations <- c("06011000")
#peakdata <- readNWISpeak(stations, convertType = FALSE)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=constructNWISURL]{constructNWISURL()}}, \code{\link[=importRDB1]{importRDB1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whatNWISsites.R
\name{whatNWISsites}
\alias{whatNWISsites}
\title{Site Data Import from NWIS}
\usage{
whatNWISsites(...)
}
\arguments{
\item{\dots}{see \url{https://waterservices.usgs.gov/docs/site-service/}
for a complete list of options. A list (or lists) can also be supplied.}
}
\value{
A data frame with at least the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
station_nm \tab character \tab Station name \cr
site_tp_cd \tab character \tab Site type code \cr
dec_lat_va \tab numeric \tab Decimal latitude \cr
dec_long_va \tab numeric \tab Decimal longitude \cr
queryTime \tab POSIXct \tab Query time \cr
}

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Returns a list of sites from the NWIS web service. This function gets the data from:
\url{https://waterservices.usgs.gov/docs/site-service/}.
Mapper format is used
}
\examples{

# see ?read_waterdata_monitoring_location
#siteListPhos <- whatNWISsites(stateCd = "OH", parameterCd = "00665")
#oneSite <- whatNWISsites(sites = "05114000")

}
\seealso{
\code{\link[=read_waterdata_monitoring_location]{read_waterdata_monitoring_location()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWQPqw.R
\name{readWQPqw}
\alias{readWQPqw}
\title{Raw Data Import for Water Quality Portal}
\usage{
readWQPqw(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  tz = "UTC",
  legacy = TRUE,
  querySummary = FALSE,
  ignore_attributes = FALSE,
  convertType = TRUE
)
}
\arguments{
\item{siteNumbers}{character site number. This needs to include the full agency code prefix.}

\item{parameterCd}{vector of USGS 5-digit parameter code or characteristicNames.
Leaving this blank will return all of the measured values during the specified time period.}

\item{startDate}{character starting date for data retrieval in the form YYYY-MM-DD.
Default is "" which indicates
retrieval for the earliest possible record. Date arguments are always specified in local time.}

\item{endDate}{character ending date for data retrieval in the form YYYY-MM-DD.
Default is "" which indicates
retrieval for the latest possible record. Date arguments are always specified in local time.}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the
data provided tz_cd column.
Possible values to provide are "America/New_York","America/Chicago",
"America/Denver","America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings
time: "America/Honolulu",
"America/Jamaica","America/Managua","America/Phoenix", and "America/Metlakatla".
See also  \code{OlsonNames()}
for more information on time zones.}

\item{legacy}{Logical. If TRUE, uses legacy WQP services. Default is TRUE.
Setting legacy = FALSE uses WQX3.0 WQP services, which are in-development, use with caution.}

\item{querySummary}{logical to look at number of records and unique sites that
will be returned from this query.}

\item{ignore_attributes}{logical to choose to ignore fetching site and parameter
attributes. Default is \code{FALSE}.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character.}
}
\value{
A data frame derived from the default data profile.

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Imports data from the Water Quality Portal.
This function gets the data from here: \url{https://www.waterqualitydata.us}. There
are four required input arguments: siteNumbers, parameterCd, startDate, and endDate.
parameterCd can either be a USGS 5-digit code, or a characteristic name. The sites can be
either USGS, or other Water Quality Portal offered sites. It is required to use the 'full'
site name, such as 'USGS-01234567'.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
#rawPcode <- readWQPqw("USGS-01594440", "01075", "", "")

#attr(rawPcode, "siteInfo")
#attr(rawPcode, "queryTime")
#attr(rawPcode, "url")

#rawCharacteristicName <- readWQPqw("WIDNR_WQX-10032762", "Specific conductance", "", "")
#pHsites_legacy <- readWQPqw(c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#                         "pH", "", "")
#ncol(pHsites_legacy)
#attr(pHsites_legacy, "url")

# pHsites_modern <- readWQPqw(c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#                         "pH", "", "", legacy = FALSE)
# ncol(pHsites_modern)
# attr(pHsites_modern, "url")

# nwisEx <- readWQPqw("USGS-04024000", c("34247", "30234", "32104", "34220"), "", "2022-12-20")

# DO <- readWQPqw(siteNumbers = "USGS-05288705",
#                 parameterCd = "00300",
#                 convertType = FALSE)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=readWQPdata]{readWQPdata()}}, \code{\link[=whatWQPsites]{whatWQPsites()}},
and \code{\link[=importWQP]{importWQP()}}
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importWQP.R
\name{parse_WQP}
\alias{parse_WQP}
\title{Convert WQP columns to correct types}
\usage{
parse_WQP(retval, tz = "UTC")
}
\arguments{
\item{retval}{Data frame from WQP}

\item{tz}{character to set timezone attribute of datetime. Default is UTC
(properly accounting for daylight savings times based on the associated "TimeZone" column).
Possible values include "America/New_York","America/Chicago", "America/Denver","America/Los_Angeles",
"America/Anchorage","America/Honolulu","America/Jamaica","America/Managua",
"America/Phoenix", and "America/Metlakatla"}
}
\value{
data frame retval with converted columns
}
\description{
Takes the character results and converts to numeric and dates.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
# These examples require an internet connection to run
rawSampleURL <- constructWQPURL("USGS-01594440", "01075", "", "")

## Examples take longer than 5 seconds:

\donttest{

rawSample <- importWQP(rawSampleURL, convertType = FALSE)
convertedSample <- parse_WQP(rawSample)

}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISsite.R
\name{readNWISsite}
\alias{readNWISsite}
\title{USGS Site File Data Retrieval}
\usage{
readNWISsite(siteNumbers)
}
\arguments{
\item{siteNumbers}{character USGS site number (or multiple sites).  This is usually an 8 digit number}
}
\value{
A data frame with at least the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
station_nm \tab character \tab Site name \cr
site_tp_cd \tab character \tab Site type \cr
lat_va \tab numeric \tab DMS latitude \cr
long_va \tab numeric \tab DMS longitude \cr
dec_lat_va \tab numeric \tab Decimal latitude \cr
dec_long_va \tab numeric \tab Decimal longitude \cr
coord_meth_cd  \tab character \tab Latitude-longitude method \cr
coord_acy_cd \tab character \tab Latitude-longitude accuracy \cr
coord_datum_cd \tab character \tab Latitude-longitude datum \cr
dec_coord_datum_cd \tab character \tab Decimal Latitude-longitude datum \cr
district_cd \tab character \tab District code \cr
state_cd \tab character \tab State code \cr
county_cd \tab character \tab County code \cr
country_cd \tab character \tab Country code \cr
land_net_ds \tab character \tab Land net location description \cr
map_nm \tab character \tab Name of location map \cr
map_scale_fc \tab character \tab Scale of location map \cr
alt_va \tab numeric \tab Altitude of Gage/land surface \cr
alt_meth_cd \tab character \tab Method altitude determined \cr
alt_acy_va  \tab numeric \tab Altitude accuracy \cr
alt_datum_cd \tab character \tab Altitude datum \cr
huc_cd \tab character \tab Hydrologic unit code \cr
basin_cd \tab character \tab Drainage basin code \cr
topo_cd \tab character \tab Topographic setting code \cr
instruments_cd \tab character \tab Flags for instruments at site \cr
construction_dt \tab character \tab Date of first construction \cr
inventory_dt \tab character \tab Date site established or inventoried \cr
drain_area_va \tab numeric \tab Drainage area \cr
contrib_drain_area_va \tab numeric \tab Contributing drainage area \cr
tz_cd  \tab character \tab Time Zone abbreviation \cr
local_time_fg \tab character \tab Site honors Daylight Savings Time \cr
reliability_cd \tab character \tab Data reliability code \cr
gw_file_cd \tab character \tab Data-other GW files \cr
nat_aqfr_cd \tab character \tab National aquifer code \cr
aqfr_cd \tab character \tab Local aquifer code \cr
aqfr_type_cd \tab character \tab Local aquifer type code \cr
well_depth_va \tab numeric \tab Well depth \cr
hole_depth_va \tab numeric \tab Hole depth \cr
depth_src_cd \tab character \tab Source of depth data \cr
project_no \tab character \tab Project number \cr
}

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
comment \tab character \tab Header comments from the RDB file \cr
}
}
\description{
Imports data from USGS site file site. This function gets data from here: \url{https://waterservices.usgs.gov/}
}
\examples{

# see ?read_waterdata_monitoring_location
# siteINFOMulti <- readNWISsite(c("05114000", "09423350"))

}
\seealso{
\code{\link[=read_waterdata_monitoring_location]{read_waterdata_monitoring_location()}}
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_metadata.R
\name{read_waterdata_metadata}
\alias{read_waterdata_metadata}
\title{Generalized USGS Water Meta Data API retrieval function}
\usage{
read_waterdata_metadata(
  collection,
  ...,
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
)
}
\arguments{
\item{collection}{character, can be any existing collection such
as "parameter-codes", "agency-codes", "altitude-datums", "aquifer-codes",
"aquifer-types", "coordinate-accuracy-codes", "coordinate-datum-codes",
"coordinate-method-codes", "hydrologic-unit-codes", "medium-codes",
"national-aquifer-codes", "reliability-codes", "site-types", "statistic-codes",
"topographic-codes", "time-zone-codes".}

\item{\dots}{Optional arguments to pass to the query. Available parameters
can be found with the \code{get_ogc_params} function.}

\item{limit}{The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}
}
\description{
Function to get metadata from Water Data API. These are useful to get the
human readable words and other metadata associated with USGS codes.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
agency_codes <- read_waterdata_metadata("agency-codes")
altitude_datums <- read_waterdata_metadata("altitude-datums")
aquifer_codes <- read_waterdata_metadata("aquifer-codes")
aquifer_types <- read_waterdata_metadata("aquifer-types")
counties <- read_waterdata_metadata("counties")
countries <- read_waterdata_metadata("countries")
us_counties <- read_waterdata_metadata("counties", country_code = "US")
coordinate_accuracy_codes <- read_waterdata_metadata("coordinate-accuracy-codes")
coordinate_datum_codes <- read_waterdata_metadata("coordinate-datum-codes")
coordinate_method_codes <- read_waterdata_metadata("coordinate-method-codes")
huc_codes <- read_waterdata_metadata("hydrologic-unit-codes")
methods <- read_waterdata_metadata("methods")
method_categories <- read_waterdata_metadata("method-categories")
method_citations <- read_waterdata_metadata("method-citations")
citations <- read_waterdata_metadata("citations")
national_aquifer_codes <- read_waterdata_metadata("national-aquifer-codes")
parameter_codes <- read_waterdata_metadata("parameter-codes")
reliability_codes <- read_waterdata_metadata("reliability-codes")
site_types <- read_waterdata_metadata("site-types")
states <- read_waterdata_metadata("states")
us_states_territories <- read_waterdata_metadata("states", country_code = "US")
statistic_codes <- read_waterdata_metadata("statistic-codes")
topographic_codes <- read_waterdata_metadata("topographic-codes")
time_zone_codes <- read_waterdata_metadata("time-zone-codes")
time_zone_limited <- read_waterdata_metadata("time-zone-codes",
                        time_zone_description = c("Alaska", "Hawaii", "Pacific North America"))
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whatNWISdata.R
\name{whatNWISdata}
\alias{whatNWISdata}
\title{USGS data availability}
\usage{
whatNWISdata(..., convertType = TRUE)
}
\arguments{
\item{\dots}{see \url{https://waterservices.usgs.gov/docs/site-service/}
for a complete list of options.  A list of arguments can also be supplied.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function will
convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
station_nm \tab character \tab Site name \cr
site_tp_cd \tab character \tab Site type \cr
dec_lat_va \tab numeric \tab Decimal latitude\cr
dec_long_va \tab numeric \tab Decimal longitude \cr
coord_acy_cd \tab character \tab Latitude-longitude accuracy \cr
dec_coord_datum_cd \tab character \tab Decimal Latitude-longitude datum \cr
alt_va \tab character \tab Altitude of Gage or land surface \cr
alt_acy_va \tab character \tab Altitude accuracy \cr
alt_datum_cd \tab character \tab Altitude datum \cr
huc_cd \tab character \tab Hydrologic unit code \cr
data_type_cd \tab character \tab Data type \cr
parm_cd \tab character \tab Parameter code \cr
stat_cd \tab character \tab Statistical code \cr
dd_nu \tab character \tab Internal database key \cr
loc_web_ds \tab character \tab Additional measurement description \cr
medium_grp_cd \tab character \tab Medium group code \cr
parm_grp_cd \tab character \tab Parameter group code \cr
srs_id \tab character \tab SRS ID \cr
access_cd \tab character \tab Access code \cr
begin_date \tab Date \tab Begin date \cr
end_date \tab Date \tab End date \cr
count_nu \tab integer \tab Record count\cr
parameter_group_nm \tab character \tab Parameter group name \cr
parameter_nm \tab character \tab Parameter name \cr
casrn \tab character \tab Chemical Abstracts Service (CAS) Registry Number \cr
srsname \tab character \tab Substance Registry Services \cr
parameter_units \tab character \tab Parameter units \cr
}

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
comment \tab character \tab Header comments from the RDB file \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Imports a table of available parameters, period of record, and count. See
\url{https://waterservices.usgs.gov/docs/site-service/}
for more information.
}
\details{
This function requires users to create their own arguments
based on the NWIS web services. It is a more complicated function to use
compared to other NWIS functions such as \code{\link[=readNWISdv]{readNWISdv()}}, \code{\link[=readNWISuv]{readNWISuv()}},
etc. However, this function adds a lot of
flexibility to the possible queries. If the "service" argument is included,
the results will be filtered to the proper data_type_cd. This is a great
function to use before a large data set, by filtering down the number
of sites that have useful data.
}
\examples{

# see ?read_waterdata_ts_meta

#site1 <- whatWQPsamples(siteid = "USGS-01594440")

#type <- "Stream"

#sites <- whatWQPsamples(countycode = "US:55:025", siteType = type)

#lakeSites_samples <- whatWQPsamples(siteType = "Lake, Reservoir, Impoundment",
#                                    countycode = "US:55:025")


}
\seealso{
\code{\link[=read_waterdata_ts_meta]{read_waterdata_ts_meta()}}
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNGWMNdata.R
\name{readNGWMNsites}
\alias{readNGWMNsites}
\title{Retrieve site data from the National Ground Water Monitoring Network.}
\usage{
readNGWMNsites(siteNumbers)
}
\arguments{
\item{siteNumbers}{character Vector of feature IDs formatted with agency code and site number
separated by a period or semicolon, e.g. \code{USGS.404159100494601}.}
}
\value{
A data frame the following columns:
#' \tabular{lll}{
Name \tab Type \tab Description \cr
site \tab char \tab Site FID \cr
description \tab char \tab Site description \cr
dec_lat_va, dec_lon_va \tab numeric \tab Site latitude and longitude \cr
}
}
\description{
Retrieve site data from the National Ground Water Monitoring Network.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
# one site
site <- "USGS.430427089284901"
#oneSite <- readNGWMNsites(siteNumbers = site)

# non-USGS site
site <- "MBMG.103306"
#siteInfo <- readNGWMNsites(siteNumbers = site)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkWQPdates.R
\name{checkWQPdates}
\alias{checkWQPdates}
\title{Date Check for Water Quality Portal}
\usage{
checkWQPdates(values)
}
\arguments{
\item{values}{named list with arguments to send to the Water Quality Portal}
}
\value{
values named list with corrected arguments to send to the Water Quality Portal
}
\description{
Checks date format for inputs to the Water Quality Portal. Used in \code{readWQPqw}
and \code{readWQPdata}.
}
\examples{
values <- list(
  startDateLo = "01-01-2002",
  characteristicName = "Phosphorous",
  endDate = as.Date("2014-01-01")
)
values <- checkWQPdates(values)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importNGWMN_wml2.R
\name{importWaterML2}
\alias{importWaterML2}
\title{Parse the WaterML2 timeseries portion of a waterML2 file}
\usage{
importWaterML2(input, asDateTime = FALSE, tz = "UTC")
}
\arguments{
\item{input}{XML with only the wml2:MeasurementTimeseries node and children}

\item{asDateTime}{logical, if \code{TRUE} returns date and time as POSIXct, if \code{FALSE}, character}

\item{tz}{character to set timezone attribute of datetime. Default is an empty quote, which converts the
datetimes to UTC (properly accounting for daylight savings times based on the data's provided time zone offset).
Possible values are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
"America/Anchorage", "America/Honolulu", "America/Jamaica", "America/Managua",
"America/Phoenix", and "America/Metlakatla"}
}
\description{
Returns data frame columns of all information with each time series measurement;
Anything defined as a default, is returned as an attribute of that data frame.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
baseURL <- httr2::request("https://waterservices.usgs.gov/nwis/dv")
baseURL <- httr2::req_url_query(baseURL,
                                format = "waterml,2.0",
                                sites = "01646500",
                                startDT = "2014-09-01",
                                endDT = "2014-09-08",
                                statCd = "00003",
                                parameterCd = "00060" )

timeseries <- importWaterML2(baseURL, asDateTime = TRUE, tz = "UTC")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zeroPad.R
\name{zeroPad}
\alias{zeroPad}
\title{Pad string with leading zeros}
\usage{
zeroPad(x, padTo)
}
\arguments{
\item{x}{character}

\item{padTo}{number Final desired length of the character}
}
\value{
x character returned with leading zeros
}
\description{
Function to pad a string with leading zeros. Useful for parameter codes and USGS site IDs.
}
\examples{
pCode <- "10"
correctPCode <- zeroPad(pCode, 5)
pCodes <- c("100", "1000", "0", "12345", "1565465465465465")
correctPCodes <- zeroPad(pCodes, 5)
pCodeNA <- c(1, 2, NA)
padPCodeNA <- zeroPad(pCodeNA, 4)
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISdata.R
\name{countyCdLookup}
\alias{countyCdLookup}
\title{US county code look up}
\usage{
countyCdLookup(state, county, outputType = "fips")
}
\arguments{
\item{state}{could be character (full name, abbreviation, id), or numeric (id)}

\item{county}{could be character (name, with or without "County") or numeric (id)}

\item{outputType}{character can be "fullName", "tableIndex", "id", or "fullEntry".}
}
\description{
Function to simplify finding county and county code definitions. Used in \code{readNWISdata}
and \code{readNWISuse}. Currently only has US counties.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{

fips <- countyCdLookup(state = "WI", county = "Dane")
id <- countyCdLookup(state = "WI", county = "Dane", outputType = "id")
name <- countyCdLookup(state = "OH", county = 13, output = "fullName")
entry <- countyCdLookup(state = "Pennsylvania", county = "ALLEGHENY COUNTY", output = "fullEntry")
fromIDs <- countyCdLookup(state = 13, county = 5, output = "fullName")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addWaterYear.R
\name{addWaterYear}
\alias{addWaterYear}
\title{add a water year column}
\usage{
addWaterYear(rawData)
}
\arguments{
\item{rawData}{the daily- or unit-values datset retrieved from NWISweb. Must
have at least one of the following columns to add the new water year columns:
\code{dateTime}, \code{Date}, \code{ActivityStartDate}, or \code{ActivityEndDate}. The date column(s)
can be character, POSIXct, Date. They cannot be numeric.}
}
\value{
data.frame with an additional integer column with "WY" appended to the
date column name. For WQP, there will be 2 columns: \code{ActivityStartDateWY} and
\code{ActivityEndDateWY}.
}
\description{
Add a column to the dataRetrieval data frame with the water year. WQP
queries will return a water year column for the start and end dates
of the data.
}
\examples{
\donttest{
nwisData <- readNWISdv("04085427", "00060", "2022-01-01", "2022-06-30")
nwisData <- addWaterYear(nwisData)

wqpData <- readWQPqw("USGS-01594440", "01075", "", "")
wqpData <- addWaterYear(wqpData)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getWebServiceData.R
\name{getWebServiceData}
\alias{getWebServiceData}
\title{Function to return data from web services}
\usage{
getWebServiceData(obs_url, ...)
}
\arguments{
\item{obs_url}{character containing the url for the retrieval}

\item{\dots}{information to pass to header request}
}
\value{
Returns xml, json, or text depending on the requested data.
}
\description{
This function accepts a url parameter, and returns the raw data.
}
\details{
To add a custom user agent, create an environmental variable: CUSTOM_DR_UA
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
siteNumber <- "02177000"
startDate <- "2012-09-01"
endDate <- "2012-10-01"
offering <- "00003"
property <- "00060"
obs_url <- constructNWISURL(siteNumber, property, startDate, endDate, "dv")
\donttest{
rawData <- getWebServiceData(obs_url)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ogc_documentation.R
\name{get_ogc_params}
\alias{get_ogc_params}
\title{Get parameter descriptions}
\usage{
get_ogc_params(service)
}
\arguments{
\item{service}{Character, can be any of the USGS Waterdata API endpoints or collections.}
}
\value{
list
}
\description{
This function returns a list of properties available from an endpoint. When
available, it will also contain a description.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
ml <- dataRetrieval::get_ogc_params("monitoring-locations")
ml$national_aquifer_code
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRetrieval-package.R
\docType{package}
\name{dataRetrieval}
\alias{dataRetrieval}
\alias{dataRetrieval-package}
\title{Retrieval functions for USGS and EPA data}
\description{
\tabular{ll}{
Package: \tab dataRetrieval\cr
Type: \tab Package\cr
License: \tab Unlimited for this package, dependencies have more restrictive licensing.\cr
Copyright: \tab This software is in the public domain because it contains materials
that originally came from the United States Geological Survey, an agency of
the United States Department of Interior. \cr
LazyLoad: \tab yes\cr
}
}
\details{
Retrieval functions for USGS and EPA hydrologic and water quality data.

Please see \url{https://doi-usgs.github.io/dataRetrieval/} for more information.
}
\seealso{
Useful links:
\itemize{
  \item Report bugs at \url{https://github.com/DOI-USGS/dataRetrieval/issues}
}

}
\author{
Laura De Cicco \email{ldecicco@usgs.gov}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRetrieval-package.R
\docType{data}
\name{countyCd}
\alias{countyCd}
\title{US County Code Lookup Table}
\value{
countyCd data frame.

\tabular{lll}{
Name \tab Type \tab Description\cr
STUSAB \tab character \tab State abbreviation \cr
STATE \tab character \tab two-digit ANSI code  \cr
COUNTY \tab character \tab three-digit county code \cr
COUNTY_NAME \tab character \tab County full name \cr
COUNTY_ID \tab character \tab County id \cr
}
}
\description{
Classic lookup table for counties. Has been replaced in functions with
\code{check_waterdata_sample_params("counties")}.
}
\examples{
head(countyCd)
}
\keyword{USGS}
\keyword{countyCd}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructNWISURL.R
\name{constructNWISURL}
\alias{constructNWISURL}
\title{Construct NWIS url for data retrieval}
\usage{
constructNWISURL(
  siteNumbers,
  parameterCd = "00060",
  startDate = "",
  endDate = "",
  service,
  statCd = "00003",
  format = "xml",
  expanded = TRUE,
  ratingType = "base",
  statReportType = "daily",
  statType = "mean"
)
}
\arguments{
\item{siteNumbers}{string or vector of strings USGS site number.  This is usually an 8 digit number}

\item{parameterCd}{string or vector of USGS parameter code.  This is usually an 5 digit number.}

\item{startDate}{character starting date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the earliest possible record.}

\item{endDate}{character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the latest possible record.}

\item{service}{string USGS service to call. Possible values are "dv" (daily values),
"uv" (unit/instantaneous values), and "rating" (rating curve),
"peak", "stat" (statistics web service BETA).}

\item{statCd}{string or vector USGS statistic code only used for daily value service.
This is usually 5 digits.  Daily mean (00003) is the default.}

\item{format}{string, can be "tsv" or "xml", and is only applicable for daily
and unit value requests.  "tsv" returns results faster, but there is a possibility
that an incomplete file is returned without warning. XML is slower,
but will offer a warning if the file was incomplete (for example, if there was a
momentary problem with the internet connection). It is possible to safely use the "tsv" option,
but the user must carefully check the results to see if the data returns matches
what is expected. The default is therefore "xml".}

\item{expanded}{logical defaults to \code{TRUE}. If \code{TRUE}, retrieves additional
information, only applicable for qw data.}

\item{ratingType}{can be "base", "corr", or "exsa". Only applies to rating curve data.}

\item{statReportType}{character Only used for statistics service requests.  Time
division for statistics: daily, monthly, or annual.  Default is daily.
Note that daily provides statistics for each calendar day over the specified
range of water years, i.e. no more than 366 data points will be returned for
each site/parameter.  Use \code{readNWISdata} or \code{readNWISdv} for daily averages.
Also note that "annual" returns statistics for the calendar year.  Use
\code{readNWISdata} for water years. Monthly and yearly
provide statistics for each month and year within the range individually.}

\item{statType}{character Only used for statistics service requests. Type(s)
of statistics to output for daily values.  Default is mean, which is the only
option for monthly and yearly report types.}
}
\value{
url string
}
\description{
Using USGS water web services to construct urls.
}
\examples{
site_id <- "01594440"
startDate <- "1985-01-01"
endDate <- ""
pCode <- c("00060", "00010")
url_daily <- constructNWISURL(site_id, pCode,
  startDate, endDate, "dv",
  statCd = c("00003", "00001")
)
url_unit <- constructNWISURL(site_id, pCode, "2012-06-28", "2012-06-30", "iv")

url_daily_tsv <- constructNWISURL(site_id, pCode, startDate, endDate, "dv",
  statCd = c("00003", "00001"), format = "tsv"
)
url_rating <- constructNWISURL(site_id, service = "rating", ratingType = "base")
url_peak <- constructNWISURL(site_id, service = "peak")
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNGWMNdata.R
\name{readNGWMNlevels}
\alias{readNGWMNlevels}
\title{Retrieve groundwater levels from the National Ground Water Monitoring Network.}
\usage{
readNGWMNlevels(siteNumbers, asDateTime = TRUE, tz = "UTC")
}
\arguments{
\item{siteNumbers}{character Vector of feature IDs formatted with agency code and site number
separated by a period or semicolon, e.g. \code{USGS.404159100494601}.}

\item{asDateTime}{logical Should dates and times be converted to date/time objects,
or returned as character?  Defaults to \code{TRUE}.  Must be set to \code{FALSE} if a site
contains non-standard dates.}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the data's provided time zone offset.
Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
"America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  \code{OlsonNames()}
for more information on time zones.}
}
\description{
Retrieve groundwater levels from the National Ground Water Monitoring Network.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
# one site
site <- "USGS.430427089284901"
# oneSite <- readNGWMNlevels(siteNumbers = site)

# multiple sites
sites <- c("USGS:272838082142201", "USGS:404159100494601", "USGS:401216080362703")
# multiSiteData <- readNGWMNlevels(sites)

# non-USGS site
site <- "MBMG.103306"
# data <- readNGWMNlevels(siteNumbers = site, asDateTime = FALSE)

# site with no data returns empty data frame
noDataSite <- "UTGS.401544112060301"
# noDataSite <- readNGWMNlevels(siteNumbers = noDataSite)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importNGWMN_wml2.R
\name{importNGWMN}
\alias{importNGWMN}
\title{Function to return data from the National Ground Water Monitoring Network waterML2 format}
\usage{
importNGWMN(input, asDateTime = FALSE, tz = "UTC")
}
\arguments{
\item{input}{character or raw, containing the url for the retrieval or a path to the data file, or raw XML.}

\item{asDateTime}{logical, if \code{TRUE} returns date and time as POSIXct, if \code{FALSE}, character}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the data's provided time zone offset.
Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
"America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  \code{OlsonNames()}
for more information on time zones.}
}
\value{
mergedDF a data frame source, time, value, uom, uomTitle, comment, gmlID
}
\description{
This function accepts a url parameter for a WaterML2 getObservation. This function is still under development,
but the general functionality is correct.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{

params <- list(request = "GetObservation",
               service = "SOS",
               version = "2.0.0",
               observedProperty = "urn:ogc:def:property:OGC:GroundWaterLevel",
               responseFormat = "text/xml",
               featureOfInterest = "VW_GWDP_GEOSERVER.USGS.403836085374401")

obs_url <- httr2::request("https://cida.usgs.gov") |>
 httr2::req_url_path_append("ngwmn_cache") |>
 httr2::req_url_path_append("sos") |>
 httr2::req_url_query(!!!params)

#data_returned <- importNGWMN(obs_url)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/citations.R
\name{create_NWIS_bib}
\alias{create_NWIS_bib}
\title{Create NWIS data citation}
\usage{
create_NWIS_bib(x)
}
\arguments{
\item{x}{Any data returned from an NWIS function, must
include "queryTime" and "url" attributes, which should
come with the data by default.}
}
\value{
bibentry object to use for citing the data.
}
\description{
Uses attributes from the NWIS functions to create data citations.
}
\details{
See \code{?bibentry} for more information.
}
\examples{
\donttest{
nwisData <- read_waterdata_daily(monitoring_location_id = "USGS-04085427",
                                 parameter_code = "00060",
                                 time = c("2012-01-01", "2012-06-30"))
nwis_citation <- create_NWIS_bib(nwisData)
nwis_citation

print(nwis_citation, style = "Bibtex")
print(nwis_citation, style = "citation")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_channel.R
\name{read_waterdata_channel}
\alias{read_waterdata_channel}
\title{Get USGS Channel Measurements}
\usage{
read_waterdata_channel(
  monitoring_location_id = NA_character_,
  field_visit_id = NA_character_,
  measurement_number = NA_character_,
  time = NA_character_,
  channel_name = NA_character_,
  channel_flow = NA_character_,
  channel_flow_unit = NA_character_,
  channel_width = NA_character_,
  channel_width_unit = NA_character_,
  channel_area = NA_character_,
  channel_area_unit = NA_character_,
  channel_velocity = NA_character_,
  channel_velocity_unit = NA_character_,
  channel_location_distance = NA_character_,
  channel_location_distance_unit = NA_character_,
  channel_location_direction = NA_character_,
  channel_stability = NA_character_,
  channel_material = NA_character_,
  channel_evenness = NA_character_,
  horizontal_velocity_description = NA_character_,
  vertical_velocity_description = NA_character_,
  longitudinal_velocity_description = NA_character_,
  measurement_type = NA_character_,
  last_modified = NA_character_,
  channel_measurement_type = NA_character_,
  properties = NA_character_,
  skipGeometry = NA,
  bbox = NA,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  no_paging = getOption("dataRetrieval.no_paging"),
  limit = getOption("dataRetrieval.limit"),
  chunk_size = getOption("dataRetrieval.site_chunk_size_data"),
  attach_request = getOption("dataRetrieval.attach_request")
)
}
\arguments{
\item{monitoring_location_id}{A unique identifier representing a single monitoring location. This corresponds to the \code{id} field in the \code{monitoring-locations} endpoint. Monitoring location IDs are created by combining the agency code of the agency responsible for the monitoring location (e.g. USGS) with the ID number of the monitoring location (e.g. 02238500), separated by a hyphen (e.g. USGS-02238500).

Multiple monitoring_location_ids can be requested as a character vector.}

\item{field_visit_id}{A universally unique identifier (UUID) for the field visit. Multiple measurements may be made during a single field visit.}

\item{measurement_number}{Measurement number.}

\item{time}{The date an observation represents. You can query this field using date-times or intervals, adhering to RFC 3339, or using ISO 8601 duration objects. Intervals may be bounded or half-bounded (double-dots at start or end).
Examples:
\itemize{
\item A date-time: "2018-02-12T23:20:50Z"
\item A bounded interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
\item Half-bounded intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"
\item Duration objects: "P1M" for data from the past month or "PT36H" for the last 36 hours
}

Only features that have a \code{time} that intersects the value of datetime are selected. If a feature has multiple temporal properties, it is the decision of the server whether only a single temporal property is used to determine the extent or all relevant temporal properties.}

\item{channel_name}{The channel name.}

\item{channel_flow}{Channel discharge.}

\item{channel_flow_unit}{The units for channel discharge.}

\item{channel_width}{The channel width.}

\item{channel_width_unit}{The units for channel width.}

\item{channel_area}{The channel area.}

\item{channel_area_unit}{The units for channel area.}

\item{channel_velocity}{The mean channel velocity.}

\item{channel_velocity_unit}{The units for channel velocity.}

\item{channel_location_distance}{The channel location distance.}

\item{channel_location_distance_unit}{The units for channel location distance.}

\item{channel_location_direction}{Location of the measurement from the gage.}

\item{channel_stability}{The stability of the channel material.}

\item{channel_material}{The channel material.}

\item{channel_evenness}{The channel evenness from bank to bank.}

\item{horizontal_velocity_description}{The horizontal velocity description.}

\item{vertical_velocity_description}{The vertical velocity description.}

\item{longitudinal_velocity_description}{The longitudinal velocity description.}

\item{measurement_type}{The measurement type.}

\item{last_modified}{The last time a record was refreshed in our database. This may happen due to regular operational processes and does not necessarily indicate anything about the measurement has changed.
You can query this field using date-times or intervals, adhering to RFC 3339, or using ISO 8601 duration objects. Intervals may be bounded or half-bounded (double-dots at start or end).
Examples:
\itemize{
\item A date-time: "2018-02-12T23:20:50Z"
\item A bounded interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
\item Half-bounded intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"
\item Duration objects: "P1M" for data from the past month or "PT36H" for the last 36 hours
}

Only features that have a \code{last_modified} that intersects the value of datetime are selected.

field_visit_id = NA_character_,}

\item{channel_measurement_type}{The channel measurement type.}

\item{properties}{A vector of requested columns to be returned from the query.
Available options are:
geometry, channel_measurements_id, monitoring_location_id, field_visit_id, measurement_number, time, channel_name, channel_flow, channel_flow_unit, channel_width, channel_width_unit, channel_area, channel_area_unit, channel_velocity, channel_velocity_unit, channel_location_distance, channel_location_distance_unit, channel_stability, channel_material, channel_evenness, horizontal_velocity_description, vertical_velocity_description, longitudinal_velocity_description, measurement_type, last_modified, channel_measurement_type, channel_location_direction.
The default (\code{NA}) will return all columns of the data.}

\item{skipGeometry}{This parameter can be used to skip response geometries for
each feature. The returning object will be a data frame with no spatial
information. The default \code{NA} will not specify the argument in the request.}

\item{bbox}{Only features that have a geometry that intersects the bounding
box are selected.The bounding box is provided as four or six numbers, depending
on whether the coordinate reference system includes a vertical axis (height or
depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
vector structured: c(xmin,ymin,xmax,ymax).
Another way to think of it is c(Western-most longitude,
Southern-most latitude, Eastern-most longitude, Northern-most longitude).}

\item{...}{Not used. Included to help differentiate official Water Data API arguments
from more seldom used, optional dataRetrieval-specific arguments.}

\item{convertType}{logical, defaults to TRUE.
If \code{TRUE}, the function will convert the data to dates, any qualifiers to string
vector and reorder the returned data frame.}

\item{no_paging}{logical, defaults to FALSE.
If \code{TRUE}, the data will
be requested from a native csv format. This can be dangerous because the
data will cut off at 50,000 rows without indication that more data
is available. Use \code{TRUE} with caution.}

\item{limit}{numeric, The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50,000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{chunk_size}{Number of monitoring_location_ids to chunk requests into.
The default for functions that don't generally return long-term data records
is 250, while
the default for time series functions is
10.
Setting to \code{NA} will eliminate site chunking, giving users full control.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}
}
\description{
Channel measurements taken as part of streamflow field measurements.
}
\details{
You can also use a vector of length 2 for any time queries (such as time
or last_modified). The first value is the starting date (or datetime),
the second value is the ending date(or datetime).
NA's within the vector indicate a half-bound date.
For example, \code{time = c("2024-01-01", NA)} will return all data starting
at 2024-01-01.
\code{time = c(NA, "2024-01-01")} will return all data from the beginning of
the timeseries until 2024-01-01.
By default, time is assumed UTC, although time zone attributes
will be accommodated. As an example, setting \code{time = as.POSIXct(c("2021-01-01 12:00:00",
"2021-01-01 14:00"), tz = "America/New_York")} will request data that between
noon and 2pm eastern time on 2021-01-01.
All time values RETURNED from the service are UTC with the exception of
daily data, which returns time values in local dates.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
site <- "USGS-02238500"
df <- read_waterdata_channel(monitoring_location_id = site)

}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_ratings.R
\name{read_waterdata_ratings}
\alias{read_waterdata_ratings}
\title{Get USGS Rating Curve Data}
\usage{
read_waterdata_ratings(
  monitoring_location_id = NA_character_,
  file_type = c("exsa", "base", "corr"),
  file_path = tempdir(),
  bbox = NA,
  datetime = NA_character_,
  ...,
  limit = 10000,
  download_and_parse = TRUE
)
}
\arguments{
\item{monitoring_location_id}{A unique identifier representing a single
monitoring location. Monitoring location IDs are created by combining the
agency code of the agency responsible for the monitoring location (e.g. USGS)
with the ID number of the monitoring location (e.g. 02238500), separated by
a hyphen (e.g. USGS-02238500).}

\item{file_type}{Rating file time. Could be any of "exsa", "corr", or "base".
If \code{file_type} is "base" then the columns are
INDEP, typically the gage height, in feet; DEP, typically the streamflow,
in cubic feet per second; and STOR, where "*" indicates that the pair are
a fixed point of the rating curve. If \code{file_type} is "exsa" then an
additional column, SHIFT, is included that indicates the current shift in
the rating for that value of INDEP. If \code{file_type} is "corr" then the
columns are INDEP, typically the gage height, in feet; CORR, the correction
for that value; and CORRINDEP, the corrected value for CORR.}

\item{file_path}{Path to save the rating curve rdb files. The
default is \code{tempdir()}, which will wipe out the files.}

\item{bbox}{Only features that have a geometry that intersects the bounding
box are selected.The bounding box is provided as four or six numbers, depending
on whether the coordinate reference system includes a vertical axis (height or
depth). Coordinates are assumed to be in crs 4326. The expected format is a numeric
vector structured: c(xmin,ymin,xmax,ymax).
Another way to think of it is c(Western-most longitude,
Southern-most latitude, Eastern-most longitude, Northern-most longitude).}

\item{datetime}{Only return items that have a temporal property that
intersects this value. Either a date-time or an interval, open or closed.
See Details below.}

\item{\dots}{Not used.}

\item{limit}{Limits the number of results that are included in each page of
the response (capped at the default 10,000).}

\item{download_and_parse}{Logical to define whether or not to download, parse,
and return a list of data frames with rating curve data (\code{TRUE}), or to return
just a list of available rating curve files (\code{FALSE}). Default is \code{TRUE}.}
}
\value{
List of data frames which contain the requested rating curves.
}
\description{
Reads current rating table for an active USGS streamgages. More information
can be found at https://api.waterdata.usgs.gov/docs/stac/.
}
\details{
You can also use a vector of length 2 for any time queries (such as time
or last_modified). The first value is the starting date (or datetime),
the second value is the ending date(or datetime).
NA's within the vector indicate a half-bound date.
For example, \code{time = c("2024-01-01", NA)} will return all data starting
at 2024-01-01.
\code{time = c(NA, "2024-01-01")} will return all data from the beginning of
the timeseries until 2024-01-01.
By default, time is assumed UTC, although time zone attributes
will be accommodated. As an example, setting \code{time = as.POSIXct(c("2021-01-01 12:00:00",
"2021-01-01 14:00"), tz = "America/New_York")} will request data that between
noon and 2pm eastern time on 2021-01-01.
All time values RETURNED from the service are UTC with the exception of
daily data, which returns time values in local dates.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{

monitoring_location_id <- c("USGS-01104475", "USGS-01104460")
ratings_exsa <- read_waterdata_ratings(
      monitoring_location_id = monitoring_location_id,
      file_type = "exsa")

head(ratings_exsa[["USGS-01104475.exsa.rdb"]])
comment(ratings_exsa[["USGS-01104475.exsa.rdb"]])[1:15]

ratings_corr <- read_waterdata_ratings(
      monitoring_location_id = monitoring_location_id,
      file_type = "corr")

head(ratings_corr[["USGS-01104460.corr.rdb"]])
comment(ratings_corr[["USGS-01104460.corr.rdb"]])[1:15]

rating_2 <- read_waterdata_ratings(
      monitoring_location_id = monitoring_location_id,
      file_type = c("corr", "exsa"))
names(rating_2)

bbox <- c(-95.00, 40.0, -92.0, 42)

bbox_query <- read_waterdata_ratings(bbox = bbox,
                                     download_and_parse = FALSE)
length(bbox_query)
recent_query <- read_waterdata_ratings(bbox = bbox,
                                       datetime = c(Sys.Date()-7, NA),
                                       download_and_parse = FALSE)
length(recent_query)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWQPdata.R
\name{wqp_check_status}
\alias{wqp_check_status}
\title{Get WQP service metadata}
\usage{
wqp_check_status(wqp_request_id)
}
\arguments{
\item{wqp_request_id}{A character returned from the header
of a WQP request.}
}
\value{
a list generated from the WQP describing what data
was returned.
}
\description{
The information from this request is only available for a
limited time after the original query from the WQP. In the
readWQPdata and readWQPqw functions, the results from this
function will be attached as an attribute to the data.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
#rawPcode <- readWQPqw("USGS-01594440", "01075",
#                       ignore_attributes = TRUE, legacy = FALSE)
#  headerInfo <- attr(rawPcode, "headerInfo")
#wqp_request_id <- headerInfo$`wqp-request-id`
#count_info <- wqp_check_status(wqp_request_id)
#count_info[["dataProviders"]]
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISdata.R
\name{readNWISdots}
\alias{readNWISdots}
\title{Format and organize NWIS arguments that are passed in as \code{...}.}
\usage{
readNWISdots(...)
}
\description{
Format and organize NWIS arguments that are passed in as \code{...}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/findNLDI.R
\name{findNLDI}
\alias{findNLDI}
\title{R Client for the Network Linked Data Index}
\usage{
findNLDI(
  comid = NULL,
  nwis = NULL,
  wqp = NULL,
  huc12 = NULL,
  location = NULL,
  origin = NULL,
  nav = NULL,
  find = c("flowlines"),
  distance_km = 100,
  no_sf = FALSE,
  warn = TRUE
)
}
\arguments{
\item{comid}{numeric or character. An NHDPlusV2 COMID}

\item{nwis}{numeric or character. A USGS NWIS surface water siteID}

\item{wqp}{numeric or character. A water quality point ID}

\item{huc12}{numeric or character. A WBD HUC12 unit ID}

\item{location}{numeric vector. Coordinate pair in WGS84
SRS ordered lng/lat (X,Y)}

\item{origin}{named list. Specifying a feature type and ID
(e.g. list("comid" = 101))}

\item{nav}{character vector. where to navigate from the starting point.
Options include along the upper mainsteam (UM), upstream tributary (UT),
downstream mainstem (DM) and downstream divergences (DD). You may select
one or more of the abbreviations ("UM", "UT", DM", "DD").}

\item{find}{character vector. Define what resources to find along the
navigation path(s) (see get_nldi_sources()$source). Can also include 'basin'
or 'flowline', which will return the upstream basin of the starting feature
or flowlines along the navigation respectively. The default is "flowlines".
If you provide any other resource, AND want flowlines, then flowlines must
be explicitly requested.}

\item{distance_km}{numeric. Define how far to look along the navigation path in
kilometers (default = 100)}

\item{no_sf}{if available, should \code{sf} be used for parsing,
defaults to \code{TRUE} if \code{sf} is locally installed}

\item{warn}{(default TRUE) should warnings be printed}
}
\value{
a list of data.frames if sf is not installed, a list of sf objects if it is
}
\description{
Provides a formal client to the USGS
Network Linked Data Index.
}
\details{
The function is useful for topology and location based
feature discovery. A user must specify an origin feature, optional navigation
direction(s) along the network, as well as features to identify along the
navigated paths. Valid starting options can be given by one of the following
arguments: comid, nwis, huc12, wqp, location, and start.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
# Find Features / Define origin features

## Find feature by COMID
findNLDI(comid = 101)

## Find feature by NWIS ID
findNLDI(nwis = "11120000")


## Find feature by LOCATION
findNLDI(location = c(-115, 40))

## GENERAL ORIGIN: COMID
findNLDI(origin = list("comid" = 101))

# Navigation (flowlines will be returned if find is unspecified)
# UPPER MAINSTEM of USGS-11120000
findNLDI(nwis = "11120000", nav = "UM")

# MULTI-REQUEST
# UPPER MAINSTEM and TRIBUTARY of USGS-11120000
findNLDI(nwis = "11120000", nav = c("UT", "UM"))

# Discover Features(flowlines will not be returned unless included in find)

## Find feature(s) on the upper tributary of USGS-11120000
findNLDI(nwis = "11120000", nav = "UT", find = c("nwis"))

## Find upstream basin boundary and  of USGS-11120000
findNLDI(nwis = "11120000", find = "basin")

# Control Distance
## Limit search to 50 km
findNLDI(comid = 101, nav = "DM", find = c("nwis", "flowlines"), distance_km = 50)
}
\dontshow{\}) # examplesIf}
}
\keyword{nldi}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISunit.R
\name{readNWISuse}
\alias{readNWISuse}
\title{Water use data retrieval from USGS (NWIS)}
\usage{
readNWISuse(
  stateCd,
  countyCd,
  years = "ALL",
  categories = "ALL",
  convertType = TRUE,
  transform = FALSE
)
}
\arguments{
\item{stateCd}{could be character (full name, abbreviation, id), or numeric (id).
Only one is accepted per query.}

\item{countyCd}{could be character (name, with or without "County", or "ALL"),
numeric (id), or \code{NULL}, which will
return state or national data depending on the stateCd argument.  "ALL" may
also be supplied, which will return data
for every county in a state. Can be a vector of counties in the same state.}

\item{years}{integer Years for data retrieval. Must be years ending in 0 or 5.
Default is all available years.}

\item{categories}{character categories of water use.  Defaults to "ALL".
Specific categories must be supplied as two-
letter abbreviations as seen in the URL when using the NWIS water use web interface.  Note that
there are different codes for national and state level data.}

\item{convertType}{logical defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to
numerics based on a standard algorithm. Years, months, and days (if appliccable) are
also returned as numerics
in separate columns.  If convertType is false, everything is returned as a character.}

\item{transform}{logical only intended for use with national data.  Defaults to
\code{FALSE}, with data being returned as
presented by the web service.  If \code{TRUE}, data will be transformed and
returned with column names, which will reformat
national data to be similar to state data.}
}
\value{
A data frame with at least the year of record, and all available
statistics for the given geographic parameters.
County and state fields will be included as appropriate.
}
\description{
Retrieves water use data from USGS Water Use Data for the Nation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/citations.R
\name{create_WQP_bib}
\alias{create_WQP_bib}
\title{Create WQP data citation}
\usage{
create_WQP_bib(x)
}
\arguments{
\item{x}{Any data returned from an NWIS function, must
include "queryTime" and "url" attributes, which should
come with the data by default.}
}
\value{
bibentry object to use for citing the data.
}
\description{
Uses attributes from the WQP functions to create data citations.
}
\details{
See \code{?bibentry} for more information.
}
\examples{
\donttest{
WQPData <- readWQPqw("USGS-05288705",
                     parameterCd = "00300")
wqp_citation <- create_WQP_bib(WQPData)
wqp_citation

print(wqp_citation, style = "Bibtex")
print(wqp_citation, style = "citation")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_parameter_codes.R
\name{read_waterdata_parameter_codes}
\alias{read_waterdata_parameter_codes}
\title{Get USGS Parameter Code Information}
\usage{
read_waterdata_parameter_codes(
  parameter_code = NA_character_,
  parameter_name = NA_character_,
  unit_of_measure = NA_character_,
  parameter_group_code = NA_character_,
  parameter_description = NA_character_,
  medium = NA_character_,
  statistical_basis = NA_character_,
  weight_basis = NA_character_,
  sample_fraction = NA_character_,
  temperature_basis = NA_character_,
  epa_equivalence = NA_character_,
  properties = NA_character_,
  ...,
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
)
}
\arguments{
\item{parameter_code}{Parameter code.}

\item{parameter_name}{Parameter short name.}

\item{unit_of_measure}{Parameter reporting units defined to cooperate with descriptions by USEPA.}

\item{parameter_group_code}{Categorical groupings of parameters by water-quality data type for display and report ordering}

\item{parameter_description}{Parameter description.}

\item{medium}{Parameter medium.}

\item{statistical_basis}{Parameter statistical basis.}

\item{weight_basis}{Parameter weight basis.}

\item{sample_fraction}{Parameter fraction.}

\item{temperature_basis}{Parameter temperature basis.}

\item{epa_equivalence}{Indicates the relationship of the USGS parameter code to the EPA code.}

\item{properties}{A vector of requested columns to be returned from the query.
Available options are:
geometry, parameter_code_id, parameter_name, unit_of_measure, parameter_group_code, parameter_description, medium, statistical_basis, time_basis, weight_basis, particle_size_basis, sample_fraction, temperature_basis, epa_equivalence.
The default (\code{NA}) will return all columns of the data.}

\item{\dots}{Not used. Included to help differentiate official Water Data API arguments
from more seldom used, optional dataRetrieval-specific arguments.}

\item{limit}{The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}
}
\description{
Parameter codes are 5-digit codes and associated descriptions used to identify the constituent measured and the units of measure. Some parameter code definitions include information about the sampling matrix, fraction, and methods used to measure the constituent. Some parameters are fixed-value (fxd) numeric codes having textual meaning (for example: parameter code 00041 is a weather code parameter, code of 60 means rain), but more commonly represent a numeric value for chemical, physical, or biological data.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
pcode <- "00060"
pcode_info <- read_waterdata_parameter_codes(parameter_code = pcode)

pcodes <- read_waterdata_parameter_codes(parameter_code = c("00660", "00060"))

# equivalent to read_waterdata_metadata("parameter-codes")
all_pcodes <- read_waterdata_parameter_codes()

total_nutrients <- read_waterdata_parameter_codes(parameter_group_code = "NUT",
                                           sample_fraction = "Total")

group_of_nutrients <- read_waterdata_parameter_codes(parameter_group_code = "NUT",
                                                    unit_of_measure = c("mg/l",
                                                    "mg/l as N", "mg/l NO3",
                                                    "mg/l asNO2"))

}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRetrieval-package.R
\docType{data}
\name{pCodeToName}
\alias{pCodeToName}
\title{Data to convert USGS parameter code to characteristic names}
\value{
pCodeToName data frame with information about USGS parameters and how they
relate to characteristic names (useful for WQP requests).

\tabular{lll}{
Name \tab Type \tab Description\cr
parm_cd \tab character \tab 5-digit USGS parameter code \cr
description \tab character \tab Parameter description\cr
characteristicname \tab character \tab Characteristic Name \cr
measureunitcode \tab character \tab Parameter units\cr
resultsamplefraction \tab character \tab Result sample fraction text\cr
resulttemperaturebasis \tab character \tab Temperature basis information\cr
resultstatisticalbasis \tab character \tab Statistical basis\cr
resulttimebasis \tab character \tab Time basis\cr
resultweightbasis \tab character \tab Weight basis\cr
resultparticlesizebasis \tab character \tab Particle size basis\cr
last_rev_dt \tab character \tab Latest revision of information\cr
}
}
\description{
Data pulled from Water Quality Portal on December 20, 2021. The data was pulled from
\url{https://www.waterqualitydata.us/Codes/public_srsnames/?mimeType=csv}.
}
\examples{
head(pCodeToName[, 1:2])
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_samples.R
\name{read_USGS_samples}
\alias{read_USGS_samples}
\title{Deprecated: Use \code{read_waterdata_samples} instead}
\usage{
read_USGS_samples(...)
}
\arguments{
\item{...}{Arguments passed to \code{\link{read_waterdata_samples}}.}
}
\value{
data frame returned from web service call.
}
\description{
This function has been renamed to \code{\link{read_waterdata_samples}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whatWQPdata.R
\name{whatWQPdata}
\alias{whatWQPdata}
\title{Data Available from Water Quality Portal}
\usage{
whatWQPdata(..., convertType = TRUE)
}
\arguments{
\item{\dots}{see \url{https://www.waterqualitydata.us/webservices_documentation} for
a complete list of options. A list of arguments can also be supplied.
One way to figure out how to construct a WQP query is to go to the "Advanced"
form in the Water Quality Portal:
\url{https://www.waterqualitydata.us/#mimeType=csv&providers=NWIS&providers=STORET}
Use the form to discover what parameters are available. Once the query is
set in the form, scroll down to the "Query URL". You will see the parameters
after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
in the Characteristic Group dropdown, you will see characteristicType=Nutrient
in the Query URL. The corresponding argument for dataRetrieval is
characteristicType = "Nutrient". dataRetrieval users do not need to include
mimeType, and providers is optional (these arguments are picked automatically).}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character.}
}
\value{
A data frame that returns basic data availability such as
sites, number of results, and number of sampling activities from the
query parameters for the Water Quality Portal.
}
\description{
Returns a list of sites from the Water Quality Portal web service. This function gets
the data from: \url{https://www.waterqualitydata.us}.
Arguments to the function should be based on
\url{https://www.waterqualitydata.us/webservices_documentation}.
The information returned from whatWQPdata describes the
available data at the WQP sites, and some metadata on the sites themselves.
For example, a row is returned for each individual site that fulfills this
query. In that we can learn how many sampling activities and results
are available for the query. It does not break those results down by any finer
grain. For example, if you ask for "Nutrients" (characteristicGroup), you will
not learn what specific nutrients are available at that site. For that
kind of data discovery see \code{readWQPsummary}.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
#site1 <- whatWQPdata(siteid = "USGS-01594440")

#type <- "Stream"
#sites <- whatWQPdata(countycode = "US:55:025", siteType = type)

#lakeSites <- whatWQPdata(siteType = "Lake, Reservoir, Impoundment",
#                        countycode = "US:55:025")
#lakeSites_chars <- whatWQPdata(
#   siteType = "Lake, Reservoir, Impoundment",
#   countycode = "US:55:025", convertType = FALSE)


#bbox <- c(-86.9736, 34.4883, -86.6135, 34.6562)
#what_bb <- whatWQPdata(bBox = bbox)
}
\dontshow{\}) # examplesIf}
}
\seealso{
whatWQPsites readWQPsummary readWQPdata
}
\keyword{WQP}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISpCode.R
\name{readNWISpCode}
\alias{readNWISpCode}
\title{USGS Parameter Data Retrieval}
\usage{
readNWISpCode(parameterCd)
}
\arguments{
\item{parameterCd}{character of USGS parameter codes (or multiple parameter codes).
These are 5 digit number codes. To get a complete list of all current parameter
codes in the USGS, use "all" as the input.}
}
\value{
parameterData data frame with the following information:
\tabular{lll}{
Name \tab Type \tab Description\cr
parameter_cd \tab character \tab 5-digit USGS parameter code \cr
parameter_group_nm \tab character \tab USGS parameter group name\cr
parameter_nm \tab character \tab USGS parameter name\cr
casrn \tab character \tab Chemical Abstracts Service (CAS) Registry Number\cr
srsname \tab character \tab Substance Registry Services Name\cr
parameter_units \tab character \tab Parameter units\cr
}
}
\description{
Imports data from NWIS about measured parameter based on user-supplied parameter code or codes.
}
\seealso{
\code{\link[=importRDB1]{importRDB1()}}
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISdata.R
\name{readNWISdata}
\alias{readNWISdata}
\title{General Data Import from NWIS}
\usage{
readNWISdata(..., asDateTime = TRUE, convertType = TRUE, tz = "UTC")
}
\arguments{
\item{\dots}{see \url{https://waterservices.usgs.gov/docs/site-service/} for
a complete list of options.  A
list of arguments can also be supplied. One important argument to include is
"service". Possible values are "iv"
(for instantaneous),
"dv" (for daily values),
"site" (for site service), and "stat" (for
statistics service).
The statistics service has a limited selection of arguments
(see \url{https://waterservices.usgs.gov/docs/site-service/}).}

\item{asDateTime}{logical, if \code{TRUE} returns date and time as POSIXct, if \code{FALSE}, Date}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the
function will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the data's provided tz_cd column.
Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
"America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  \code{OlsonNames()}
for more information on time zones.}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency \tab character \tab The NWIS code for the agency reporting the data\cr
site \tab character \tab The USGS site number \cr
dateTime \tab POSIXct \tab The date and time (if applicable) of the measurement,
converted to UTC for unit value data. R only allows one time zone attribute per column. For unit data
spanning a time zone change, converting the data to UTC solves this problem. For daily data,
the time zone attribute is the time zone of the first returned measurement.
\cr
tz_cd \tab character \tab The time zone code for dateTime column\cr
code \tab character \tab Any codes that qualify the corresponding value\cr
value \tab numeric \tab The numeric value for the parameter \cr
}
Note that code and value are repeated for the parameters requested. The names are of the form
X_D_P_S, where X is literal,
D is an option description of the parameter,
P is the parameter code,
and S is the statistic code (if applicable).

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Returns data from the NWIS web service.
Arguments to the function should be based on \url{https://waterservices.usgs.gov} service calls.
See examples below for ideas of constructing queries.
}
\details{
This function requires users to create their own arguments
based on the NWIS web services. It is a more complicated function to use
compared to other NWIS functions such as \code{\link[=readNWISdv]{readNWISdv()}}, \code{\link[=readNWISuv]{readNWISuv()}},
etc. However, this function adds a lot of
flexibility to the possible queries. This function will also behave exactly
as NWIS when it comes to date queries. NWIS by default will only return the latest
value for the daily and instantaneous services. So if you do not provide
a starting date, you will only get back the latest value. If you want
the full period of record, you can use "startDate = '1900-01-01'". Other options
for dates are periods, such as "period = 'P7D'" which translates to a period
of 7 days. For period, use only a positive ISO-8601 duration format, which should
not be expressed in periods of less than a day, or in increments of months M or years Y.
period returns data for a site generally from now to a time in the past.
Note that when period is used all data up to the most recent value are returned.
}
\seealso{
\code{\link[=read_waterdata]{read_waterdata()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getWebServiceData.R
\name{default_ua}
\alias{default_ua}
\title{Create user agent}
\usage{
default_ua()
}
\description{
Create user agent
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISdata.R
\name{convertLists}
\alias{convertLists}
\title{convert variables in dots to usable format}
\usage{
convertLists(...)
}
\description{
convert variables in dots to usable format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISunit.R
\name{readNWISstat}
\alias{readNWISstat}
\title{Site statistics retrieval from USGS (NWIS)}
\usage{
readNWISstat(
  siteNumbers,
  parameterCd,
  startDate = "",
  endDate = "",
  convertType = TRUE,
  statReportType = "daily",
  statType = "mean"
)
}
\arguments{
\item{siteNumbers}{character USGS site number (or multiple sites).  This is usually an 8 digit number.}

\item{parameterCd}{character USGS parameter code.  This is usually a 5 digit number.}

\item{startDate}{character starting date for data retrieval in the form YYYY, YYYY-MM, or YYYY-MM-DD. Dates cannot
be more specific than the statReportType, i.e. startDate for monthly statReportTypes cannot include days, and annual
statReportTypes cannot include days or months.  Months and days are optional for the daily statReportType.
Default is "" which indicates retrieval for the earliest possible record.  For daily data, this indicates the
start of the period the statistics will be computed over.}

\item{endDate}{character ending date for data retrieval in the form YYYY, YYYY-MM, or YYYY-MM-DD. Default is ""
which indicates retrieval for the latest possible record.  For daily data, this
indicates the end of the period
the statistics will be computed over.  The same restrictions as startDate apply.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function will convert the data to
numerics based on a standard algorithm. Years, months, and days (if appliccable) are also returned as numerics
in separate columns.  If convertType is false, everything is returned as a character.}

\item{statReportType}{character time division for statistics: daily, monthly, or annual.  Default is daily.
Note that daily provides statistics for each calendar day over the specified range
of water years, i.e. no more than 366
data points will be returned for each site/parameter.  Use readNWISdata or readNWISdv for daily averages.
Also note that 'annual' returns statistics for the calendar year.  Use readNWISdata
for water years. Monthly and yearly
provide statistics for each month and year within the range indivually.}

\item{statType}{character type(s) of statistics to output for daily values.
Default is mean, which is the only option for monthly and yearly report types.}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
parameter_cd \tab character \tab The USGS parameter code \cr

Other columns will be present depending on statReportType and statType
}
}
\description{
Retrieves site statistics from the USGS Statistics Web Service beta.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
# x1 <- readNWISstat(
#   siteNumbers = c("02319394"),
#   parameterCd = c("00060"),
#   statReportType = "annual"
# )

# all the annual mean discharge data for two sites
#x2 <- readNWISstat(
#   siteNumbers = c("02319394", "02171500"),
#   parameterCd = c("00010", "00060"),
#   statReportType = "annual"
# )

# Request p25, p75, and mean values for temperature and discharge for the 2000s
# Note that p25 and p75 were not available for temperature, and return NAs
#x <- readNWISstat(
#   siteNumbers = c("02171500"),
#   parameterCd = c("00010", "00060"),
#   statReportType = "daily",
#   statType = c("mean", "median"),
#   startDate = "2000", endDate = "2010"
# )
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=constructNWISURL]{constructNWISURL()}}, \code{\link[=importRDB1]{importRDB1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importWaterML1.R
\name{importWaterML1}
\alias{importWaterML1}
\title{Function to return data from the NWISWeb WaterML1.1 service}
\usage{
importWaterML1(obs_url, asDateTime = FALSE, tz = "UTC")
}
\arguments{
\item{obs_url}{character or raw, containing the url for the retrieval or a
file path to the data file, or raw XML.}

\item{asDateTime}{logical, if \code{TRUE} returns date and time as POSIXct, if \code{FALSE}, Date}

\item{tz}{character to set timezone attribute of datetime. Default converts the datetimes to UTC
(properly accounting for daylight savings times based on the data's provided tz_cd column).
Recommended US values include "UTC", "America/New_York", "America/Chicago", "America/Denver",
"America/Los_Angeles", "America/Anchorage", "America/Honolulu", "America/Jamaica", "America/Managua",
"America/Phoenix", and "America/Metlakatla".
For a complete list, see \url{https://en.wikipedia.org/wiki/List_of_tz_database_time_zones}}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
\tab POSIXct \tab The date and time of the value converted to UTC (if asDateTime = TRUE), \cr
\tab character \tab or raw character string (if asDateTime = FALSE) \cr
tz_cd \tab character \tab The time zone code for  \cr
code \tab character \tab Any codes that qualify the corresponding value\cr
value \tab numeric \tab The numeric value for the parameter \cr
}
Note that code and value are repeated for the parameters requested. The names are of the form
X_D_P_S, where X is literal,
D is an option description of the parameter,
P is the parameter code,
and S is the statistic code (if applicable).

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
This function accepts a url parameter that already contains the desired
NWIS site, parameter code, statistic, startdate and enddate.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
site_id <- "02177000"
startDate <- "2012-09-01"
endDate <- "2012-10-01"
offering <- "00003"
property <- "00060"
obs_url <- constructNWISURL(site_id, property, startDate, endDate, "dv")
\donttest{
data <- importWaterML1(obs_url, asDateTime = TRUE)


unitDataURL <- constructNWISURL(
  site_id, property,
  "2013-11-03", "2013-11-03", "uv"
)
unitData <- importWaterML1(unitDataURL, TRUE)

# Two sites, two pcodes, one site has two data descriptors:
siteNumber <- c("01480015", "04085427")
obs_url <- constructNWISURL(
  siteNumber, c("00060", "00010"),
  startDate, endDate, "dv"
)
data <- importWaterML1(obs_url)
data$dateTime <- as.Date(data$dateTime)
data <- renameNWISColumns(data)
names(attributes(data))
attr(data, "url")
attr(data, "disclaimer")

inactiveSite <- "05212700"
inactiveSite <- constructNWISURL(inactiveSite, "00060",
                                 "2014-01-01", "2014-01-10", "dv")
inactiveSite <- importWaterML1(inactiveSite)

inactiveAndAcitive <- c("07334200", "05212700")
inactiveAndAcitive <- constructNWISURL(inactiveAndAcitive,
                         "00060", "2014-01-01", "2014-01-10", "dv")
inactiveAndAcitive <- importWaterML1(inactiveAndAcitive)

# Timezone change with specified local timezone:
tzURL <- constructNWISURL("04027000", c("00300", "63680"),
                          "2011-11-05", "2011-11-07", "uv")
tzIssue <- importWaterML1(tzURL,
  asDateTime = TRUE, tz = "America/Chicago"
)

}
filePath <- system.file("extdata", package = "dataRetrieval")
fileName <- "WaterML1Example.xml"
fullPath <- file.path(filePath, fileName)
importFile <- importWaterML1(fullPath, TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=renameNWISColumns]{renameNWISColumns()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWQPdots.R
\name{readWQPdots}
\alias{readWQPdots}
\title{Format and organize WQP arguments that are passed in as \code{...}.}
\usage{
readWQPdots(..., legacy = TRUE)
}
\description{
Format and organize WQP arguments that are passed in as \code{...}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISunit.R
\name{readNWISuv}
\alias{readNWISuv}
\title{Instantaneous value data retrieval from USGS (NWIS)}
\usage{
readNWISuv(siteNumbers, parameterCd, startDate = "", endDate = "", tz = "UTC")
}
\arguments{
\item{siteNumbers}{character USGS site number (or multiple sites).  This is usually an 8 digit number}

\item{parameterCd}{character USGS parameter code.  This is usually an 5 digit number.}

\item{startDate}{character starting date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the earliest possible record. Simple date arguments are specified in local time.}

\item{endDate}{character ending date for data retrieval in the form YYYY-MM-DD. Default is "" which indicates
retrieval for the latest possible record. Simple date arguments are specified in local time.}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the data provided tz_cd column.
Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
"America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  \code{OlsonNames()}
for more information on time zones.}
}
\value{
A data frame with the following columns:
\tabular{lll}{
Name \tab Type \tab Description \cr
agency_cd \tab character \tab The NWIS code for the agency reporting the data\cr
site_no \tab character \tab The USGS site number \cr
dateTime \tab POSIXct \tab The date and time of the value converted to UTC \cr
tz_cd \tab character \tab The time zone code for dateTime \cr
code \tab character \tab Any codes that qualify the corresponding value\cr
value \tab numeric \tab The numeric value for the parameter \cr
}
Note that code and value are repeated for the parameters requested. The names are of the form:
X_D_P_S, where X is literal,
D is an option description of the parameter,
P is the parameter code,
and S is the statistic code (if applicable).

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
variableInfo \tab data.frame \tab A data frame containing information on the requested parameters \cr
statisticInfo \tab data.frame \tab A data frame containing information on the requested statistics on the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Imports data from NWIS web service.
Inputs to this function are USGS site ids, USGS parameter codes,
and start and end date. For a more complex query, use \code{\link[=readNWISdata]{readNWISdata()}},
including an arguement service="uv".
Not all parameter codes are available for all data.
Use the function \code{\link[=whatNWISdata]{whatNWISdata()}} to discover what data
is available for a USGS site. The column data_type_cd with the values "uv"
returned from \code{\link[=whatNWISdata]{whatNWISdata()}}) are available from this service.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
site_id <- "05114000"
parameterCd <- "00060"
startDate <- "2014-10-10"
endDate <- "2014-10-10"
\donttest{

#rawData <- readNWISuv(site_id, parameterCd, startDate, endDate)

#rawData_today <- readNWISuv(site_id, parameterCd, Sys.Date(), Sys.Date())

}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=renameNWISColumns]{renameNWISColumns()}}, \code{\link[=importWaterML1]{importWaterML1()}}
}
\keyword{USGS}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ogc_documentation.R
\name{check_OGC_requests}
\alias{check_OGC_requests}
\title{Check OGC requests}
\usage{
check_OGC_requests(endpoint = "daily", type = "queryables")
}
\arguments{
\item{endpoint}{Character, can be any existing collection}

\item{type}{Character, can be "queryables", "schema"}
}
\value{
list
}
\description{
Check OGC requests
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{

dv_queryables <- check_OGC_requests(endpoint = "daily",
                                type = "queryables")
dv_schema <- check_OGC_requests(endpoint = "daily",
                            type = "schema")
ts_meta_queryables <- check_OGC_requests(endpoint = "time-series-metadata",
                                type = "queryables")
ts_meta_schema <- check_OGC_requests(endpoint = "time-series-metadata",
                                type = "schema")
}
\dontshow{\}) # examplesIf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNWISunit.R
\name{readNWISrating}
\alias{readNWISrating}
\title{Rating table for an active USGS streamgage retrieval}
\usage{
readNWISrating(siteNumber, type = "base", convertType = TRUE)
}
\arguments{
\item{siteNumber}{character USGS site number.  This is usually an 8 digit number}

\item{type}{character can be "base", "corr", or "exsa"}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character}
}
\value{
A data frame. If \code{type} is "base, " then the columns are
INDEP, typically the gage height, in feet; DEP, typically the streamflow,
in cubic feet per second; and STOR, where "*" indicates that the pair are
a fixed point of the rating curve. If \code{type} is "exsa, " then an
additional column, SHIFT, is included that indicates the current shift in
the rating for that value of INDEP. If \code{type} is "corr, " then the
columns are INDEP, typically the gage height, in feet; CORR, the correction
for that value; and CORRINDEP, the corrected value for CORR.\cr
If \code{type} is "base, " then the data frame has an attribute called "RATING"
that describes the rating curve is included.

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
comment \tab character \tab Header comments from the RDB file \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
RATING \tab character \tab Rating information \cr
}
}
\description{
Reads current rating table for an active USGS streamgage from NWISweb.
}
\note{
Not all active USGS streamgages have traditional rating curves that
relate flow to stage.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
site_id <- "01594440"
\donttest{
#data <- readNWISrating(site_id, "base")
#attr(data, "RATING")
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=constructNWISURL]{constructNWISURL()}}, \code{\link[=importRDB1]{importRDB1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importWQP.R
\name{importWQP}
\alias{importWQP}
\title{Basic Water Quality Portal Data parser}
\usage{
importWQP(obs_url, tz = "UTC", csv = TRUE, convertType = TRUE)
}
\arguments{
\item{obs_url}{character URL to Water Quality Portal#' @keywords data import USGS web service}

\item{tz}{character to set timezone attribute of datetime. Default is UTC
(properly accounting for daylight savings times based on the data's provided tz_cd column).
Possible values include "America/New_York","America/Chicago", "America/Denver","America/Los_Angeles",
"America/Anchorage","America/Honolulu","America/Jamaica","America/Managua",
"America/Phoenix", and "America/Metlakatla"}

\item{csv}{logical. Is the data coming back with a csv or tsv format. Default is \code{FALSE}.
Currently, the summary service does not support tsv, for other services tsv is the safer choice.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character.}
}
\value{
retval dataframe raw data returned from the Water Quality Portal. Additionally,
a POSIXct dateTime column is supplied for
start and end times, and converted to UTC. See
\url{https://www.waterqualitydata.us/portal_userguide/} for more information.
}
\description{
Imports data from the Water Quality Portal based on a specified url.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
# These examples require an internet connection to run

## Examples take longer than 5 seconds:
\donttest{
rawSampleURL <- constructWQPURL("USGS-01594440", "01075", "", "")

rawSample <- importWQP(rawSampleURL)


STORETex <- constructWQPURL("WIDNR_WQX-10032762", "Specific conductance", "", "")

STORETdata <- importWQP(STORETex)

STORETdata_char <- importWQP(STORETex, convertType = FALSE)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=readWQPdata]{readWQPdata()}}, \code{\link[=readWQPqw]{readWQPqw()}}, \code{\link[=whatWQPsites]{whatWQPsites()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_samples.R
\name{summarize_waterdata_samples}
\alias{summarize_waterdata_samples}
\alias{summarize_USGS_samples}
\title{USGS Samples Summary Data}
\usage{
summarize_waterdata_samples(monitoringLocationIdentifier)

summarize_USGS_samples(monitoringLocationIdentifier)
}
\arguments{
\item{monitoringLocationIdentifier}{A single monitoring location identifier
with two parts, separated by a dash (-): the agency code and the location
number. Examples: USGS-040851385, AZ014-320821110580701, CAX01-15304600.
The summary service accepts only one site at a time; supplying a vector of
length > 1 raises an error. The agency prefix is required: bare location
numbers (e.g. "040851385") are accepted by the service but return an empty
result.}
}
\value{
data frame with summary of data available based on the monitoringLocationIdentifier
}
\description{
This function creates the call and gets the data for discrete water quality samples summary data
service described at \url{https://api.waterdata.usgs.gov/samples-data/docs}.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
monitoringLocationIdentifier <- "USGS-04074950"

what_data <- summarize_waterdata_samples(monitoringLocationIdentifier)

}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getWebServiceData.R
\name{has_internet_2}
\alias{has_internet_2}
\title{has_internet_2}
\usage{
has_internet_2(obs_url)
}
\arguments{
\item{obs_url}{character obs_url to check}
}
\description{
Function to check for internet even if the user
is behind a proxy
}
\details{
If this is giving you problems, override the false negative:
Sys.setenv("OVERRIDE_INTERNET_TEST" = TRUE)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWQPdata.R
\name{readWQPdata}
\alias{readWQPdata}
\title{General Data Import from Water Quality Portal}
\usage{
readWQPdata(
  ...,
  service = "Result",
  querySummary = FALSE,
  tz = "UTC",
  ignore_attributes = FALSE,
  convertType = TRUE
)
}
\arguments{
\item{\dots}{see \url{https://www.waterqualitydata.us/webservices_documentation} for a complete list of options.
A list of arguments can also be supplied. For more information see the above
description for this help file. One way to figure out how to construct a WQP query is to go to the "Advanced"
form in the Water Quality Portal. Use the form to discover what parameters are available. Once the query is
set in the form, scroll down to the "Query URL". You will see the parameters
after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
in the Characteristic Group dropdown, you will see characteristicType=Nutrient
in the Query URL. The corresponding argument for dataRetrieval is
characteristicType = "Nutrient". dataRetrieval users do not need to include
mimeType, and providers is optional (these arguments are picked automatically).}

\item{service}{character. See Details for more information.}

\item{querySummary}{logical to only return the number of records and unique sites that
will be returned from this query. Choosing TRUE is deprecated, readWQPsummary
is recommended instead.}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the
data's provided tz_cd column.
Possible values to provide are "America/New_York","America/Chicago",
"America/Denver","America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings
time: "America/Honolulu",
"America/Jamaica","America/Managua","America/Phoenix", and "America/Metlakatla".
See also  \code{OlsonNames()}
for more information on time zones.}

\item{ignore_attributes}{logical to choose to ignore fetching site and status
attributes. Default is \code{FALSE}.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character.}
}
\value{
A data frame, the specific columns will depend on the "service" and/or "dataProfile".

There are also several useful attributes attached to the data frame:
\tabular{lll}{
Name \tab Type \tab Description \cr
url \tab character \tab The url used to generate the data \cr
siteInfo \tab data.frame \tab A data frame containing information on the requested sites \cr
headerInfo \tab data.frame \tab A data frame returned from the WQP status service \cr
queryTime \tab POSIXct \tab The time the data was returned \cr
}
}
\description{
Imports data from Water Quality Portal web service. This function gets the data from here:
\url{https://www.waterqualitydata.us}.
}
\details{
This function uses \dots as a query input, which can be very flexible, but also
has a steeper learning curve. For a quick overview, scroll down to the Examples
in this help file to see many query options.

There are currently 10 legacy options
for data provided by the Water Quality Portal:

Legacy:
\tabular{lll}{
WQP Radio Button \tab service argument \tab Base URL  \cr
Sample Results \tab Result \tab /data/Result/search \cr
Site Data Only \tab Station \tab /data/Station/search \cr
Sampling Activity \tab Activity \tab /data/Activity/search \cr
Sampling Activity Metrics \tab ActivityMetric \tab /data/ActivityMetric/search \cr
Site Summary (not advertised on WQP)  \tab SiteSummary \tab /data/summary/monitoringLocation/search \cr
Project Data \tab Project \tab /data/Project/search \cr
Project Monitoring Location Weighting Data \tab ProjectMonitoringLocationWeighting \tab /data/ProjectMonitoringLocationWeighting/search \cr
Result Detection Quantitation Limit Data \tab ResultDetectionQuantitationLimit \tab /data/ResultDetectionQuantitationLimit/search \cr
Biological Habitat Metrics \tab BiologicalMetric \tab /data/BiologicalMetric/search \cr
Organization Data \tab Organization \tab /data/Organization/search \cr
}

There are 4 WQX3 options. These are still in-development,
and should be used with caution.

\tabular{llll}{
WQP Radio Button \tab service argument \tab Base URL \tab dataProfile \cr
Monitoring Locations \tab StationWQX3 \tab /wqx3/Station/search \tab \cr
Full Physical Chemical \tab ResultWQX3 \tab /wqx3/Result/search \tab fullPhysChem \cr
Narrow \tab ResultWQX3 \tab /wqx3/Result/search \tab narrow \cr
Basic Physical Chemical \tab ResultWQX3 \tab /wqx3/Result/search \tab basicPhysChem \cr
Sampling Activity \tab ActivityWQX3 \tab /wqx3/Activity/search \cr
}
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{

# Legacy:
#nameToUse <- "pH"
#pHData <- readWQPdata(siteid = "USGS-04024315",
#                       characteristicName = nameToUse)
#ncol(pHData)
#attr(pHData, "siteInfo")
#attr(pHData, "queryTime")
#attr(pHData, "url")

# WQX3:
#pHData_wqx3 <- readWQPdata(siteid = "USGS-04024315",
#                            characteristicName = nameToUse,
#                            service = "ResultWQX3",
#                            dataProfile = "basicPhysChem")
# attr(pHData_wqx3, "url")

# More examples:
# querying by county
#DeWitt <- readWQPdata(
#   statecode = "Illinois",
#   countycode = "DeWitt",
#   characteristicName = "Nitrogen"
# )

#attr(DeWitt, "url")

#DeWitt_wqx3 <- readWQPdata(
#    statecode = "Illinois",
#    countycode = "DeWitt",
#    characteristicName = "Nitrogen",
#    service = "ResultWQX3",
#    dataProfile = "basicPhysChem",
#    ignore_attributes = TRUE)

#attr(DeWitt_wqx3, "url")

# Data profile: "Sampling Activity"
#activity <- readWQPdata(
#   siteid = "USGS-04024315",
#   service = "Activity"
#)
#attr(activity, "url")

# activity_wqx3 <- readWQPdata(
#   siteid = "USGS-04024315",
#   service = "ActivityWQX3"
# )
# attr(activity_wqx3, "url")

#Dane_activity <- readWQPdata(
#   statecode = "Wisconsin",
#   countycode = "Dane",
#   startDateLo = "2023-01-01",
#   startDateHi = "2023-12-31",
#   service = "Activity"
# )
#attr(Dane_activity, "url")

# Dane_activity_wqx3 <- readWQPdata(
#   statecode = "Wisconsin",
#   countycode = "Dane",
#   startDateLo = "2023-01-01",
#   startDateHi = "2023-12-31",
#   service = "ActivityWQX3"
# )
# attr(Dane_activity_wqx3, "url")

########################################################
# Additional examples:


# Data profiles: "Organization Data"
#org_data <- readWQPdata(
#   statecode = "WI",
#   countycode = "Dane",
#   service = "Organization"
# )

# Data profiles: "Project Data"
#project_data <- readWQPdata(
#   statecode = "WI",
#   countycode = "Dane",
#   service = "Project"
# )

# Data profiles: "Project Monitoring Location Weighting Data"
#proj_mlwd <- readWQPdata(
#   statecode = "WI",
#   countycode = "Dane",
#   service = "ProjectMonitoringLocationWeighting"
# )

# Data profiles: "Sample Results (physical/chemical metadata)"
# samp_data <- readWQPdata(
#   siteid = "USGS-04024315",
#   dataProfile = "resultPhysChem",
#   service = "Result"
# )

# Data profiles: "Sample Results (biological metadata)"
#samp_bio <- readWQPdata(
#   siteid = "USGS-04024315",
#   dataProfile = "biological",
#   service = "Result"
# )


# Data profiles: "Sample Results (narrow)"
# samp_narrow <- readWQPdata(
#   siteid = "USGS-04024315",
#   service = "Result",
#   dataProfile = "narrowResult"
# )

# samp_narrow_wqx3 <- readWQPdata(
#   siteid = "USGS-04024315",
#   service = "ResultWQX3",
#   dataProfile = "narrow"
# )


# Data profiles: "Sampling Activity"
#samp_activity <- readWQPdata(
#   siteid = "USGS-04024315",
#   dataProfile = "activityAll",
#   service = "Activity"
#  )

# Data profile: "Sampling Activity Metrics"
# act_metrics <- readWQPdata(
#   statecode = "WI",
#   countycode = "Dane",
#   service = "ActivityMetric"
# )

# Data profile: "Result Detection Quantitation Limit Data"
# dl_data <- readWQPdata(
#   siteid = "USGS-04024315",
#   service = "ResultDetectionQuantitationLimit"
# )

# other options:
# Phosphorus <- readWQPdata(
#   statecode = "WI", countycode = "Dane",
#   characteristicName = "Phosphorus",
#   startDateLo = "2023-01-01",
#   ignore_attributes = TRUE,
#   convertType = FALSE
# )

#rawPHsites_legacy <- readWQPdata(siteid = c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#                         characteristicName = "pH",
#                         service = "Result",
#                         dataProfile = "narrowResult" )

# rawPHsites <- readWQPdata(siteid = c("USGS-05406450", "USGS-05427949", "WIDNR_WQX-133040"),
#                           characteristicName = "pH",
#                           service = "ResultWQX3",
#                           dataProfile = "narrow" )

}
\dontshow{\}) # examplesIf}
}
\keyword{WQP}
\keyword{data}
\keyword{import}
\keyword{service}
\keyword{web}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getWebServiceData.R
\name{getQuerySummary}
\alias{getQuerySummary}
\title{getting header information from a WQP query}
\usage{
getQuerySummary(url)
}
\arguments{
\item{url}{the query url}
}
\description{
getting header information from a WQP query
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whatWQPsites.R
\name{readWQPsummary}
\alias{readWQPsummary}
\title{Summary of Data Available from Water Quality Portal}
\usage{
readWQPsummary(...)
}
\arguments{
\item{\dots}{see \url{https://www.waterqualitydata.us/webservices_documentation}
for a complete list of options. A list of arguments can also be supplied.
One way to figure out how to construct a WQP query is to go to the "Advanced"
form in the Water Quality Portal:
\url{https://www.waterqualitydata.us/#mimeType=csv&providers=NWIS&providers=STORET}
Use the form to discover what parameters are available. Once the query is
set in the form, scroll down to the "Query URL". You will see the parameters
after "https://www.waterqualitydata.us/#". For example, if you chose "Nutrient"
in the Characteristic Group dropdown, you will see characteristicType=Nutrient
in the Query URL. The corresponding argument for dataRetrieval is
characteristicType = "Nutrient". dataRetrieval users do not need to include
mimeType, and providers is optional (these arguments are picked automatically).}
}
\value{
A data frame from the data returned from the Water Quality Portal
about the data available for the query parameters.
}
\description{
Returns a list of sites with year-by-year information on what data is available.
The function gets the data from: \url{https://www.waterqualitydata.us}.
Arguments to the function should be based on
\url{https://www.waterqualitydata.us/webservices_documentation}.
The information returned from this function describes the
available data at the WQP sites, and some metadata on the sites themselves.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
# Summary of a single site for the last 5 years:
#site_5 <- readWQPsummary(
#   siteid = "USGS-07144100",
#   summaryYears = 5
#  )

# Summary of a single site for the full period of record:
#site_all <- readWQPsummary(
#   siteid = "USGS-07144100",
#   summaryYears = "all"
# )

# Summary of the data available from streams in a single county:
# dane_county_data <- readWQPsummary(
#   countycode = "US:55:025",
#   summaryYears = 5,
#   siteType = "Stream"
# )

# Summary of the data all available from lakes in a single county:
# lake_sites <- readWQPsummary(
#   siteType = "Lake, Reservoir, Impoundment",
#   countycode = "US:55:025"
# )

# Summary of the data available for the last 5 years in New Jersey:
#state1 <- readWQPsummary(
#   statecode = "NJ",
#   summaryYears = 5,
#    siteType = "Stream"
# )
}
\dontshow{\}) # examplesIf}
}
\seealso{
whatWQPsites whatWQPdata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renameColumns.R
\name{renameNWISColumns}
\alias{renameNWISColumns}
\title{renameColumns}
\usage{
renameNWISColumns(
  rawData,
  p00010 = "Wtemp",
  p00045 = "Precip",
  p00060 = "Flow",
  p00065 = "GH",
  p00095 = "SpecCond",
  p00300 = "DO",
  p00400 = "pH",
  p62611 = "GWL",
  p63680 = "Turb",
  p72019 = "WLBLS",
  ...
)
}
\arguments{
\item{rawData}{the daily- or unit-values datset retrieved from NWISweb.}

\item{p00010}{the base name for parameter code 00010.}

\item{p00045}{the base name for parameter code 00045.}

\item{p00060}{the base name for parameter code 00060.}

\item{p00065}{the base name for parameter code 00065.}

\item{p00095}{the base name for parameter code 00095.}

\item{p00300}{the base name for parameter code 00300.}

\item{p00400}{the base name for parameter code 00400.}

\item{p62611}{the base name for parameter code 62611.}

\item{p63680}{the base name for parameter code 63680.}

\item{p72019}{the base name for parameter code 72019.}

\item{\dots}{named arguments for the base name for any other parameter code. The
form of the name must be like pXXXXX, where XXXXX is the parameter code.}
}
\value{
A dataset like \code{data} with selected columns renamed.
}
\description{
Rename columns coming back from NWIS data retrievals.  Daily and unit value columns
have names derived from their data descriptor, parameter, and statistic codes. This
function reads information from the header and the arguments in the call to
to rename those columns.
}
\note{
The following statistics codes are converted by \code{renameNWISColumns}.
\describe{
\item{00000}{Instantaneous Value, suffix: Inst}
\item{00001}{Maximum value, suffix: Max}
\item{00002}{Minimum value, suffix: Min}
\item{00003}{Mean value, no suffix}
\item{00006}{Sum of values, suffix: Sum}
\item{00007}{Modal value, suffix: Mode}
\item{00008}{Median value, suffix: Median}
\item{00012}{Equivalent mean value, suffix: EqMean}
\item{00021}{Tidal high-high value, suffix: HiHiTide}
\item{00022}{Tidal low-high value, suffix: LoHiTide}
\item{00023}{Tidal high-low value, suffix: HiLoTide}
\item{00024}{Tidal low-low value, suffix: LoLoTide}
}
}
\examples{
siteWithTwo <- "01480015"
startDate <- "2012-09-01"
endDate <- "2012-10-01"
\donttest{
twoResults <- readNWISdv(siteWithTwo, "00060", startDate, endDate)
names(twoResults)
renamedCols <- renameNWISColumns(twoResults)
names(renamedCols)
# Custom names:
newNames <- renameNWISColumns(twoResults, p00060 = "Discharge")
names(newNames)
}
}
\seealso{
\code{\link[=readNWISdv]{readNWISdv()}}, \code{\link[=readNWISuv]{readNWISuv()}}
}
\keyword{IO}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata_samples.R
\name{read_waterdata_samples}
\alias{read_waterdata_samples}
\title{USGS Samples Data}
\usage{
read_waterdata_samples(
  monitoringLocationIdentifier = NA,
  siteTypeCode = NA,
  boundingBox = NA,
  hydrologicUnit = NA,
  activityMediaName = NA,
  characteristicGroup = NA,
  characteristic = NA,
  characteristicUserSupplied = NA,
  activityStartDateLower = NA,
  activityStartDateUpper = NA,
  countryFips = NA,
  stateFips = NA,
  countyFips = NA,
  projectIdentifier = NA,
  recordIdentifierUserSupplied = NA,
  siteTypeName = NA,
  usgsPCode = NA,
  pointLocationLatitude = NA,
  pointLocationLongitude = NA,
  pointLocationWithinMiles = NA,
  dataType = "results",
  dataProfile = NA,
  tz = "UTC",
  convertType = TRUE
)
}
\arguments{
\item{monitoringLocationIdentifier}{A monitoring location identifier has two parts: the agency code
and the location number, separated by a dash (-). Location identifiers should be separated with commas,
for example: AZ014-320821110580701, CAX01-15304600, USGS-040851385. Location
numbers without an agency prefix are assumed to have the prefix USGS.}

\item{siteTypeCode}{Site type code query parameter. See available
options by running \code{check_waterdata_sample_params("sitetype")$typeCode}.}

\item{boundingBox}{North and South are latitude values; East and West are longitude values.
A vector of 4 (west, south, east, north) is expected.
An example would be: c(-92.8, 44.2, -88.9, 46.0).}

\item{hydrologicUnit}{Hydrologic Unit Codes (HUCs) identify physical areas
within the US that drain to a certain portion of the stream network.
This filter accepts values containing 2, 4, 6, 8, 10 or 12 digits.}

\item{activityMediaName}{Sample media refers to the environmental medium that
was sampled or analyzed.}

\item{characteristicGroup}{Characteristic group is a broad category describing the sample.
See available options by running
\code{check_waterdata_sample_params("characteristicgroup")$characteristicGroup}.}

\item{characteristic}{Characteristic is a specific category describing the sample.
See available options by running
\code{check_waterdata_sample_params("characteristics")$characteristicName}.}

\item{characteristicUserSupplied}{Observed property is the USGS term for the
constituent sampled and the property name gives a detailed description of what
was sampled. Observed property is mapped to characteristicUserSupplied and replaces
the parameter name and pcode USGS
previously used to describe discrete sample data. Find more information in the
Observed Properties and Parameter Codes section of the Code Dictionary found here:
\url{https://waterdata.usgs.gov/code-dictionary/}.}

\item{activityStartDateLower}{The service will return records with dates earlier
than the value entered for activityStartDateUpper. Can be an R Date object, or
a string with format YYYY-MM-DD. The logic is inclusive, i.e. it will also return
records that match the date.}

\item{activityStartDateUpper}{The service will return records with dates later
than the value entered for activityStartDateLower. Can be an R Date object, or
a string with format YYYY-MM-DD. The logic is inclusive, i.e. it will also return
records that match the date.}

\item{countryFips}{Country query parameter. Do not set redundant parameters.
If another query parameter contains the country information, leave this parameter
set to the default NA. See available options by running \code{check_waterdata_sample_params("countries")},
where the "id" field contains the value to use in the countryFips input.}

\item{stateFips}{State query parameter. To get a list of available state fips,
run \code{check_waterdata_sample_params("states")}. The "fips" can be created using the function
\code{stateCdLookup} - for example: \code{stateCdLookup("WI", "fips")}.
FIPs codes for states take the format:
CountryAbbrev:StateNumber, like US:55 for Wisconsin.}

\item{countyFips}{County query parameter. To get a list of available counties,
run \code{check_waterdata_sample_params("counties")}. The "Fips" can be created using the function
\code{countyCdLookup} - for example: \code{countyCdLookup("WI", "Dane", "fips")}
for Dane County, WI.
FIPs codes for counties take the format:
CountryAbbrev:StateNumber:CountyNumber, like US:55:025 for Dane County, WI.}

\item{projectIdentifier}{Project identifier query parameter. This information
would be needed from prior project information.}

\item{recordIdentifierUserSupplied}{Record identifier, user supplied identifier. This
information would be needed from the data supplier.}

\item{siteTypeName}{Site type name query parameter. See available
options by running \code{check_waterdata_sample_params("sitetype")$typeName}.}

\item{usgsPCode}{USGS parameter code. See available options by running
\code{check_waterdata_sample_params("characteristics")$parameterCode}.}

\item{pointLocationLatitude}{Latitude for a point/radius query (decimal degrees). Must be used
with pointLocationLongitude and pointLocationWithinMiles.}

\item{pointLocationLongitude}{Longitude for a point/radius query (decimal degrees). Must be used
with pointLocationLatitude and pointLocationWithinMiles.}

\item{pointLocationWithinMiles}{Radius for a point/radius query. Must be used
with pointLocationLatitude and pointLocationLongitude}

\item{dataType}{Options include: "Results", "Monitoring locations", "Activities",
"Projects", and "Organizations".}

\item{dataProfile}{Profile depends on type. Options for "results" dataType are:
"fullphyschem", "basicphyschem", "fullbio", "basicbio", "narrow",
"resultdetectionquantitationlimit", "labsampleprep", "count". Options for "locations" are:
"site" and "count". Options for "activities" are "sampact", "actmetric", "actgroup",
and "count". Options for "projects" are:
"project" and "projectmonitoringlocationweight". Options for "organizations" are:
"organization" and "count".}

\item{tz}{character to set timezone attribute of datetime. Default is UTC
(properly accounting for daylight savings times based on the data's provided tz_cd column).
Possible values include "America/New_York","America/Chicago", "America/Denver","America/Los_Angeles",
"America/Anchorage","America/Honolulu","America/Jamaica","America/Managua",
"America/Phoenix", and "America/Metlakatla"}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates, datetimes,
numerics based on a standard algorithm. If false, everything is returned as a character.}
}
\description{
This function creates the call and gets the data for discrete water quality samples data
service described at \url{https://waterdata.usgs.gov/download-samples/}.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
ph_data <- read_waterdata_samples(
               monitoringLocationIdentifier = "USGS-04074950",
               characteristicUserSupplied = "pH, water, unfiltered, field",
               activityStartDateUpper = "2000-01-01",
               dataProfile = "narrow")

nameToUse <- "pH"
pHData <- read_waterdata_samples(monitoringLocationIdentifier = "USGS-04024315",
                         characteristic = nameToUse)
ncol(pHData)
attr(pHData, "url")
attr(pHData, "queryTime")

summary_data <- read_waterdata_samples(monitoringLocationIdentifier = "USGS-04024315",
                               dataType = "projects")

}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readNGWMNdata.R
\name{readNGWMNdata}
\alias{readNGWMNdata}
\title{Import data from the National Groundwater Monitoring Network.}
\usage{
readNGWMNdata(service, ..., asDateTime = TRUE, tz = "UTC")
}
\arguments{
\item{service}{char Service for the request - "observation" and "featureOfInterest" are implemented.}

\item{\dots}{Other parameters to supply, namely \code{siteNumbers} or \code{bbox}}

\item{asDateTime}{logical if \code{TRUE}, will convert times to POSIXct format.  Currently defaults to
\code{FALSE} since time zone information is not included.}

\item{tz}{character to set timezone attribute of dateTime. Default is "UTC", and converts the
date times to UTC, properly accounting for daylight savings times based on the data's provided time zone offset.
Possible values to provide are "America/New_York", "America/Chicago", "America/Denver", "America/Los_Angeles",
"America/Anchorage", as well as the following which do not use daylight savings time: "America/Honolulu",
"America/Jamaica", "America/Managua", "America/Phoenix", and "America/Metlakatla". See also  \code{OlsonNames()}
for more information on time zones.}
}
\description{
Only water level data and site locations and names are currently available through the web service.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}
\donttest{
# one site
site <- "USGS.430427089284901"
#oneSite <- readNGWMNdata(siteNumbers = site, service = "observation")

# multiple sites
sites <- c("USGS.272838082142201", "USGS.404159100494601", "USGS.401216080362703")
# Very slow:
# multiSiteData <- readNGWMNdata(siteNumbers = sites, service = "observation")
# attributes(multiSiteData)

# non-USGS site
# accepts colon or period between agency and ID
site <- "MBMG:702934"
# data <- readNGWMNdata(siteNumbers = site, service = "featureOfInterest")

# bounding box
# bboxSites <- readNGWMNdata(service = "featureOfInterest", bbox = c(30, -102, 31, 99))
# retrieve  sites.  Set asDateTime to false since one site has an invalid date
# Very slow:
# bboxData <- readNGWMNdata(service = "observation", siteNumbers = bboxSites$site[1:3],
#                           asDateTime = FALSE)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_waterdata.R
\name{read_waterdata}
\alias{read_waterdata}
\title{Generalized USGS Water Data API retrieval function}
\usage{
read_waterdata(
  service,
  CQL,
  ...,
  convertType = getOption("dataRetrieval.convertType"),
  limit = getOption("dataRetrieval.limit"),
  attach_request = getOption("dataRetrieval.attach_request")
)
}
\arguments{
\item{service}{character, can be any existing collection.}

\item{CQL}{A string in a Common Query Language format.}

\item{\dots}{Additional arguments to send to the request.}

\item{convertType}{logical, defaults to \code{TRUE}. If \code{TRUE}, the function
will convert the data to dates and qualifier to string vector.}

\item{limit}{numeric, The optional limit parameter is used to control the subset of the
selected features that should be returned in each page. The maximum allowable
limit is 50,000. It may be beneficial to set this number lower if your internet
connection is spotty. The default (\code{NA}) will set the limit to the maximum
allowable limit for the service.}

\item{attach_request}{logical, defaults to TRUE.
If set to \code{TRUE}, the full request sent to the Water Data API is attached
as an attribute to the data set.}
}
\description{
Function that allows complex CQL queries.
See \url{https://api.waterdata.usgs.gov/docs/ogcapi/complex-queries/}
for more information.
}
\examples{
\dontshow{if (is_dataRetrieval_user()) withAutoprint(\{ # examplesIf}

\donttest{
cql <- '{
"op": "and",
"args": [
  {
    "op": "in",
    "args": [
      { "property": "parameter_code" },
      [ "00060", "00065" ]
    ]
  },
 {
    "op": "in",
    "args": [
      { "property": "monitoring_location_id" },
      [ "USGS-07367300", "USGS-03277200" ]
    ]
  }
]
}'

dv_data <- read_waterdata(service = "daily",
                          CQL = cql,
                          time = c("2023-01-01", "2024-01-01"))

# A wildcard in CQL2 is \%
# Here's how to get HUCs that fall within 02070010
cql_huc_wildcard <- '{
"op": "like",
"args": [
  { "property": "hydrologic_unit_code" },
  "02070010\%"
]
}'

what_huc_sites <- read_waterdata(service = "monitoring-locations",
                                 CQL = cql_huc_wildcard)

}
\dontshow{\}) # examplesIf}
}
